/* Generated by Cython 0.29.5 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/home/sean/local/include/cpl_conv.h",
            "/home/sean/local/include/cpl_string.h",
            "/home/sean/local/include/cpl_vsi.h",
            "/home/sean/local/include/gdal.h",
            "/home/sean/local/include/gdal_version.h",
            "/home/sean/local/include/ogr_api.h",
            "/home/sean/local/include/ogr_core.h",
            "/home/sean/local/include/ogr_srs_api.h"
        ],
        "include_dirs": [
            "/home/sean/local/include"
        ],
        "libraries": [
            "gdal"
        ],
        "library_dirs": [
            "/home/sean/local/lib"
        ],
        "name": "fiona.ogrext",
        "sources": [
            "fiona/ogrext.pyx"
        ]
    },
    "module_name": "fiona.ogrext"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_5"
#define CYTHON_HEX_VERSION 0x001D05F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__fiona__ogrext
#define __PYX_HAVE_API__fiona__ogrext
/* Early includes */
#include <string.h>
#include <stdio.h>
#include "ogr_core.h"
#include "gdal.h"
#include "gdal_version.h"
#include "cpl_conv.h"
#include "cpl_string.h"
#include "cpl_vsi.h"
#include "ogr_srs_api.h"
#include "ogr_api.h"
#include <stdlib.h>
#include "pythread.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "fiona/ogrext.pyx",
  "stringsource",
  "fiona/_geometry.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
};
/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif


/* "fiona/_geometry.pxd":3
 * # Geometry API functions.
 * 
 * ctypedef int OGRErr             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef int __pyx_t_5fiona_9_geometry_OGRErr;

/*--- Type declarations ---*/
struct __pyx_obj_5fiona_9_geometry_GeomBuilder;
struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder;
struct __pyx_obj_5fiona_6ogrext_FeatureBuilder;
struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder;
struct __pyx_obj_5fiona_6ogrext_Session;
struct __pyx_obj_5fiona_6ogrext_WritingSession;
struct __pyx_obj_5fiona_6ogrext_Iterator;
struct __pyx_obj_5fiona_6ogrext_ItemsIterator;
struct __pyx_obj_5fiona_6ogrext_KeysIterator;
struct __pyx_obj_5fiona_6ogrext_MemoryFileBase;
struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode;
struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs;
struct __pyx_t_5fiona_9_geometry_OGREnvelope;
typedef struct __pyx_t_5fiona_9_geometry_OGREnvelope __pyx_t_5fiona_9_geometry_OGREnvelope;

/* "fiona/_geometry.pxd":81
 * 
 * 
 * ctypedef struct OGREnvelope:             # <<<<<<<<<<<<<<
 *     double MinX
 *     double MaxX
 */
struct __pyx_t_5fiona_9_geometry_OGREnvelope {
  double MinX;
  double MaxX;
  double MinY;
  double MaxY;
};
struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build;

/* "fiona/ogrext.pyx":129
 *     """
 * 
 *     cdef build(self, void *feature, encoding='utf-8', bbox=False, driver=None, ignore_fields=None, ignore_geometry=False):             # <<<<<<<<<<<<<<
 *         """Build a Fiona feature object from an OGR feature
 * 
 */
struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build {
  int __pyx_n;
  PyObject *encoding;
  PyObject *bbox;
  PyObject *driver;
  PyObject *ignore_fields;
  PyObject *ignore_geometry;
};

/* "fiona/_geometry.pxd":110
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 *     cdef void *geom
 *     cdef object code
 */
struct __pyx_obj_5fiona_9_geometry_GeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder *__pyx_vtab;
  void *geom;
  PyObject *code;
  PyObject *geomtypename;
  PyObject *ndims;
};


/* "fiona/_geometry.pxd":129
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 *     cdef void * _createOgrGeometry(self, int geom_type) except NULL
 *     cdef _addPointToGeometry(self, void *cogr_geometry, object coordinate)
 */
struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder *__pyx_vtab;
};


/* "fiona/ogrext.pyx":122
 * # Feature extension classes and functions follow.
 * 
 * cdef class FeatureBuilder:             # <<<<<<<<<<<<<<
 *     """Build Fiona features from OGR feature pointers.
 * 
 */
struct __pyx_obj_5fiona_6ogrext_FeatureBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *__pyx_vtab;
};


/* "fiona/ogrext.pyx":304
 * 
 * 
 * cdef class OGRFeatureBuilder:             # <<<<<<<<<<<<<<
 * 
 *     """Builds an OGR Feature from a Fiona feature mapping.
 */
struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder *__pyx_vtab;
};


/* "fiona/ogrext.pyx":442
 * # Collection-related extension classes and functions
 * 
 * cdef class Session:             # <<<<<<<<<<<<<<
 * 
 *     cdef void *cogr_ds
 */
struct __pyx_obj_5fiona_6ogrext_Session {
  PyObject_HEAD
  struct __pyx_vtabstruct_5fiona_6ogrext_Session *__pyx_vtab;
  void *cogr_ds;
  void *cogr_layer;
  PyObject *_fileencoding;
  PyObject *_encoding;
  PyObject *collection;
};


/* "fiona/ogrext.pyx":875
 * 
 * 
 * cdef class WritingSession(Session):             # <<<<<<<<<<<<<<
 * 
 *     cdef object _schema_mapping
 */
struct __pyx_obj_5fiona_6ogrext_WritingSession {
  struct __pyx_obj_5fiona_6ogrext_Session __pyx_base;
  PyObject *_schema_mapping;
};


/* "fiona/ogrext.pyx":1229
 *         log.debug("Flushed data source cache")
 * 
 * cdef class Iterator:             # <<<<<<<<<<<<<<
 * 
 *     """Provides iterated access to feature data.
 */
struct __pyx_obj_5fiona_6ogrext_Iterator {
  PyObject_HEAD
  PyObject *collection;
  PyObject *encoding;
  int next_index;
  PyObject *stop;
  PyObject *start;
  PyObject *step;
  PyObject *fastindex;
  PyObject *stepsign;
};


/* "fiona/ogrext.pyx":1381
 * 
 * 
 * cdef class ItemsIterator(Iterator):             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
struct __pyx_obj_5fiona_6ogrext_ItemsIterator {
  struct __pyx_obj_5fiona_6ogrext_Iterator __pyx_base;
};


/* "fiona/ogrext.pyx":1412
 * 
 * 
 * cdef class KeysIterator(Iterator):             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
struct __pyx_obj_5fiona_6ogrext_KeysIterator {
  struct __pyx_obj_5fiona_6ogrext_Iterator __pyx_base;
};


/* "fiona/ogrext.pyx":1555
 * 
 * 
 * cdef class MemoryFileBase(object):             # <<<<<<<<<<<<<<
 *     """Base for a BytesIO-like class backed by an in-memory file."""
 * 
 */
struct __pyx_obj_5fiona_6ogrext_MemoryFileBase {
  PyObject_HEAD
};


/* "fiona/ogrext.pyx":95
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode {
  PyObject_HEAD
  PyObject *__pyx_v_coords;
  PyObject *__pyx_v_e;
  PyObject *__pyx_v_f;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
  PyObject *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
};


/* "fiona/ogrext.pyx":1154
 *         log.debug("Writing started")
 * 
 *     def writerecs(self, records, collection):             # <<<<<<<<<<<<<<
 *         """Writes buffered records to OGR."""
 *         cdef void *cogr_driver
 */
struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs {
  PyObject_HEAD
  PyObject *__pyx_v_valid_geom_types;
};



/* "fiona/_geometry.pxd":110
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 *     cdef void *geom
 *     cdef object code
 */

struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder {
  PyObject *(*_buildCoords)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, void *);
  PyObject *(*_buildPoint)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLineString)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLinearRing)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildParts)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, void *);
  PyObject *(*_buildPolygon)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiPoint)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiLineString)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiPolygon)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildGeometryCollection)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*build)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, void *);
  PyObject *(*build_wkb)(struct __pyx_obj_5fiona_9_geometry_GeomBuilder *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder *__pyx_vtabptr_5fiona_9_geometry_GeomBuilder;


/* "fiona/_geometry.pxd":129
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 *     cdef void * _createOgrGeometry(self, int geom_type) except NULL
 *     cdef _addPointToGeometry(self, void *cogr_geometry, object coordinate)
 */

struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder {
  void *(*_createOgrGeometry)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, int);
  PyObject *(*_addPointToGeometry)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, void *, PyObject *);
  void *(*_buildPoint)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildLineString)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildLinearRing)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildPolygon)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildMultiPoint)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildMultiLineString)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildMultiPolygon)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*_buildGeometryCollection)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
  void *(*build)(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *, PyObject *);
};
static struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder *__pyx_vtabptr_5fiona_9_geometry_OGRGeomBuilder;


/* "fiona/ogrext.pyx":122
 * # Feature extension classes and functions follow.
 * 
 * cdef class FeatureBuilder:             # <<<<<<<<<<<<<<
 *     """Build Fiona features from OGR feature pointers.
 * 
 */

struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder {
  PyObject *(*build)(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *, void *, struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build *__pyx_optional_args);
};
static struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *__pyx_vtabptr_5fiona_6ogrext_FeatureBuilder;


/* "fiona/ogrext.pyx":304
 * 
 * 
 * cdef class OGRFeatureBuilder:             # <<<<<<<<<<<<<<
 * 
 *     """Builds an OGR Feature from a Fiona feature mapping.
 */

struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder {
  void *(*build)(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder *__pyx_vtabptr_5fiona_6ogrext_OGRFeatureBuilder;


/* "fiona/ogrext.pyx":442
 * # Collection-related extension classes and functions
 * 
 * cdef class Session:             # <<<<<<<<<<<<<<
 * 
 *     cdef void *cogr_ds
 */

struct __pyx_vtabstruct_5fiona_6ogrext_Session {
  PyObject *(*stop)(struct __pyx_obj_5fiona_6ogrext_Session *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_5fiona_6ogrext_Session *__pyx_vtabptr_5fiona_6ogrext_Session;


/* "fiona/ogrext.pyx":875
 * 
 * 
 * cdef class WritingSession(Session):             # <<<<<<<<<<<<<<
 * 
 *     cdef object _schema_mapping
 */

struct __pyx_vtabstruct_5fiona_6ogrext_WritingSession {
  struct __pyx_vtabstruct_5fiona_6ogrext_Session __pyx_base;
};
static struct __pyx_vtabstruct_5fiona_6ogrext_WritingSession *__pyx_vtabptr_5fiona_6ogrext_WritingSession;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectFormatAndDecref.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f);
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f);

/* CBIntToPyUnicode.proto */
#define __Pyx_PyUnicode_FromBInt_int(value)\
    ((value) ? __Pyx_NewRef(__pyx_n_u_True) : __Pyx_NewRef(__pyx_n_u_False))

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* dict_getitem_default.proto */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod1.proto */
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* CallUnboundCMethod2.proto */
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* py_dict_pop.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* PyIntCompare.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_RemainderObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceRemainder(op1, op2) : PyNumber_Remainder(op1, op2))
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_Check(obj)  (__Pyx_TypeCheck(obj, __pyx_CyFunctionType))
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_OGRErr(OGRErr value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_vsi_l_offset(vsi_l_offset value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE OGRFieldType __Pyx_PyInt_As_OGRFieldType(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE vsi_l_offset __Pyx_PyInt_As_vsi_l_offset(PyObject *);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_5fiona_6ogrext_14FeatureBuilder_build(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self, void *__pyx_v_feature, struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build *__pyx_optional_args); /* proto*/
static void *__pyx_f_5fiona_6ogrext_17OGRFeatureBuilder_build(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self, PyObject *__pyx_v_feature, PyObject *__pyx_v_collection); /* proto*/
static PyObject *__pyx_f_5fiona_6ogrext_7Session_stop(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'fiona._shim' */
static int (*__pyx_f_5fiona_5_shim_is_field_null)(void *, int); /*proto*/
static void (*__pyx_f_5fiona_5_shim_set_field_null)(void *, int); /*proto*/
static void (*__pyx_f_5fiona_5_shim_gdal_flush_cache)(void *); /*proto*/
static void *(*__pyx_f_5fiona_5_shim_gdal_open_vector)(char const *, int, PyObject *, PyObject *); /*proto*/
static void *(*__pyx_f_5fiona_5_shim_gdal_create)(void *, char const *, PyObject *); /*proto*/
static OGRErr (*__pyx_f_5fiona_5_shim_gdal_start_transaction)(void *, int); /*proto*/
static OGRErr (*__pyx_f_5fiona_5_shim_gdal_commit_transaction)(void *); /*proto*/
static OGRFieldSubType (*__pyx_f_5fiona_5_shim_get_field_subtype)(void *); /*proto*/
static void (*__pyx_f_5fiona_5_shim_set_field_subtype)(void *, OGRFieldSubType); /*proto*/
static int (*__pyx_f_5fiona_5_shim_check_capability_create_layer)(void *); /*proto*/
static void *(*__pyx_f_5fiona_5_shim_get_linear_geometry)(void *); /*proto*/
static char const *(*__pyx_f_5fiona_5_shim_osr_get_name)(OGRSpatialReferenceH); /*proto*/
static void (*__pyx_f_5fiona_5_shim_osr_set_traditional_axis_mapping_strategy)(OGRSpatialReferenceH); /*proto*/

/* Module declarations from 'fiona._geometry' */
static PyTypeObject *__pyx_ptype_5fiona_9_geometry_GeomBuilder = 0;
static PyTypeObject *__pyx_ptype_5fiona_9_geometry_OGRGeomBuilder = 0;
static unsigned int (*__pyx_f_5fiona_9_geometry_geometry_type_code)(PyObject *); /*proto*/
static PyObject *(*__pyx_f_5fiona_9_geometry_normalize_geometry_type_code)(unsigned int); /*proto*/
static unsigned int (*__pyx_f_5fiona_9_geometry_base_geometry_type_code)(unsigned int); /*proto*/

/* Module declarations from 'fiona._err' */
static int (*__pyx_f_5fiona_4_err_exc_wrap_int)(int); /*proto*/
static void *(*__pyx_f_5fiona_4_err_exc_wrap_pointer)(void *); /*proto*/
static VSILFILE *(*__pyx_f_5fiona_4_err_exc_wrap_vsilfile)(VSILFILE *); /*proto*/

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'cpython.version' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'cpython.exc' */

/* Module declarations from 'cpython.module' */

/* Module declarations from 'cpython.mem' */

/* Module declarations from 'cpython.tuple' */

/* Module declarations from 'cpython.list' */

/* Module declarations from 'cpython.sequence' */

/* Module declarations from 'cpython.mapping' */

/* Module declarations from 'cpython.iterator' */

/* Module declarations from 'cpython.number' */

/* Module declarations from 'cpython.int' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.bool' */
static PyTypeObject *__pyx_ptype_7cpython_4bool_bool = 0;

/* Module declarations from 'cpython.long' */

/* Module declarations from 'cpython.float' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.complex' */
static PyTypeObject *__pyx_ptype_7cpython_7complex_complex = 0;

/* Module declarations from 'cpython.string' */

/* Module declarations from 'cpython.unicode' */

/* Module declarations from 'cpython.dict' */

/* Module declarations from 'cpython.instance' */

/* Module declarations from 'cpython.function' */

/* Module declarations from 'cpython.method' */

/* Module declarations from 'cpython.weakref' */

/* Module declarations from 'cpython.getargs' */

/* Module declarations from 'cpython.pythread' */

/* Module declarations from 'cpython.pystate' */

/* Module declarations from 'cpython.cobject' */

/* Module declarations from 'cpython.oldbuffer' */

/* Module declarations from 'cpython.set' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from 'cpython.bytes' */

/* Module declarations from 'cpython.pycapsule' */

/* Module declarations from 'cpython' */

/* Module declarations from 'fiona.ogrext' */
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_FeatureBuilder = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_OGRFeatureBuilder = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_Session = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_WritingSession = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_Iterator = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_ItemsIterator = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_KeysIterator = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext_MemoryFileBase = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext___pyx_scope_struct___explode = 0;
static PyTypeObject *__pyx_ptype_5fiona_6ogrext___pyx_scope_struct_1_writerecs = 0;
static char const *__pyx_v_5fiona_6ogrext_ODrCCreateDataSource;
static char const *__pyx_v_5fiona_6ogrext_ODrCDeleteDataSource;
static char const *__pyx_v_5fiona_6ogrext_OLC_RANDOMREAD;
static char const *__pyx_v_5fiona_6ogrext_OLC_SEQUENTIALWRITE;
static char const *__pyx_v_5fiona_6ogrext_OLC_RANDOMWRITE;
static char const *__pyx_v_5fiona_6ogrext_OLC_FASTSPATIALFILTER;
static char const *__pyx_v_5fiona_6ogrext_OLC_FASTFEATURECOUNT;
static char const *__pyx_v_5fiona_6ogrext_OLC_FASTGETEXTENT;
static char const *__pyx_v_5fiona_6ogrext_OLC_FASTSETNEXTBYINDEX;
static char const *__pyx_v_5fiona_6ogrext_OLC_CREATEFIELD;
static char const *__pyx_v_5fiona_6ogrext_OLC_CREATEGEOMFIELD;
static char const *__pyx_v_5fiona_6ogrext_OLC_DELETEFIELD;
static char const *__pyx_v_5fiona_6ogrext_OLC_REORDERFIELDS;
static char const *__pyx_v_5fiona_6ogrext_OLC_ALTERFIELDDEFN;
static char const *__pyx_v_5fiona_6ogrext_OLC_DELETEFEATURE;
static char const *__pyx_v_5fiona_6ogrext_OLC_STRINGSASUTF8;
static char const *__pyx_v_5fiona_6ogrext_OLC_TRANSACTIONS;
static int __pyx_v_5fiona_6ogrext_GDAL_VERSION_NUM;
static PyObject *__pyx_f_5fiona_6ogrext__deleteOgrFeature(void *); /*proto*/
static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_FeatureBuilder__set_state(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *, PyObject *); /*proto*/
static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_OGRFeatureBuilder__set_state(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *, PyObject *); /*proto*/
static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_MemoryFileBase__set_state(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "fiona.ogrext"
extern int __pyx_module_is_main_fiona__ogrext;
int __pyx_module_is_main_fiona__ogrext = 0;

/* Implementation of 'fiona.ogrext' */
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_KeyError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_UnicodeDecodeError;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_OSError;
static PyObject *__pyx_builtin_map;
static PyObject *__pyx_builtin_UnicodeEncodeError;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_RuntimeWarning;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_IOError;
static const char __pyx_k_a[] = "a";
static const char __pyx_k_d[] = ":%d";
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_w[] = "w";
static const char __pyx_k_3D[] = "3D ";
static const char __pyx_k_ON[] = "ON";
static const char __pyx_k__4[] = "{";
  static const char __pyx_k__5[] = ", ";
  static const char __pyx_k__8[] = ": ";
  static const char __pyx_k_id[] = "id";
  static const char __pyx_k_os[] = "os";
  static const char __pyx_k_Any[] = "Any";
  static const char __pyx_k_OFF[] = "OFF";
  static const char __pyx_k__12[] = "";
  static const char __pyx_k__13[] = "{}";
  static const char __pyx_k__15[] = "=";
  static const char __pyx_k__16[] = "+";
  static const char __pyx_k__21[] = ":";
  static const char __pyx_k__22[] = ".";
  static const char __pyx_k__25[] = " != ";
  static const char __pyx_k__40[] = "\"{}\"";
  static const char __pyx_k_d_2[] = ".%d";
  static const char __pyx_k_day[] = "day";
  static const char __pyx_k_err[] = "_err";
  static const char __pyx_k_ext[] = "ext";
  static const char __pyx_k_get[] = "get";
  static const char __pyx_k_hex[] = "hex";
  static const char __pyx_k_int[] = "int";
  static const char __pyx_k_len[] = "_len";
  static const char __pyx_k_log[] = "log";
  static const char __pyx_k_map[] = "map";
  static const char __pyx_k_max[] = "max";
  static const char __pyx_k_min[] = "min";
  static const char __pyx_k_new[] = "__new__";
  static const char __pyx_k_pop[] = "pop";
  static const char __pyx_k_pos[] = "_pos";
  static const char __pyx_k_six[] = "six";
  static const char __pyx_k_str[] = "str";
  static const char __pyx_k_xyz[] = "xyz";
  static const char __pyx_k_zip[] = "zip";
  static const char __pyx_k_EPSG[] = "EPSG";
  static const char __pyx_k_True[] = "True";
  static const char __pyx_k_args[] = "args";
  static const char __pyx_k_bbox[] = "bbox";
  static const char __pyx_k_bool[] = "bool";
  static const char __pyx_k_date[] = "date";
  static const char __pyx_k_dict[] = "__dict__";
  static const char __pyx_k_epsg[] = "epsg:";
  static const char __pyx_k_hour[] = "hour";
  static const char __pyx_k_info[] = "info";
  static const char __pyx_k_init[] = "init";
  static const char __pyx_k_json[] = "json";
  static const char __pyx_k_keys[] = "keys";
  static const char __pyx_k_long[] = "long";
  static const char __pyx_k_main[] = "__main__";
  static const char __pyx_k_mask[] = "mask";
  static const char __pyx_k_math[] = "math";
  static const char __pyx_k_mode[] = "mode";
  static const char __pyx_k_name[] = "name";
  static const char __pyx_k_next[] = "_next";
  static const char __pyx_k_path[] = "path";
  static const char __pyx_k_read[] = "read";
  static const char __pyx_k_send[] = "send";
  static const char __pyx_k_size[] = "size";
  static const char __pyx_k_step[] = "step";
  static const char __pyx_k_stop[] = "stop";
  static const char __pyx_k_test[] = "__test__";
  static const char __pyx_k_time[] = "time";
  static const char __pyx_k_type[] = "type";
  static const char __pyx_k_uuid[] = "uuid";
  static const char __pyx_k_warn[] = "warn";
  static const char __pyx_k_year[] = "year";
  static const char __pyx_k_False[] = "False";
  static const char __pyx_k_bytes[] = "bytes";
  static const char __pyx_k_class[] = "__class__";
  static const char __pyx_k_close[] = "close";
  static const char __pyx_k_debug[] = "debug";
  static const char __pyx_k_dumps[] = "dumps";
  static const char __pyx_k_fiona[] = "fiona";
  static const char __pyx_k_float[] = "float";
  static const char __pyx_k_index[] = "index";
  static const char __pyx_k_int32[] = "int32";
  static const char __pyx_k_int64[] = "int64";
  static const char __pyx_k_items[] = "items";
  static const char __pyx_k_layer[] = "layer";
  static const char __pyx_k_loads[] = "loads";
  static const char __pyx_k_major[] = "major";
  static const char __pyx_k_month[] = "month";
  static const char __pyx_k_range[] = "range";
  static const char __pyx_k_split[] = "split";
  static const char __pyx_k_start[] = "start";
  static const char __pyx_k_strip[] = "strip";
  static const char __pyx_k_throw[] = "throw";
  static const char __pyx_k_upper[] = "upper";
  static const char __pyx_k_utf_8[] = "utf-8";
  static const char __pyx_k_uuid4[] = "uuid4";
  static const char __pyx_k_bounds[] = "_bounds";
  static const char __pyx_k_closed[] = "closed";
  static const char __pyx_k_compat[] = "compat";
  static const char __pyx_k_coords[] = "coords";
  static const char __pyx_k_decode[] = "decode";
  static const char __pyx_k_driver[] = "driver";
  static const char __pyx_k_encode[] = "encode";
  static const char __pyx_k_exists[] = "exists";
  static const char __pyx_k_format[] = "format";
  static const char __pyx_k_import[] = "__import__";
  static const char __pyx_k_kwargs[] = "kwargs";
  static const char __pyx_k_locale[] = "locale";
  static const char __pyx_k_lstrip[] = "lstrip";
  static const char __pyx_k_minute[] = "minute";
  static const char __pyx_k_name_2[] = "__name__";
  static const char __pyx_k_name_b[] = "name_b";
  static const char __pyx_k_name_c[] = "name_c";
  static const char __pyx_k_offset[] = "offset";
  static const char __pyx_k_path_b[] = "path_b";
  static const char __pyx_k_path_c[] = "path_c";
  static const char __pyx_k_pickle[] = "pickle";
  static const char __pyx_k_record[] = "record";
  static const char __pyx_k_reduce[] = "__reduce__";
  static const char __pyx_k_remove[] = "_remove";
  static const char __pyx_k_result[] = "result";
  static const char __pyx_k_schema[] = "schema";
  static const char __pyx_k_second[] = "second";
  static const char __pyx_k_unlink[] = "unlink";
  static const char __pyx_k_update[] = "update";
  static const char __pyx_k_vsimem[] = "/vsimem/{}";
  static const char __pyx_k_whence[] = "whence";
  static const char __pyx_k_Feature[] = "Feature";
  static const char __pyx_k_GeoJSON[] = "GeoJSON";
  static const char __pyx_k_IOError[] = "IOError";
  static const char __pyx_k_Index_d[] = "Index: %d";
  static const char __pyx_k_OSError[] = "OSError";
  static const char __pyx_k_Session[] = "Session";
  static const char __pyx_k_Slice_r[] = "Slice: %r";
  static const char __pyx_k_Unknown[] = "Unknown";
  static const char __pyx_k_cogr_ds[] = "cogr_ds";
  static const char __pyx_k_crs_wkt[] = "_crs_wkt";
  static const char __pyx_k_explode[] = "_explode";
  static const char __pyx_k_feature[] = "feature";
  static const char __pyx_k_logging[] = "logging";
  static const char __pyx_k_records[] = "records";
  static const char __pyx_k_session[] = "session";
  static const char __pyx_k_warning[] = "warning";
  static const char __pyx_k_CRSError[] = "CRSError";
  static const char __pyx_k_Iterator[] = "Iterator";
  static const char __pyx_k_KeyError[] = "KeyError";
  static const char __pyx_k_Params_s[] = "Params: %s";
  static const char __pyx_k_bytesbuf[] = "bytesbuf";
  static const char __pyx_k_copysign[] = "copysign";
  static const char __pyx_k_cpl_errs[] = "cpl_errs";
  static const char __pyx_k_datetime[] = "datetime";
  static const char __pyx_k_driver_2[] = "_driver";
  static const char __pyx_k_driver_c[] = "driver_c";
  static const char __pyx_k_encoding[] = "encoding";
  static const char __pyx_k_filename[] = "filename";
  static const char __pyx_k_geometry[] = "geometry";
  static const char __pyx_k_getstate[] = "__getstate__";
  static const char __pyx_k_isactive[] = "isactive";
  static const char __pyx_k_pyx_type[] = "__pyx_type";
  static const char __pyx_k_setstate[] = "__setstate__";
  static const char __pyx_k_vsi_path[] = "vsi_path";
  static const char __pyx_k_vsimem_0[] = "/vsimem/{0}";
  static const char __pyx_k_warnings[] = "warnings";
  static const char __pyx_k_Set_field[] = "Set field ";
  static const char __pyx_k_Shapefile[] = "Shapefile";
  static const char __pyx_k_TypeError[] = "TypeError";
  static const char __pyx_k_exception[] = "exception";
  static const char __pyx_k_featureRT[] = "featureRT";
  static const char __pyx_k_getLogger[] = "getLogger";
  static const char __pyx_k_isoformat[] = "isoformat";
  static const char __pyx_k_layer_str[] = "layer_str";
  static const char __pyx_k_pyx_state[] = "__pyx_state";
  static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
  static const char __pyx_k_text_type[] = "text_type";
  static const char __pyx_k_IndexError[] = "IndexError";
  static const char __pyx_k_Null_layer[] = "Null layer";
  static const char __pyx_k_ValueError[] = "ValueError";
  static const char __pyx_k_cogr_layer[] = "cogr_layer";
  static const char __pyx_k_collection[] = "collection";
  static const char __pyx_k_fiona__env[] = "fiona._env";
  static const char __pyx_k_fiona__err[] = "fiona._err";
  static const char __pyx_k_fiona_path[] = "fiona.path";
  static const char __pyx_k_get_driver[] = "get_driver";
  static const char __pyx_k_get_schema[] = "get_schema";
  static const char __pyx_k_iso_8859_1[] = "iso-8859-1";
  static const char __pyx_k_listlayers[] = "_listlayers";
  static const char __pyx_k_namedtuple[] = "namedtuple";
  static const char __pyx_k_parse_date[] = "parse_date";
  static const char __pyx_k_parse_time[] = "parse_time";
  static const char __pyx_k_properties[] = "properties";
  static const char __pyx_k_pyx_result[] = "__pyx_result";
  static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
  static const char __pyx_k_startswith[] = "startswith";
  static const char __pyx_k_vsi_handle[] = "vsi_handle";
  static const char __pyx_k_vsimem_0_1[] = "/vsimem/{0}.{1}";
  static const char __pyx_k_DriverError[] = "DriverError";
  static const char __pyx_k_FIELD_TYPES[] = "FIELD_TYPES";
  static const char __pyx_k_GDALVersion[] = "GDALVersion";
  static const char __pyx_k_Null_driver[] = "Null driver";
  static const char __pyx_k_OGRERR_NONE[] = "OGRERR_NONE";
  static const char __pyx_k_OrderedDict[] = "OrderedDict";
  static const char __pyx_k_PickleError[] = "PickleError";
  static const char __pyx_k_SchemaError[] = "SchemaError";
  static const char __pyx_k_cogr_driver[] = "cogr_driver";
  static const char __pyx_k_collections[] = "collections";
  static const char __pyx_k_coordinates[] = "coordinates";
  static const char __pyx_k_get_feature[] = "get_feature";
  static const char __pyx_k_layer_count[] = "layer_count";
  static const char __pyx_k_layer_index[] = "layer_index";
  static const char __pyx_k_layer_names[] = "layer_names";
  static const char __pyx_k_KeysIterator[] = "KeysIterator";
  static const char __pyx_k_Null_feature[] = "Null feature";
  static const char __pyx_k_Null_layer_2[] = "Null layer: ";
  static const char __pyx_k_RuntimeError[] = "RuntimeError";
  static const char __pyx_k_cogr_feature[] = "cogr_feature";
  static const char __pyx_k_fiona_compat[] = "fiona.compat";
  static const char __pyx_k_fiona_errors[] = "fiona.errors";
  static const char __pyx_k_fiona_ogrext[] = "fiona.ogrext";
  static const char __pyx_k_fiona_schema[] = "fiona.schema";
  static const char __pyx_k_gdal_version[] = "gdal_version";
  static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
  static const char __pyx_k_remove_layer[] = "_remove_layer";
  static const char __pyx_k_string_types[] = "string_types";
  static const char __pyx_k_stringsource[] = "stringsource";
  static const char __pyx_k_vsi_filename[] = "vsi_filename";
  static const char __pyx_k_DriverIOError[] = "DriverIOError";
  static const char __pyx_k_FionaDateType[] = "FionaDateType";
  static const char __pyx_k_FionaTimeType[] = "FionaTimeType";
  static const char __pyx_k_ItemsIterator[] = "ItemsIterator";
  static const char __pyx_k_Null_geometry[] = "Null geometry";
  static const char __pyx_k_StopIteration[] = "StopIteration";
  static const char __pyx_k_cogr_geometry[] = "cogr_geometry";
  static const char __pyx_k_file_or_bytes[] = "file_or_bytes";
  static const char __pyx_k_fiona_rfc3339[] = "fiona.rfc3339";
  static const char __pyx_k_ignore_fields[] = "ignore_fields";
  static const char __pyx_k_initial_bytes[] = "_initial_bytes";
  static const char __pyx_k_integer_types[] = "integer_types";
  static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
  static const char __pyx_k_vsi_cfilename[] = "vsi_cfilename";
  static const char __pyx_k_AttributeError[] = "AttributeError";
  static const char __pyx_k_CPLE_BaseError[] = "CPLE_BaseError";
  static const char __pyx_k_FeatureBuilder[] = "FeatureBuilder";
  static const char __pyx_k_GEOMETRY_TYPES[] = "GEOMETRY_TYPES";
  static const char __pyx_k_MemoryFileBase[] = "MemoryFileBase";
  static const char __pyx_k_None_fieldtype[] = ": None, fieldtype: ";
  static const char __pyx_k_OGRERR_FAILURE[] = "OGRERR_FAILURE";
  static const char __pyx_k_RuntimeWarning[] = "RuntimeWarning";
  static const char __pyx_k_Set_option_r_r[] = "Set option %r: %r";
  static const char __pyx_k_WritingSession[] = "WritingSession";
  static const char __pyx_k_parse_datetime[] = "parse_datetime";
  static const char __pyx_k_Created_layer_s[] = "Created layer %s";
  static const char __pyx_k_FIELD_TYPES_MAP[] = "FIELD_TYPES_MAP";
  static const char __pyx_k_FionaValueError[] = "FionaValueError";
  static const char __pyx_k_Null_projection[] = "Null projection";
  static const char __pyx_k_Writing_started[] = "Writing started";
  static const char __pyx_k_enabled_drivers[] = "enabled_drivers";
  static const char __pyx_k_fiona__geometry[] = "fiona._geometry";
  static const char __pyx_k_ignore_geometry[] = "ignore_geometry";
  static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
  static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
  static const char __pyx_k_Null_data_source[] = "Null data source";
  static const char __pyx_k_TransactionError[] = "TransactionError";
  static const char __pyx_k_fiona_ogrext_pyx[] = "fiona/ogrext.pyx";
  static const char __pyx_k_valid_geom_types[] = "_valid_geom_types";
  static const char __pyx_k_FionaDateTimeType[] = "FionaDateTimeType";
  static const char __pyx_k_Looking_up_s_in_s[] = "Looking up %s in %s";
  static const char __pyx_k_OGRFeatureBuilder[] = "OGRFeatureBuilder";
  static const char __pyx_k_DatasetDeleteError[] = "DatasetDeleteError";
  static const char __pyx_k_UnicodeDecodeError[] = "UnicodeDecodeError";
  static const char __pyx_k_UnicodeEncodeError[] = "UnicodeEncodeError";
  static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
  static const char __pyx_k_OGRERR_CORRUPT_DATA[] = "OGRERR_CORRUPT_DATA";
  static const char __pyx_k_remove_virtual_file[] = "remove_virtual_file";
  static const char __pyx_k_CPLE_OpenFailedError[] = "CPLE_OpenFailedError";
  static const char __pyx_k_End_creating_field_r[] = "End creating field %r";
  static const char __pyx_k_Invalid_field_type_s[] = "Invalid field type %s";
  static const char __pyx_k_get_gdal_version_num[] = "get_gdal_version_num";
  static const char __pyx_k_getpreferredencoding[] = "getpreferredencoding";
  static const char __pyx_k_normalize_field_type[] = "normalize_field_type";
  static const char __pyx_k_FionaNullPointerError[] = "FionaNullPointerError";
  static const char __pyx_k_Got_coordinate_system[] = "Got coordinate system";
  static const char __pyx_k_Null_field_definition[] = "Null field definition";
  static const char __pyx_k_OGRERR_INVALID_HANDLE[] = "OGRERR_INVALID_HANDLE";
  static const char __pyx_k_get_fallback_encoding[] = "_get_fallback_encoding";
  static const char __pyx_k_get_internal_encoding[] = "_get_internal_encoding";
  static const char __pyx_k_OGRERR_NOT_ENOUGH_DATA[] = "OGRERR_NOT_ENOUGH_DATA";
  static const char __pyx_k_OGRERR_UNSUPPORTED_SRS[] = "OGRERR_UNSUPPORTED_SRS";
  static const char __pyx_k_buffer_to_virtual_file[] = "buffer_to_virtual_file";
  static const char __pyx_k_negative_seek_position[] = "negative seek position: {}";
  static const char __pyx_k_validate_geometry_type[] = "validate_geometry_type";
  static const char __pyx_k_FionaDeprecationWarning[] = "FionaDeprecationWarning";
  static const char __pyx_k_Null_feature_definition[] = "Null feature definition";
  static const char __pyx_k_DEFAULT_TRANSACTION_SIZE[] = "DEFAULT_TRANSACTION_SIZE";
  static const char __pyx_k_Failed_to_write_record_s[] = "Failed to write record: %s";
  static const char __pyx_k_Invalid_field_name_ref_s[] = "Invalid field name ref: %s";
  static const char __pyx_k_OGRERR_NOT_ENOUGH_MEMORY[] = "OGRERR_NOT_ENOUGH_MEMORY";
  static const char __pyx_k_Flushed_data_source_cache[] = "Flushed data source cache";
  static const char __pyx_k_Null_field_name_reference[] = "Null field name reference";
  static const char __pyx_k_slice_step_cannot_be_zero[] = "slice step cannot be zero";
  static const char __pyx_k_By_request_ignoring_field_r[] = "By request, ignoring field %r";
  static const char __pyx_k_Comitting_transaction_final[] = "Comitting transaction (final)";
  static const char __pyx_k_Creating_feature_in_layer_s[] = "Creating feature in layer: %s";
  static const char __pyx_k_Failed_to_start_transaction[] = "Failed to start transaction";
  static const char __pyx_k_GeometryTypeValidationError[] = "GeometryTypeValidationError";
  static const char __pyx_k_No_such_file_or_directory_s[] = "No such file or directory %s";
  static const char __pyx_k_Unexpected_proj_parameter_s[] = "Unexpected proj parameter %s";
  static const char __pyx_k_pyx_unpickle_FeatureBuilder[] = "__pyx_unpickle_FeatureBuilder";
  static const char __pyx_k_pyx_unpickle_MemoryFileBase[] = "__pyx_unpickle_MemoryFileBase";
  static const char __pyx_k_Failed_to_commit_transaction[] = "Failed to commit transaction";
  static const char __pyx_k_Failed_to_remove_data_source[] = "Failed to remove data source {}";
  static const char __pyx_k_OGRERR_UNSUPPORTED_OPERATION[] = "OGRERR_UNSUPPORTED_OPERATION";
  static const char __pyx_k_Starting_transaction_initial[] = "Starting transaction (initial)";
  static const char __pyx_k_failed_to_map_buffer_to_file[] = "failed to map buffer to file";
  static const char __pyx_k_Begin_creating_field_r_value_r[] = "Begin creating field: %r value: %r";
  static const char __pyx_k_Failed_to_auto_identify_EPSG_d[] = "Failed to auto identify EPSG: %d";
  static const char __pyx_k_Layer_has_no_coordinate_system[] = "Layer has no coordinate system";
  static const char __pyx_k_There_is_no_feature_with_fid_r[] = "There is no feature with fid {!r}";
  static const char __pyx_k_pyx_unpickle_OGRFeatureBuilder[] = "__pyx_unpickle_OGRFeatureBuilder";
  static const char __pyx_k_seek_position_past_end_of_file[] = "seek position past end of file: {}";
  static const char __pyx_k_Deleted_pre_existing_layer_at_s[] = "Deleted pre-existing layer at %s";
  static const char __pyx_k_Driver_does_not_support_dataset[] = "Driver does not support dataset removal operation";
  static const char __pyx_k_Failed_to_create_in_memory_file[] = "Failed to create in-memory file using initial bytes.";
  static const char __pyx_k_Failed_to_seek_to_offset_s_in_s[] = "Failed to seek to offset %s in %s.";
  static const char __pyx_k_Ignored_field_has_type_expected[] = "Ignored field \"{}\" has type \"{}\", expected string";
  static const char __pyx_k_Normalizing_schema_type_for_key[] = "Normalizing schema type for key %r in schema %r to %r";
  static const char __pyx_k_Record_s_geometry_type_does_not[] = "Record's geometry type does not match collection schema's geometry type: ";
  static const char __pyx_k_Setting_field_r_type_r_to_value[] = "Setting field %r, type %r, to value %r";
  static const char __pyx_k_Skipping_field_s_invalid_type_s[] = "Skipping field %s: invalid type %s";
  static const char __pyx_k_A_support_file_wasn_t_opened_See[] = "A support file wasn't opened. See the preceding ERROR level message.";
  static const char __pyx_k_Called_OGR_L_GetSpatialRef_again[] = "Called OGR_L_GetSpatialRef() again without error checking.";
  static const char __pyx_k_Collection_slicing_is_deprecated[] = "Collection slicing is deprecated and will be disabled in a future version.";
  static const char __pyx_k_Comitting_transaction_intermedia[] = "Comitting transaction (intermediate)";
  static const char __pyx_k_Constructor_argument_must_be_a_f[] = "Constructor argument must be a file opened in binary mode or bytes/bytearray.";
  static const char __pyx_k_Failed_to_decode_s_using_s_codec[] = "Failed to decode %s using %s codec";
  static const char __pyx_k_Failed_to_encode_s_using_s_codec[] = "Failed to encode %s using %s codec";
  static const char __pyx_k_Failed_to_properly_close_in_memo[] = "Failed to properly close in-memory file.";
  static const char __pyx_k_Failed_to_remove_layer_from_data[] = "Failed to remove layer {} from datasource: {}";
  static const char __pyx_k_Field_format_parsing_value_r_fmt[] = "Field format parsing, value: %r, fmt: %r";
  static const char __pyx_k_File_doesn_t_exist_Creating_a_ne[] = "File doesn't exist. Creating a new one...";
  static const char __pyx_k_I_O_operation_on_closed_collecti[] = "I/O operation on closed collection";
  static const char __pyx_k_Incompatible_checksums_s_vs_0xd4[] = "Incompatible checksums (%s vs 0xd41d8cd = ())";
  static const char __pyx_k_Layer_does_not_exist_in_datasour[] = "Layer \"{}\" does not exist in datasource: {}";
  static const char __pyx_k_Layer_does_not_supportOLCFastSet[] = "Layer does not supportOLCFastSetNextByIndex, negative step size may be slow";
  static const char __pyx_k_Null_driver_when_attempting_to_d[] = "Null driver when attempting to delete {}";
  static const char __pyx_k_OGRERR_UNSUPPORTED_GEOMETRY_TYPE[] = "OGRERR_UNSUPPORTED_GEOMETRY_TYPE";
  static const char __pyx_k_Projection_not_found_cogr_crs_wa[] = "Projection not found (cogr_crs was NULL)";
  static const char __pyx_k_Record_does_not_match_collection[] = "Record does not match collection schema: ";
  static const char __pyx_k_Removal_of_layer_not_supported_b[] = "Removal of layer {} not supported by driver";
  static const char __pyx_k_Session_is_inactive_dataset_is_c[] = "Session is inactive, dataset is closed or layer is unavailable.";
  static const char __pyx_k_Starting_transaction_intermediat[] = "Starting transaction (intermediate)";
  static const char __pyx_k_collection_s_dataset_does_not_su[] = "collection's dataset does not support negative indexes";
  static const char __pyx_k_failed_to_close_mapped_file_hand[] = "failed to close mapped file handle";
  static const char __pyx_k_get_fileencoding_is_deprecated_a[] = "get_fileencoding is deprecated and will be removed in a future version.";
  static const char __pyx_k_mask_and_bbox_can_not_be_set_tog[] = "mask and bbox can not be set together";
  static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
  static const char __pyx_k_self_cogr_ds_self_cogr_layer_can[] = "self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling";
  static const char __pyx_k_writerecs_locals_validate_geomet[] = "writerecs.<locals>.validate_geometry_type";
  static const char __pyx_k_collection_s_dataset_does_not_su_2[] = "collection's dataset does not support negative slice indexes";
  static PyObject *__pyx_kp_u_3D;
  static PyObject *__pyx_kp_u_A_support_file_wasn_t_opened_See;
  static PyObject *__pyx_n_u_Any;
  static PyObject *__pyx_n_s_AttributeError;
  static PyObject *__pyx_kp_u_Begin_creating_field_r_value_r;
  static PyObject *__pyx_kp_u_By_request_ignoring_field_r;
  static PyObject *__pyx_n_s_CPLE_BaseError;
  static PyObject *__pyx_n_s_CPLE_OpenFailedError;
  static PyObject *__pyx_n_s_CRSError;
  static PyObject *__pyx_kp_u_Called_OGR_L_GetSpatialRef_again;
  static PyObject *__pyx_kp_u_Collection_slicing_is_deprecated;
  static PyObject *__pyx_kp_u_Comitting_transaction_final;
  static PyObject *__pyx_kp_u_Comitting_transaction_intermedia;
  static PyObject *__pyx_kp_u_Constructor_argument_must_be_a_f;
  static PyObject *__pyx_kp_u_Created_layer_s;
  static PyObject *__pyx_kp_u_Creating_feature_in_layer_s;
  static PyObject *__pyx_n_s_DEFAULT_TRANSACTION_SIZE;
  static PyObject *__pyx_n_s_DatasetDeleteError;
  static PyObject *__pyx_kp_u_Deleted_pre_existing_layer_at_s;
  static PyObject *__pyx_n_s_DriverError;
  static PyObject *__pyx_n_s_DriverIOError;
  static PyObject *__pyx_kp_u_Driver_does_not_support_dataset;
  static PyObject *__pyx_n_u_EPSG;
  static PyObject *__pyx_kp_u_End_creating_field_r;
  static PyObject *__pyx_n_s_FIELD_TYPES;
  static PyObject *__pyx_n_s_FIELD_TYPES_MAP;
  static PyObject *__pyx_kp_u_Failed_to_auto_identify_EPSG_d;
  static PyObject *__pyx_kp_u_Failed_to_commit_transaction;
  static PyObject *__pyx_kp_u_Failed_to_create_in_memory_file;
  static PyObject *__pyx_kp_u_Failed_to_decode_s_using_s_codec;
  static PyObject *__pyx_kp_u_Failed_to_encode_s_using_s_codec;
  static PyObject *__pyx_kp_u_Failed_to_properly_close_in_memo;
  static PyObject *__pyx_kp_u_Failed_to_remove_data_source;
  static PyObject *__pyx_kp_u_Failed_to_remove_layer_from_data;
  static PyObject *__pyx_kp_u_Failed_to_seek_to_offset_s_in_s;
  static PyObject *__pyx_kp_u_Failed_to_start_transaction;
  static PyObject *__pyx_kp_u_Failed_to_write_record_s;
  static PyObject *__pyx_n_u_False;
  static PyObject *__pyx_n_u_Feature;
  static PyObject *__pyx_n_s_FeatureBuilder;
  static PyObject *__pyx_kp_u_Field_format_parsing_value_r_fmt;
  static PyObject *__pyx_kp_u_File_doesn_t_exist_Creating_a_ne;
  static PyObject *__pyx_n_s_FionaDateTimeType;
  static PyObject *__pyx_n_s_FionaDateType;
  static PyObject *__pyx_n_s_FionaDeprecationWarning;
  static PyObject *__pyx_n_s_FionaNullPointerError;
  static PyObject *__pyx_n_s_FionaTimeType;
  static PyObject *__pyx_n_s_FionaValueError;
  static PyObject *__pyx_kp_u_Flushed_data_source_cache;
  static PyObject *__pyx_n_s_GDALVersion;
  static PyObject *__pyx_n_s_GEOMETRY_TYPES;
  static PyObject *__pyx_n_u_GeoJSON;
  static PyObject *__pyx_n_s_GeometryTypeValidationError;
  static PyObject *__pyx_kp_u_Got_coordinate_system;
  static PyObject *__pyx_n_s_IOError;
  static PyObject *__pyx_kp_u_I_O_operation_on_closed_collecti;
  static PyObject *__pyx_kp_u_Ignored_field_has_type_expected;
  static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xd4;
  static PyObject *__pyx_n_s_IndexError;
  static PyObject *__pyx_kp_u_Index_d;
  static PyObject *__pyx_kp_u_Invalid_field_name_ref_s;
  static PyObject *__pyx_kp_u_Invalid_field_type_s;
  static PyObject *__pyx_n_s_ItemsIterator;
  static PyObject *__pyx_n_s_Iterator;
  static PyObject *__pyx_n_s_KeyError;
  static PyObject *__pyx_n_s_KeysIterator;
  static PyObject *__pyx_kp_u_Layer_does_not_exist_in_datasour;
  static PyObject *__pyx_kp_u_Layer_does_not_supportOLCFastSet;
  static PyObject *__pyx_kp_u_Layer_has_no_coordinate_system;
  static PyObject *__pyx_kp_u_Looking_up_s_in_s;
  static PyObject *__pyx_n_s_MemoryFileBase;
  static PyObject *__pyx_kp_u_No_such_file_or_directory_s;
  static PyObject *__pyx_kp_u_None_fieldtype;
  static PyObject *__pyx_kp_u_Normalizing_schema_type_for_key;
  static PyObject *__pyx_kp_u_Null_data_source;
  static PyObject *__pyx_kp_u_Null_driver;
  static PyObject *__pyx_kp_u_Null_driver_when_attempting_to_d;
  static PyObject *__pyx_kp_u_Null_feature;
  static PyObject *__pyx_kp_u_Null_feature_definition;
  static PyObject *__pyx_kp_u_Null_field_definition;
  static PyObject *__pyx_kp_u_Null_field_name_reference;
  static PyObject *__pyx_kp_u_Null_geometry;
  static PyObject *__pyx_kp_u_Null_layer;
  static PyObject *__pyx_kp_u_Null_layer_2;
  static PyObject *__pyx_kp_u_Null_projection;
  static PyObject *__pyx_n_u_OFF;
  static PyObject *__pyx_n_s_OGRERR_CORRUPT_DATA;
  static PyObject *__pyx_n_s_OGRERR_FAILURE;
  static PyObject *__pyx_n_s_OGRERR_INVALID_HANDLE;
  static PyObject *__pyx_n_s_OGRERR_NONE;
  static PyObject *__pyx_n_s_OGRERR_NOT_ENOUGH_DATA;
  static PyObject *__pyx_n_s_OGRERR_NOT_ENOUGH_MEMORY;
  static PyObject *__pyx_n_s_OGRERR_UNSUPPORTED_GEOMETRY_TYPE;
  static PyObject *__pyx_n_s_OGRERR_UNSUPPORTED_OPERATION;
  static PyObject *__pyx_n_s_OGRERR_UNSUPPORTED_SRS;
  static PyObject *__pyx_n_s_OGRFeatureBuilder;
  static PyObject *__pyx_n_u_ON;
  static PyObject *__pyx_n_s_OSError;
  static PyObject *__pyx_n_s_OrderedDict;
  static PyObject *__pyx_kp_u_Params_s;
  static PyObject *__pyx_n_s_PickleError;
  static PyObject *__pyx_kp_u_Projection_not_found_cogr_crs_wa;
  static PyObject *__pyx_kp_u_Record_does_not_match_collection;
  static PyObject *__pyx_kp_u_Record_s_geometry_type_does_not;
  static PyObject *__pyx_kp_u_Removal_of_layer_not_supported_b;
  static PyObject *__pyx_n_s_RuntimeError;
  static PyObject *__pyx_n_s_RuntimeWarning;
  static PyObject *__pyx_n_s_SchemaError;
  static PyObject *__pyx_n_s_Session;
  static PyObject *__pyx_kp_u_Session_is_inactive_dataset_is_c;
  static PyObject *__pyx_kp_u_Set_field;
  static PyObject *__pyx_kp_u_Set_option_r_r;
  static PyObject *__pyx_kp_u_Setting_field_r_type_r_to_value;
  static PyObject *__pyx_n_u_Shapefile;
  static PyObject *__pyx_kp_u_Skipping_field_s_invalid_type_s;
  static PyObject *__pyx_kp_u_Slice_r;
  static PyObject *__pyx_kp_u_Starting_transaction_initial;
  static PyObject *__pyx_kp_u_Starting_transaction_intermediat;
  static PyObject *__pyx_n_s_StopIteration;
  static PyObject *__pyx_kp_u_There_is_no_feature_with_fid_r;
  static PyObject *__pyx_n_s_TransactionError;
  static PyObject *__pyx_n_u_True;
  static PyObject *__pyx_n_s_TypeError;
  static PyObject *__pyx_kp_u_Unexpected_proj_parameter_s;
  static PyObject *__pyx_n_s_UnicodeDecodeError;
  static PyObject *__pyx_n_s_UnicodeEncodeError;
  static PyObject *__pyx_n_u_Unknown;
  static PyObject *__pyx_n_s_ValueError;
  static PyObject *__pyx_n_s_WritingSession;
  static PyObject *__pyx_kp_u_Writing_started;
  static PyObject *__pyx_kp_b__12;
  static PyObject *__pyx_kp_u__12;
  static PyObject *__pyx_kp_u__13;
  static PyObject *__pyx_kp_u__15;
  static PyObject *__pyx_kp_u__16;
  static PyObject *__pyx_kp_u__21;
  static PyObject *__pyx_kp_u__22;
  static PyObject *__pyx_kp_u__25;
  static PyObject *__pyx_kp_u__4;
  static PyObject *__pyx_kp_u__40;
  static PyObject *__pyx_kp_u__5;
  static PyObject *__pyx_kp_u__8;
  static PyObject *__pyx_n_u_a;
  static PyObject *__pyx_n_s_args;
  static PyObject *__pyx_n_s_bbox;
  static PyObject *__pyx_n_u_bool;
  static PyObject *__pyx_n_s_bounds;
  static PyObject *__pyx_n_s_buffer_to_virtual_file;
  static PyObject *__pyx_n_u_bytes;
  static PyObject *__pyx_n_s_bytesbuf;
  static PyObject *__pyx_n_s_class;
  static PyObject *__pyx_n_s_cline_in_traceback;
  static PyObject *__pyx_n_s_close;
  static PyObject *__pyx_n_s_closed;
  static PyObject *__pyx_n_s_cogr_driver;
  static PyObject *__pyx_n_s_cogr_ds;
  static PyObject *__pyx_n_s_cogr_feature;
  static PyObject *__pyx_n_s_cogr_geometry;
  static PyObject *__pyx_n_s_cogr_layer;
  static PyObject *__pyx_n_s_collection;
  static PyObject *__pyx_kp_u_collection_s_dataset_does_not_su;
  static PyObject *__pyx_kp_u_collection_s_dataset_does_not_su_2;
  static PyObject *__pyx_n_s_collections;
  static PyObject *__pyx_n_s_compat;
  static PyObject *__pyx_n_u_coordinates;
  static PyObject *__pyx_n_s_coords;
  static PyObject *__pyx_n_s_copysign;
  static PyObject *__pyx_n_s_cpl_errs;
  static PyObject *__pyx_n_s_crs_wkt;
  static PyObject *__pyx_kp_u_d;
  static PyObject *__pyx_kp_u_d_2;
  static PyObject *__pyx_n_s_date;
  static PyObject *__pyx_n_u_date;
  static PyObject *__pyx_n_s_datetime;
  static PyObject *__pyx_n_u_datetime;
  static PyObject *__pyx_n_s_day;
  static PyObject *__pyx_n_s_debug;
  static PyObject *__pyx_n_s_decode;
  static PyObject *__pyx_n_s_dict;
  static PyObject *__pyx_n_s_driver;
  static PyObject *__pyx_n_s_driver_2;
  static PyObject *__pyx_n_s_driver_c;
  static PyObject *__pyx_n_s_dumps;
  static PyObject *__pyx_n_s_e;
  static PyObject *__pyx_n_s_enabled_drivers;
  static PyObject *__pyx_n_s_encode;
  static PyObject *__pyx_n_s_encoding;
  static PyObject *__pyx_n_u_encoding;
  static PyObject *__pyx_kp_u_epsg;
  static PyObject *__pyx_n_s_err;
  static PyObject *__pyx_n_s_exception;
  static PyObject *__pyx_n_s_exists;
  static PyObject *__pyx_n_s_explode;
  static PyObject *__pyx_n_s_ext;
  static PyObject *__pyx_n_s_f;
  static PyObject *__pyx_kp_u_failed_to_close_mapped_file_hand;
  static PyObject *__pyx_kp_u_failed_to_map_buffer_to_file;
  static PyObject *__pyx_n_s_feature;
  static PyObject *__pyx_n_s_featureRT;
  static PyObject *__pyx_n_s_file_or_bytes;
  static PyObject *__pyx_n_s_filename;
  static PyObject *__pyx_n_s_fiona;
  static PyObject *__pyx_n_s_fiona__env;
  static PyObject *__pyx_n_s_fiona__err;
  static PyObject *__pyx_n_s_fiona__geometry;
  static PyObject *__pyx_n_s_fiona_compat;
  static PyObject *__pyx_n_s_fiona_errors;
  static PyObject *__pyx_n_s_fiona_ogrext;
  static PyObject *__pyx_kp_s_fiona_ogrext_pyx;
  static PyObject *__pyx_n_s_fiona_path;
  static PyObject *__pyx_n_s_fiona_rfc3339;
  static PyObject *__pyx_n_s_fiona_schema;
  static PyObject *__pyx_n_u_float;
  static PyObject *__pyx_n_s_format;
  static PyObject *__pyx_n_s_gdal_version;
  static PyObject *__pyx_n_s_geometry;
  static PyObject *__pyx_n_u_geometry;
  static PyObject *__pyx_n_s_get;
  static PyObject *__pyx_n_s_getLogger;
  static PyObject *__pyx_n_s_get_driver;
  static PyObject *__pyx_n_s_get_fallback_encoding;
  static PyObject *__pyx_n_s_get_feature;
  static PyObject *__pyx_kp_u_get_fileencoding_is_deprecated_a;
  static PyObject *__pyx_n_s_get_gdal_version_num;
  static PyObject *__pyx_n_s_get_internal_encoding;
  static PyObject *__pyx_n_s_get_schema;
  static PyObject *__pyx_n_s_getpreferredencoding;
  static PyObject *__pyx_n_s_getstate;
  static PyObject *__pyx_n_s_hex;
  static PyObject *__pyx_n_s_hour;
  static PyObject *__pyx_n_s_i;
  static PyObject *__pyx_n_u_id;
  static PyObject *__pyx_n_s_ignore_fields;
  static PyObject *__pyx_n_s_ignore_geometry;
  static PyObject *__pyx_n_s_import;
  static PyObject *__pyx_n_s_index;
  static PyObject *__pyx_n_s_info;
  static PyObject *__pyx_n_u_init;
  static PyObject *__pyx_n_s_initial_bytes;
  static PyObject *__pyx_n_u_int;
  static PyObject *__pyx_n_u_int32;
  static PyObject *__pyx_n_u_int64;
  static PyObject *__pyx_n_s_integer_types;
  static PyObject *__pyx_n_s_isactive;
  static PyObject *__pyx_kp_u_iso_8859_1;
  static PyObject *__pyx_n_s_isoformat;
  static PyObject *__pyx_n_s_items;
  static PyObject *__pyx_n_s_json;
  static PyObject *__pyx_n_s_keys;
  static PyObject *__pyx_n_s_kwargs;
  static PyObject *__pyx_n_s_layer;
  static PyObject *__pyx_n_s_layer_count;
  static PyObject *__pyx_n_s_layer_index;
  static PyObject *__pyx_n_s_layer_names;
  static PyObject *__pyx_n_s_layer_str;
  static PyObject *__pyx_n_s_len;
  static PyObject *__pyx_n_s_listlayers;
  static PyObject *__pyx_n_s_loads;
  static PyObject *__pyx_n_s_locale;
  static PyObject *__pyx_n_s_log;
  static PyObject *__pyx_n_s_logging;
  static PyObject *__pyx_n_u_long;
  static PyObject *__pyx_n_s_lstrip;
  static PyObject *__pyx_n_s_main;
  static PyObject *__pyx_n_s_major;
  static PyObject *__pyx_n_s_map;
  static PyObject *__pyx_n_s_mask;
  static PyObject *__pyx_kp_u_mask_and_bbox_can_not_be_set_tog;
  static PyObject *__pyx_n_s_math;
  static PyObject *__pyx_n_s_max;
  static PyObject *__pyx_n_s_min;
  static PyObject *__pyx_n_s_minute;
  static PyObject *__pyx_n_s_mode;
  static PyObject *__pyx_n_s_month;
  static PyObject *__pyx_n_s_name;
  static PyObject *__pyx_n_s_name_2;
  static PyObject *__pyx_n_s_name_b;
  static PyObject *__pyx_n_s_name_c;
  static PyObject *__pyx_n_s_namedtuple;
  static PyObject *__pyx_kp_u_negative_seek_position;
  static PyObject *__pyx_n_s_new;
  static PyObject *__pyx_n_s_next;
  static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
  static PyObject *__pyx_n_s_normalize_field_type;
  static PyObject *__pyx_n_s_offset;
  static PyObject *__pyx_n_s_os;
  static PyObject *__pyx_n_s_parse_date;
  static PyObject *__pyx_n_s_parse_datetime;
  static PyObject *__pyx_n_s_parse_time;
  static PyObject *__pyx_n_s_path;
  static PyObject *__pyx_n_s_path_b;
  static PyObject *__pyx_n_s_path_c;
  static PyObject *__pyx_n_s_pickle;
  static PyObject *__pyx_n_s_pop;
  static PyObject *__pyx_n_s_pos;
  static PyObject *__pyx_n_u_properties;
  static PyObject *__pyx_n_s_pyx_PickleError;
  static PyObject *__pyx_n_s_pyx_checksum;
  static PyObject *__pyx_n_s_pyx_result;
  static PyObject *__pyx_n_s_pyx_state;
  static PyObject *__pyx_n_s_pyx_type;
  static PyObject *__pyx_n_s_pyx_unpickle_FeatureBuilder;
  static PyObject *__pyx_n_s_pyx_unpickle_MemoryFileBase;
  static PyObject *__pyx_n_s_pyx_unpickle_OGRFeatureBuilder;
  static PyObject *__pyx_n_s_pyx_vtable;
  static PyObject *__pyx_n_s_range;
  static PyObject *__pyx_n_s_read;
  static PyObject *__pyx_n_u_read;
  static PyObject *__pyx_n_s_record;
  static PyObject *__pyx_n_s_records;
  static PyObject *__pyx_n_s_reduce;
  static PyObject *__pyx_n_s_reduce_cython;
  static PyObject *__pyx_n_s_reduce_ex;
  static PyObject *__pyx_n_s_remove;
  static PyObject *__pyx_n_s_remove_layer;
  static PyObject *__pyx_n_s_remove_virtual_file;
  static PyObject *__pyx_n_s_result;
  static PyObject *__pyx_n_s_schema;
  static PyObject *__pyx_n_s_second;
  static PyObject *__pyx_kp_u_seek_position_past_end_of_file;
  static PyObject *__pyx_kp_s_self_cogr_ds_self_cogr_layer_can;
  static PyObject *__pyx_n_s_send;
  static PyObject *__pyx_n_s_session;
  static PyObject *__pyx_n_s_setstate;
  static PyObject *__pyx_n_s_setstate_cython;
  static PyObject *__pyx_n_s_six;
  static PyObject *__pyx_n_s_size;
  static PyObject *__pyx_kp_u_slice_step_cannot_be_zero;
  static PyObject *__pyx_n_s_split;
  static PyObject *__pyx_n_s_start;
  static PyObject *__pyx_n_s_startswith;
  static PyObject *__pyx_n_s_step;
  static PyObject *__pyx_n_s_stop;
  static PyObject *__pyx_n_u_str;
  static PyObject *__pyx_n_s_string_types;
  static PyObject *__pyx_kp_s_stringsource;
  static PyObject *__pyx_n_s_strip;
  static PyObject *__pyx_n_s_test;
  static PyObject *__pyx_n_s_text_type;
  static PyObject *__pyx_n_s_throw;
  static PyObject *__pyx_n_s_time;
  static PyObject *__pyx_n_u_time;
  static PyObject *__pyx_n_u_type;
  static PyObject *__pyx_n_s_unlink;
  static PyObject *__pyx_n_s_update;
  static PyObject *__pyx_n_s_upper;
  static PyObject *__pyx_kp_u_utf_8;
  static PyObject *__pyx_n_s_uuid;
  static PyObject *__pyx_n_s_uuid4;
  static PyObject *__pyx_n_s_valid_geom_types;
  static PyObject *__pyx_n_s_validate_geometry_type;
  static PyObject *__pyx_n_s_vsi_cfilename;
  static PyObject *__pyx_n_s_vsi_filename;
  static PyObject *__pyx_n_s_vsi_handle;
  static PyObject *__pyx_n_s_vsi_path;
  static PyObject *__pyx_kp_u_vsimem;
  static PyObject *__pyx_kp_u_vsimem_0;
  static PyObject *__pyx_kp_u_vsimem_0_1;
  static PyObject *__pyx_n_u_w;
  static PyObject *__pyx_n_s_warn;
  static PyObject *__pyx_n_s_warning;
  static PyObject *__pyx_n_s_warnings;
  static PyObject *__pyx_n_s_whence;
  static PyObject *__pyx_n_s_writerecs_locals_validate_geomet;
  static PyObject *__pyx_n_s_xyz;
  static PyObject *__pyx_n_s_year;
  static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_5fiona_6ogrext__explode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_3_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_geometry); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14FeatureBuilder___reduce_cython__(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14FeatureBuilder_2__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder___reduce_cython__(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder_2__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_5featureRT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_feature, PyObject *__pyx_v_collection); /* proto */
static int __pyx_pf_5fiona_6ogrext_7Session___init__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static void __pyx_pf_5fiona_6ogrext_7Session_2__dealloc__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_4start(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_6stop(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_8get_fileencoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_10_get_fallback_encoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_12_get_internal_encoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_14get_length(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_16get_driver(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_18get_schema(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_20get_crs(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_22get_crs_wkt(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_24get_extent(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_26has_feature(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_fid); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_28get_feature(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_fid); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_30__getitem__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_32isactive(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_34__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7Session_36__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_start(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_9writerecs_validate_geometry_type(PyObject *__pyx_self, PyObject *__pyx_v_record); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_2writerecs(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, PyObject *__pyx_v_records, PyObject *__pyx_v_collection); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_4sync(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_collection); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_5fiona_6ogrext_8Iterator___cinit__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_start, PyObject *__pyx_v_stop, PyObject *__pyx_v_step, PyObject *__pyx_v_bbox, PyObject *__pyx_v_mask); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_2__iter__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_4_next(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_6__next__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator___next__(struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator___next__(struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_7_remove(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_driver); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_9_remove_layer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_layer, CYTHON_UNUSED PyObject *__pyx_v_driver); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_11_listlayers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_13buffer_to_virtual_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytesbuf, PyObject *__pyx_v_ext); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_15remove_virtual_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_vsi_filename); /* proto */
static int __pyx_pf_5fiona_6ogrext_14MemoryFileBase___init__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_file_or_bytes, PyObject *__pyx_v_filename, PyObject *__pyx_v_ext); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_2exists(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_pf_5fiona_6ogrext_14MemoryFileBase_4__len__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_6close(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_8read(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_size); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_10seek(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_12tell(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_14write(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_16__reduce_cython__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_18__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_17__pyx_unpickle_FeatureBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_19__pyx_unpickle_OGRFeatureBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_5fiona_6ogrext_21__pyx_unpickle_MemoryFileBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_5fiona_6ogrext_FeatureBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_OGRFeatureBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_Session(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_WritingSession(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_Iterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_ItemsIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_KeysIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext_MemoryFileBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct___explode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct_1_writerecs(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_get = {0, &__pyx_n_s_get, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop = {0, &__pyx_n_s_pop, 0, 0, 0};
static __Pyx_CachedCFunction __pyx_umethod_PyList_Type_index = {0, &__pyx_n_s_index, 0, 0, 0};
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_20000;
static PyObject *__pyx_int_222419149;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_codeobj_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__42;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__44;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__46;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
/* Late includes */
static PyObject *__pyx_gb_5fiona_6ogrext_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "fiona/ogrext.pyx":95
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_1_explode(PyObject *__pyx_self, PyObject *__pyx_v_coords); /*proto*/
static char __pyx_doc_5fiona_6ogrext__explode[] = "Explode a GeoJSON geometry's coordinates object and yield\n    coordinate tuples. As long as the input is conforming, the type of\n    the geometry doesn't matter.";
static PyMethodDef __pyx_mdef_5fiona_6ogrext_1_explode = {"_explode", (PyCFunction)__pyx_pw_5fiona_6ogrext_1_explode, METH_O, __pyx_doc_5fiona_6ogrext__explode};
static PyObject *__pyx_pw_5fiona_6ogrext_1_explode(PyObject *__pyx_self, PyObject *__pyx_v_coords) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explode (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext__explode(__pyx_self, ((PyObject *)__pyx_v_coords));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext__explode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords) {
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explode", 0);
  __pyx_cur_scope = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct___explode(__pyx_ptype_5fiona_6ogrext___pyx_scope_struct___explode, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 95, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_coords = __pyx_v_coords;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_coords);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_coords);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_5fiona_6ogrext_2generator, __pyx_codeobj_, (PyObject *) __pyx_cur_scope, __pyx_n_s_explode, __pyx_n_s_explode, __pyx_n_s_fiona_ogrext); if (unlikely(!gen)) __PYX_ERR(0, 95, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("fiona.ogrext._explode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_5fiona_6ogrext_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *__pyx_cur_scope = ((struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explode", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L9_resume_from_yield;
    case 2: goto __pyx_L12_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 95, __pyx_L1_error)

  /* "fiona/ogrext.pyx":99
 *     coordinate tuples. As long as the input is conforming, the type of
 *     the geometry doesn't matter."""
 *     for e in coords:             # <<<<<<<<<<<<<<
 *         if isinstance(e, (float, int)):
 *             yield coords
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_coords)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_coords)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_coords; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 99, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 99, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 99, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 99, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":100
 *     the geometry doesn't matter."""
 *     for e in coords:
 *         if isinstance(e, (float, int)):             # <<<<<<<<<<<<<<
 *             yield coords
 *             break
 */
    __pyx_t_6 = PyFloat_Check(__pyx_cur_scope->__pyx_v_e); 
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = PyInt_Check(__pyx_cur_scope->__pyx_v_e); 
    __pyx_t_6 = (__pyx_t_7 != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "fiona/ogrext.pyx":101
 *     for e in coords:
 *         if isinstance(e, (float, int)):
 *             yield coords             # <<<<<<<<<<<<<<
 *             break
 *         else:
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_coords);
      __pyx_r = __pyx_cur_scope->__pyx_v_coords;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L9_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 101, __pyx_L1_error)

      /* "fiona/ogrext.pyx":102
 *         if isinstance(e, (float, int)):
 *             yield coords
 *             break             # <<<<<<<<<<<<<<
 *         else:
 *             for f in _explode(e):
 */
      goto __pyx_L5_break;

      /* "fiona/ogrext.pyx":100
 *     the geometry doesn't matter."""
 *     for e in coords:
 *         if isinstance(e, (float, int)):             # <<<<<<<<<<<<<<
 *             yield coords
 *             break
 */
    }

    /* "fiona/ogrext.pyx":104
 *             break
 *         else:
 *             for f in _explode(e):             # <<<<<<<<<<<<<<
 *                 yield f
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_explode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_9, __pyx_cur_scope->__pyx_v_e) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_e);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_8 = __pyx_t_4; __Pyx_INCREF(__pyx_t_8); __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 104, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_11 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 104, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 104, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 104, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_11(__pyx_t_8);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 104, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_f);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":105
 *         else:
 *             for f in _explode(e):
 *                 yield f             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
        __pyx_r = __pyx_cur_scope->__pyx_v_f;
        __Pyx_XGIVEREF(__pyx_t_1);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
        __Pyx_XGIVEREF(__pyx_t_8);
        __pyx_cur_scope->__pyx_t_3 = __pyx_t_8;
        __pyx_cur_scope->__pyx_t_4 = __pyx_t_10;
        __pyx_cur_scope->__pyx_t_5 = __pyx_t_11;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
        /* return from generator, yielding value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L12_resume_from_yield:;
        __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
        __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
        __pyx_t_8 = __pyx_cur_scope->__pyx_t_3;
        __pyx_cur_scope->__pyx_t_3 = 0;
        __Pyx_XGOTREF(__pyx_t_8);
        __pyx_t_10 = __pyx_cur_scope->__pyx_t_4;
        __pyx_t_11 = __pyx_cur_scope->__pyx_t_5;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 105, __pyx_L1_error)

        /* "fiona/ogrext.pyx":104
 *             break
 *         else:
 *             for f in _explode(e):             # <<<<<<<<<<<<<<
 *                 yield f
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }

    /* "fiona/ogrext.pyx":99
 *     coordinate tuples. As long as the input is conforming, the type of
 *     the geometry doesn't matter."""
 *     for e in coords:             # <<<<<<<<<<<<<<
 *         if isinstance(e, (float, int)):
 *             yield coords
 */
  }
  __pyx_L5_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "fiona/ogrext.pyx":95
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("_explode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":108
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry"""
 *     try:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_4_bounds(PyObject *__pyx_self, PyObject *__pyx_v_geometry); /*proto*/
static char __pyx_doc_5fiona_6ogrext_3_bounds[] = "Bounding box of a GeoJSON geometry";
static PyMethodDef __pyx_mdef_5fiona_6ogrext_4_bounds = {"_bounds", (PyCFunction)__pyx_pw_5fiona_6ogrext_4_bounds, METH_O, __pyx_doc_5fiona_6ogrext_3_bounds};
static PyObject *__pyx_pw_5fiona_6ogrext_4_bounds(PyObject *__pyx_self, PyObject *__pyx_v_geometry) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bounds (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_3_bounds(__pyx_self, ((PyObject *)__pyx_v_geometry));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_3_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_geometry) {
  PyObject *__pyx_v_xyz = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  __Pyx_RefNannySetupContext("_bounds", 0);

  /* "fiona/ogrext.pyx":110
 * def _bounds(geometry):
 *     """Bounding box of a GeoJSON geometry"""
 *     try:             # <<<<<<<<<<<<<<
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fiona/ogrext.pyx":111
 *     """Bounding box of a GeoJSON geometry"""
 *     try:
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))             # <<<<<<<<<<<<<<
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 *     except (KeyError, TypeError):
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_explode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyObject_Dict_GetItem(__pyx_v_geometry, __pyx_n_u_coordinates); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 111, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_xyz = ((PyObject*)__pyx_t_4);
      __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":112
 *     try:
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])             # <<<<<<<<<<<<<<
 *     except (KeyError, TypeError):
 *         return None
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_min, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_builtin_max, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 112, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_8);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_8 = 0;
      __pyx_r = __pyx_t_4;
      __pyx_t_4 = 0;
      goto __pyx_L7_try_return;

      /* "fiona/ogrext.pyx":110
 * def _bounds(geometry):
 *     """Bounding box of a GeoJSON geometry"""
 *     try:             # <<<<<<<<<<<<<<
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":113
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 *     except (KeyError, TypeError):             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
    __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_KeyError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
    if (__pyx_t_9) {
      __Pyx_AddTraceback("fiona.ogrext._bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_8, &__pyx_t_7) < 0) __PYX_ERR(0, 113, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);

      /* "fiona/ogrext.pyx":114
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 *     except (KeyError, TypeError):
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L6_except_return;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fiona/ogrext.pyx":110
 * def _bounds(geometry):
 *     """Bounding box of a GeoJSON geometry"""
 *     try:             # <<<<<<<<<<<<<<
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":108
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry"""
 *     try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fiona.ogrext._bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xyz);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":129
 *     """
 * 
 *     cdef build(self, void *feature, encoding='utf-8', bbox=False, driver=None, ignore_fields=None, ignore_geometry=False):             # <<<<<<<<<<<<<<
 *         """Build a Fiona feature object from an OGR feature
 * 
 */

static PyObject *__pyx_f_5fiona_6ogrext_14FeatureBuilder_build(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self, void *__pyx_v_feature, struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build *__pyx_optional_args) {
  PyObject *__pyx_v_encoding = ((PyObject *)__pyx_kp_u_utf_8);
  PyObject *__pyx_v_driver = ((PyObject *)Py_None);
  PyObject *__pyx_v_ignore_fields = ((PyObject *)Py_None);
  PyObject *__pyx_v_ignore_geometry = ((PyObject *)Py_False);
  void *__pyx_v_fdefn;
  int __pyx_v_i;
  int __pyx_v_y;
  int __pyx_v_m;
  int __pyx_v_d;
  int __pyx_v_hh;
  int __pyx_v_mm;
  int __pyx_v_ss;
  int __pyx_v_tz;
  unsigned char *__pyx_v_data;
  int __pyx_v_l;
  CYTHON_UNUSED int __pyx_v_retval;
  int __pyx_v_fieldsubtype;
  char const *__pyx_v_key_c;
  long __pyx_v_fid;
  PyObject *__pyx_v_props = NULL;
  PyObject *__pyx_v_fiona_feature = NULL;
  char const *__pyx_v_key_b;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_fieldtypename = NULL;
  PyObject *__pyx_v_fieldtype = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_err = NULL;
  void *__pyx_v_cogr_geometry;
  void *__pyx_v_org_geometry;
  unsigned int __pyx_v_code;
  PyObject *__pyx_v_geom = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  char const *__pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  PyObject *__pyx_t_27 = NULL;
  PyObject *__pyx_t_28 = NULL;
  char const *__pyx_t_29;
  Py_ssize_t __pyx_t_30;
  Py_UCS4 __pyx_t_31;
  __Pyx_RefNannySetupContext("build", 0);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_encoding = __pyx_optional_args->encoding;
      if (__pyx_optional_args->__pyx_n > 2) {
        __pyx_v_driver = __pyx_optional_args->driver;
        if (__pyx_optional_args->__pyx_n > 3) {
          __pyx_v_ignore_fields = __pyx_optional_args->ignore_fields;
          if (__pyx_optional_args->__pyx_n > 4) {
            __pyx_v_ignore_geometry = __pyx_optional_args->ignore_geometry;
          }
        }
      }
    }
  }
  __Pyx_INCREF(__pyx_v_ignore_fields);

  /* "fiona/ogrext.pyx":152
 *         dict
 *         """
 *         cdef void *fdefn = NULL             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef int y = 0
 */
  __pyx_v_fdefn = NULL;

  /* "fiona/ogrext.pyx":154
 *         cdef void *fdefn = NULL
 *         cdef int i
 *         cdef int y = 0             # <<<<<<<<<<<<<<
 *         cdef int m = 0
 *         cdef int d = 0
 */
  __pyx_v_y = 0;

  /* "fiona/ogrext.pyx":155
 *         cdef int i
 *         cdef int y = 0
 *         cdef int m = 0             # <<<<<<<<<<<<<<
 *         cdef int d = 0
 *         cdef int hh = 0
 */
  __pyx_v_m = 0;

  /* "fiona/ogrext.pyx":156
 *         cdef int y = 0
 *         cdef int m = 0
 *         cdef int d = 0             # <<<<<<<<<<<<<<
 *         cdef int hh = 0
 *         cdef int mm = 0
 */
  __pyx_v_d = 0;

  /* "fiona/ogrext.pyx":157
 *         cdef int m = 0
 *         cdef int d = 0
 *         cdef int hh = 0             # <<<<<<<<<<<<<<
 *         cdef int mm = 0
 *         cdef int ss = 0
 */
  __pyx_v_hh = 0;

  /* "fiona/ogrext.pyx":158
 *         cdef int d = 0
 *         cdef int hh = 0
 *         cdef int mm = 0             # <<<<<<<<<<<<<<
 *         cdef int ss = 0
 *         cdef int tz = 0
 */
  __pyx_v_mm = 0;

  /* "fiona/ogrext.pyx":159
 *         cdef int hh = 0
 *         cdef int mm = 0
 *         cdef int ss = 0             # <<<<<<<<<<<<<<
 *         cdef int tz = 0
 *         cdef unsigned char *data = NULL
 */
  __pyx_v_ss = 0;

  /* "fiona/ogrext.pyx":160
 *         cdef int mm = 0
 *         cdef int ss = 0
 *         cdef int tz = 0             # <<<<<<<<<<<<<<
 *         cdef unsigned char *data = NULL
 *         cdef int l
 */
  __pyx_v_tz = 0;

  /* "fiona/ogrext.pyx":161
 *         cdef int ss = 0
 *         cdef int tz = 0
 *         cdef unsigned char *data = NULL             # <<<<<<<<<<<<<<
 *         cdef int l
 *         cdef int retval
 */
  __pyx_v_data = NULL;

  /* "fiona/ogrext.pyx":165
 *         cdef int retval
 *         cdef int fieldsubtype
 *         cdef const char *key_c = NULL             # <<<<<<<<<<<<<<
 * 
 *         # Skeleton of the feature to be returned.
 */
  __pyx_v_key_c = NULL;

  /* "fiona/ogrext.pyx":168
 * 
 *         # Skeleton of the feature to be returned.
 *         fid = OGR_F_GetFID(feature)             # <<<<<<<<<<<<<<
 *         props = OrderedDict()
 *         fiona_feature = {
 */
  __pyx_v_fid = OGR_F_GetFID(__pyx_v_feature);

  /* "fiona/ogrext.pyx":169
 *         # Skeleton of the feature to be returned.
 *         fid = OGR_F_GetFID(feature)
 *         props = OrderedDict()             # <<<<<<<<<<<<<<
 *         fiona_feature = {
 *             "type": "Feature",
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_props = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":171
 *         props = OrderedDict()
 *         fiona_feature = {
 *             "type": "Feature",             # <<<<<<<<<<<<<<
 *             "id": str(fid),
 *             "properties": props,
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_type, __pyx_n_u_Feature) < 0) __PYX_ERR(0, 171, __pyx_L1_error)

  /* "fiona/ogrext.pyx":172
 *         fiona_feature = {
 *             "type": "Feature",
 *             "id": str(fid),             # <<<<<<<<<<<<<<
 *             "properties": props,
 *         }
 */
  __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v_fid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_id, __pyx_t_3) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":173
 *             "type": "Feature",
 *             "id": str(fid),
 *             "properties": props,             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_properties, __pyx_v_props) < 0) __PYX_ERR(0, 171, __pyx_L1_error)
  __pyx_v_fiona_feature = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":176
 *         }
 * 
 *         ignore_fields = set(ignore_fields or [])             # <<<<<<<<<<<<<<
 * 
 *         # Iterate over the fields of the OGR feature.
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_ignore_fields); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 176, __pyx_L1_error)
  if (!__pyx_t_4) {
  } else {
    __Pyx_INCREF(__pyx_v_ignore_fields);
    __pyx_t_1 = __pyx_v_ignore_fields;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __pyx_t_3 = PySet_New(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_ignore_fields, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":179
 * 
 *         # Iterate over the fields of the OGR feature.
 *         for i in range(OGR_F_GetFieldCount(feature)):             # <<<<<<<<<<<<<<
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)
 *             if fdefn == NULL:
 */
  __pyx_t_5 = OGR_F_GetFieldCount(__pyx_v_feature);
  __pyx_t_6 = __pyx_t_5;
  for (__pyx_t_7 = 0; __pyx_t_7 < __pyx_t_6; __pyx_t_7+=1) {
    __pyx_v_i = __pyx_t_7;

    /* "fiona/ogrext.pyx":180
 *         # Iterate over the fields of the OGR feature.
 *         for i in range(OGR_F_GetFieldCount(feature)):
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)             # <<<<<<<<<<<<<<
 *             if fdefn == NULL:
 *                 raise ValueError("Null feature definition")
 */
    __pyx_v_fdefn = OGR_F_GetFieldDefnRef(__pyx_v_feature, __pyx_v_i);

    /* "fiona/ogrext.pyx":181
 *         for i in range(OGR_F_GetFieldCount(feature)):
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)
 *             if fdefn == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null feature definition")
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 */
    __pyx_t_4 = ((__pyx_v_fdefn == NULL) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "fiona/ogrext.pyx":182
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)
 *             if fdefn == NULL:
 *                 raise ValueError("Null feature definition")             # <<<<<<<<<<<<<<
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 182, __pyx_L1_error)

      /* "fiona/ogrext.pyx":181
 *         for i in range(OGR_F_GetFieldCount(feature)):
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)
 *             if fdefn == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null feature definition")
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 */
    }

    /* "fiona/ogrext.pyx":183
 *             if fdefn == NULL:
 *                 raise ValueError("Null feature definition")
 *             key_c = OGR_Fld_GetNameRef(fdefn)             # <<<<<<<<<<<<<<
 *             if key_c == NULL:
 *                 raise ValueError("Null field name reference")
 */
    __pyx_v_key_c = OGR_Fld_GetNameRef(__pyx_v_fdefn);

    /* "fiona/ogrext.pyx":184
 *                 raise ValueError("Null feature definition")
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null field name reference")
 *             key_b = key_c
 */
    __pyx_t_4 = ((__pyx_v_key_c == NULL) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "fiona/ogrext.pyx":185
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:
 *                 raise ValueError("Null field name reference")             # <<<<<<<<<<<<<<
 *             key_b = key_c
 *             key = key_b.decode(encoding)
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 185, __pyx_L1_error)

      /* "fiona/ogrext.pyx":184
 *                 raise ValueError("Null feature definition")
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null field name reference")
 *             key_b = key_c
 */
    }

    /* "fiona/ogrext.pyx":186
 *             if key_c == NULL:
 *                 raise ValueError("Null field name reference")
 *             key_b = key_c             # <<<<<<<<<<<<<<
 *             key = key_b.decode(encoding)
 * 
 */
    __pyx_v_key_b = __pyx_v_key_c;

    /* "fiona/ogrext.pyx":187
 *                 raise ValueError("Null field name reference")
 *             key_b = key_c
 *             key = key_b.decode(encoding)             # <<<<<<<<<<<<<<
 * 
 *             if key in ignore_fields:
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_key_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_encoding);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":189
 *             key = key_b.decode(encoding)
 * 
 *             if key in ignore_fields:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_ignore_fields, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
    __pyx_t_8 = (__pyx_t_4 != 0);
    if (__pyx_t_8) {

      /* "fiona/ogrext.pyx":190
 * 
 *             if key in ignore_fields:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(fdefn)]
 */
      goto __pyx_L5_continue;

      /* "fiona/ogrext.pyx":189
 *             key = key_b.decode(encoding)
 * 
 *             if key in ignore_fields:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    }

    /* "fiona/ogrext.pyx":192
 *                 continue
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(fdefn)]             # <<<<<<<<<<<<<<
 *             fieldsubtype = get_field_subtype(fdefn)
 *             if not fieldtypename:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_FIELD_TYPES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = OGR_Fld_GetType(__pyx_v_fdefn);
    __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_3, __pyx_t_9, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fieldtypename, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":193
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(fdefn)]
 *             fieldsubtype = get_field_subtype(fdefn)             # <<<<<<<<<<<<<<
 *             if not fieldtypename:
 *                 log.warning(
 */
    __pyx_v_fieldsubtype = __pyx_f_5fiona_5_shim_get_field_subtype(__pyx_v_fdefn);

    /* "fiona/ogrext.pyx":194
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(fdefn)]
 *             fieldsubtype = get_field_subtype(fdefn)
 *             if not fieldtypename:             # <<<<<<<<<<<<<<
 *                 log.warning(
 *                     "Skipping field %s: invalid type %s",
 */
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_fieldtypename); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 194, __pyx_L1_error)
    __pyx_t_4 = ((!__pyx_t_8) != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":195
 *             fieldsubtype = get_field_subtype(fdefn)
 *             if not fieldtypename:
 *                 log.warning(             # <<<<<<<<<<<<<<
 *                     "Skipping field %s: invalid type %s",
 *                     key,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":198
 *                     "Skipping field %s: invalid type %s",
 *                     key,
 *                     OGR_Fld_GetType(fdefn))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(OGR_Fld_GetType(__pyx_v_fdefn)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_kp_u_Skipping_field_s_invalid_type_s, __pyx_v_key, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_kp_u_Skipping_field_s_invalid_type_s, __pyx_v_key, __pyx_t_3};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_11 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 195, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (__pyx_t_10) {
          __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Skipping_field_s_invalid_type_s);
        __Pyx_GIVEREF(__pyx_kp_u_Skipping_field_s_invalid_type_s);
        PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_9, __pyx_kp_u_Skipping_field_s_invalid_type_s);
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_GIVEREF(__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_9, __pyx_v_key);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_11, 2+__pyx_t_9, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 195, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "fiona/ogrext.pyx":199
 *                     key,
 *                     OGR_Fld_GetType(fdefn))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # TODO: other types
 */
      goto __pyx_L5_continue;

      /* "fiona/ogrext.pyx":194
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(fdefn)]
 *             fieldsubtype = get_field_subtype(fdefn)
 *             if not fieldtypename:             # <<<<<<<<<<<<<<
 *                 log.warning(
 *                     "Skipping field %s: invalid type %s",
 */
    }

    /* "fiona/ogrext.pyx":202
 * 
 *             # TODO: other types
 *             fieldtype = FIELD_TYPES_MAP[fieldtypename]             # <<<<<<<<<<<<<<
 * 
 *             if is_field_null(feature, i):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FIELD_TYPES_MAP); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_fieldtypename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fieldtype, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":204
 *             fieldtype = FIELD_TYPES_MAP[fieldtypename]
 * 
 *             if is_field_null(feature, i):             # <<<<<<<<<<<<<<
 *                 props[key] = None
 * 
 */
    __pyx_t_4 = (__pyx_f_5fiona_5_shim_is_field_null(__pyx_v_feature, __pyx_v_i) != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":205
 * 
 *             if is_field_null(feature, i):
 *                 props[key] = None             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtypename is 'int32':
 */
      if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, Py_None) < 0)) __PYX_ERR(0, 205, __pyx_L1_error)

      /* "fiona/ogrext.pyx":204
 *             fieldtype = FIELD_TYPES_MAP[fieldtypename]
 * 
 *             if is_field_null(feature, i):             # <<<<<<<<<<<<<<
 *                 props[key] = None
 * 
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":207
 *                 props[key] = None
 * 
 *             elif fieldtypename is 'int32':             # <<<<<<<<<<<<<<
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))
 */
    __pyx_t_4 = (__pyx_v_fieldtypename == __pyx_n_u_int32);
    __pyx_t_8 = (__pyx_t_4 != 0);
    if (__pyx_t_8) {

      /* "fiona/ogrext.pyx":208
 * 
 *             elif fieldtypename is 'int32':
 *                 if fieldsubtype == OFSTBoolean:             # <<<<<<<<<<<<<<
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))
 *                 else:
 */
      __pyx_t_8 = ((__pyx_v_fieldsubtype == OFSTBoolean) != 0);
      if (__pyx_t_8) {

        /* "fiona/ogrext.pyx":209
 *             elif fieldtypename is 'int32':
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))             # <<<<<<<<<<<<<<
 *                 else:
 *                     props[key] = OGR_F_GetFieldAsInteger(feature, i)
 */
        __pyx_t_1 = __Pyx_PyInt_From_int(OGR_F_GetFieldAsInteger(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_8))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_1) < 0)) __PYX_ERR(0, 209, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":208
 * 
 *             elif fieldtypename is 'int32':
 *                 if fieldsubtype == OFSTBoolean:             # <<<<<<<<<<<<<<
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))
 *                 else:
 */
        goto __pyx_L12;
      }

      /* "fiona/ogrext.pyx":211
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))
 *                 else:
 *                     props[key] = OGR_F_GetFieldAsInteger(feature, i)             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtype is int:
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_PyInt_From_int(OGR_F_GetFieldAsInteger(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_1) < 0)) __PYX_ERR(0, 211, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L12:;

      /* "fiona/ogrext.pyx":207
 *                 props[key] = None
 * 
 *             elif fieldtypename is 'int32':             # <<<<<<<<<<<<<<
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger(feature, i))
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":213
 *                     props[key] = OGR_F_GetFieldAsInteger(feature, i)
 * 
 *             elif fieldtype is int:             # <<<<<<<<<<<<<<
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))
 */
    __pyx_t_8 = (__pyx_v_fieldtype == ((PyObject *)(&PyInt_Type)));
    __pyx_t_4 = (__pyx_t_8 != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":214
 * 
 *             elif fieldtype is int:
 *                 if fieldsubtype == OFSTBoolean:             # <<<<<<<<<<<<<<
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))
 *                 else:
 */
      __pyx_t_4 = ((__pyx_v_fieldsubtype == OFSTBoolean) != 0);
      if (__pyx_t_4) {

        /* "fiona/ogrext.pyx":215
 *             elif fieldtype is int:
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))             # <<<<<<<<<<<<<<
 *                 else:
 *                     props[key] = OGR_F_GetFieldAsInteger64(feature, i)
 */
        __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(OGR_F_GetFieldAsInteger64(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 215, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_4))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_1) < 0)) __PYX_ERR(0, 215, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":214
 * 
 *             elif fieldtype is int:
 *                 if fieldsubtype == OFSTBoolean:             # <<<<<<<<<<<<<<
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))
 *                 else:
 */
        goto __pyx_L13;
      }

      /* "fiona/ogrext.pyx":217
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))
 *                 else:
 *                     props[key] = OGR_F_GetFieldAsInteger64(feature, i)             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtype is float:
 */
      /*else*/ {
        __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(OGR_F_GetFieldAsInteger64(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_1) < 0)) __PYX_ERR(0, 217, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L13:;

      /* "fiona/ogrext.pyx":213
 *                     props[key] = OGR_F_GetFieldAsInteger(feature, i)
 * 
 *             elif fieldtype is int:             # <<<<<<<<<<<<<<
 *                 if fieldsubtype == OFSTBoolean:
 *                     props[key] = bool(OGR_F_GetFieldAsInteger64(feature, i))
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":219
 *                     props[key] = OGR_F_GetFieldAsInteger64(feature, i)
 * 
 *             elif fieldtype is float:             # <<<<<<<<<<<<<<
 *                 props[key] = OGR_F_GetFieldAsDouble(feature, i)
 * 
 */
    __pyx_t_4 = (__pyx_v_fieldtype == ((PyObject *)(&PyFloat_Type)));
    __pyx_t_8 = (__pyx_t_4 != 0);
    if (__pyx_t_8) {

      /* "fiona/ogrext.pyx":220
 * 
 *             elif fieldtype is float:
 *                 props[key] = OGR_F_GetFieldAsDouble(feature, i)             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtype is text_type:
 */
      __pyx_t_1 = PyFloat_FromDouble(OGR_F_GetFieldAsDouble(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_1) < 0)) __PYX_ERR(0, 220, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":219
 *                     props[key] = OGR_F_GetFieldAsInteger64(feature, i)
 * 
 *             elif fieldtype is float:             # <<<<<<<<<<<<<<
 *                 props[key] = OGR_F_GetFieldAsDouble(feature, i)
 * 
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":222
 *                 props[key] = OGR_F_GetFieldAsDouble(feature, i)
 * 
 *             elif fieldtype is text_type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     val = OGR_F_GetFieldAsString(feature, i)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_text_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = (__pyx_v_fieldtype == __pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = (__pyx_t_8 != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":223
 * 
 *             elif fieldtype is text_type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     val = OGR_F_GetFieldAsString(feature, i)
 *                     val = val.decode(encoding)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {

          /* "fiona/ogrext.pyx":224
 *             elif fieldtype is text_type:
 *                 try:
 *                     val = OGR_F_GetFieldAsString(feature, i)             # <<<<<<<<<<<<<<
 *                     val = val.decode(encoding)
 *                 except UnicodeDecodeError:
 */
          __pyx_t_1 = __Pyx_PyBytes_FromString(OGR_F_GetFieldAsString(__pyx_v_feature, __pyx_v_i)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 224, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "fiona/ogrext.pyx":225
 *                 try:
 *                     val = OGR_F_GetFieldAsString(feature, i)
 *                     val = val.decode(encoding)             # <<<<<<<<<<<<<<
 *                 except UnicodeDecodeError:
 *                     log.warning(
 */
          __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 225, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
            }
          }
          __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_11, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_encoding);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 225, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "fiona/ogrext.pyx":223
 * 
 *             elif fieldtype is text_type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     val = OGR_F_GetFieldAsString(feature, i)
 *                     val = val.decode(encoding)
 */
        }
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L21_try_end;
        __pyx_L14_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "fiona/ogrext.pyx":226
 *                     val = OGR_F_GetFieldAsString(feature, i)
 *                     val = val.decode(encoding)
 *                 except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *                     log.warning(
 *                         "Failed to decode %s using %s codec", val, encoding)
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
        if (__pyx_t_9) {
          __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_2, &__pyx_t_11) < 0) __PYX_ERR(0, 226, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_11);

          /* "fiona/ogrext.pyx":227
 *                     val = val.decode(encoding)
 *                 except UnicodeDecodeError:
 *                     log.warning(             # <<<<<<<<<<<<<<
 *                         "Failed to decode %s using %s codec", val, encoding)
 * 
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_log); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 227, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_warning); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 227, __pyx_L16_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "fiona/ogrext.pyx":228
 *                 except UnicodeDecodeError:
 *                     log.warning(
 *                         "Failed to decode %s using %s codec", val, encoding)             # <<<<<<<<<<<<<<
 * 
 *                 # Does the text contain a JSON object? Let's check.
 */
          if (unlikely(!__pyx_v_val)) { __Pyx_RaiseUnboundLocalError("val"); __PYX_ERR(0, 228, __pyx_L16_except_error) }
          __pyx_t_10 = NULL;
          __pyx_t_9 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_15);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_15, function);
              __pyx_t_9 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_15)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_kp_u_Failed_to_decode_s_using_s_codec, __pyx_v_val, __pyx_v_encoding};
            __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_15)) {
            PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_kp_u_Failed_to_decode_s_using_s_codec, __pyx_v_val, __pyx_v_encoding};
            __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_15, __pyx_temp+1-__pyx_t_9, 3+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L16_except_error)
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            __Pyx_GOTREF(__pyx_t_3);
          } else
          #endif
          {
            __pyx_t_16 = PyTuple_New(3+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 227, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_16);
            if (__pyx_t_10) {
              __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_10); __pyx_t_10 = NULL;
            }
            __Pyx_INCREF(__pyx_kp_u_Failed_to_decode_s_using_s_codec);
            __Pyx_GIVEREF(__pyx_kp_u_Failed_to_decode_s_using_s_codec);
            PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_9, __pyx_kp_u_Failed_to_decode_s_using_s_codec);
            __Pyx_INCREF(__pyx_v_val);
            __Pyx_GIVEREF(__pyx_v_val);
            PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_9, __pyx_v_val);
            __Pyx_INCREF(__pyx_v_encoding);
            __Pyx_GIVEREF(__pyx_v_encoding);
            PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_9, __pyx_v_encoding);
            __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_15, __pyx_t_16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 227, __pyx_L16_except_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L15_exception_handled;
        }
        goto __pyx_L16_except_error;
        __pyx_L16_except_error:;

        /* "fiona/ogrext.pyx":223
 * 
 *             elif fieldtype is text_type:
 *                 try:             # <<<<<<<<<<<<<<
 *                     val = OGR_F_GetFieldAsString(feature, i)
 *                     val = val.decode(encoding)
 */
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L15_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        __pyx_L21_try_end:;
      }

      /* "fiona/ogrext.pyx":232
 *                 # Does the text contain a JSON object? Let's check.
 *                 # Let's check as cheaply as we can.
 *                 if driver == 'GeoJSON' and val.startswith('{'):             # <<<<<<<<<<<<<<
 *                     try:
 *                         val = json.loads(val)
 */
      __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_driver, __pyx_n_u_GeoJSON, Py_EQ)); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 232, __pyx_L1_error)
      if (__pyx_t_8) {
      } else {
        __pyx_t_4 = __pyx_t_8;
        goto __pyx_L25_bool_binop_done;
      }
      if (unlikely(!__pyx_v_val)) { __Pyx_RaiseUnboundLocalError("val"); __PYX_ERR(0, 232, __pyx_L1_error) }
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_val, __pyx_n_s_startswith); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_11 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_kp_u__4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u__4);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_4 = __pyx_t_8;
      __pyx_L25_bool_binop_done:;
      if (__pyx_t_4) {

        /* "fiona/ogrext.pyx":233
 *                 # Let's check as cheaply as we can.
 *                 if driver == 'GeoJSON' and val.startswith('{'):
 *                     try:             # <<<<<<<<<<<<<<
 *                         val = json.loads(val)
 *                     except ValueError as err:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_12);
          /*try:*/ {

            /* "fiona/ogrext.pyx":234
 *                 if driver == 'GeoJSON' and val.startswith('{'):
 *                     try:
 *                         val = json.loads(val)             # <<<<<<<<<<<<<<
 *                     except ValueError as err:
 *                         log.warning(str(err))
 */
            __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_json); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 234, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_loads); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 234, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_v_val)) { __Pyx_RaiseUnboundLocalError("val"); __PYX_ERR(0, 234, __pyx_L27_error) }
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_11 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_2, __pyx_v_val) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_val);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 234, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_val, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "fiona/ogrext.pyx":233
 *                 # Let's check as cheaply as we can.
 *                 if driver == 'GeoJSON' and val.startswith('{'):
 *                     try:             # <<<<<<<<<<<<<<
 *                         val = json.loads(val)
 *                     except ValueError as err:
 */
          }
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          goto __pyx_L34_try_end;
          __pyx_L27_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "fiona/ogrext.pyx":235
 *                     try:
 *                         val = json.loads(val)
 *                     except ValueError as err:             # <<<<<<<<<<<<<<
 *                         log.warning(str(err))
 * 
 */
          __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
          if (__pyx_t_9) {
            __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_11, &__pyx_t_1, &__pyx_t_2) < 0) __PYX_ERR(0, 235, __pyx_L29_except_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_1);
            __pyx_v_err = __pyx_t_1;
            /*try:*/ {

              /* "fiona/ogrext.pyx":236
 *                         val = json.loads(val)
 *                     except ValueError as err:
 *                         log.warning(str(err))             # <<<<<<<<<<<<<<
 * 
 *                 # Now add to the properties object.
 */
              __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_log); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 236, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_warning); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 236, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __pyx_t_15 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_err); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 236, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_10 = NULL;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
                __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_16);
                if (likely(__pyx_t_10)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                  __Pyx_INCREF(__pyx_t_10);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_16, function);
                }
              }
              __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_10, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_t_15);
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 236, __pyx_L40_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            }

            /* "fiona/ogrext.pyx":235
 *                     try:
 *                         val = json.loads(val)
 *                     except ValueError as err:             # <<<<<<<<<<<<<<
 *                         log.warning(str(err))
 * 
 */
            /*finally:*/ {
              /*normal exit:*/{
                __Pyx_DECREF(__pyx_v_err);
                __pyx_v_err = NULL;
                goto __pyx_L41;
              }
              __pyx_L40_error:;
              /*exception exit:*/{
                __Pyx_PyThreadState_declare
                __Pyx_PyThreadState_assign
                __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
                __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
                __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
                if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_23, &__pyx_t_24);
                if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21) < 0)) __Pyx_ErrFetch(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_19);
                __Pyx_XGOTREF(__pyx_t_20);
                __Pyx_XGOTREF(__pyx_t_21);
                __Pyx_XGOTREF(__pyx_t_22);
                __Pyx_XGOTREF(__pyx_t_23);
                __Pyx_XGOTREF(__pyx_t_24);
                __pyx_t_9 = __pyx_lineno; __pyx_t_17 = __pyx_clineno; __pyx_t_18 = __pyx_filename;
                {
                  __Pyx_DECREF(__pyx_v_err);
                  __pyx_v_err = NULL;
                }
                if (PY_MAJOR_VERSION >= 3) {
                  __Pyx_XGIVEREF(__pyx_t_22);
                  __Pyx_XGIVEREF(__pyx_t_23);
                  __Pyx_XGIVEREF(__pyx_t_24);
                  __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_23, __pyx_t_24);
                }
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_ErrRestore(__pyx_t_19, __pyx_t_20, __pyx_t_21);
                __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0;
                __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_17; __pyx_filename = __pyx_t_18;
                goto __pyx_L29_except_error;
              }
              __pyx_L41:;
            }
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            goto __pyx_L28_exception_handled;
          }
          goto __pyx_L29_except_error;
          __pyx_L29_except_error:;

          /* "fiona/ogrext.pyx":233
 *                 # Let's check as cheaply as we can.
 *                 if driver == 'GeoJSON' and val.startswith('{'):
 *                     try:             # <<<<<<<<<<<<<<
 *                         val = json.loads(val)
 *                     except ValueError as err:
 */
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
          goto __pyx_L1_error;
          __pyx_L28_exception_handled:;
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_13, __pyx_t_12);
          __pyx_L34_try_end:;
        }

        /* "fiona/ogrext.pyx":232
 *                 # Does the text contain a JSON object? Let's check.
 *                 # Let's check as cheaply as we can.
 *                 if driver == 'GeoJSON' and val.startswith('{'):             # <<<<<<<<<<<<<<
 *                     try:
 *                         val = json.loads(val)
 */
      }

      /* "fiona/ogrext.pyx":239
 * 
 *                 # Now add to the properties object.
 *                 props[key] = val             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtype in (FionaDateType, FionaTimeType, FionaDateTimeType):
 */
      if (unlikely(!__pyx_v_val)) { __Pyx_RaiseUnboundLocalError("val"); __PYX_ERR(0, 239, __pyx_L1_error) }
      if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_v_val) < 0)) __PYX_ERR(0, 239, __pyx_L1_error)

      /* "fiona/ogrext.pyx":222
 *                 props[key] = OGR_F_GetFieldAsDouble(feature, i)
 * 
 *             elif fieldtype is text_type:             # <<<<<<<<<<<<<<
 *                 try:
 *                     val = OGR_F_GetFieldAsString(feature, i)
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":241
 *                 props[key] = val
 * 
 *             elif fieldtype in (FionaDateType, FionaTimeType, FionaDateTimeType):             # <<<<<<<<<<<<<<
 *                 retval = OGR_F_GetFieldAsDateTime(
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 */
    __Pyx_INCREF(__pyx_v_fieldtype);
    __pyx_t_2 = __pyx_v_fieldtype;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_FionaDateType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (!__pyx_t_8) {
    } else {
      __pyx_t_4 = __pyx_t_8;
      goto __pyx_L46_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_FionaTimeType); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (!__pyx_t_8) {
    } else {
      __pyx_t_4 = __pyx_t_8;
      goto __pyx_L46_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_FionaDateTimeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_8 < 0)) __PYX_ERR(0, 241, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_4 = __pyx_t_8;
    __pyx_L46_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = (__pyx_t_4 != 0);
    if (__pyx_t_8) {

      /* "fiona/ogrext.pyx":242
 * 
 *             elif fieldtype in (FionaDateType, FionaTimeType, FionaDateTimeType):
 *                 retval = OGR_F_GetFieldAsDateTime(             # <<<<<<<<<<<<<<
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:
 */
      __pyx_v_retval = OGR_F_GetFieldAsDateTime(__pyx_v_feature, __pyx_v_i, (&__pyx_v_y), (&__pyx_v_m), (&__pyx_v_d), (&__pyx_v_hh), (&__pyx_v_mm), (&__pyx_v_ss), (&__pyx_v_tz));

      /* "fiona/ogrext.pyx":244
 *                 retval = OGR_F_GetFieldAsDateTime(
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:             # <<<<<<<<<<<<<<
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {

          /* "fiona/ogrext.pyx":245
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:
 *                     if fieldtype is FionaDateType:             # <<<<<<<<<<<<<<
 *                         props[key] = datetime.date(y, m, d).isoformat()
 *                     elif fieldtype is FionaTimeType:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FionaDateType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 245, __pyx_L49_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_8 = (__pyx_v_fieldtype == __pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_4 = (__pyx_t_8 != 0);
          if (__pyx_t_4) {

            /* "fiona/ogrext.pyx":246
 *                 try:
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()             # <<<<<<<<<<<<<<
 *                     elif fieldtype is FionaTimeType:
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()
 */
            __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_date); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_y); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_m); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_d); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_10 = NULL;
            __pyx_t_17 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_10)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_10);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
                __pyx_t_17 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_1, __pyx_t_16, __pyx_t_15};
              __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 246, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
              PyObject *__pyx_temp[4] = {__pyx_t_10, __pyx_t_1, __pyx_t_16, __pyx_t_15};
              __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 246, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            {
              __pyx_t_25 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 246, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_25);
              if (__pyx_t_10) {
                __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_25, 0, __pyx_t_10); __pyx_t_10 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_25, 0+__pyx_t_17, __pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_16);
              PyTuple_SET_ITEM(__pyx_t_25, 1+__pyx_t_17, __pyx_t_16);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_25, 2+__pyx_t_17, __pyx_t_15);
              __pyx_t_1 = 0;
              __pyx_t_16 = 0;
              __pyx_t_15 = 0;
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_25, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 246, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            }
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_isoformat); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
              __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
              if (likely(__pyx_t_11)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
                __Pyx_INCREF(__pyx_t_11);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_3, function);
              }
            }
            __pyx_t_2 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_2) < 0)) __PYX_ERR(0, 246, __pyx_L49_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "fiona/ogrext.pyx":245
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:
 *                     if fieldtype is FionaDateType:             # <<<<<<<<<<<<<<
 *                         props[key] = datetime.date(y, m, d).isoformat()
 *                     elif fieldtype is FionaTimeType:
 */
            goto __pyx_L57;
          }

          /* "fiona/ogrext.pyx":247
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()
 *                     elif fieldtype is FionaTimeType:             # <<<<<<<<<<<<<<
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()
 *                     else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FionaTimeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 247, __pyx_L49_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_4 = (__pyx_v_fieldtype == __pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_8 = (__pyx_t_4 != 0);
          if (__pyx_t_8) {

            /* "fiona/ogrext.pyx":248
 *                         props[key] = datetime.date(y, m, d).isoformat()
 *                     elif fieldtype is FionaTimeType:
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()             # <<<<<<<<<<<<<<
 *                     else:
 *                         props[key] = datetime.datetime(
 */
            __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_datetime); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_time); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_25);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_hh); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_mm); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_ss); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_1 = NULL;
            __pyx_t_17 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_25))) {
              __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_25);
              if (likely(__pyx_t_1)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_25);
                __Pyx_INCREF(__pyx_t_1);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_25, function);
                __pyx_t_17 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_25)) {
              PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_11, __pyx_t_15, __pyx_t_16};
              __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_25, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_25)) {
              PyObject *__pyx_temp[4] = {__pyx_t_1, __pyx_t_11, __pyx_t_15, __pyx_t_16};
              __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_25, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            } else
            #endif
            {
              __pyx_t_10 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 248, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_10);
              if (__pyx_t_1) {
                __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_1); __pyx_t_1 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_11);
              PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_17, __pyx_t_11);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_17, __pyx_t_15);
              __Pyx_GIVEREF(__pyx_t_16);
              PyTuple_SET_ITEM(__pyx_t_10, 2+__pyx_t_17, __pyx_t_16);
              __pyx_t_11 = 0;
              __pyx_t_15 = 0;
              __pyx_t_16 = 0;
              __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_25, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 248, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_3);
              __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            }
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            __pyx_t_25 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_isoformat); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_25);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
            __pyx_t_3 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_25))) {
              __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_25);
              if (likely(__pyx_t_3)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_25);
                __Pyx_INCREF(__pyx_t_3);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_25, function);
              }
            }
            __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_25, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_25);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_2) < 0)) __PYX_ERR(0, 248, __pyx_L49_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

            /* "fiona/ogrext.pyx":247
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()
 *                     elif fieldtype is FionaTimeType:             # <<<<<<<<<<<<<<
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()
 *                     else:
 */
            goto __pyx_L57;
          }

          /* "fiona/ogrext.pyx":251
 *                     else:
 *                         props[key] = datetime.datetime(
 *                             y, m, d, hh, mm, ss).isoformat()             # <<<<<<<<<<<<<<
 *                 except ValueError as err:
 *                     log.exception(err)
 */
          /*else*/ {

            /* "fiona/ogrext.pyx":250
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()
 *                     else:
 *                         props[key] = datetime.datetime(             # <<<<<<<<<<<<<<
 *                             y, m, d, hh, mm, ss).isoformat()
 *                 except ValueError as err:
 */
            __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 250, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_datetime); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 250, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

            /* "fiona/ogrext.pyx":251
 *                     else:
 *                         props[key] = datetime.datetime(
 *                             y, m, d, hh, mm, ss).isoformat()             # <<<<<<<<<<<<<<
 *                 except ValueError as err:
 *                     log.exception(err)
 */
            __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_3);
            __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_m); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_d); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_hh); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_mm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_26 = __Pyx_PyInt_From_int(__pyx_v_ss); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_26);
            __pyx_t_27 = NULL;
            __pyx_t_17 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_27 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_27)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_27);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
                __pyx_t_17 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[7] = {__pyx_t_27, __pyx_t_3, __pyx_t_16, __pyx_t_15, __pyx_t_11, __pyx_t_1, __pyx_t_26};
              __pyx_t_25 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_17, 6+__pyx_t_17); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 250, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
              __Pyx_GOTREF(__pyx_t_25);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
              PyObject *__pyx_temp[7] = {__pyx_t_27, __pyx_t_3, __pyx_t_16, __pyx_t_15, __pyx_t_11, __pyx_t_1, __pyx_t_26};
              __pyx_t_25 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_17, 6+__pyx_t_17); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 250, __pyx_L49_error)
              __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
              __Pyx_GOTREF(__pyx_t_25);
              __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
            } else
            #endif
            {
              __pyx_t_28 = PyTuple_New(6+__pyx_t_17); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 250, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_28);
              if (__pyx_t_27) {
                __Pyx_GIVEREF(__pyx_t_27); PyTuple_SET_ITEM(__pyx_t_28, 0, __pyx_t_27); __pyx_t_27 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_3);
              PyTuple_SET_ITEM(__pyx_t_28, 0+__pyx_t_17, __pyx_t_3);
              __Pyx_GIVEREF(__pyx_t_16);
              PyTuple_SET_ITEM(__pyx_t_28, 1+__pyx_t_17, __pyx_t_16);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_28, 2+__pyx_t_17, __pyx_t_15);
              __Pyx_GIVEREF(__pyx_t_11);
              PyTuple_SET_ITEM(__pyx_t_28, 3+__pyx_t_17, __pyx_t_11);
              __Pyx_GIVEREF(__pyx_t_1);
              PyTuple_SET_ITEM(__pyx_t_28, 4+__pyx_t_17, __pyx_t_1);
              __Pyx_GIVEREF(__pyx_t_26);
              PyTuple_SET_ITEM(__pyx_t_28, 5+__pyx_t_17, __pyx_t_26);
              __pyx_t_3 = 0;
              __pyx_t_16 = 0;
              __pyx_t_15 = 0;
              __pyx_t_11 = 0;
              __pyx_t_1 = 0;
              __pyx_t_26 = 0;
              __pyx_t_25 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_28, NULL); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 250, __pyx_L49_error)
              __Pyx_GOTREF(__pyx_t_25);
              __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
            }
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_25, __pyx_n_s_isoformat); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_10);
            __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
            __pyx_t_25 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
              __pyx_t_25 = PyMethod_GET_SELF(__pyx_t_10);
              if (likely(__pyx_t_25)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
                __Pyx_INCREF(__pyx_t_25);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_10, function);
              }
            }
            __pyx_t_2 = (__pyx_t_25) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_25) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 251, __pyx_L49_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

            /* "fiona/ogrext.pyx":250
 *                         props[key] = datetime.time(hh, mm, ss).isoformat()
 *                     else:
 *                         props[key] = datetime.datetime(             # <<<<<<<<<<<<<<
 *                             y, m, d, hh, mm, ss).isoformat()
 *                 except ValueError as err:
 */
            if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_2) < 0)) __PYX_ERR(0, 250, __pyx_L49_error)
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_L57:;

          /* "fiona/ogrext.pyx":244
 *                 retval = OGR_F_GetFieldAsDateTime(
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:             # <<<<<<<<<<<<<<
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()
 */
        }
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L56_try_end;
        __pyx_L49_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
        __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
        __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
        __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "fiona/ogrext.pyx":252
 *                         props[key] = datetime.datetime(
 *                             y, m, d, hh, mm, ss).isoformat()
 *                 except ValueError as err:             # <<<<<<<<<<<<<<
 *                     log.exception(err)
 *                     props[key] = None
 */
        __pyx_t_17 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
        if (__pyx_t_17) {
          __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_10, &__pyx_t_25) < 0) __PYX_ERR(0, 252, __pyx_L51_except_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GOTREF(__pyx_t_25);
          __Pyx_INCREF(__pyx_t_10);
          __pyx_v_err = __pyx_t_10;
          /*try:*/ {

            /* "fiona/ogrext.pyx":253
 *                             y, m, d, hh, mm, ss).isoformat()
 *                 except ValueError as err:
 *                     log.exception(err)             # <<<<<<<<<<<<<<
 *                     props[key] = None
 * 
 */
            __Pyx_GetModuleGlobalName(__pyx_t_26, __pyx_n_s_log); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 253, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_26);
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_26, __pyx_n_s_exception); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
            __pyx_t_26 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_26 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_26)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_26);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_28 = (__pyx_t_26) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_26, __pyx_v_err) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_err);
            __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
            if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 253, __pyx_L63_error)
            __Pyx_GOTREF(__pyx_t_28);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;

            /* "fiona/ogrext.pyx":254
 *                 except ValueError as err:
 *                     log.exception(err)
 *                     props[key] = None             # <<<<<<<<<<<<<<
 * 
 *             elif fieldtype is bytes:
 */
            if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, Py_None) < 0)) __PYX_ERR(0, 254, __pyx_L63_error)
          }

          /* "fiona/ogrext.pyx":252
 *                         props[key] = datetime.datetime(
 *                             y, m, d, hh, mm, ss).isoformat()
 *                 except ValueError as err:             # <<<<<<<<<<<<<<
 *                     log.exception(err)
 *                     props[key] = None
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_err);
              __pyx_v_err = NULL;
              goto __pyx_L64;
            }
            __pyx_L63_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
              __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
              __Pyx_XDECREF(__pyx_t_27); __pyx_t_27 = 0;
              __Pyx_XDECREF(__pyx_t_28); __pyx_t_28 = 0;
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_21, &__pyx_t_20, &__pyx_t_19);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_24, &__pyx_t_23, &__pyx_t_22) < 0)) __Pyx_ErrFetch(&__pyx_t_24, &__pyx_t_23, &__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_24);
              __Pyx_XGOTREF(__pyx_t_23);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_19);
              __pyx_t_17 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_29 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_err);
                __pyx_v_err = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_ExceptionReset(__pyx_t_21, __pyx_t_20, __pyx_t_19);
              }
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_ErrRestore(__pyx_t_24, __pyx_t_23, __pyx_t_22);
              __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0;
              __pyx_lineno = __pyx_t_17; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_29;
              goto __pyx_L51_except_error;
            }
            __pyx_L64:;
          }
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
          goto __pyx_L50_exception_handled;
        }
        goto __pyx_L51_except_error;
        __pyx_L51_except_error:;

        /* "fiona/ogrext.pyx":244
 *                 retval = OGR_F_GetFieldAsDateTime(
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 *                 try:             # <<<<<<<<<<<<<<
 *                     if fieldtype is FionaDateType:
 *                         props[key] = datetime.date(y, m, d).isoformat()
 */
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L50_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
        __pyx_L56_try_end:;
      }

      /* "fiona/ogrext.pyx":241
 *                 props[key] = val
 * 
 *             elif fieldtype in (FionaDateType, FionaTimeType, FionaDateTimeType):             # <<<<<<<<<<<<<<
 *                 retval = OGR_F_GetFieldAsDateTime(
 *                     feature, i, &y, &m, &d, &hh, &mm, &ss, &tz)
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":256
 *                     props[key] = None
 * 
 *             elif fieldtype is bytes:             # <<<<<<<<<<<<<<
 *                 data = OGR_F_GetFieldAsBinary(feature, i, &l)
 *                 props[key] = data[:l]
 */
    __pyx_t_8 = (__pyx_v_fieldtype == ((PyObject *)(&PyBytes_Type)));
    __pyx_t_4 = (__pyx_t_8 != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":257
 * 
 *             elif fieldtype is bytes:
 *                 data = OGR_F_GetFieldAsBinary(feature, i, &l)             # <<<<<<<<<<<<<<
 *                 props[key] = data[:l]
 * 
 */
      __pyx_v_data = OGR_F_GetFieldAsBinary(__pyx_v_feature, __pyx_v_i, (&__pyx_v_l));

      /* "fiona/ogrext.pyx":258
 *             elif fieldtype is bytes:
 *                 data = OGR_F_GetFieldAsBinary(feature, i, &l)
 *                 props[key] = data[:l]             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __pyx_t_25 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_data) + 0, __pyx_v_l - 0); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, __pyx_t_25) < 0)) __PYX_ERR(0, 258, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

      /* "fiona/ogrext.pyx":256
 *                     props[key] = None
 * 
 *             elif fieldtype is bytes:             # <<<<<<<<<<<<<<
 *                 data = OGR_F_GetFieldAsBinary(feature, i, &l)
 *                 props[key] = data[:l]
 */
      goto __pyx_L11;
    }

    /* "fiona/ogrext.pyx":261
 * 
 *             else:
 *                 log.debug("%s: None, fieldtype: %r, %r" % (key, fieldtype, fieldtype in string_types))             # <<<<<<<<<<<<<<
 *                 props[key] = None
 * 
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_log); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_debug); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PyTuple_New(5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_30 = 0;
      __pyx_t_31 = 127;
      __pyx_t_28 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_key), __pyx_empty_unicode); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      __pyx_t_31 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) > __pyx_t_31) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) : __pyx_t_31;
      __pyx_t_30 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_28);
      __Pyx_GIVEREF(__pyx_t_28);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_28);
      __pyx_t_28 = 0;
      __Pyx_INCREF(__pyx_kp_u_None_fieldtype);
      __pyx_t_30 += 19;
      __Pyx_GIVEREF(__pyx_kp_u_None_fieldtype);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_kp_u_None_fieldtype);
      __pyx_t_28 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_fieldtype), __pyx_empty_unicode); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      __pyx_t_31 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) > __pyx_t_31) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_28) : __pyx_t_31;
      __pyx_t_30 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_28);
      __Pyx_GIVEREF(__pyx_t_28);
      PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_t_28);
      __pyx_t_28 = 0;
      __Pyx_INCREF(__pyx_kp_u__5);
      __pyx_t_30 += 2;
      __Pyx_GIVEREF(__pyx_kp_u__5);
      PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_kp_u__5);
      __Pyx_GetModuleGlobalName(__pyx_t_28, __pyx_n_s_string_types); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_v_fieldtype, __pyx_t_28, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
      __pyx_t_28 = __Pyx_PyUnicode_FromBInt_int(__pyx_t_4); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      __pyx_t_30 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_28);
      __Pyx_GIVEREF(__pyx_t_28);
      PyTuple_SET_ITEM(__pyx_t_10, 4, __pyx_t_28);
      __pyx_t_28 = 0;
      __pyx_t_28 = __Pyx_PyUnicode_Join(__pyx_t_10, 5, __pyx_t_30, __pyx_t_31); if (unlikely(!__pyx_t_28)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_28);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      __pyx_t_25 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_10, __pyx_t_28) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_28);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_28); __pyx_t_28 = 0;
      if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_25);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;

      /* "fiona/ogrext.pyx":262
 *             else:
 *                 log.debug("%s: None, fieldtype: %r, %r" % (key, fieldtype, fieldtype in string_types))
 *                 props[key] = None             # <<<<<<<<<<<<<<
 * 
 *         cdef void *cogr_geometry = NULL
 */
      if (unlikely(PyObject_SetItem(__pyx_v_props, __pyx_v_key, Py_None) < 0)) __PYX_ERR(0, 262, __pyx_L1_error)
    }
    __pyx_L11:;
    __pyx_L5_continue:;
  }

  /* "fiona/ogrext.pyx":264
 *                 props[key] = None
 * 
 *         cdef void *cogr_geometry = NULL             # <<<<<<<<<<<<<<
 *         cdef void *org_geometry = NULL
 * 
 */
  __pyx_v_cogr_geometry = NULL;

  /* "fiona/ogrext.pyx":265
 * 
 *         cdef void *cogr_geometry = NULL
 *         cdef void *org_geometry = NULL             # <<<<<<<<<<<<<<
 * 
 *         if not ignore_geometry:
 */
  __pyx_v_org_geometry = NULL;

  /* "fiona/ogrext.pyx":267
 *         cdef void *org_geometry = NULL
 * 
 *         if not ignore_geometry:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGR_F_GetGeometryRef(feature)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_ignore_geometry); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 267, __pyx_L1_error)
  __pyx_t_8 = ((!__pyx_t_4) != 0);
  if (__pyx_t_8) {

    /* "fiona/ogrext.pyx":268
 * 
 *         if not ignore_geometry:
 *             cogr_geometry = OGR_F_GetGeometryRef(feature)             # <<<<<<<<<<<<<<
 * 
 *             if cogr_geometry is not NULL:
 */
    __pyx_v_cogr_geometry = OGR_F_GetGeometryRef(__pyx_v_feature);

    /* "fiona/ogrext.pyx":270
 *             cogr_geometry = OGR_F_GetGeometryRef(feature)
 * 
 *             if cogr_geometry is not NULL:             # <<<<<<<<<<<<<<
 * 
 *                 code = base_geometry_type_code(OGR_G_GetGeometryType(cogr_geometry))
 */
    __pyx_t_8 = ((__pyx_v_cogr_geometry != NULL) != 0);
    if (__pyx_t_8) {

      /* "fiona/ogrext.pyx":272
 *             if cogr_geometry is not NULL:
 * 
 *                 code = base_geometry_type_code(OGR_G_GetGeometryType(cogr_geometry))             # <<<<<<<<<<<<<<
 * 
 *                 if 8 <= code <= 14:  # Curves.
 */
      __pyx_v_code = __pyx_f_5fiona_9_geometry_base_geometry_type_code(OGR_G_GetGeometryType(__pyx_v_cogr_geometry));

      /* "fiona/ogrext.pyx":274
 *                 code = base_geometry_type_code(OGR_G_GetGeometryType(cogr_geometry))
 * 
 *                 if 8 <= code <= 14:  # Curves.             # <<<<<<<<<<<<<<
 *                     cogr_geometry = get_linear_geometry(cogr_geometry)
 *                     geom = GeomBuilder().build(cogr_geometry)
 */
      __pyx_t_8 = (8 <= __pyx_v_code);
      if (__pyx_t_8) {
        __pyx_t_8 = (__pyx_v_code <= 14);
      }
      __pyx_t_4 = (__pyx_t_8 != 0);
      if (__pyx_t_4) {

        /* "fiona/ogrext.pyx":275
 * 
 *                 if 8 <= code <= 14:  # Curves.
 *                     cogr_geometry = get_linear_geometry(cogr_geometry)             # <<<<<<<<<<<<<<
 *                     geom = GeomBuilder().build(cogr_geometry)
 *                     OGR_G_DestroyGeometry(cogr_geometry)
 */
        __pyx_v_cogr_geometry = __pyx_f_5fiona_5_shim_get_linear_geometry(__pyx_v_cogr_geometry);

        /* "fiona/ogrext.pyx":276
 *                 if 8 <= code <= 14:  # Curves.
 *                     cogr_geometry = get_linear_geometry(cogr_geometry)
 *                     geom = GeomBuilder().build(cogr_geometry)             # <<<<<<<<<<<<<<
 *                     OGR_G_DestroyGeometry(cogr_geometry)
 * 
 */
        __pyx_t_25 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_9_geometry_GeomBuilder)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_25);
        __pyx_t_2 = ((struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder *)((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_25)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_25), __pyx_v_cogr_geometry); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 276, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __pyx_v_geom = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":277
 *                     cogr_geometry = get_linear_geometry(cogr_geometry)
 *                     geom = GeomBuilder().build(cogr_geometry)
 *                     OGR_G_DestroyGeometry(cogr_geometry)             # <<<<<<<<<<<<<<
 * 
 *                 elif 15 <= code <= 17:
 */
        OGR_G_DestroyGeometry(__pyx_v_cogr_geometry);

        /* "fiona/ogrext.pyx":274
 *                 code = base_geometry_type_code(OGR_G_GetGeometryType(cogr_geometry))
 * 
 *                 if 8 <= code <= 14:  # Curves.             # <<<<<<<<<<<<<<
 *                     cogr_geometry = get_linear_geometry(cogr_geometry)
 *                     geom = GeomBuilder().build(cogr_geometry)
 */
        goto __pyx_L71;
      }

      /* "fiona/ogrext.pyx":279
 *                     OGR_G_DestroyGeometry(cogr_geometry)
 * 
 *                 elif 15 <= code <= 17:             # <<<<<<<<<<<<<<
 *                     # We steal the geometry: the geometry of the in-memory feature is now null
 *                     # and we are responsible for cogr_geometry.
 */
      __pyx_t_4 = (15 <= __pyx_v_code);
      if (__pyx_t_4) {
        __pyx_t_4 = (__pyx_v_code <= 17);
      }
      __pyx_t_8 = (__pyx_t_4 != 0);
      if (__pyx_t_8) {

        /* "fiona/ogrext.pyx":282
 *                     # We steal the geometry: the geometry of the in-memory feature is now null
 *                     # and we are responsible for cogr_geometry.
 *                     org_geometry = OGR_F_StealGeometry(feature)             # <<<<<<<<<<<<<<
 * 
 *                     if code in (15, 16):
 */
        __pyx_v_org_geometry = OGR_F_StealGeometry(__pyx_v_feature);

        /* "fiona/ogrext.pyx":284
 *                     org_geometry = OGR_F_StealGeometry(feature)
 * 
 *                     if code in (15, 16):             # <<<<<<<<<<<<<<
 *                         cogr_geometry = OGR_G_ForceToMultiPolygon(org_geometry)
 *                     elif code == 17:
 */
        switch (__pyx_v_code) {
          case 15:
          case 16:

          /* "fiona/ogrext.pyx":285
 * 
 *                     if code in (15, 16):
 *                         cogr_geometry = OGR_G_ForceToMultiPolygon(org_geometry)             # <<<<<<<<<<<<<<
 *                     elif code == 17:
 *                         cogr_geometry = OGR_G_ForceToPolygon(org_geometry)
 */
          __pyx_v_cogr_geometry = OGR_G_ForceToMultiPolygon(__pyx_v_org_geometry);

          /* "fiona/ogrext.pyx":284
 *                     org_geometry = OGR_F_StealGeometry(feature)
 * 
 *                     if code in (15, 16):             # <<<<<<<<<<<<<<
 *                         cogr_geometry = OGR_G_ForceToMultiPolygon(org_geometry)
 *                     elif code == 17:
 */
          break;
          case 17:

          /* "fiona/ogrext.pyx":287
 *                         cogr_geometry = OGR_G_ForceToMultiPolygon(org_geometry)
 *                     elif code == 17:
 *                         cogr_geometry = OGR_G_ForceToPolygon(org_geometry)             # <<<<<<<<<<<<<<
 * 
 *                     geom = GeomBuilder().build(cogr_geometry)
 */
          __pyx_v_cogr_geometry = OGR_G_ForceToPolygon(__pyx_v_org_geometry);

          /* "fiona/ogrext.pyx":286
 *                     if code in (15, 16):
 *                         cogr_geometry = OGR_G_ForceToMultiPolygon(org_geometry)
 *                     elif code == 17:             # <<<<<<<<<<<<<<
 *                         cogr_geometry = OGR_G_ForceToPolygon(org_geometry)
 * 
 */
          break;
          default: break;
        }

        /* "fiona/ogrext.pyx":289
 *                         cogr_geometry = OGR_G_ForceToPolygon(org_geometry)
 * 
 *                     geom = GeomBuilder().build(cogr_geometry)             # <<<<<<<<<<<<<<
 *                     OGR_G_DestroyGeometry(cogr_geometry)
 * 
 */
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_9_geometry_GeomBuilder)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_25 = ((struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder *)((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_2)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_2), __pyx_v_cogr_geometry); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 289, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_25);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_geom = __pyx_t_25;
        __pyx_t_25 = 0;

        /* "fiona/ogrext.pyx":290
 * 
 *                     geom = GeomBuilder().build(cogr_geometry)
 *                     OGR_G_DestroyGeometry(cogr_geometry)             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
        OGR_G_DestroyGeometry(__pyx_v_cogr_geometry);

        /* "fiona/ogrext.pyx":279
 *                     OGR_G_DestroyGeometry(cogr_geometry)
 * 
 *                 elif 15 <= code <= 17:             # <<<<<<<<<<<<<<
 *                     # We steal the geometry: the geometry of the in-memory feature is now null
 *                     # and we are responsible for cogr_geometry.
 */
        goto __pyx_L71;
      }

      /* "fiona/ogrext.pyx":293
 * 
 *                 else:
 *                     geom = GeomBuilder().build(cogr_geometry)             # <<<<<<<<<<<<<<
 * 
 *                 fiona_feature["geometry"] = geom
 */
      /*else*/ {
        __pyx_t_25 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_9_geometry_GeomBuilder)); if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_25);
        __pyx_t_2 = ((struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder *)((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_25)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_9_geometry_GeomBuilder *)__pyx_t_25), __pyx_v_cogr_geometry); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_25); __pyx_t_25 = 0;
        __pyx_v_geom = __pyx_t_2;
        __pyx_t_2 = 0;
      }
      __pyx_L71:;

      /* "fiona/ogrext.pyx":295
 *                     geom = GeomBuilder().build(cogr_geometry)
 * 
 *                 fiona_feature["geometry"] = geom             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      if (unlikely(PyDict_SetItem(__pyx_v_fiona_feature, __pyx_n_u_geometry, __pyx_v_geom) < 0)) __PYX_ERR(0, 295, __pyx_L1_error)

      /* "fiona/ogrext.pyx":270
 *             cogr_geometry = OGR_F_GetGeometryRef(feature)
 * 
 *             if cogr_geometry is not NULL:             # <<<<<<<<<<<<<<
 * 
 *                 code = base_geometry_type_code(OGR_G_GetGeometryType(cogr_geometry))
 */
      goto __pyx_L70;
    }

    /* "fiona/ogrext.pyx":299
 *             else:
 * 
 *                 fiona_feature["geometry"] = None             # <<<<<<<<<<<<<<
 * 
 *         return fiona_feature
 */
    /*else*/ {
      if (unlikely(PyDict_SetItem(__pyx_v_fiona_feature, __pyx_n_u_geometry, Py_None) < 0)) __PYX_ERR(0, 299, __pyx_L1_error)
    }
    __pyx_L70:;

    /* "fiona/ogrext.pyx":267
 *         cdef void *org_geometry = NULL
 * 
 *         if not ignore_geometry:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGR_F_GetGeometryRef(feature)
 * 
 */
  }

  /* "fiona/ogrext.pyx":301
 *                 fiona_feature["geometry"] = None
 * 
 *         return fiona_feature             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_fiona_feature);
  __pyx_r = __pyx_v_fiona_feature;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":129
 *     """
 * 
 *     cdef build(self, void *feature, encoding='utf-8', bbox=False, driver=None, ignore_fields=None, ignore_geometry=False):             # <<<<<<<<<<<<<<
 *         """Build a Fiona feature object from an OGR feature
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_25);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_XDECREF(__pyx_t_27);
  __Pyx_XDECREF(__pyx_t_28);
  __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_props);
  __Pyx_XDECREF(__pyx_v_fiona_feature);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_fieldtypename);
  __Pyx_XDECREF(__pyx_v_fieldtype);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XDECREF(__pyx_v_geom);
  __Pyx_XDECREF(__pyx_v_ignore_fields);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14FeatureBuilder_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14FeatureBuilder_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14FeatureBuilder___reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14FeatureBuilder___reduce_cython__(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_FeatureBuilder); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FeatureBuilder__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_FeatureBuilder); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FeatureBuilder__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14FeatureBuilder_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14FeatureBuilder_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14FeatureBuilder_2__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14FeatureBuilder_2__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_FeatureBuilder__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5fiona_6ogrext___pyx_unpickle_FeatureBuilder__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_FeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_FeatureBuilder__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.FeatureBuilder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":312
 *     """
 * 
 *     cdef void * build(self, feature, collection) except NULL:             # <<<<<<<<<<<<<<
 *         cdef void *cogr_geometry = NULL
 *         cdef const char *string_c = NULL
 */

static void *__pyx_f_5fiona_6ogrext_17OGRFeatureBuilder_build(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self, PyObject *__pyx_v_feature, PyObject *__pyx_v_collection) {
  void *__pyx_v_cogr_geometry;
  char const *__pyx_v_string_c;
  struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_session = 0;
  void *__pyx_v_cogr_layer;
  void *__pyx_v_cogr_featuredefn;
  void *__pyx_v_cogr_feature;
  PyObject *__pyx_v_encoding = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_ogr_key = NULL;
  PyObject *__pyx_v_schema_type = NULL;
  PyObject *__pyx_v_key_bytes = NULL;
  PyObject *__pyx_v_key_c = NULL;
  int __pyx_v_i;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_m = NULL;
  PyObject *__pyx_v_d = NULL;
  PyObject *__pyx_v_hh = NULL;
  PyObject *__pyx_v_mm = NULL;
  PyObject *__pyx_v_ss = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_ff = NULL;
  PyObject *__pyx_v_value_bytes = NULL;
  void *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  void *__pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  char *__pyx_t_19;
  PY_LONG_LONG __pyx_t_20;
  double __pyx_t_21;
  int __pyx_t_22;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *(*__pyx_t_25)(PyObject *);
  int __pyx_t_26;
  int __pyx_t_27;
  int __pyx_t_28;
  int __pyx_t_29;
  int __pyx_t_30;
  char const *__pyx_t_31;
  Py_ssize_t __pyx_t_32;
  Py_UCS4 __pyx_t_33;
  __Pyx_RefNannySetupContext("build", 0);

  /* "fiona/ogrext.pyx":313
 * 
 *     cdef void * build(self, feature, collection) except NULL:
 *         cdef void *cogr_geometry = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *string_c = NULL
 *         cdef WritingSession session
 */
  __pyx_v_cogr_geometry = NULL;

  /* "fiona/ogrext.pyx":314
 *     cdef void * build(self, feature, collection) except NULL:
 *         cdef void *cogr_geometry = NULL
 *         cdef const char *string_c = NULL             # <<<<<<<<<<<<<<
 *         cdef WritingSession session
 *         session = collection.session
 */
  __pyx_v_string_c = NULL;

  /* "fiona/ogrext.pyx":316
 *         cdef const char *string_c = NULL
 *         cdef WritingSession session
 *         session = collection.session             # <<<<<<<<<<<<<<
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_WritingSession))))) __PYX_ERR(0, 316, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":317
 *         cdef WritingSession session
 *         session = collection.session
 *         cdef void *cogr_layer = session.cogr_layer             # <<<<<<<<<<<<<<
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")
 */
  __pyx_t_2 = __pyx_v_session->__pyx_base.cogr_layer;
  __pyx_v_cogr_layer = __pyx_t_2;

  /* "fiona/ogrext.pyx":318
 *         session = collection.session
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 */
  __pyx_t_3 = ((__pyx_v_cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":319
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 *         if cogr_featuredefn == NULL:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 319, __pyx_L1_error)

    /* "fiona/ogrext.pyx":318
 *         session = collection.session
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 */
  }

  /* "fiona/ogrext.pyx":320
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)             # <<<<<<<<<<<<<<
 *         if cogr_featuredefn == NULL:
 *             raise ValueError("Null feature definition")
 */
  __pyx_v_cogr_featuredefn = OGR_L_GetLayerDefn(__pyx_v_cogr_layer);

  /* "fiona/ogrext.pyx":321
 *             raise ValueError("Null layer")
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 *         if cogr_featuredefn == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature definition")
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 */
  __pyx_t_3 = ((__pyx_v_cogr_featuredefn == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":322
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 *         if cogr_featuredefn == NULL:
 *             raise ValueError("Null feature definition")             # <<<<<<<<<<<<<<
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 *         if cogr_feature == NULL:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 322, __pyx_L1_error)

    /* "fiona/ogrext.pyx":321
 *             raise ValueError("Null layer")
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 *         if cogr_featuredefn == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature definition")
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 */
  }

  /* "fiona/ogrext.pyx":323
 *         if cogr_featuredefn == NULL:
 *             raise ValueError("Null feature definition")
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)             # <<<<<<<<<<<<<<
 *         if cogr_feature == NULL:
 *             raise ValueError("Null feature")
 */
  __pyx_v_cogr_feature = OGR_F_Create(__pyx_v_cogr_featuredefn);

  /* "fiona/ogrext.pyx":324
 *             raise ValueError("Null feature definition")
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature")
 * 
 */
  __pyx_t_3 = ((__pyx_v_cogr_feature == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":325
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 *         if cogr_feature == NULL:
 *             raise ValueError("Null feature")             # <<<<<<<<<<<<<<
 * 
 *         if feature['geometry'] is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 325, __pyx_L1_error)

    /* "fiona/ogrext.pyx":324
 *             raise ValueError("Null feature definition")
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature")
 * 
 */
  }

  /* "fiona/ogrext.pyx":327
 *             raise ValueError("Null feature")
 * 
 *         if feature['geometry'] is not None:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGRGeomBuilder().build(
 *                                 feature['geometry'])
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_feature, __pyx_n_u_geometry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "fiona/ogrext.pyx":328
 * 
 *         if feature['geometry'] is not None:
 *             cogr_geometry = OGRGeomBuilder().build(             # <<<<<<<<<<<<<<
 *                                 feature['geometry'])
 *         OGR_F_SetGeometryDirectly(cogr_feature, cogr_geometry)
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_9_geometry_OGRGeomBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "fiona/ogrext.pyx":329
 *         if feature['geometry'] is not None:
 *             cogr_geometry = OGRGeomBuilder().build(
 *                                 feature['geometry'])             # <<<<<<<<<<<<<<
 *         OGR_F_SetGeometryDirectly(cogr_feature, cogr_geometry)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_v_feature, __pyx_n_u_geometry); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "fiona/ogrext.pyx":328
 * 
 *         if feature['geometry'] is not None:
 *             cogr_geometry = OGRGeomBuilder().build(             # <<<<<<<<<<<<<<
 *                                 feature['geometry'])
 *         OGR_F_SetGeometryDirectly(cogr_feature, cogr_geometry)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder *)((struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *)__pyx_t_1), __pyx_t_5); if (unlikely(__pyx_t_2 == ((void *)NULL))) __PYX_ERR(0, 328, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cogr_geometry = __pyx_t_2;

    /* "fiona/ogrext.pyx":327
 *             raise ValueError("Null feature")
 * 
 *         if feature['geometry'] is not None:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGRGeomBuilder().build(
 *                                 feature['geometry'])
 */
  }

  /* "fiona/ogrext.pyx":330
 *             cogr_geometry = OGRGeomBuilder().build(
 *                                 feature['geometry'])
 *         OGR_F_SetGeometryDirectly(cogr_feature, cogr_geometry)             # <<<<<<<<<<<<<<
 * 
 *         # OGR_F_SetFieldString takes encoded strings ('bytes' in Python 3).
 */
  (void)(OGR_F_SetGeometryDirectly(__pyx_v_cogr_feature, __pyx_v_cogr_geometry));

  /* "fiona/ogrext.pyx":333
 * 
 *         # OGR_F_SetFieldString takes encoded strings ('bytes' in Python 3).
 *         encoding = session._get_internal_encoding()             # <<<<<<<<<<<<<<
 * 
 *         for key, value in feature['properties'].items():
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_session), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_encoding = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":335
 *         encoding = session._get_internal_encoding()
 * 
 *         for key, value in feature['properties'].items():             # <<<<<<<<<<<<<<
 *             log.debug(
 *                 "Looking up %s in %s", key, repr(session._schema_mapping))
 */
  __pyx_t_7 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_feature, __pyx_n_u_properties); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 335, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_1, 0, __pyx_n_s_items, (&__pyx_t_8), (&__pyx_t_9)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XDECREF(__pyx_t_5);
  __pyx_t_5 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_10 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_8, &__pyx_t_7, &__pyx_t_6, &__pyx_t_1, NULL, __pyx_t_9);
    if (unlikely(__pyx_t_10 == 0)) break;
    if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 335, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":336
 * 
 *         for key, value in feature['properties'].items():
 *             log.debug(             # <<<<<<<<<<<<<<
 *                 "Looking up %s in %s", key, repr(session._schema_mapping))
 *             ogr_key = session._schema_mapping[key]
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_log); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 336, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fiona/ogrext.pyx":337
 *         for key, value in feature['properties'].items():
 *             log.debug(
 *                 "Looking up %s in %s", key, repr(session._schema_mapping))             # <<<<<<<<<<<<<<
 *             ogr_key = session._schema_mapping[key]
 * 
 */
    __pyx_t_6 = __pyx_v_session->_schema_mapping;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_12 = PyObject_Repr(__pyx_t_6); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_kp_u_Looking_up_s_in_s, __pyx_v_key, __pyx_t_12};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
      PyObject *__pyx_temp[4] = {__pyx_t_6, __pyx_kp_u_Looking_up_s_in_s, __pyx_v_key, __pyx_t_12};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_Looking_up_s_in_s);
      __Pyx_GIVEREF(__pyx_kp_u_Looking_up_s_in_s);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_10, __pyx_kp_u_Looking_up_s_in_s);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_10, __pyx_v_key);
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_10, __pyx_t_12);
      __pyx_t_12 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 336, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":338
 *             log.debug(
 *                 "Looking up %s in %s", key, repr(session._schema_mapping))
 *             ogr_key = session._schema_mapping[key]             # <<<<<<<<<<<<<<
 * 
 *             schema_type = normalize_field_type(collection.schema['properties'][key])
 */
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_session->_schema_mapping, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_ogr_key, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":340
 *             ogr_key = session._schema_mapping[key]
 * 
 *             schema_type = normalize_field_type(collection.schema['properties'][key])             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Normalizing schema type for key %r in schema %r to %r", key, collection.schema['properties'], schema_type)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_normalize_field_type); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_12 = __Pyx_PyObject_Dict_GetItem(__pyx_t_13, __pyx_n_u_properties); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_13 = __Pyx_PyObject_GetItem(__pyx_t_12, __pyx_v_key); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_1 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_13);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_schema_type, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":342
 *             schema_type = normalize_field_type(collection.schema['properties'][key])
 * 
 *             log.debug("Normalizing schema type for key %r in schema %r to %r", key, collection.schema['properties'], schema_type)             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyObject_Dict_GetItem(__pyx_t_11, __pyx_n_u_properties); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_13);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_13, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_kp_u_Normalizing_schema_type_for_key, __pyx_v_key, __pyx_t_12, __pyx_v_schema_type};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
      PyObject *__pyx_temp[5] = {__pyx_t_11, __pyx_kp_u_Normalizing_schema_type_for_key, __pyx_v_key, __pyx_t_12, __pyx_v_schema_type};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_Normalizing_schema_type_for_key);
      __Pyx_GIVEREF(__pyx_kp_u_Normalizing_schema_type_for_key);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_10, __pyx_kp_u_Normalizing_schema_type_for_key);
      __Pyx_INCREF(__pyx_v_key);
      __Pyx_GIVEREF(__pyx_v_key);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_10, __pyx_v_key);
      __Pyx_GIVEREF(__pyx_t_12);
      PyTuple_SET_ITEM(__pyx_t_6, 2+__pyx_t_10, __pyx_t_12);
      __Pyx_INCREF(__pyx_v_schema_type);
      __Pyx_GIVEREF(__pyx_v_schema_type);
      PyTuple_SET_ITEM(__pyx_t_6, 3+__pyx_t_10, __pyx_v_schema_type);
      __pyx_t_12 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":344
 *             log.debug("Normalizing schema type for key %r in schema %r to %r", key, collection.schema['properties'], schema_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 key_bytes = ogr_key.encode(encoding)
 *             except UnicodeDecodeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      /*try:*/ {

        /* "fiona/ogrext.pyx":345
 * 
 *             try:
 *                 key_bytes = ogr_key.encode(encoding)             # <<<<<<<<<<<<<<
 *             except UnicodeDecodeError:
 *                 log.warning("Failed to encode %s using %s codec", key, encoding)
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_ogr_key, __pyx_n_s_encode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 345, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
          }
        }
        __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_6, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_encoding);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 345, __pyx_L9_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF_SET(__pyx_v_key_bytes, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":344
 *             log.debug("Normalizing schema type for key %r in schema %r to %r", key, collection.schema['properties'], schema_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 key_bytes = ogr_key.encode(encoding)
 *             except UnicodeDecodeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      goto __pyx_L16_try_end;
      __pyx_L9_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fiona/ogrext.pyx":346
 *             try:
 *                 key_bytes = ogr_key.encode(encoding)
 *             except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *                 log.warning("Failed to encode %s using %s codec", key, encoding)
 *                 key_bytes = ogr_key
 */
      __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
      if (__pyx_t_10) {
        __Pyx_AddTraceback("fiona.ogrext.OGRFeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_13, &__pyx_t_6) < 0) __PYX_ERR(0, 346, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_13);
        __Pyx_GOTREF(__pyx_t_6);

        /* "fiona/ogrext.pyx":347
 *                 key_bytes = ogr_key.encode(encoding)
 *             except UnicodeDecodeError:
 *                 log.warning("Failed to encode %s using %s codec", key, encoding)             # <<<<<<<<<<<<<<
 *                 key_bytes = ogr_key
 *             key_c = key_bytes
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 347, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_warning); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 347, __pyx_L11_except_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_17);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_17, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_17)) {
          PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_kp_u_Failed_to_encode_s_using_s_codec, __pyx_v_key, __pyx_v_encoding};
          __pyx_t_12 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 347, __pyx_L11_except_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_12);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
          PyObject *__pyx_temp[4] = {__pyx_t_11, __pyx_kp_u_Failed_to_encode_s_using_s_codec, __pyx_v_key, __pyx_v_encoding};
          __pyx_t_12 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 347, __pyx_L11_except_error)
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_GOTREF(__pyx_t_12);
        } else
        #endif
        {
          __pyx_t_18 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 347, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_18);
          if (__pyx_t_11) {
            __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_11); __pyx_t_11 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Failed_to_encode_s_using_s_codec);
          __Pyx_GIVEREF(__pyx_kp_u_Failed_to_encode_s_using_s_codec);
          PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_10, __pyx_kp_u_Failed_to_encode_s_using_s_codec);
          __Pyx_INCREF(__pyx_v_key);
          __Pyx_GIVEREF(__pyx_v_key);
          PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_10, __pyx_v_key);
          __Pyx_INCREF(__pyx_v_encoding);
          __Pyx_GIVEREF(__pyx_v_encoding);
          PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_10, __pyx_v_encoding);
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_18, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 347, __pyx_L11_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        }
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

        /* "fiona/ogrext.pyx":348
 *             except UnicodeDecodeError:
 *                 log.warning("Failed to encode %s using %s codec", key, encoding)
 *                 key_bytes = ogr_key             # <<<<<<<<<<<<<<
 *             key_c = key_bytes
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)
 */
        __Pyx_INCREF(__pyx_v_ogr_key);
        __Pyx_XDECREF_SET(__pyx_v_key_bytes, __pyx_v_ogr_key);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L10_exception_handled;
      }
      goto __pyx_L11_except_error;
      __pyx_L11_except_error:;

      /* "fiona/ogrext.pyx":344
 *             log.debug("Normalizing schema type for key %r in schema %r to %r", key, collection.schema['properties'], schema_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 key_bytes = ogr_key.encode(encoding)
 *             except UnicodeDecodeError:
 */
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      goto __pyx_L1_error;
      __pyx_L10_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ExceptionReset(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_L16_try_end:;
    }

    /* "fiona/ogrext.pyx":349
 *                 log.warning("Failed to encode %s using %s codec", key, encoding)
 *                 key_bytes = ogr_key
 *             key_c = key_bytes             # <<<<<<<<<<<<<<
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)
 *             if i < 0:
 */
    __Pyx_INCREF(__pyx_v_key_bytes);
    __Pyx_XDECREF_SET(__pyx_v_key_c, __pyx_v_key_bytes);

    /* "fiona/ogrext.pyx":350
 *                 key_bytes = ogr_key
 *             key_c = key_bytes
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)             # <<<<<<<<<<<<<<
 *             if i < 0:
 *                 continue
 */
    __pyx_t_19 = __Pyx_PyObject_AsWritableString(__pyx_v_key_c); if (unlikely((!__pyx_t_19) && PyErr_Occurred())) __PYX_ERR(0, 350, __pyx_L1_error)
    __pyx_v_i = OGR_F_GetFieldIndex(__pyx_v_cogr_feature, __pyx_t_19);

    /* "fiona/ogrext.pyx":351
 *             key_c = key_bytes
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)
 *             if i < 0:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    __pyx_t_4 = ((__pyx_v_i < 0) != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":352
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)
 *             if i < 0:
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             # Special case: serialize dicts to assist OGR.
 */
      goto __pyx_L7_continue;

      /* "fiona/ogrext.pyx":351
 *             key_c = key_bytes
 *             i = OGR_F_GetFieldIndex(cogr_feature, key_c)
 *             if i < 0:             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
    }

    /* "fiona/ogrext.pyx":355
 * 
 *             # Special case: serialize dicts to assist OGR.
 *             if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *                 value = json.dumps(value)
 * 
 */
    __pyx_t_4 = PyDict_Check(__pyx_v_value); 
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":356
 *             # Special case: serialize dicts to assist OGR.
 *             if isinstance(value, dict):
 *                 value = json.dumps(value)             # <<<<<<<<<<<<<<
 * 
 *             # Continue over the standard OGR types.
 */
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_json); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_dumps); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_13, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_value);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 356, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "fiona/ogrext.pyx":355
 * 
 *             # Special case: serialize dicts to assist OGR.
 *             if isinstance(value, dict):             # <<<<<<<<<<<<<<
 *                 value = json.dumps(value)
 * 
 */
    }

    /* "fiona/ogrext.pyx":359
 * 
 *             # Continue over the standard OGR types.
 *             if isinstance(value, integer_types):             # <<<<<<<<<<<<<<
 * 
 *                 log.debug("Setting field %r, type %r, to value %r", i, schema_type, value)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_integer_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 359, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":361
 *             if isinstance(value, integer_types):
 * 
 *                 log.debug("Setting field %r, type %r, to value %r", i, schema_type, value)             # <<<<<<<<<<<<<<
 * 
 *                 if schema_type == 'int32':
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_12 = NULL;
      __pyx_t_10 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_13);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_13, function);
          __pyx_t_10 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_13)) {
        PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_kp_u_Setting_field_r_type_r_to_value, __pyx_t_1, __pyx_v_schema_type, __pyx_v_value};
        __pyx_t_6 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
        PyObject *__pyx_temp[5] = {__pyx_t_12, __pyx_kp_u_Setting_field_r_type_r_to_value, __pyx_t_1, __pyx_v_schema_type, __pyx_v_value};
        __pyx_t_6 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_10, 4+__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      } else
      #endif
      {
        __pyx_t_17 = PyTuple_New(4+__pyx_t_10); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_17);
        if (__pyx_t_12) {
          __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_12); __pyx_t_12 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Setting_field_r_type_r_to_value);
        __Pyx_GIVEREF(__pyx_kp_u_Setting_field_r_type_r_to_value);
        PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_10, __pyx_kp_u_Setting_field_r_type_r_to_value);
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_10, __pyx_t_1);
        __Pyx_INCREF(__pyx_v_schema_type);
        __Pyx_GIVEREF(__pyx_v_schema_type);
        PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_10, __pyx_v_schema_type);
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_17, 3+__pyx_t_10, __pyx_v_value);
        __pyx_t_1 = 0;
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_17, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 361, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
      }
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fiona/ogrext.pyx":363
 *                 log.debug("Setting field %r, type %r, to value %r", i, schema_type, value)
 * 
 *                 if schema_type == 'int32':             # <<<<<<<<<<<<<<
 *                     OGR_F_SetFieldInteger(cogr_feature, i, value)
 *                 else:
 */
      __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_int32, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 363, __pyx_L1_error)
      if (__pyx_t_4) {

        /* "fiona/ogrext.pyx":364
 * 
 *                 if schema_type == 'int32':
 *                     OGR_F_SetFieldInteger(cogr_feature, i, value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     OGR_F_SetFieldInteger64(cogr_feature, i, value)
 */
        __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_value); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 364, __pyx_L1_error)
        OGR_F_SetFieldInteger(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_10);

        /* "fiona/ogrext.pyx":363
 *                 log.debug("Setting field %r, type %r, to value %r", i, schema_type, value)
 * 
 *                 if schema_type == 'int32':             # <<<<<<<<<<<<<<
 *                     OGR_F_SetFieldInteger(cogr_feature, i, value)
 *                 else:
 */
        goto __pyx_L22;
      }

      /* "fiona/ogrext.pyx":366
 *                     OGR_F_SetFieldInteger(cogr_feature, i, value)
 *                 else:
 *                     OGR_F_SetFieldInteger64(cogr_feature, i, value)             # <<<<<<<<<<<<<<
 * 
 *             elif isinstance(value, float):
 */
      /*else*/ {
        __pyx_t_20 = __Pyx_PyInt_As_PY_LONG_LONG(__pyx_v_value); if (unlikely((__pyx_t_20 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(0, 366, __pyx_L1_error)
        OGR_F_SetFieldInteger64(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_20);
      }
      __pyx_L22:;

      /* "fiona/ogrext.pyx":359
 * 
 *             # Continue over the standard OGR types.
 *             if isinstance(value, integer_types):             # <<<<<<<<<<<<<<
 * 
 *                 log.debug("Setting field %r, type %r, to value %r", i, schema_type, value)
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":368
 *                     OGR_F_SetFieldInteger64(cogr_feature, i, value)
 * 
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)
 */
    __pyx_t_4 = PyFloat_Check(__pyx_v_value); 
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":369
 * 
 *             elif isinstance(value, float):
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)             # <<<<<<<<<<<<<<
 *             elif (isinstance(value, string_types)
 *             and schema_type in ['date', 'time', 'datetime']):
 */
      __pyx_t_21 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_21 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 369, __pyx_L1_error)
      OGR_F_SetFieldDouble(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_21);

      /* "fiona/ogrext.pyx":368
 *                     OGR_F_SetFieldInteger64(cogr_feature, i, value)
 * 
 *             elif isinstance(value, float):             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":370
 *             elif isinstance(value, float):
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)             # <<<<<<<<<<<<<<
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_string_types); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 370, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_22 = (__pyx_t_4 != 0);
    if (__pyx_t_22) {
    } else {
      __pyx_t_3 = __pyx_t_22;
      goto __pyx_L23_bool_binop_done;
    }

    /* "fiona/ogrext.pyx":371
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)
 *             and schema_type in ['date', 'time', 'datetime']):             # <<<<<<<<<<<<<<
 *                 if schema_type == 'date':
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 */
    __Pyx_INCREF(__pyx_v_schema_type);
    __pyx_t_6 = __pyx_v_schema_type;
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_date, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_22 = __pyx_t_4;
      goto __pyx_L25_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_time, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
    if (!__pyx_t_4) {
    } else {
      __pyx_t_22 = __pyx_t_4;
      goto __pyx_L25_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_t_6, __pyx_n_u_datetime, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 371, __pyx_L1_error)
    __pyx_t_22 = __pyx_t_4;
    __pyx_L25_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__pyx_t_22 != 0);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L23_bool_binop_done:;

    /* "fiona/ogrext.pyx":370
 *             elif isinstance(value, float):
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)             # <<<<<<<<<<<<<<
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':
 */
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":372
 *             elif (isinstance(value, string_types)
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':             # <<<<<<<<<<<<<<
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 *                 elif schema_type == 'time':
 */
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_date, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 372, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "fiona/ogrext.pyx":373
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)             # <<<<<<<<<<<<<<
 *                 elif schema_type == 'time':
 *                     y, m, d, hh, mm, ss, ff = parse_time(value)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_parse_date); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_17 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_17)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_17);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
          }
        }
        __pyx_t_6 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_17, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_value);
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 373, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 7)) {
            if (size > 7) __Pyx_RaiseTooManyValuesError(7);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 373, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
            __pyx_t_18 = PyTuple_GET_ITEM(sequence, 4); 
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 5); 
            __pyx_t_23 = PyTuple_GET_ITEM(sequence, 6); 
          } else {
            __pyx_t_13 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
            __pyx_t_18 = PyList_GET_ITEM(sequence, 4); 
            __pyx_t_11 = PyList_GET_ITEM(sequence, 5); 
            __pyx_t_23 = PyList_GET_ITEM(sequence, 6); 
          }
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_23);
          #else
          {
            Py_ssize_t i;
            PyObject** temps[7] = {&__pyx_t_13,&__pyx_t_17,&__pyx_t_1,&__pyx_t_12,&__pyx_t_18,&__pyx_t_11,&__pyx_t_23};
            for (i=0; i < 7; i++) {
              PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 373, __pyx_L1_error)
              __Pyx_GOTREF(item);
              *(temps[i]) = item;
            }
          }
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          PyObject** temps[7] = {&__pyx_t_13,&__pyx_t_17,&__pyx_t_1,&__pyx_t_12,&__pyx_t_18,&__pyx_t_11,&__pyx_t_23};
          __pyx_t_24 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 373, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_24);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_25 = Py_TYPE(__pyx_t_24)->tp_iternext;
          for (index=0; index < 7; index++) {
            PyObject* item = __pyx_t_25(__pyx_t_24); if (unlikely(!item)) goto __pyx_L29_unpacking_failed;
            __Pyx_GOTREF(item);
            *(temps[index]) = item;
          }
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_25(__pyx_t_24), 7) < 0) __PYX_ERR(0, 373, __pyx_L1_error)
          __pyx_t_25 = NULL;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          goto __pyx_L30_unpacking_done;
          __pyx_L29_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          __pyx_t_25 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 373, __pyx_L1_error)
          __pyx_L30_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_1);
        __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_hh, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_XDECREF_SET(__pyx_v_mm, __pyx_t_18);
        __pyx_t_18 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ss, __pyx_t_11);
        __pyx_t_11 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ff, __pyx_t_23);
        __pyx_t_23 = 0;

        /* "fiona/ogrext.pyx":372
 *             elif (isinstance(value, string_types)
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':             # <<<<<<<<<<<<<<
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 *                 elif schema_type == 'time':
 */
        goto __pyx_L28;
      }

      /* "fiona/ogrext.pyx":374
 *                 if schema_type == 'date':
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 *                 elif schema_type == 'time':             # <<<<<<<<<<<<<<
 *                     y, m, d, hh, mm, ss, ff = parse_time(value)
 *                 else:
 */
      __pyx_t_3 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_time, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 374, __pyx_L1_error)
      if (__pyx_t_3) {

        /* "fiona/ogrext.pyx":375
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 *                 elif schema_type == 'time':
 *                     y, m, d, hh, mm, ss, ff = parse_time(value)             # <<<<<<<<<<<<<<
 *                 else:
 *                     y, m, d, hh, mm, ss, ff = parse_datetime(value)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_parse_time); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_23);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_23);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_23, function);
          }
        }
        __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_23, __pyx_t_11, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_23, __pyx_v_value);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 375, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 7)) {
            if (size > 7) __Pyx_RaiseTooManyValuesError(7);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 375, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_23 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_18 = PyTuple_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 4); 
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 5); 
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 6); 
          } else {
            __pyx_t_23 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_11 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_18 = PyList_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 4); 
            __pyx_t_17 = PyList_GET_ITEM(sequence, 5); 
            __pyx_t_13 = PyList_GET_ITEM(sequence, 6); 
          }
          __Pyx_INCREF(__pyx_t_23);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_13);
          #else
          {
            Py_ssize_t i;
            PyObject** temps[7] = {&__pyx_t_23,&__pyx_t_11,&__pyx_t_18,&__pyx_t_12,&__pyx_t_1,&__pyx_t_17,&__pyx_t_13};
            for (i=0; i < 7; i++) {
              PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 375, __pyx_L1_error)
              __Pyx_GOTREF(item);
              *(temps[i]) = item;
            }
          }
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          PyObject** temps[7] = {&__pyx_t_23,&__pyx_t_11,&__pyx_t_18,&__pyx_t_12,&__pyx_t_1,&__pyx_t_17,&__pyx_t_13};
          __pyx_t_24 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 375, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_24);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_25 = Py_TYPE(__pyx_t_24)->tp_iternext;
          for (index=0; index < 7; index++) {
            PyObject* item = __pyx_t_25(__pyx_t_24); if (unlikely(!item)) goto __pyx_L31_unpacking_failed;
            __Pyx_GOTREF(item);
            *(temps[index]) = item;
          }
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_25(__pyx_t_24), 7) < 0) __PYX_ERR(0, 375, __pyx_L1_error)
          __pyx_t_25 = NULL;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          goto __pyx_L32_unpacking_done;
          __pyx_L31_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          __pyx_t_25 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 375, __pyx_L1_error)
          __pyx_L32_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_23);
        __pyx_t_23 = 0;
        __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_11);
        __pyx_t_11 = 0;
        __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_18);
        __pyx_t_18 = 0;
        __Pyx_XDECREF_SET(__pyx_v_hh, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_XDECREF_SET(__pyx_v_mm, __pyx_t_1);
        __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ss, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ff, __pyx_t_13);
        __pyx_t_13 = 0;

        /* "fiona/ogrext.pyx":374
 *                 if schema_type == 'date':
 *                     y, m, d, hh, mm, ss, ff = parse_date(value)
 *                 elif schema_type == 'time':             # <<<<<<<<<<<<<<
 *                     y, m, d, hh, mm, ss, ff = parse_time(value)
 *                 else:
 */
        goto __pyx_L28;
      }

      /* "fiona/ogrext.pyx":377
 *                     y, m, d, hh, mm, ss, ff = parse_time(value)
 *                 else:
 *                     y, m, d, hh, mm, ss, ff = parse_datetime(value)             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_parse_datetime); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_17 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
          __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_13);
          if (likely(__pyx_t_17)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
            __Pyx_INCREF(__pyx_t_17);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_13, function);
          }
        }
        __pyx_t_6 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_17, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_value);
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 377, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
          PyObject* sequence = __pyx_t_6;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 7)) {
            if (size > 7) __Pyx_RaiseTooManyValuesError(7);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 377, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyTuple_GET_ITEM(sequence, 1); 
            __pyx_t_1 = PyTuple_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyTuple_GET_ITEM(sequence, 3); 
            __pyx_t_18 = PyTuple_GET_ITEM(sequence, 4); 
            __pyx_t_11 = PyTuple_GET_ITEM(sequence, 5); 
            __pyx_t_23 = PyTuple_GET_ITEM(sequence, 6); 
          } else {
            __pyx_t_13 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_17 = PyList_GET_ITEM(sequence, 1); 
            __pyx_t_1 = PyList_GET_ITEM(sequence, 2); 
            __pyx_t_12 = PyList_GET_ITEM(sequence, 3); 
            __pyx_t_18 = PyList_GET_ITEM(sequence, 4); 
            __pyx_t_11 = PyList_GET_ITEM(sequence, 5); 
            __pyx_t_23 = PyList_GET_ITEM(sequence, 6); 
          }
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_18);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_23);
          #else
          {
            Py_ssize_t i;
            PyObject** temps[7] = {&__pyx_t_13,&__pyx_t_17,&__pyx_t_1,&__pyx_t_12,&__pyx_t_18,&__pyx_t_11,&__pyx_t_23};
            for (i=0; i < 7; i++) {
              PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 377, __pyx_L1_error)
              __Pyx_GOTREF(item);
              *(temps[i]) = item;
            }
          }
          #endif
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        } else {
          Py_ssize_t index = -1;
          PyObject** temps[7] = {&__pyx_t_13,&__pyx_t_17,&__pyx_t_1,&__pyx_t_12,&__pyx_t_18,&__pyx_t_11,&__pyx_t_23};
          __pyx_t_24 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_24)) __PYX_ERR(0, 377, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_24);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_25 = Py_TYPE(__pyx_t_24)->tp_iternext;
          for (index=0; index < 7; index++) {
            PyObject* item = __pyx_t_25(__pyx_t_24); if (unlikely(!item)) goto __pyx_L33_unpacking_failed;
            __Pyx_GOTREF(item);
            *(temps[index]) = item;
          }
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_25(__pyx_t_24), 7) < 0) __PYX_ERR(0, 377, __pyx_L1_error)
          __pyx_t_25 = NULL;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          goto __pyx_L34_unpacking_done;
          __pyx_L33_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_24); __pyx_t_24 = 0;
          __pyx_t_25 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 377, __pyx_L1_error)
          __pyx_L34_unpacking_done:;
        }
        __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_13);
        __pyx_t_13 = 0;
        __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_17);
        __pyx_t_17 = 0;
        __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_1);
        __pyx_t_1 = 0;
        __Pyx_XDECREF_SET(__pyx_v_hh, __pyx_t_12);
        __pyx_t_12 = 0;
        __Pyx_XDECREF_SET(__pyx_v_mm, __pyx_t_18);
        __pyx_t_18 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ss, __pyx_t_11);
        __pyx_t_11 = 0;
        __Pyx_XDECREF_SET(__pyx_v_ff, __pyx_t_23);
        __pyx_t_23 = 0;
      }
      __pyx_L28:;

      /* "fiona/ogrext.pyx":379
 *                     y, m, d, hh, mm, ss, ff = parse_datetime(value)
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)             # <<<<<<<<<<<<<<
 *             elif (isinstance(value, datetime.date)
 *             and schema_type == 'date'):
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_y); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_26 = __Pyx_PyInt_As_int(__pyx_v_m); if (unlikely((__pyx_t_26 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_27 = __Pyx_PyInt_As_int(__pyx_v_d); if (unlikely((__pyx_t_27 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_28 = __Pyx_PyInt_As_int(__pyx_v_hh); if (unlikely((__pyx_t_28 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_29 = __Pyx_PyInt_As_int(__pyx_v_mm); if (unlikely((__pyx_t_29 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)
      __pyx_t_30 = __Pyx_PyInt_As_int(__pyx_v_ss); if (unlikely((__pyx_t_30 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 379, __pyx_L1_error)

      /* "fiona/ogrext.pyx":378
 *                 else:
 *                     y, m, d, hh, mm, ss, ff = parse_datetime(value)
 *                 OGR_F_SetFieldDateTime(             # <<<<<<<<<<<<<<
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.date)
 */
      OGR_F_SetFieldDateTime(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_10, __pyx_t_26, __pyx_t_27, __pyx_t_28, __pyx_t_29, __pyx_t_30, 0);

      /* "fiona/ogrext.pyx":370
 *             elif isinstance(value, float):
 *                 OGR_F_SetFieldDouble(cogr_feature, i, value)
 *             elif (isinstance(value, string_types)             # <<<<<<<<<<<<<<
 *             and schema_type in ['date', 'time', 'datetime']):
 *                 if schema_type == 'date':
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":380
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.date)             # <<<<<<<<<<<<<<
 *             and schema_type == 'date'):
 *                 y, m, d = value.year, value.month, value.day
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_date); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_23); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 380, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_22 = (__pyx_t_4 != 0);
    if (__pyx_t_22) {
    } else {
      __pyx_t_3 = __pyx_t_22;
      goto __pyx_L35_bool_binop_done;
    }

    /* "fiona/ogrext.pyx":381
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.date)
 *             and schema_type == 'date'):             # <<<<<<<<<<<<<<
 *                 y, m, d = value.year, value.month, value.day
 *                 OGR_F_SetFieldDateTime(
 */
    __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_date, Py_EQ)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 381, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_22;
    __pyx_L35_bool_binop_done:;

    /* "fiona/ogrext.pyx":380
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.date)             # <<<<<<<<<<<<<<
 *             and schema_type == 'date'):
 *                 y, m, d = value.year, value.month, value.day
 */
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":382
 *             elif (isinstance(value, datetime.date)
 *             and schema_type == 'date'):
 *                 y, m, d = value.year, value.month, value.day             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 */
      __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_year); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_month); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_day); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 382, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_23);
      __pyx_t_23 = 0;
      __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "fiona/ogrext.pyx":384
 *                 y, m, d = value.year, value.month, value.day
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)             # <<<<<<<<<<<<<<
 *             elif (isinstance(value, datetime.datetime)
 *             and schema_type == 'datetime'):
 */
      __pyx_t_30 = __Pyx_PyInt_As_int(__pyx_v_y); if (unlikely((__pyx_t_30 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 384, __pyx_L1_error)
      __pyx_t_29 = __Pyx_PyInt_As_int(__pyx_v_m); if (unlikely((__pyx_t_29 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 384, __pyx_L1_error)
      __pyx_t_28 = __Pyx_PyInt_As_int(__pyx_v_d); if (unlikely((__pyx_t_28 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 384, __pyx_L1_error)

      /* "fiona/ogrext.pyx":383
 *             and schema_type == 'date'):
 *                 y, m, d = value.year, value.month, value.day
 *                 OGR_F_SetFieldDateTime(             # <<<<<<<<<<<<<<
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 *             elif (isinstance(value, datetime.datetime)
 */
      OGR_F_SetFieldDateTime(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_30, __pyx_t_29, __pyx_t_28, 0, 0, 0, 0);

      /* "fiona/ogrext.pyx":380
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.date)             # <<<<<<<<<<<<<<
 *             and schema_type == 'date'):
 *                 y, m, d = value.year, value.month, value.day
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":385
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 *             elif (isinstance(value, datetime.datetime)             # <<<<<<<<<<<<<<
 *             and schema_type == 'datetime'):
 *                 y, m, d = value.year, value.month, value.day
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_datetime); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_22 = PyObject_IsInstance(__pyx_v_value, __pyx_t_6); if (unlikely(__pyx_t_22 == ((int)-1))) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = (__pyx_t_22 != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L37_bool_binop_done;
    }

    /* "fiona/ogrext.pyx":386
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 *             elif (isinstance(value, datetime.datetime)
 *             and schema_type == 'datetime'):             # <<<<<<<<<<<<<<
 *                 y, m, d = value.year, value.month, value.day
 *                 hh, mm, ss = value.hour, value.minute, value.second
 */
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_datetime, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 386, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_4;
    __pyx_L37_bool_binop_done:;

    /* "fiona/ogrext.pyx":385
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 *             elif (isinstance(value, datetime.datetime)             # <<<<<<<<<<<<<<
 *             and schema_type == 'datetime'):
 *                 y, m, d = value.year, value.month, value.day
 */
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":387
 *             elif (isinstance(value, datetime.datetime)
 *             and schema_type == 'datetime'):
 *                 y, m, d = value.year, value.month, value.day             # <<<<<<<<<<<<<<
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_year); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_month); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_day); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 387, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_XDECREF_SET(__pyx_v_y, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_m, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_d, __pyx_t_23);
      __pyx_t_23 = 0;

      /* "fiona/ogrext.pyx":388
 *             and schema_type == 'datetime'):
 *                 y, m, d = value.year, value.month, value.day
 *                 hh, mm, ss = value.hour, value.minute, value.second             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 */
      __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_hour); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_minute); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_second); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 388, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_hh, __pyx_t_23);
      __pyx_t_23 = 0;
      __Pyx_XDECREF_SET(__pyx_v_mm, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ss, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "fiona/ogrext.pyx":390
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)             # <<<<<<<<<<<<<<
 *             elif (isinstance(value, datetime.time)
 *             and schema_type == 'time'):
 */
      __pyx_t_28 = __Pyx_PyInt_As_int(__pyx_v_y); if (unlikely((__pyx_t_28 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
      __pyx_t_29 = __Pyx_PyInt_As_int(__pyx_v_m); if (unlikely((__pyx_t_29 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
      __pyx_t_30 = __Pyx_PyInt_As_int(__pyx_v_d); if (unlikely((__pyx_t_30 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
      __pyx_t_27 = __Pyx_PyInt_As_int(__pyx_v_hh); if (unlikely((__pyx_t_27 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
      __pyx_t_26 = __Pyx_PyInt_As_int(__pyx_v_mm); if (unlikely((__pyx_t_26 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_ss); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 390, __pyx_L1_error)

      /* "fiona/ogrext.pyx":389
 *                 y, m, d = value.year, value.month, value.day
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(             # <<<<<<<<<<<<<<
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.time)
 */
      OGR_F_SetFieldDateTime(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_28, __pyx_t_29, __pyx_t_30, __pyx_t_27, __pyx_t_26, __pyx_t_10, 0);

      /* "fiona/ogrext.pyx":385
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, 0, 0, 0, 0)
 *             elif (isinstance(value, datetime.datetime)             # <<<<<<<<<<<<<<
 *             and schema_type == 'datetime'):
 *                 y, m, d = value.year, value.month, value.day
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":391
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.time)             # <<<<<<<<<<<<<<
 *             and schema_type == 'time'):
 *                 hh, mm, ss = value.hour, value.minute, value.second
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_datetime); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_time); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_4 = PyObject_IsInstance(__pyx_v_value, __pyx_t_11); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_22 = (__pyx_t_4 != 0);
    if (__pyx_t_22) {
    } else {
      __pyx_t_3 = __pyx_t_22;
      goto __pyx_L39_bool_binop_done;
    }

    /* "fiona/ogrext.pyx":392
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.time)
 *             and schema_type == 'time'):             # <<<<<<<<<<<<<<
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(
 */
    __pyx_t_22 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_time, Py_EQ)); if (unlikely(__pyx_t_22 < 0)) __PYX_ERR(0, 392, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_22;
    __pyx_L39_bool_binop_done:;

    /* "fiona/ogrext.pyx":391
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.time)             # <<<<<<<<<<<<<<
 *             and schema_type == 'time'):
 *                 hh, mm, ss = value.hour, value.minute, value.second
 */
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":393
 *             elif (isinstance(value, datetime.time)
 *             and schema_type == 'time'):
 *                 hh, mm, ss = value.hour, value.minute, value.second             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_hour); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_minute); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_second); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 393, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_23);
      __Pyx_XDECREF_SET(__pyx_v_hh, __pyx_t_11);
      __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_mm, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ss, __pyx_t_23);
      __pyx_t_23 = 0;

      /* "fiona/ogrext.pyx":395
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)             # <<<<<<<<<<<<<<
 *             elif isinstance(value, bytes) and schema_type == "bytes":
 *                 string_c = value
 */
      __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_hh); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L1_error)
      __pyx_t_26 = __Pyx_PyInt_As_int(__pyx_v_mm); if (unlikely((__pyx_t_26 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L1_error)
      __pyx_t_27 = __Pyx_PyInt_As_int(__pyx_v_ss); if (unlikely((__pyx_t_27 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 395, __pyx_L1_error)

      /* "fiona/ogrext.pyx":394
 *             and schema_type == 'time'):
 *                 hh, mm, ss = value.hour, value.minute, value.second
 *                 OGR_F_SetFieldDateTime(             # <<<<<<<<<<<<<<
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)
 *             elif isinstance(value, bytes) and schema_type == "bytes":
 */
      OGR_F_SetFieldDateTime(__pyx_v_cogr_feature, __pyx_v_i, 0, 0, 0, __pyx_t_10, __pyx_t_26, __pyx_t_27, 0);

      /* "fiona/ogrext.pyx":391
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, y, m, d, hh, mm, ss, 0)
 *             elif (isinstance(value, datetime.time)             # <<<<<<<<<<<<<<
 *             and schema_type == 'time'):
 *                 hh, mm, ss = value.hour, value.minute, value.second
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":396
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)
 *             elif isinstance(value, bytes) and schema_type == "bytes":             # <<<<<<<<<<<<<<
 *                 string_c = value
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 */
    __pyx_t_22 = PyBytes_Check(__pyx_v_value); 
    __pyx_t_4 = (__pyx_t_22 != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L41_bool_binop_done;
    }
    __pyx_t_4 = (__Pyx_PyUnicode_Equals(__pyx_v_schema_type, __pyx_n_u_bytes, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 396, __pyx_L1_error)
    __pyx_t_3 = __pyx_t_4;
    __pyx_L41_bool_binop_done:;
    if (__pyx_t_3) {

      /* "fiona/ogrext.pyx":397
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)
 *             elif isinstance(value, bytes) and schema_type == "bytes":
 *                 string_c = value             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 *                     <unsigned char*>string_c)
 */
      __pyx_t_31 = __Pyx_PyObject_AsString(__pyx_v_value); if (unlikely((!__pyx_t_31) && PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
      __pyx_v_string_c = __pyx_t_31;

      /* "fiona/ogrext.pyx":398
 *             elif isinstance(value, bytes) and schema_type == "bytes":
 *                 string_c = value
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),             # <<<<<<<<<<<<<<
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):
 */
      __pyx_t_32 = PyObject_Length(__pyx_v_value); if (unlikely(__pyx_t_32 == ((Py_ssize_t)-1))) __PYX_ERR(0, 398, __pyx_L1_error)

      /* "fiona/ogrext.pyx":399
 *                 string_c = value
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 *                     <unsigned char*>string_c)             # <<<<<<<<<<<<<<
 *             elif isinstance(value, string_types):
 *                 try:
 */
      OGR_F_SetFieldBinary(__pyx_v_cogr_feature, __pyx_v_i, __pyx_t_32, ((unsigned char *)__pyx_v_string_c));

      /* "fiona/ogrext.pyx":396
 *                 OGR_F_SetFieldDateTime(
 *                     cogr_feature, i, 0, 0, 0, hh, mm, ss, 0)
 *             elif isinstance(value, bytes) and schema_type == "bytes":             # <<<<<<<<<<<<<<
 *                 string_c = value
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":400
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value_bytes = value.encode(encoding)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_23, __pyx_n_s_string_types); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __pyx_t_3 = PyObject_IsInstance(__pyx_v_value, __pyx_t_23); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "fiona/ogrext.pyx":401
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value_bytes = value.encode(encoding)
 *                 except UnicodeDecodeError:
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_16, &__pyx_t_15, &__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_16);
        __Pyx_XGOTREF(__pyx_t_15);
        __Pyx_XGOTREF(__pyx_t_14);
        /*try:*/ {

          /* "fiona/ogrext.pyx":402
 *             elif isinstance(value, string_types):
 *                 try:
 *                     value_bytes = value.encode(encoding)             # <<<<<<<<<<<<<<
 *                 except UnicodeDecodeError:
 *                     log.warning(
 */
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 402, __pyx_L43_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_11 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_6);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_6, function);
            }
          }
          __pyx_t_23 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_11, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_encoding);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 402, __pyx_L43_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF_SET(__pyx_v_value_bytes, __pyx_t_23);
          __pyx_t_23 = 0;

          /* "fiona/ogrext.pyx":401
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value_bytes = value.encode(encoding)
 *                 except UnicodeDecodeError:
 */
        }
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        goto __pyx_L50_try_end;
        __pyx_L43_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "fiona/ogrext.pyx":403
 *                 try:
 *                     value_bytes = value.encode(encoding)
 *                 except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *                     log.warning(
 *                         "Failed to encode %s using %s codec", value, encoding)
 */
        __pyx_t_27 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
        if (__pyx_t_27) {
          __Pyx_AddTraceback("fiona.ogrext.OGRFeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_23, &__pyx_t_6, &__pyx_t_11) < 0) __PYX_ERR(0, 403, __pyx_L45_except_error)
          __Pyx_GOTREF(__pyx_t_23);
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_GOTREF(__pyx_t_11);

          /* "fiona/ogrext.pyx":404
 *                     value_bytes = value.encode(encoding)
 *                 except UnicodeDecodeError:
 *                     log.warning(             # <<<<<<<<<<<<<<
 *                         "Failed to encode %s using %s codec", value, encoding)
 *                     value_bytes = value
 */
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_log); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 404, __pyx_L45_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 404, __pyx_L45_except_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "fiona/ogrext.pyx":405
 *                 except UnicodeDecodeError:
 *                     log.warning(
 *                         "Failed to encode %s using %s codec", value, encoding)             # <<<<<<<<<<<<<<
 *                     value_bytes = value
 *                 string_c = value_bytes
 */
          __pyx_t_12 = NULL;
          __pyx_t_27 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_27 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_kp_u_Failed_to_encode_s_using_s_codec, __pyx_v_value, __pyx_v_encoding};
            __pyx_t_18 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_27, 3+__pyx_t_27); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 404, __pyx_L45_except_error)
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_GOTREF(__pyx_t_18);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
            PyObject *__pyx_temp[4] = {__pyx_t_12, __pyx_kp_u_Failed_to_encode_s_using_s_codec, __pyx_v_value, __pyx_v_encoding};
            __pyx_t_18 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_27, 3+__pyx_t_27); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 404, __pyx_L45_except_error)
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_GOTREF(__pyx_t_18);
          } else
          #endif
          {
            __pyx_t_17 = PyTuple_New(3+__pyx_t_27); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 404, __pyx_L45_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            if (__pyx_t_12) {
              __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_12); __pyx_t_12 = NULL;
            }
            __Pyx_INCREF(__pyx_kp_u_Failed_to_encode_s_using_s_codec);
            __Pyx_GIVEREF(__pyx_kp_u_Failed_to_encode_s_using_s_codec);
            PyTuple_SET_ITEM(__pyx_t_17, 0+__pyx_t_27, __pyx_kp_u_Failed_to_encode_s_using_s_codec);
            __Pyx_INCREF(__pyx_v_value);
            __Pyx_GIVEREF(__pyx_v_value);
            PyTuple_SET_ITEM(__pyx_t_17, 1+__pyx_t_27, __pyx_v_value);
            __Pyx_INCREF(__pyx_v_encoding);
            __Pyx_GIVEREF(__pyx_v_encoding);
            PyTuple_SET_ITEM(__pyx_t_17, 2+__pyx_t_27, __pyx_v_encoding);
            __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_17, NULL); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 404, __pyx_L45_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;

          /* "fiona/ogrext.pyx":406
 *                     log.warning(
 *                         "Failed to encode %s using %s codec", value, encoding)
 *                     value_bytes = value             # <<<<<<<<<<<<<<
 *                 string_c = value_bytes
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)
 */
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_XDECREF_SET(__pyx_v_value_bytes, __pyx_v_value);
          __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          goto __pyx_L44_exception_handled;
        }
        goto __pyx_L45_except_error;
        __pyx_L45_except_error:;

        /* "fiona/ogrext.pyx":401
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):
 *                 try:             # <<<<<<<<<<<<<<
 *                     value_bytes = value.encode(encoding)
 *                 except UnicodeDecodeError:
 */
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_15, __pyx_t_14);
        goto __pyx_L1_error;
        __pyx_L44_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_16);
        __Pyx_XGIVEREF(__pyx_t_15);
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_ExceptionReset(__pyx_t_16, __pyx_t_15, __pyx_t_14);
        __pyx_L50_try_end:;
      }

      /* "fiona/ogrext.pyx":407
 *                         "Failed to encode %s using %s codec", value, encoding)
 *                     value_bytes = value
 *                 string_c = value_bytes             # <<<<<<<<<<<<<<
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)
 *             elif value is None:
 */
      __pyx_t_31 = __Pyx_PyObject_AsString(__pyx_v_value_bytes); if (unlikely((!__pyx_t_31) && PyErr_Occurred())) __PYX_ERR(0, 407, __pyx_L1_error)
      __pyx_v_string_c = __pyx_t_31;

      /* "fiona/ogrext.pyx":408
 *                     value_bytes = value
 *                 string_c = value_bytes
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)             # <<<<<<<<<<<<<<
 *             elif value is None:
 *                 set_field_null(cogr_feature, i)
 */
      OGR_F_SetFieldString(__pyx_v_cogr_feature, __pyx_v_i, __pyx_v_string_c);

      /* "fiona/ogrext.pyx":400
 *                 OGR_F_SetFieldBinary(cogr_feature, i, len(value),
 *                     <unsigned char*>string_c)
 *             elif isinstance(value, string_types):             # <<<<<<<<<<<<<<
 *                 try:
 *                     value_bytes = value.encode(encoding)
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":409
 *                 string_c = value_bytes
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)
 *             elif value is None:             # <<<<<<<<<<<<<<
 *                 set_field_null(cogr_feature, i)
 *             else:
 */
    __pyx_t_4 = (__pyx_v_value == Py_None);
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (likely(__pyx_t_3)) {

      /* "fiona/ogrext.pyx":410
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)
 *             elif value is None:
 *                 set_field_null(cogr_feature, i)             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError("Invalid field type %s" % type(value))
 */
      __pyx_f_5fiona_5_shim_set_field_null(__pyx_v_cogr_feature, __pyx_v_i);

      /* "fiona/ogrext.pyx":409
 *                 string_c = value_bytes
 *                 OGR_F_SetFieldString(cogr_feature, i, string_c)
 *             elif value is None:             # <<<<<<<<<<<<<<
 *                 set_field_null(cogr_feature, i)
 *             else:
 */
      goto __pyx_L21;
    }

    /* "fiona/ogrext.pyx":412
 *                 set_field_null(cogr_feature, i)
 *             else:
 *                 raise ValueError("Invalid field type %s" % type(value))             # <<<<<<<<<<<<<<
 *             log.debug("Set field %s: %r" % (key, value))
 *         return cogr_feature
 */
    /*else*/ {
      __pyx_t_11 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Invalid_field_type_s, ((PyObject *)Py_TYPE(__pyx_v_value))); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 412, __pyx_L1_error)
    }
    __pyx_L21:;

    /* "fiona/ogrext.pyx":413
 *             else:
 *                 raise ValueError("Invalid field type %s" % type(value))
 *             log.debug("Set field %s: %r" % (key, value))             # <<<<<<<<<<<<<<
 *         return cogr_feature
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_23 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_23)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_23);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = PyTuple_New(4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_32 = 0;
    __pyx_t_33 = 127;
    __Pyx_INCREF(__pyx_kp_u_Set_field);
    __pyx_t_32 += 10;
    __Pyx_GIVEREF(__pyx_kp_u_Set_field);
    PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_kp_u_Set_field);
    __pyx_t_18 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Unicode(__pyx_v_key), __pyx_empty_unicode); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) : __pyx_t_33;
    __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_18);
    __pyx_t_18 = 0;
    __Pyx_INCREF(__pyx_kp_u__8);
    __pyx_t_32 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__8);
    PyTuple_SET_ITEM(__pyx_t_11, 2, __pyx_kp_u__8);
    __pyx_t_18 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_v_value), __pyx_empty_unicode); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __pyx_t_33 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) > __pyx_t_33) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_18) : __pyx_t_33;
    __pyx_t_32 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_18);
    __Pyx_GIVEREF(__pyx_t_18);
    PyTuple_SET_ITEM(__pyx_t_11, 3, __pyx_t_18);
    __pyx_t_18 = 0;
    __pyx_t_18 = __Pyx_PyUnicode_Join(__pyx_t_11, 4, __pyx_t_32, __pyx_t_33); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_18);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_23))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_23);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_23);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_23, function);
      }
    }
    __pyx_t_6 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_23, __pyx_t_11, __pyx_t_18) : __Pyx_PyObject_CallOneArg(__pyx_t_23, __pyx_t_18);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 413, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_23); __pyx_t_23 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_L7_continue:;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":414
 *                 raise ValueError("Invalid field type %s" % type(value))
 *             log.debug("Set field %s: %r" % (key, value))
 *         return cogr_feature             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_cogr_feature;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":312
 *     """
 * 
 *     cdef void * build(self, feature, collection) except NULL:             # <<<<<<<<<<<<<<
 *         cdef void *cogr_geometry = NULL
 *         cdef const char *string_c = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_23);
  __Pyx_XDECREF(__pyx_t_24);
  __Pyx_AddTraceback("fiona.ogrext.OGRFeatureBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_ogr_key);
  __Pyx_XDECREF(__pyx_v_schema_type);
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_key_c);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_m);
  __Pyx_XDECREF(__pyx_v_d);
  __Pyx_XDECREF(__pyx_v_hh);
  __Pyx_XDECREF(__pyx_v_mm);
  __Pyx_XDECREF(__pyx_v_ss);
  __Pyx_XDECREF(__pyx_v_ff);
  __Pyx_XDECREF(__pyx_v_value_bytes);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder___reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder___reduce_cython__(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_OGRFeatureBuilder); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OGRFeatureBuilder__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_OGRFeatureBuilder); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.OGRFeatureBuilder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OGRFeatureBuilder__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder_2__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_17OGRFeatureBuilder_2__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OGRFeatureBuilder__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5fiona_6ogrext___pyx_unpickle_OGRFeatureBuilder__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OGRFeatureBuilder, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OGRFeatureBuilder__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.OGRFeatureBuilder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":417
 * 
 * 
 * cdef _deleteOgrFeature(void *cogr_feature):             # <<<<<<<<<<<<<<
 *     """Delete an OGR feature"""
 *     if cogr_feature is not NULL:
 */

static PyObject *__pyx_f_5fiona_6ogrext__deleteOgrFeature(void *__pyx_v_cogr_feature) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_deleteOgrFeature", 0);

  /* "fiona/ogrext.pyx":419
 * cdef _deleteOgrFeature(void *cogr_feature):
 *     """Delete an OGR feature"""
 *     if cogr_feature is not NULL:             # <<<<<<<<<<<<<<
 *         OGR_F_Destroy(cogr_feature)
 *     cogr_feature = NULL
 */
  __pyx_t_1 = ((__pyx_v_cogr_feature != NULL) != 0);
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":420
 *     """Delete an OGR feature"""
 *     if cogr_feature is not NULL:
 *         OGR_F_Destroy(cogr_feature)             # <<<<<<<<<<<<<<
 *     cogr_feature = NULL
 * 
 */
    OGR_F_Destroy(__pyx_v_cogr_feature);

    /* "fiona/ogrext.pyx":419
 * cdef _deleteOgrFeature(void *cogr_feature):
 *     """Delete an OGR feature"""
 *     if cogr_feature is not NULL:             # <<<<<<<<<<<<<<
 *         OGR_F_Destroy(cogr_feature)
 *     cogr_feature = NULL
 */
  }

  /* "fiona/ogrext.pyx":421
 *     if cogr_feature is not NULL:
 *         OGR_F_Destroy(cogr_feature)
 *     cogr_feature = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_cogr_feature = NULL;

  /* "fiona/ogrext.pyx":417
 * 
 * 
 * cdef _deleteOgrFeature(void *cogr_feature):             # <<<<<<<<<<<<<<
 *     """Delete an OGR feature"""
 *     if cogr_feature is not NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":424
 * 
 * 
 * def featureRT(feature, collection):             # <<<<<<<<<<<<<<
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_6featureRT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_6featureRT = {"featureRT", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_6featureRT, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_6featureRT(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_feature = 0;
  PyObject *__pyx_v_collection = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("featureRT (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_feature,&__pyx_n_s_collection,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_feature)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collection)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("featureRT", 1, 2, 2, 1); __PYX_ERR(0, 424, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "featureRT") < 0)) __PYX_ERR(0, 424, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_feature = values[0];
    __pyx_v_collection = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("featureRT", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 424, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.featureRT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_5featureRT(__pyx_self, __pyx_v_feature, __pyx_v_collection);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_5featureRT(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_feature, PyObject *__pyx_v_collection) {
  void *__pyx_v_cogr_feature;
  void *__pyx_v_cogr_geometry;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  void *__pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build __pyx_t_6;
  __Pyx_RefNannySetupContext("featureRT", 0);

  /* "fiona/ogrext.pyx":426
 * def featureRT(feature, collection):
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)             # <<<<<<<<<<<<<<
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)
 *     if cogr_geometry == NULL:
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_OGRFeatureBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_t_1), __pyx_v_feature, __pyx_v_collection); if (unlikely(__pyx_t_2 == ((void *)NULL))) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cogr_feature = __pyx_t_2;

  /* "fiona/ogrext.pyx":427
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)             # <<<<<<<<<<<<<<
 *     if cogr_geometry == NULL:
 *         raise ValueError("Null geometry")
 */
  __pyx_v_cogr_geometry = OGR_F_GetGeometryRef(__pyx_v_cogr_feature);

  /* "fiona/ogrext.pyx":428
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)
 *     if cogr_geometry == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Null geometry")
 *     result = FeatureBuilder().build(
 */
  __pyx_t_3 = ((__pyx_v_cogr_geometry == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":429
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)
 *     if cogr_geometry == NULL:
 *         raise ValueError("Null geometry")             # <<<<<<<<<<<<<<
 *     result = FeatureBuilder().build(
 *         cogr_feature,
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 429, __pyx_L1_error)

    /* "fiona/ogrext.pyx":428
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)
 *     if cogr_geometry == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("Null geometry")
 *     result = FeatureBuilder().build(
 */
  }

  /* "fiona/ogrext.pyx":430
 *     if cogr_geometry == NULL:
 *         raise ValueError("Null geometry")
 *     result = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *         cogr_feature,
 *         encoding='utf-8',
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fiona/ogrext.pyx":434
 *         encoding='utf-8',
 *         bbox=False,
 *         driver=collection.driver             # <<<<<<<<<<<<<<
 *     )
 *     _deleteOgrFeature(cogr_feature)
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "fiona/ogrext.pyx":430
 *     if cogr_geometry == NULL:
 *         raise ValueError("Null geometry")
 *     result = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *         cogr_feature,
 *         encoding='utf-8',
 */
  __pyx_t_6.__pyx_n = 3;
  __pyx_t_6.encoding = __pyx_kp_u_utf_8;
  __pyx_t_6.bbox = Py_False;
  __pyx_t_6.driver = __pyx_t_4;
  __pyx_t_5 = ((struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1), __pyx_v_cogr_feature, &__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_result = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":436
 *         driver=collection.driver
 *     )
 *     _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_t_5 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":437
 *     )
 *     _deleteOgrFeature(cogr_feature)
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":424
 * 
 * 
 * def featureRT(feature, collection):             # <<<<<<<<<<<<<<
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.featureRT", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":450
 *     cdef object collection
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.cogr_ds = NULL
 *         self.cogr_layer = NULL
 */

/* Python wrapper */
static int __pyx_pw_5fiona_6ogrext_7Session_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5fiona_6ogrext_7Session_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  if (unlikely(PyTuple_GET_SIZE(__pyx_args) > 0)) {
    __Pyx_RaiseArgtupleInvalid("__init__", 1, 0, 0, PyTuple_GET_SIZE(__pyx_args)); return -1;}
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__init__", 0))) return -1;
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session___init__(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5fiona_6ogrext_7Session___init__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fiona/ogrext.pyx":451
 * 
 *     def __init__(self):
 *         self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 *         self.cogr_layer = NULL
 *         self._fileencoding = None
 */
  __pyx_v_self->cogr_ds = NULL;

  /* "fiona/ogrext.pyx":452
 *     def __init__(self):
 *         self.cogr_ds = NULL
 *         self.cogr_layer = NULL             # <<<<<<<<<<<<<<
 *         self._fileencoding = None
 *         self._encoding = None
 */
  __pyx_v_self->cogr_layer = NULL;

  /* "fiona/ogrext.pyx":453
 *         self.cogr_ds = NULL
 *         self.cogr_layer = NULL
 *         self._fileencoding = None             # <<<<<<<<<<<<<<
 *         self._encoding = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_fileencoding);
  __Pyx_DECREF(__pyx_v_self->_fileencoding);
  __pyx_v_self->_fileencoding = Py_None;

  /* "fiona/ogrext.pyx":454
 *         self.cogr_layer = NULL
 *         self._fileencoding = None
 *         self._encoding = None             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_encoding);
  __Pyx_DECREF(__pyx_v_self->_encoding);
  __pyx_v_self->_encoding = Py_None;

  /* "fiona/ogrext.pyx":450
 *     cdef object collection
 * 
 *     def __init__(self):             # <<<<<<<<<<<<<<
 *         self.cogr_ds = NULL
 *         self.cogr_layer = NULL
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":456
 *         self._encoding = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.stop()
 * 
 */

/* Python wrapper */
static void __pyx_pw_5fiona_6ogrext_7Session_3__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_5fiona_6ogrext_7Session_3__dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_pf_5fiona_6ogrext_7Session_2__dealloc__(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5fiona_6ogrext_7Session_2__dealloc__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "fiona/ogrext.pyx":457
 * 
 *     def __dealloc__(self):
 *         self.stop()             # <<<<<<<<<<<<<<
 * 
 *     def start(self, collection, **kwargs):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_5fiona_6ogrext_Session *)__pyx_v_self->__pyx_vtab)->stop(__pyx_v_self, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":456
 *         self._encoding = None
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         self.stop()
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("fiona.ogrext.Session.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
}

/* "fiona/ogrext.pyx":459
 *         self.stop()
 * 
 *     def start(self, collection, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef const char *path_c = NULL
 *         cdef const char *name_c = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_5start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_5start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_collection = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_collection,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "start") < 0)) __PYX_ERR(0, 459, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_collection = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 459, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("fiona.ogrext.Session.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_4start(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self), __pyx_v_collection, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_4start(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_kwargs) {
  char const *__pyx_v_path_c;
  char const *__pyx_v_name_c;
  CYTHON_UNUSED void *__pyx_v_drv;
  CYTHON_UNUSED void *__pyx_v_ds;
  char **__pyx_v_ignore_fields;
  PyObject *__pyx_v_path_b = NULL;
  PyObject *__pyx_v_drivers = NULL;
  PyObject *__pyx_v_encoding = NULL;
  PyObject *__pyx_v_name_b = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  char const *__pyx_t_4;
  int __pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  int __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  char const *__pyx_t_21;
  int __pyx_t_22;
  char const *__pyx_t_23;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  __Pyx_RefNannySetupContext("start", 0);

  /* "fiona/ogrext.pyx":460
 * 
 *     def start(self, collection, **kwargs):
 *         cdef const char *path_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *name_c = NULL
 *         cdef void *drv = NULL
 */
  __pyx_v_path_c = NULL;

  /* "fiona/ogrext.pyx":461
 *     def start(self, collection, **kwargs):
 *         cdef const char *path_c = NULL
 *         cdef const char *name_c = NULL             # <<<<<<<<<<<<<<
 *         cdef void *drv = NULL
 *         cdef void *ds = NULL
 */
  __pyx_v_name_c = NULL;

  /* "fiona/ogrext.pyx":462
 *         cdef const char *path_c = NULL
 *         cdef const char *name_c = NULL
 *         cdef void *drv = NULL             # <<<<<<<<<<<<<<
 *         cdef void *ds = NULL
 *         cdef char **ignore_fields = NULL
 */
  __pyx_v_drv = NULL;

  /* "fiona/ogrext.pyx":463
 *         cdef const char *name_c = NULL
 *         cdef void *drv = NULL
 *         cdef void *ds = NULL             # <<<<<<<<<<<<<<
 *         cdef char **ignore_fields = NULL
 * 
 */
  __pyx_v_ds = NULL;

  /* "fiona/ogrext.pyx":464
 *         cdef void *drv = NULL
 *         cdef void *ds = NULL
 *         cdef char **ignore_fields = NULL             # <<<<<<<<<<<<<<
 * 
 *         path_b = collection.path.encode('utf-8')
 */
  __pyx_v_ignore_fields = NULL;

  /* "fiona/ogrext.pyx":466
 *         cdef char **ignore_fields = NULL
 * 
 *         path_b = collection.path.encode('utf-8')             # <<<<<<<<<<<<<<
 *         path_c = path_b
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_path_b = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":467
 * 
 *         path_b = collection.path.encode('utf-8')
 *         path_c = path_b             # <<<<<<<<<<<<<<
 * 
 *         self._fileencoding = kwargs.get('encoding') or collection.encoding
 */
  __pyx_t_4 = __Pyx_PyObject_AsString(__pyx_v_path_b); if (unlikely((!__pyx_t_4) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L1_error)
  __pyx_v_path_c = __pyx_t_4;

  /* "fiona/ogrext.pyx":469
 *         path_c = path_b
 * 
 *         self._fileencoding = kwargs.get('encoding') or collection.encoding             # <<<<<<<<<<<<<<
 * 
 *         # We have two ways of specifying drivers to try. Resolve the
 */
  __pyx_t_3 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_u_encoding, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 469, __pyx_L1_error)
  if (!__pyx_t_5) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_fileencoding);
  __Pyx_DECREF(__pyx_v_self->_fileencoding);
  __pyx_v_self->_fileencoding = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":473
 *         # We have two ways of specifying drivers to try. Resolve the
 *         # values into a single set of driver short names.
 *         if collection._driver:             # <<<<<<<<<<<<<<
 *             drivers = set([collection._driver])
 *         elif collection.enabled_drivers:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":474
 *         # values into a single set of driver short names.
 *         if collection._driver:
 *             drivers = set([collection._driver])             # <<<<<<<<<<<<<<
 *         elif collection.enabled_drivers:
 *             drivers = set(collection.enabled_drivers)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PySet_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PySet_Add(__pyx_t_3, __pyx_t_1) < 0) __PYX_ERR(0, 474, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_drivers = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":473
 *         # We have two ways of specifying drivers to try. Resolve the
 *         # values into a single set of driver short names.
 *         if collection._driver:             # <<<<<<<<<<<<<<
 *             drivers = set([collection._driver])
 *         elif collection.enabled_drivers:
 */
    goto __pyx_L5;
  }

  /* "fiona/ogrext.pyx":475
 *         if collection._driver:
 *             drivers = set([collection._driver])
 *         elif collection.enabled_drivers:             # <<<<<<<<<<<<<<
 *             drivers = set(collection.enabled_drivers)
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_enabled_drivers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":476
 *             drivers = set([collection._driver])
 *         elif collection.enabled_drivers:
 *             drivers = set(collection.enabled_drivers)             # <<<<<<<<<<<<<<
 *         else:
 *             drivers = None
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_enabled_drivers); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 476, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_drivers = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":475
 *         if collection._driver:
 *             drivers = set([collection._driver])
 *         elif collection.enabled_drivers:             # <<<<<<<<<<<<<<
 *             drivers = set(collection.enabled_drivers)
 *         else:
 */
    goto __pyx_L5;
  }

  /* "fiona/ogrext.pyx":478
 *             drivers = set(collection.enabled_drivers)
 *         else:
 *             drivers = None             # <<<<<<<<<<<<<<
 * 
 *         encoding = kwargs.pop('encoding', None)
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_drivers = ((PyObject*)Py_None);
  }
  __pyx_L5:;

  /* "fiona/ogrext.pyx":480
 *             drivers = None
 * 
 *         encoding = kwargs.pop('encoding', None)             # <<<<<<<<<<<<<<
 *         if encoding:
 *             kwargs['encoding'] = encoding.upper()
 */
  __pyx_t_1 = __Pyx_PyDict_Pop(__pyx_v_kwargs, __pyx_n_u_encoding, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_encoding = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":481
 * 
 *         encoding = kwargs.pop('encoding', None)
 *         if encoding:             # <<<<<<<<<<<<<<
 *             kwargs['encoding'] = encoding.upper()
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_encoding); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 481, __pyx_L1_error)
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":482
 *         encoding = kwargs.pop('encoding', None)
 *         if encoding:
 *             kwargs['encoding'] = encoding.upper()             # <<<<<<<<<<<<<<
 * 
 *         self.cogr_ds = gdal_open_vector(path_c, 0, drivers, kwargs)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_encoding, __pyx_n_s_upper); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_kwargs, __pyx_n_u_encoding, __pyx_t_1) < 0)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":481
 * 
 *         encoding = kwargs.pop('encoding', None)
 *         if encoding:             # <<<<<<<<<<<<<<
 *             kwargs['encoding'] = encoding.upper()
 * 
 */
  }

  /* "fiona/ogrext.pyx":484
 *             kwargs['encoding'] = encoding.upper()
 * 
 *         self.cogr_ds = gdal_open_vector(path_c, 0, drivers, kwargs)             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(collection.name, string_types):
 */
  __pyx_t_6 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_v_path_c, 0, __pyx_v_drivers, __pyx_v_kwargs); if (unlikely(__pyx_t_6 == ((void *)NULL))) __PYX_ERR(0, 484, __pyx_L1_error)
  __pyx_v_self->cogr_ds = __pyx_t_6;

  /* "fiona/ogrext.pyx":486
 *         self.cogr_ds = gdal_open_vector(path_c, 0, drivers, kwargs)
 * 
 *         if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *             name_b = collection.name.encode('utf-8')
 *             name_c = name_b
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_string_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyObject_IsInstance(__pyx_t_1, __pyx_t_3); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = (__pyx_t_5 != 0);
  if (__pyx_t_7) {

    /* "fiona/ogrext.pyx":487
 * 
 *         if isinstance(collection.name, string_types):
 *             name_b = collection.name.encode('utf-8')             # <<<<<<<<<<<<<<
 *             name_c = name_b
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_name_b = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":488
 *         if isinstance(collection.name, string_types):
 *             name_b = collection.name.encode('utf-8')
 *             name_c = name_b             # <<<<<<<<<<<<<<
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)
 *         elif isinstance(collection.name, int):
 */
    __pyx_t_8 = __Pyx_PyObject_AsString(__pyx_v_name_b); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 488, __pyx_L1_error)
    __pyx_v_name_c = __pyx_t_8;

    /* "fiona/ogrext.pyx":489
 *             name_b = collection.name.encode('utf-8')
 *             name_c = name_b
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)             # <<<<<<<<<<<<<<
 *         elif isinstance(collection.name, int):
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)
 */
    __pyx_v_self->cogr_layer = GDALDatasetGetLayerByName(__pyx_v_self->cogr_ds, __pyx_v_name_c);

    /* "fiona/ogrext.pyx":486
 *         self.cogr_ds = gdal_open_vector(path_c, 0, drivers, kwargs)
 * 
 *         if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *             name_b = collection.name.encode('utf-8')
 *             name_c = name_b
 */
    goto __pyx_L7;
  }

  /* "fiona/ogrext.pyx":490
 *             name_c = name_b
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)
 *         elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)
 *             name_c = OGR_L_GetName(self.cogr_layer)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = PyInt_Check(__pyx_t_3); 
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = (__pyx_t_7 != 0);
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":491
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)
 *         elif isinstance(collection.name, int):
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)             # <<<<<<<<<<<<<<
 *             name_c = OGR_L_GetName(self.cogr_layer)
 *             name_b = name_c
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->cogr_layer = GDALDatasetGetLayer(__pyx_v_self->cogr_ds, __pyx_t_9);

    /* "fiona/ogrext.pyx":492
 *         elif isinstance(collection.name, int):
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)
 *             name_c = OGR_L_GetName(self.cogr_layer)             # <<<<<<<<<<<<<<
 *             name_b = name_c
 *             collection.name = name_b.decode('utf-8')
 */
    __pyx_v_name_c = OGR_L_GetName(__pyx_v_self->cogr_layer);

    /* "fiona/ogrext.pyx":493
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)
 *             name_c = OGR_L_GetName(self.cogr_layer)
 *             name_b = name_c             # <<<<<<<<<<<<<<
 *             collection.name = name_b.decode('utf-8')
 * 
 */
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_name_c); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_name_b = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":494
 *             name_c = OGR_L_GetName(self.cogr_layer)
 *             name_b = name_c
 *             collection.name = name_b.decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *         if self.cogr_layer == NULL:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_name_b, __pyx_n_s_decode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_1, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_collection, __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":490
 *             name_c = name_b
 *             self.cogr_layer = GDALDatasetGetLayerByName(self.cogr_ds, name_c)
 *         elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *             self.cogr_layer = GDALDatasetGetLayer(self.cogr_ds, collection.name)
 *             name_c = OGR_L_GetName(self.cogr_layer)
 */
  }
  __pyx_L7:;

  /* "fiona/ogrext.pyx":496
 *             collection.name = name_b.decode('utf-8')
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer: " + repr(collection.name))
 * 
 */
  __pyx_t_5 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_5)) {

    /* "fiona/ogrext.pyx":497
 * 
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer: " + repr(collection.name))             # <<<<<<<<<<<<<<
 * 
 *         encoding = self._get_internal_encoding()
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyObject_Repr(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_ConcatSafe(__pyx_kp_u_Null_layer_2, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 497, __pyx_L1_error)

    /* "fiona/ogrext.pyx":496
 *             collection.name = name_b.decode('utf-8')
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer: " + repr(collection.name))
 * 
 */
  }

  /* "fiona/ogrext.pyx":499
 *             raise ValueError("Null layer: " + repr(collection.name))
 * 
 *         encoding = self._get_internal_encoding()             # <<<<<<<<<<<<<<
 * 
 *         if collection.ignore_fields:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_encoding, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":501
 *         encoding = self._get_internal_encoding()
 * 
 *         if collection.ignore_fields:             # <<<<<<<<<<<<<<
 *             try:
 *                 for name in collection.ignore_fields:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":502
 * 
 *         if collection.ignore_fields:
 *             try:             # <<<<<<<<<<<<<<
 *                 for name in collection.ignore_fields:
 *                     try:
 */
    /*try:*/ {

      /* "fiona/ogrext.pyx":503
 *         if collection.ignore_fields:
 *             try:
 *                 for name in collection.ignore_fields:             # <<<<<<<<<<<<<<
 *                     try:
 *                         name_b = name.encode(encoding)
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 503, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_11 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 503, __pyx_L11_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_3))) {
            if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L11_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L11_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L11_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_11(__pyx_t_3);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 503, __pyx_L11_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_XDECREF_SET(__pyx_v_name, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":504
 *             try:
 *                 for name in collection.ignore_fields:
 *                     try:             # <<<<<<<<<<<<<<
 *                         name_b = name.encode(encoding)
 *                     except AttributeError:
 */
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_12, &__pyx_t_13, &__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_12);
          __Pyx_XGOTREF(__pyx_t_13);
          __Pyx_XGOTREF(__pyx_t_14);
          /*try:*/ {

            /* "fiona/ogrext.pyx":505
 *                 for name in collection.ignore_fields:
 *                     try:
 *                         name_b = name.encode(encoding)             # <<<<<<<<<<<<<<
 *                     except AttributeError:
 *                         raise TypeError("Ignored field \"{}\" has type \"{}\", expected string".format(name, name.__class__.__name__))
 */
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_15 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
              __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_1);
              if (likely(__pyx_t_15)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
                __Pyx_INCREF(__pyx_t_15);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_1, function);
              }
            }
            __pyx_t_2 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_15, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_encoding);
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 505, __pyx_L15_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_XDECREF_SET(__pyx_v_name_b, __pyx_t_2);
            __pyx_t_2 = 0;

            /* "fiona/ogrext.pyx":504
 *             try:
 *                 for name in collection.ignore_fields:
 *                     try:             # <<<<<<<<<<<<<<
 *                         name_b = name.encode(encoding)
 *                     except AttributeError:
 */
          }
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
          goto __pyx_L22_try_end;
          __pyx_L15_error:;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":506
 *                     try:
 *                         name_b = name.encode(encoding)
 *                     except AttributeError:             # <<<<<<<<<<<<<<
 *                         raise TypeError("Ignored field \"{}\" has type \"{}\", expected string".format(name, name.__class__.__name__))
 *                     ignore_fields = CSLAddString(ignore_fields, <const char *>name_b)
 */
          __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
          if (__pyx_t_9) {
            __Pyx_AddTraceback("fiona.ogrext.Session.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_2, &__pyx_t_1, &__pyx_t_15) < 0) __PYX_ERR(0, 506, __pyx_L17_except_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_GOTREF(__pyx_t_15);

            /* "fiona/ogrext.pyx":507
 *                         name_b = name.encode(encoding)
 *                     except AttributeError:
 *                         raise TypeError("Ignored field \"{}\" has type \"{}\", expected string".format(name, name.__class__.__name__))             # <<<<<<<<<<<<<<
 *                     ignore_fields = CSLAddString(ignore_fields, <const char *>name_b)
 *                 OGR_L_SetIgnoredFields(self.cogr_layer, <const char**>ignore_fields)
 */
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Ignored_field_has_type_expected, __pyx_n_s_format); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 507, __pyx_L17_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_18 = __Pyx_PyObject_GetAttrStr(__pyx_v_name, __pyx_n_s_class); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 507, __pyx_L17_except_error)
            __Pyx_GOTREF(__pyx_t_18);
            __pyx_t_19 = __Pyx_PyObject_GetAttrStr(__pyx_t_18, __pyx_n_s_name_2); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 507, __pyx_L17_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            __pyx_t_18 = NULL;
            __pyx_t_9 = 0;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_18)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_18);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
                __pyx_t_9 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_v_name, __pyx_t_19};
              __pyx_t_16 = __Pyx_PyFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 507, __pyx_L17_except_error)
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_17)) {
              PyObject *__pyx_temp[3] = {__pyx_t_18, __pyx_v_name, __pyx_t_19};
              __pyx_t_16 = __Pyx_PyCFunction_FastCall(__pyx_t_17, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 507, __pyx_L17_except_error)
              __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
            } else
            #endif
            {
              __pyx_t_20 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 507, __pyx_L17_except_error)
              __Pyx_GOTREF(__pyx_t_20);
              if (__pyx_t_18) {
                __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_20, 0, __pyx_t_18); __pyx_t_18 = NULL;
              }
              __Pyx_INCREF(__pyx_v_name);
              __Pyx_GIVEREF(__pyx_v_name);
              PyTuple_SET_ITEM(__pyx_t_20, 0+__pyx_t_9, __pyx_v_name);
              __Pyx_GIVEREF(__pyx_t_19);
              PyTuple_SET_ITEM(__pyx_t_20, 1+__pyx_t_9, __pyx_t_19);
              __pyx_t_19 = 0;
              __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_t_20, NULL); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 507, __pyx_L17_except_error)
              __Pyx_GOTREF(__pyx_t_16);
              __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
            }
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_16); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 507, __pyx_L17_except_error)
            __Pyx_GOTREF(__pyx_t_17);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_Raise(__pyx_t_17, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __PYX_ERR(0, 507, __pyx_L17_except_error)
          }
          goto __pyx_L17_except_error;
          __pyx_L17_except_error:;

          /* "fiona/ogrext.pyx":504
 *             try:
 *                 for name in collection.ignore_fields:
 *                     try:             # <<<<<<<<<<<<<<
 *                         name_b = name.encode(encoding)
 *                     except AttributeError:
 */
          __Pyx_XGIVEREF(__pyx_t_12);
          __Pyx_XGIVEREF(__pyx_t_13);
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_ExceptionReset(__pyx_t_12, __pyx_t_13, __pyx_t_14);
          goto __pyx_L11_error;
          __pyx_L22_try_end:;
        }

        /* "fiona/ogrext.pyx":508
 *                     except AttributeError:
 *                         raise TypeError("Ignored field \"{}\" has type \"{}\", expected string".format(name, name.__class__.__name__))
 *                     ignore_fields = CSLAddString(ignore_fields, <const char *>name_b)             # <<<<<<<<<<<<<<
 *                 OGR_L_SetIgnoredFields(self.cogr_layer, <const char**>ignore_fields)
 *             finally:
 */
        __pyx_t_21 = __Pyx_PyObject_AsString(__pyx_v_name_b); if (unlikely((!__pyx_t_21) && PyErr_Occurred())) __PYX_ERR(0, 508, __pyx_L11_error)
        __pyx_v_ignore_fields = CSLAddString(__pyx_v_ignore_fields, ((char const *)__pyx_t_21));

        /* "fiona/ogrext.pyx":503
 *         if collection.ignore_fields:
 *             try:
 *                 for name in collection.ignore_fields:             # <<<<<<<<<<<<<<
 *                     try:
 *                         name_b = name.encode(encoding)
 */
      }
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":509
 *                         raise TypeError("Ignored field \"{}\" has type \"{}\", expected string".format(name, name.__class__.__name__))
 *                     ignore_fields = CSLAddString(ignore_fields, <const char *>name_b)
 *                 OGR_L_SetIgnoredFields(self.cogr_layer, <const char**>ignore_fields)             # <<<<<<<<<<<<<<
 *             finally:
 *                 CSLDestroy(ignore_fields)
 */
      (void)(OGR_L_SetIgnoredFields(__pyx_v_self->cogr_layer, ((char const **)__pyx_v_ignore_fields)));
    }

    /* "fiona/ogrext.pyx":511
 *                 OGR_L_SetIgnoredFields(self.cogr_layer, <const char**>ignore_fields)
 *             finally:
 *                 CSLDestroy(ignore_fields)             # <<<<<<<<<<<<<<
 * 
 *         self.collection = collection
 */
    /*finally:*/ {
      /*normal exit:*/{
        CSLDestroy(__pyx_v_ignore_fields);
        goto __pyx_L12;
      }
      __pyx_L11_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_24, &__pyx_t_25, &__pyx_t_26);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_13, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_14);
        __Pyx_XGOTREF(__pyx_t_13);
        __Pyx_XGOTREF(__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_24);
        __Pyx_XGOTREF(__pyx_t_25);
        __Pyx_XGOTREF(__pyx_t_26);
        __pyx_t_9 = __pyx_lineno; __pyx_t_22 = __pyx_clineno; __pyx_t_23 = __pyx_filename;
        {
          CSLDestroy(__pyx_v_ignore_fields);
        }
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_24);
          __Pyx_XGIVEREF(__pyx_t_25);
          __Pyx_XGIVEREF(__pyx_t_26);
          __Pyx_ExceptionReset(__pyx_t_24, __pyx_t_25, __pyx_t_26);
        }
        __Pyx_XGIVEREF(__pyx_t_14);
        __Pyx_XGIVEREF(__pyx_t_13);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ErrRestore(__pyx_t_14, __pyx_t_13, __pyx_t_12);
        __pyx_t_14 = 0; __pyx_t_13 = 0; __pyx_t_12 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
        __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_22; __pyx_filename = __pyx_t_23;
        goto __pyx_L1_error;
      }
      __pyx_L12:;
    }

    /* "fiona/ogrext.pyx":501
 *         encoding = self._get_internal_encoding()
 * 
 *         if collection.ignore_fields:             # <<<<<<<<<<<<<<
 *             try:
 *                 for name in collection.ignore_fields:
 */
  }

  /* "fiona/ogrext.pyx":513
 *                 CSLDestroy(ignore_fields)
 * 
 *         self.collection = collection             # <<<<<<<<<<<<<<
 * 
 *     cpdef stop(self):
 */
  __Pyx_INCREF(__pyx_v_collection);
  __Pyx_GIVEREF(__pyx_v_collection);
  __Pyx_GOTREF(__pyx_v_self->collection);
  __Pyx_DECREF(__pyx_v_self->collection);
  __pyx_v_self->collection = __pyx_v_collection;

  /* "fiona/ogrext.pyx":459
 *         self.stop()
 * 
 *     def start(self, collection, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef const char *path_c = NULL
 *         cdef const char *name_c = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("fiona.ogrext.Session.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path_b);
  __Pyx_XDECREF(__pyx_v_drivers);
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_name_b);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":515
 *         self.collection = collection
 * 
 *     cpdef stop(self):             # <<<<<<<<<<<<<<
 *         self.cogr_layer = NULL
 *         if self.cogr_ds != NULL:
 */

static PyObject *__pyx_pw_5fiona_6ogrext_7Session_7stop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_5fiona_6ogrext_7Session_stop(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("stop", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely((Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0) || (Py_TYPE(((PyObject *)__pyx_v_self))->tp_flags & (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_type_dict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_stop); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)(void*)__pyx_pw_5fiona_6ogrext_7Session_7stop)) {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_2 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 515, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_2;
        __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_type_dict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP
    }
    #endif
  }

  /* "fiona/ogrext.pyx":516
 * 
 *     cpdef stop(self):
 *         self.cogr_layer = NULL             # <<<<<<<<<<<<<<
 *         if self.cogr_ds != NULL:
 *             GDALClose(self.cogr_ds)
 */
  __pyx_v_self->cogr_layer = NULL;

  /* "fiona/ogrext.pyx":517
 *     cpdef stop(self):
 *         self.cogr_layer = NULL
 *         if self.cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *             GDALClose(self.cogr_ds)
 *         self.cogr_ds = NULL
 */
  __pyx_t_5 = ((__pyx_v_self->cogr_ds != NULL) != 0);
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":518
 *         self.cogr_layer = NULL
 *         if self.cogr_ds != NULL:
 *             GDALClose(self.cogr_ds)             # <<<<<<<<<<<<<<
 *         self.cogr_ds = NULL
 * 
 */
    GDALClose(__pyx_v_self->cogr_ds);

    /* "fiona/ogrext.pyx":517
 *     cpdef stop(self):
 *         self.cogr_layer = NULL
 *         if self.cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *             GDALClose(self.cogr_ds)
 *         self.cogr_ds = NULL
 */
  }

  /* "fiona/ogrext.pyx":519
 *         if self.cogr_ds != NULL:
 *             GDALClose(self.cogr_ds)
 *         self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 * 
 *     def get_fileencoding(self):
 */
  __pyx_v_self->cogr_ds = NULL;

  /* "fiona/ogrext.pyx":515
 *         self.collection = collection
 * 
 *     cpdef stop(self):             # <<<<<<<<<<<<<<
 *         self.cogr_layer = NULL
 *         if self.cogr_ds != NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("fiona.ogrext.Session.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_7stop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_7stop(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stop (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_6stop(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_6stop(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("stop", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_5fiona_6ogrext_7Session_stop(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 515, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Session.stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":521
 *         self.cogr_ds = NULL
 * 
 *     def get_fileencoding(self):             # <<<<<<<<<<<<<<
 *         """DEPRECATED"""
 *         warnings.warn("get_fileencoding is deprecated and will be removed in a future version.", FionaDeprecationWarning)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_9get_fileencoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_8get_fileencoding[] = "DEPRECATED";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_9get_fileencoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fileencoding (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_8get_fileencoding(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_8get_fileencoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_fileencoding", 0);

  /* "fiona/ogrext.pyx":523
 *     def get_fileencoding(self):
 *         """DEPRECATED"""
 *         warnings.warn("get_fileencoding is deprecated and will be removed in a future version.", FionaDeprecationWarning)             # <<<<<<<<<<<<<<
 *         return self._fileencoding
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_warnings); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_warn); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_FionaDeprecationWarning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 523, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_get_fileencoding_is_deprecated_a, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_get_fileencoding_is_deprecated_a, __pyx_t_2};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_5, 2+__pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else
  #endif
  {
    __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (__pyx_t_4) {
      __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_get_fileencoding_is_deprecated_a);
    __Pyx_GIVEREF(__pyx_kp_u_get_fileencoding_is_deprecated_a);
    PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_kp_u_get_fileencoding_is_deprecated_a);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 523, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":524
 *         """DEPRECATED"""
 *         warnings.warn("get_fileencoding is deprecated and will be removed in a future version.", FionaDeprecationWarning)
 *         return self._fileencoding             # <<<<<<<<<<<<<<
 * 
 *     def _get_fallback_encoding(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_fileencoding);
  __pyx_r = __pyx_v_self->_fileencoding;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":521
 *         self.cogr_ds = NULL
 * 
 *     def get_fileencoding(self):             # <<<<<<<<<<<<<<
 *         """DEPRECATED"""
 *         warnings.warn("get_fileencoding is deprecated and will be removed in a future version.", FionaDeprecationWarning)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_fileencoding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":526
 *         return self._fileencoding
 * 
 *     def _get_fallback_encoding(self):             # <<<<<<<<<<<<<<
 *         """Determine a format-specific fallback encoding to use when using OGR_F functions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_11_get_fallback_encoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_10_get_fallback_encoding[] = "Determine a format-specific fallback encoding to use when using OGR_F functions\n\n        Parameters\n        ----------\n        None\n\n        Returns\n        -------\n        str\n\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_11_get_fallback_encoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_fallback_encoding (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_10_get_fallback_encoding(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_10_get_fallback_encoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  __Pyx_RefNannySetupContext("_get_fallback_encoding", 0);

  /* "fiona/ogrext.pyx":538
 * 
 *         """
 *         if "Shapefile" in self.get_driver():             # <<<<<<<<<<<<<<
 *             return 'iso-8859-1'
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_driver); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_Shapefile, __pyx_t_1, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":539
 *         """
 *         if "Shapefile" in self.get_driver():
 *             return 'iso-8859-1'             # <<<<<<<<<<<<<<
 *         else:
 *             return locale.getpreferredencoding()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u_iso_8859_1);
    __pyx_r = __pyx_kp_u_iso_8859_1;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":538
 * 
 *         """
 *         if "Shapefile" in self.get_driver():             # <<<<<<<<<<<<<<
 *             return 'iso-8859-1'
 *         else:
 */
  }

  /* "fiona/ogrext.pyx":541
 *             return 'iso-8859-1'
 *         else:
 *             return locale.getpreferredencoding()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_locale); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_getpreferredencoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_2) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":526
 *         return self._fileencoding
 * 
 *     def _get_fallback_encoding(self):             # <<<<<<<<<<<<<<
 *         """Determine a format-specific fallback encoding to use when using OGR_F functions
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fiona.ogrext.Session._get_fallback_encoding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":544
 * 
 * 
 *     def _get_internal_encoding(self):             # <<<<<<<<<<<<<<
 *         """Determine the encoding to use when use OGR_F functions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_13_get_internal_encoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_12_get_internal_encoding[] = "Determine the encoding to use when use OGR_F functions\n\n        Parameters\n        ----------\n        None\n\n        Returns\n        -------\n        str\n\n        Notes\n        -----\n        If the layer implements RFC 23 support for UTF-8, the return\n        value will be 'utf-8' and callers can be certain that this is\n        correct.  If the layer does not have the OLC_STRINGSASUTF8\n        capability marker, it is not possible to know exactly what the\n        internal encoding is and this method returns best guesses. That\n        means ISO-8859-1 for shapefiles and the locale's preferred\n        encoding for other formats such as CSV files.\n\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_13_get_internal_encoding(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_internal_encoding (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_12_get_internal_encoding(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_12_get_internal_encoding(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_get_internal_encoding", 0);

  /* "fiona/ogrext.pyx":566
 * 
 *         """
 *         if OGR_L_TestCapability(self.cogr_layer, OLC_STRINGSASUTF8):             # <<<<<<<<<<<<<<
 *             return 'utf-8'
 *         else:
 */
  __pyx_t_1 = (OGR_L_TestCapability(__pyx_v_self->cogr_layer, __pyx_v_5fiona_6ogrext_OLC_STRINGSASUTF8) != 0);
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":567
 *         """
 *         if OGR_L_TestCapability(self.cogr_layer, OLC_STRINGSASUTF8):
 *             return 'utf-8'             # <<<<<<<<<<<<<<
 *         else:
 *             return self._fileencoding or self._get_fallback_encoding()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u_utf_8);
    __pyx_r = __pyx_kp_u_utf_8;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":566
 * 
 *         """
 *         if OGR_L_TestCapability(self.cogr_layer, OLC_STRINGSASUTF8):             # <<<<<<<<<<<<<<
 *             return 'utf-8'
 *         else:
 */
  }

  /* "fiona/ogrext.pyx":569
 *             return 'utf-8'
 *         else:
 *             return self._fileencoding or self._get_fallback_encoding()             # <<<<<<<<<<<<<<
 * 
 *     def get_length(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_self->_fileencoding); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 569, __pyx_L1_error)
    if (!__pyx_t_1) {
    } else {
      __Pyx_INCREF(__pyx_v_self->_fileencoding);
      __pyx_t_2 = __pyx_v_self->_fileencoding;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_fallback_encoding); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 569, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_L4_bool_binop_done:;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":544
 * 
 * 
 *     def _get_internal_encoding(self):             # <<<<<<<<<<<<<<
 *         """Determine the encoding to use when use OGR_F functions
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.Session._get_internal_encoding", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":571
 *             return self._fileencoding or self._get_fallback_encoding()
 * 
 *     def get_length(self):             # <<<<<<<<<<<<<<
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_15get_length(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_15get_length(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_length (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_14get_length(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_14get_length(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("get_length", 0);

  /* "fiona/ogrext.pyx":572
 * 
 *     def get_length(self):
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)
 */
  __pyx_t_1 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fiona/ogrext.pyx":573
 *     def get_length(self):
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 573, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 573, __pyx_L1_error)

    /* "fiona/ogrext.pyx":572
 * 
 *     def get_length(self):
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)
 */
  }

  /* "fiona/ogrext.pyx":574
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)             # <<<<<<<<<<<<<<
 * 
 *     def get_driver(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyInt_From_int(OGR_L_GetFeatureCount(__pyx_v_self->cogr_layer, 0)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":571
 *             return self._fileencoding or self._get_fallback_encoding()
 * 
 *     def get_length(self):             # <<<<<<<<<<<<<<
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_length", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":576
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)
 * 
 *     def get_driver(self):             # <<<<<<<<<<<<<<
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_17get_driver(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_17get_driver(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_driver (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_16get_driver(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_16get_driver(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  void *__pyx_v_cogr_driver;
  char const *__pyx_v_name;
  char const *__pyx_v_driver_name;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("get_driver", 0);

  /* "fiona/ogrext.pyx":577
 * 
 *     def get_driver(self):
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)             # <<<<<<<<<<<<<<
 *         if cogr_driver == NULL:
 *             raise ValueError("Null driver")
 */
  __pyx_v_cogr_driver = GDALGetDatasetDriver(__pyx_v_self->cogr_ds);

  /* "fiona/ogrext.pyx":578
 *     def get_driver(self):
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null driver")
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 */
  __pyx_t_1 = ((__pyx_v_cogr_driver == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fiona/ogrext.pyx":579
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:
 *             raise ValueError("Null driver")             # <<<<<<<<<<<<<<
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 *         driver_name = name
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 579, __pyx_L1_error)

    /* "fiona/ogrext.pyx":578
 *     def get_driver(self):
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null driver")
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 */
  }

  /* "fiona/ogrext.pyx":580
 *         if cogr_driver == NULL:
 *             raise ValueError("Null driver")
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)             # <<<<<<<<<<<<<<
 *         driver_name = name
 *         return driver_name.decode()
 */
  __pyx_v_name = OGR_Dr_GetName(__pyx_v_cogr_driver);

  /* "fiona/ogrext.pyx":581
 *             raise ValueError("Null driver")
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 *         driver_name = name             # <<<<<<<<<<<<<<
 *         return driver_name.decode()
 * 
 */
  __pyx_v_driver_name = __pyx_v_name;

  /* "fiona/ogrext.pyx":582
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 *         driver_name = name
 *         return driver_name.decode()             # <<<<<<<<<<<<<<
 * 
 *     def get_schema(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_driver_name, 0, strlen(__pyx_v_driver_name), NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 582, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":576
 *         return OGR_L_GetFeatureCount(self.cogr_layer, 0)
 * 
 *     def get_driver(self):             # <<<<<<<<<<<<<<
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_driver", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":584
 *         return driver_name.decode()
 * 
 *     def get_schema(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef int n
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_19get_schema(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_19get_schema(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_schema (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_18get_schema(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_18get_schema(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  int __pyx_v_i;
  int __pyx_v_n;
  void *__pyx_v_cogr_featuredefn;
  void *__pyx_v_cogr_fielddefn;
  char const *__pyx_v_key_c;
  PyObject *__pyx_v_props = NULL;
  PyObject *__pyx_v_ignore_fields = NULL;
  PyObject *__pyx_v_encoding = NULL;
  char const *__pyx_v_key_b;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_fieldtypename = NULL;
  PyObject *__pyx_v_val = NULL;
  PyObject *__pyx_v_fmt = NULL;
  PyObject *__pyx_v_width = NULL;
  PyObject *__pyx_v_precision = NULL;
  PyObject *__pyx_v_ret = NULL;
  PyObject *__pyx_v_code = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("get_schema", 0);

  /* "fiona/ogrext.pyx":587
 *         cdef int i
 *         cdef int n
 *         cdef void *cogr_featuredefn = NULL             # <<<<<<<<<<<<<<
 *         cdef void *cogr_fielddefn = NULL
 *         cdef const char *key_c
 */
  __pyx_v_cogr_featuredefn = NULL;

  /* "fiona/ogrext.pyx":588
 *         cdef int n
 *         cdef void *cogr_featuredefn = NULL
 *         cdef void *cogr_fielddefn = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *key_c
 *         props = []
 */
  __pyx_v_cogr_fielddefn = NULL;

  /* "fiona/ogrext.pyx":590
 *         cdef void *cogr_fielddefn = NULL
 *         cdef const char *key_c
 *         props = []             # <<<<<<<<<<<<<<
 * 
 *         if self.cogr_layer == NULL:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_props = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":592
 *         props = []
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":593
 * 
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 * 
 *         if self.collection.ignore_fields:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 593, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 593, __pyx_L1_error)

    /* "fiona/ogrext.pyx":592
 *         props = []
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  }

  /* "fiona/ogrext.pyx":595
 *             raise ValueError("Null layer")
 * 
 *         if self.collection.ignore_fields:             # <<<<<<<<<<<<<<
 *             ignore_fields = self.collection.ignore_fields
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 595, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":596
 * 
 *         if self.collection.ignore_fields:
 *             ignore_fields = self.collection.ignore_fields             # <<<<<<<<<<<<<<
 *         else:
 *             ignore_fields = set()
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_ignore_fields = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":595
 *             raise ValueError("Null layer")
 * 
 *         if self.collection.ignore_fields:             # <<<<<<<<<<<<<<
 *             ignore_fields = self.collection.ignore_fields
 *         else:
 */
    goto __pyx_L4;
  }

  /* "fiona/ogrext.pyx":598
 *             ignore_fields = self.collection.ignore_fields
 *         else:
 *             ignore_fields = set()             # <<<<<<<<<<<<<<
 * 
 *         cogr_featuredefn = OGR_L_GetLayerDefn(self.cogr_layer)
 */
  /*else*/ {
    __pyx_t_1 = PySet_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 598, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_ignore_fields = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L4:;

  /* "fiona/ogrext.pyx":600
 *             ignore_fields = set()
 * 
 *         cogr_featuredefn = OGR_L_GetLayerDefn(self.cogr_layer)             # <<<<<<<<<<<<<<
 *         if cogr_featuredefn == NULL:
 *             raise ValueError("Null feature definition")
 */
  __pyx_v_cogr_featuredefn = OGR_L_GetLayerDefn(__pyx_v_self->cogr_layer);

  /* "fiona/ogrext.pyx":601
 * 
 *         cogr_featuredefn = OGR_L_GetLayerDefn(self.cogr_layer)
 *         if cogr_featuredefn == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature definition")
 * 
 */
  __pyx_t_2 = ((__pyx_v_cogr_featuredefn == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":602
 *         cogr_featuredefn = OGR_L_GetLayerDefn(self.cogr_layer)
 *         if cogr_featuredefn == NULL:
 *             raise ValueError("Null feature definition")             # <<<<<<<<<<<<<<
 * 
 *         encoding = self._get_internal_encoding()
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 602, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 602, __pyx_L1_error)

    /* "fiona/ogrext.pyx":601
 * 
 *         cogr_featuredefn = OGR_L_GetLayerDefn(self.cogr_layer)
 *         if cogr_featuredefn == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null feature definition")
 * 
 */
  }

  /* "fiona/ogrext.pyx":604
 *             raise ValueError("Null feature definition")
 * 
 *         encoding = self._get_internal_encoding()             # <<<<<<<<<<<<<<
 * 
 *         n = OGR_FD_GetFieldCount(cogr_featuredefn)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_encoding = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":606
 *         encoding = self._get_internal_encoding()
 * 
 *         n = OGR_FD_GetFieldCount(cogr_featuredefn)             # <<<<<<<<<<<<<<
 * 
 *         for i from 0 <= i < n:
 */
  __pyx_v_n = OGR_FD_GetFieldCount(__pyx_v_cogr_featuredefn);

  /* "fiona/ogrext.pyx":608
 *         n = OGR_FD_GetFieldCount(cogr_featuredefn)
 * 
 *         for i from 0 <= i < n:             # <<<<<<<<<<<<<<
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)
 *             if cogr_fielddefn == NULL:
 */
  __pyx_t_5 = __pyx_v_n;
  for (__pyx_v_i = 0; __pyx_v_i < __pyx_t_5; __pyx_v_i++) {

    /* "fiona/ogrext.pyx":609
 * 
 *         for i from 0 <= i < n:
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)             # <<<<<<<<<<<<<<
 *             if cogr_fielddefn == NULL:
 *                 raise ValueError("Null field definition")
 */
    __pyx_v_cogr_fielddefn = OGR_FD_GetFieldDefn(__pyx_v_cogr_featuredefn, __pyx_v_i);

    /* "fiona/ogrext.pyx":610
 *         for i from 0 <= i < n:
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)
 *             if cogr_fielddefn == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null field definition")
 * 
 */
    __pyx_t_2 = ((__pyx_v_cogr_fielddefn == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "fiona/ogrext.pyx":611
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)
 *             if cogr_fielddefn == NULL:
 *                 raise ValueError("Null field definition")             # <<<<<<<<<<<<<<
 * 
 *             key_c = OGR_Fld_GetNameRef(cogr_fielddefn)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 611, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(0, 611, __pyx_L1_error)

      /* "fiona/ogrext.pyx":610
 *         for i from 0 <= i < n:
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)
 *             if cogr_fielddefn == NULL:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Null field definition")
 * 
 */
    }

    /* "fiona/ogrext.pyx":613
 *                 raise ValueError("Null field definition")
 * 
 *             key_c = OGR_Fld_GetNameRef(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *             key_b = key_c
 * 
 */
    __pyx_v_key_c = OGR_Fld_GetNameRef(__pyx_v_cogr_fielddefn);

    /* "fiona/ogrext.pyx":614
 * 
 *             key_c = OGR_Fld_GetNameRef(cogr_fielddefn)
 *             key_b = key_c             # <<<<<<<<<<<<<<
 * 
 *             if not bool(key_b):
 */
    __pyx_v_key_b = __pyx_v_key_c;

    /* "fiona/ogrext.pyx":616
 *             key_b = key_c
 * 
 *             if not bool(key_b):             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid field name ref: %s" % key)
 * 
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_key_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 616, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = (PyBytes_GET_SIZE(__pyx_t_1) != 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = ((!((!(!__pyx_t_2)) != 0)) != 0);
    if (unlikely(__pyx_t_6)) {

      /* "fiona/ogrext.pyx":617
 * 
 *             if not bool(key_b):
 *                 raise ValueError("Invalid field name ref: %s" % key)             # <<<<<<<<<<<<<<
 * 
 *             key = key_b.decode(encoding)
 */
      if (unlikely(!__pyx_v_key)) { __Pyx_RaiseUnboundLocalError("key"); __PYX_ERR(0, 617, __pyx_L1_error) }
      __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Invalid_field_name_ref_s, __pyx_v_key); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 617, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 617, __pyx_L1_error)

      /* "fiona/ogrext.pyx":616
 *             key_b = key_c
 * 
 *             if not bool(key_b):             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid field name ref: %s" % key)
 * 
 */
    }

    /* "fiona/ogrext.pyx":619
 *                 raise ValueError("Invalid field name ref: %s" % key)
 * 
 *             key = key_b.decode(encoding)             # <<<<<<<<<<<<<<
 * 
 *             if key in ignore_fields:
 */
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_key_b); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_decode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_1, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_encoding);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 619, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":621
 *             key = key_b.decode(encoding)
 * 
 *             if key in ignore_fields:             # <<<<<<<<<<<<<<
 *                 log.debug("By request, ignoring field %r", key)
 *                 continue
 */
    __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_v_key, __pyx_v_ignore_fields, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 621, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_6 != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":622
 * 
 *             if key in ignore_fields:
 *                 log.debug("By request, ignoring field %r", key)             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 622, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_By_request_ignoring_field_r, __pyx_v_key};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_By_request_ignoring_field_r, __pyx_v_key};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_By_request_ignoring_field_r);
        __Pyx_GIVEREF(__pyx_kp_u_By_request_ignoring_field_r);
        PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_u_By_request_ignoring_field_r);
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_GIVEREF(__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_key);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 622, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":623
 *             if key in ignore_fields:
 *                 log.debug("By request, ignoring field %r", key)
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(cogr_fielddefn)]
 */
      goto __pyx_L6_continue;

      /* "fiona/ogrext.pyx":621
 *             key = key_b.decode(encoding)
 * 
 *             if key in ignore_fields:             # <<<<<<<<<<<<<<
 *                 log.debug("By request, ignoring field %r", key)
 *                 continue
 */
    }

    /* "fiona/ogrext.pyx":625
 *                 continue
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(cogr_fielddefn)]             # <<<<<<<<<<<<<<
 *             if not fieldtypename:
 *                 log.warning(
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_FIELD_TYPES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = OGR_Fld_GetType(__pyx_v_cogr_fielddefn);
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_3, __pyx_t_7, int, 1, __Pyx_PyInt_From_int, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 625, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fieldtypename, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":626
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(cogr_fielddefn)]
 *             if not fieldtypename:             # <<<<<<<<<<<<<<
 *                 log.warning(
 *                     "Skipping field %s: invalid type %s",
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_fieldtypename); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 626, __pyx_L1_error)
    __pyx_t_6 = ((!__pyx_t_2) != 0);
    if (__pyx_t_6) {

      /* "fiona/ogrext.pyx":627
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(cogr_fielddefn)]
 *             if not fieldtypename:
 *                 log.warning(             # <<<<<<<<<<<<<<
 *                     "Skipping field %s: invalid type %s",
 *                     key,
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 627, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":630
 *                     "Skipping field %s: invalid type %s",
 *                     key,
 *                     OGR_Fld_GetType(cogr_fielddefn))             # <<<<<<<<<<<<<<
 *                 continue
 * 
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(OGR_Fld_GetType(__pyx_v_cogr_fielddefn)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 630, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = NULL;
      __pyx_t_7 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_7 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_kp_u_Skipping_field_s_invalid_type_s, __pyx_v_key, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_kp_u_Skipping_field_s_invalid_type_s, __pyx_v_key, __pyx_t_3};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-__pyx_t_7, 3+__pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(3+__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Skipping_field_s_invalid_type_s);
        __Pyx_GIVEREF(__pyx_kp_u_Skipping_field_s_invalid_type_s);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_7, __pyx_kp_u_Skipping_field_s_invalid_type_s);
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_GIVEREF(__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_7, __pyx_v_key);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_7, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 627, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":631
 *                     key,
 *                     OGR_Fld_GetType(cogr_fielddefn))
 *                 continue             # <<<<<<<<<<<<<<
 * 
 *             val = fieldtypename
 */
      goto __pyx_L6_continue;

      /* "fiona/ogrext.pyx":626
 * 
 *             fieldtypename = FIELD_TYPES[OGR_Fld_GetType(cogr_fielddefn)]
 *             if not fieldtypename:             # <<<<<<<<<<<<<<
 *                 log.warning(
 *                     "Skipping field %s: invalid type %s",
 */
    }

    /* "fiona/ogrext.pyx":633
 *                 continue
 * 
 *             val = fieldtypename             # <<<<<<<<<<<<<<
 *             if fieldtypename == 'float':
 *                 fmt = ""
 */
    __Pyx_INCREF(__pyx_v_fieldtypename);
    __Pyx_XDECREF_SET(__pyx_v_val, __pyx_v_fieldtypename);

    /* "fiona/ogrext.pyx":634
 * 
 *             val = fieldtypename
 *             if fieldtypename == 'float':             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
    __pyx_t_6 = (__Pyx_PyUnicode_Equals(__pyx_v_fieldtypename, __pyx_n_u_float, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 634, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "fiona/ogrext.pyx":635
 *             val = fieldtypename
 *             if fieldtypename == 'float':
 *                 fmt = ""             # <<<<<<<<<<<<<<
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width: # and width != 24:
 */
      __Pyx_INCREF(__pyx_kp_u__12);
      __Pyx_XDECREF_SET(__pyx_v_fmt, __pyx_kp_u__12);

      /* "fiona/ogrext.pyx":636
 *             if fieldtypename == 'float':
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *                 if width: # and width != 24:
 *                     fmt = ":%d" % width
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(OGR_Fld_GetWidth(__pyx_v_cogr_fielddefn)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 636, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_width, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":637
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width: # and width != 24:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_width); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 637, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "fiona/ogrext.pyx":638
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width: # and width != 24:
 *                     fmt = ":%d" % width             # <<<<<<<<<<<<<<
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 *                 if precision: # and precision != 15:
 */
        __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_d, __pyx_v_width); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 638, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF_SET(__pyx_v_fmt, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":637
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width: # and width != 24:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 */
      }

      /* "fiona/ogrext.pyx":639
 *                 if width: # and width != 24:
 *                     fmt = ":%d" % width
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *                 if precision: # and precision != 15:
 *                     fmt += ".%d" % precision
 */
      __pyx_t_1 = __Pyx_PyInt_From_int(OGR_Fld_GetPrecision(__pyx_v_cogr_fielddefn)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_precision, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":640
 *                     fmt = ":%d" % width
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 *                 if precision: # and precision != 15:             # <<<<<<<<<<<<<<
 *                     fmt += ".%d" % precision
 *                 val = "float" + fmt
 */
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_v_precision); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 640, __pyx_L1_error)
      if (__pyx_t_6) {

        /* "fiona/ogrext.pyx":641
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 *                 if precision: # and precision != 15:
 *                     fmt += ".%d" % precision             # <<<<<<<<<<<<<<
 *                 val = "float" + fmt
 *             elif fieldtypename in ('int32', 'int64'):
 */
        __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_d_2, __pyx_v_precision); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_v_fmt, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 641, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_fmt, ((PyObject*)__pyx_t_8));
        __pyx_t_8 = 0;

        /* "fiona/ogrext.pyx":640
 *                     fmt = ":%d" % width
 *                 precision = OGR_Fld_GetPrecision(cogr_fielddefn)
 *                 if precision: # and precision != 15:             # <<<<<<<<<<<<<<
 *                     fmt += ".%d" % precision
 *                 val = "float" + fmt
 */
      }

      /* "fiona/ogrext.pyx":642
 *                 if precision: # and precision != 15:
 *                     fmt += ".%d" % precision
 *                 val = "float" + fmt             # <<<<<<<<<<<<<<
 *             elif fieldtypename in ('int32', 'int64'):
 *                 fmt = ""
 */
      __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_n_u_float, __pyx_v_fmt); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 642, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":634
 * 
 *             val = fieldtypename
 *             if fieldtypename == 'float':             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
      goto __pyx_L12;
    }

    /* "fiona/ogrext.pyx":643
 *                     fmt += ".%d" % precision
 *                 val = "float" + fmt
 *             elif fieldtypename in ('int32', 'int64'):             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
    __Pyx_INCREF(__pyx_v_fieldtypename);
    __pyx_t_8 = __pyx_v_fieldtypename;
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_int32, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __pyx_t_6 = __pyx_t_2;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_8, __pyx_n_u_int64, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 643, __pyx_L1_error)
    __pyx_t_6 = __pyx_t_2;
    __pyx_L15_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_2 = (__pyx_t_6 != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":644
 *                 val = "float" + fmt
 *             elif fieldtypename in ('int32', 'int64'):
 *                 fmt = ""             # <<<<<<<<<<<<<<
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:
 */
      __Pyx_INCREF(__pyx_kp_u__12);
      __Pyx_XDECREF_SET(__pyx_v_fmt, __pyx_kp_u__12);

      /* "fiona/ogrext.pyx":645
 *             elif fieldtypename in ('int32', 'int64'):
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *                 if width:
 *                     fmt = ":%d" % width
 */
      __pyx_t_8 = __Pyx_PyInt_From_int(OGR_Fld_GetWidth(__pyx_v_cogr_fielddefn)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 645, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_width, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":646
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 val = 'int' + fmt
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_width); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 646, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "fiona/ogrext.pyx":647
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:
 *                     fmt = ":%d" % width             # <<<<<<<<<<<<<<
 *                 val = 'int' + fmt
 *             elif fieldtypename == 'str':
 */
        __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_d, __pyx_v_width); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 647, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF_SET(__pyx_v_fmt, ((PyObject*)__pyx_t_8));
        __pyx_t_8 = 0;

        /* "fiona/ogrext.pyx":646
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 val = 'int' + fmt
 */
      }

      /* "fiona/ogrext.pyx":648
 *                 if width:
 *                     fmt = ":%d" % width
 *                 val = 'int' + fmt             # <<<<<<<<<<<<<<
 *             elif fieldtypename == 'str':
 *                 fmt = ""
 */
      __pyx_t_8 = __Pyx_PyUnicode_Concat(__pyx_n_u_int, __pyx_v_fmt); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 648, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":643
 *                     fmt += ".%d" % precision
 *                 val = "float" + fmt
 *             elif fieldtypename in ('int32', 'int64'):             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
      goto __pyx_L12;
    }

    /* "fiona/ogrext.pyx":649
 *                     fmt = ":%d" % width
 *                 val = 'int' + fmt
 *             elif fieldtypename == 'str':             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
    __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_fieldtypename, __pyx_n_u_str, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 649, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":650
 *                 val = 'int' + fmt
 *             elif fieldtypename == 'str':
 *                 fmt = ""             # <<<<<<<<<<<<<<
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:
 */
      __Pyx_INCREF(__pyx_kp_u__12);
      __Pyx_XDECREF_SET(__pyx_v_fmt, __pyx_kp_u__12);

      /* "fiona/ogrext.pyx":651
 *             elif fieldtypename == 'str':
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *                 if width:
 *                     fmt = ":%d" % width
 */
      __pyx_t_8 = __Pyx_PyInt_From_int(OGR_Fld_GetWidth(__pyx_v_cogr_fielddefn)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 651, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_XDECREF_SET(__pyx_v_width, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":652
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 val = fieldtypename + fmt
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_width); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 652, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "fiona/ogrext.pyx":653
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:
 *                     fmt = ":%d" % width             # <<<<<<<<<<<<<<
 *                 val = fieldtypename + fmt
 * 
 */
        __pyx_t_8 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_d, __pyx_v_width); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 653, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF_SET(__pyx_v_fmt, ((PyObject*)__pyx_t_8));
        __pyx_t_8 = 0;

        /* "fiona/ogrext.pyx":652
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 *                 if width:             # <<<<<<<<<<<<<<
 *                     fmt = ":%d" % width
 *                 val = fieldtypename + fmt
 */
      }

      /* "fiona/ogrext.pyx":654
 *                 if width:
 *                     fmt = ":%d" % width
 *                 val = fieldtypename + fmt             # <<<<<<<<<<<<<<
 * 
 *             props.append((key, val))
 */
      __pyx_t_8 = PyNumber_Add(__pyx_v_fieldtypename, __pyx_v_fmt); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_val, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":649
 *                     fmt = ":%d" % width
 *                 val = 'int' + fmt
 *             elif fieldtypename == 'str':             # <<<<<<<<<<<<<<
 *                 fmt = ""
 *                 width = OGR_Fld_GetWidth(cogr_fielddefn)
 */
    }
    __pyx_L12:;

    /* "fiona/ogrext.pyx":656
 *                 val = fieldtypename + fmt
 * 
 *             props.append((key, val))             # <<<<<<<<<<<<<<
 * 
 *         ret = {"properties": OrderedDict(props)}
 */
    __pyx_t_8 = PyTuple_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_INCREF(__pyx_v_key);
    __Pyx_GIVEREF(__pyx_v_key);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_key);
    __Pyx_INCREF(__pyx_v_val);
    __Pyx_GIVEREF(__pyx_v_val);
    PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_v_val);
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_props, __pyx_t_8); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 656, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_L6_continue:;
  }

  /* "fiona/ogrext.pyx":658
 *             props.append((key, val))
 * 
 *         ret = {"properties": OrderedDict(props)}             # <<<<<<<<<<<<<<
 * 
 *         if not self.collection.ignore_geometry:
 */
  __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_3, __pyx_v_props) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_v_props);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_u_properties, __pyx_t_1) < 0) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ret = ((PyObject*)__pyx_t_8);
  __pyx_t_8 = 0;

  /* "fiona/ogrext.pyx":660
 *         ret = {"properties": OrderedDict(props)}
 * 
 *         if not self.collection.ignore_geometry:             # <<<<<<<<<<<<<<
 *             code = normalize_geometry_type_code(
 *                 OGR_FD_GetGeomType(cogr_featuredefn))
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_geometry); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_6 = ((!__pyx_t_2) != 0);
  if (__pyx_t_6) {

    /* "fiona/ogrext.pyx":661
 * 
 *         if not self.collection.ignore_geometry:
 *             code = normalize_geometry_type_code(             # <<<<<<<<<<<<<<
 *                 OGR_FD_GetGeomType(cogr_featuredefn))
 *             ret["geometry"] = GEOMETRY_TYPES[code]
 */
    __pyx_t_8 = __pyx_f_5fiona_9_geometry_normalize_geometry_type_code(OGR_FD_GetGeomType(__pyx_v_cogr_featuredefn)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 661, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_v_code = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":663
 *             code = normalize_geometry_type_code(
 *                 OGR_FD_GetGeomType(cogr_featuredefn))
 *             ret["geometry"] = GEOMETRY_TYPES[code]             # <<<<<<<<<<<<<<
 * 
 *         return ret
 */
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_GEOMETRY_TYPES); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_t_8, __pyx_v_code); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(PyDict_SetItem(__pyx_v_ret, __pyx_n_u_geometry, __pyx_t_1) < 0)) __PYX_ERR(0, 663, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":660
 *         ret = {"properties": OrderedDict(props)}
 * 
 *         if not self.collection.ignore_geometry:             # <<<<<<<<<<<<<<
 *             code = normalize_geometry_type_code(
 *                 OGR_FD_GetGeomType(cogr_featuredefn))
 */
  }

  /* "fiona/ogrext.pyx":665
 *             ret["geometry"] = GEOMETRY_TYPES[code]
 * 
 *         return ret             # <<<<<<<<<<<<<<
 * 
 *     def get_crs(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_ret);
  __pyx_r = __pyx_v_ret;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":584
 *         return driver_name.decode()
 * 
 *     def get_schema(self):             # <<<<<<<<<<<<<<
 *         cdef int i
 *         cdef int n
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_schema", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_props);
  __Pyx_XDECREF(__pyx_v_ignore_fields);
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_fieldtypename);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_fmt);
  __Pyx_XDECREF(__pyx_v_width);
  __Pyx_XDECREF(__pyx_v_precision);
  __Pyx_XDECREF(__pyx_v_ret);
  __Pyx_XDECREF(__pyx_v_code);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":667
 *         return ret
 * 
 *     def get_crs(self):             # <<<<<<<<<<<<<<
 *         """Get the layer's CRS
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_21get_crs(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_20get_crs[] = "Get the layer's CRS\n\n        Returns\n        -------\n        CRS\n\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_21get_crs(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_crs (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_20get_crs(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_20get_crs(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  char *__pyx_v_proj_c;
  char const *__pyx_v_auth_key;
  char const *__pyx_v_auth_val;
  void *__pyx_v_cogr_crs;
  PyObject *__pyx_v_crs = NULL;
  int __pyx_v_retval;
  PyObject *__pyx_v_exc = NULL;
  char const *__pyx_v_key_b;
  PyObject *__pyx_v_key = NULL;
  char const *__pyx_v_val_b;
  PyObject *__pyx_v_val = NULL;
  char *__pyx_v_proj_b;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_param = NULL;
  PyObject *__pyx_v_kv = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  void *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_t_15;
  char const *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  Py_ssize_t __pyx_t_26;
  PyObject *(*__pyx_t_27)(PyObject *);
  __Pyx_RefNannySetupContext("get_crs", 0);

  /* "fiona/ogrext.pyx":675
 * 
 *         """
 *         cdef char *proj_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *auth_key = NULL
 *         cdef const char *auth_val = NULL
 */
  __pyx_v_proj_c = NULL;

  /* "fiona/ogrext.pyx":676
 *         """
 *         cdef char *proj_c = NULL
 *         cdef const char *auth_key = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *auth_val = NULL
 *         cdef void *cogr_crs = NULL
 */
  __pyx_v_auth_key = NULL;

  /* "fiona/ogrext.pyx":677
 *         cdef char *proj_c = NULL
 *         cdef const char *auth_key = NULL
 *         cdef const char *auth_val = NULL             # <<<<<<<<<<<<<<
 *         cdef void *cogr_crs = NULL
 * 
 */
  __pyx_v_auth_val = NULL;

  /* "fiona/ogrext.pyx":678
 *         cdef const char *auth_key = NULL
 *         cdef const char *auth_val = NULL
 *         cdef void *cogr_crs = NULL             # <<<<<<<<<<<<<<
 * 
 *         if self.cogr_layer == NULL:
 */
  __pyx_v_cogr_crs = NULL;

  /* "fiona/ogrext.pyx":680
 *         cdef void *cogr_crs = NULL
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fiona/ogrext.pyx":681
 * 
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 681, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 681, __pyx_L1_error)

    /* "fiona/ogrext.pyx":680
 *         cdef void *cogr_crs = NULL
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  }

  /* "fiona/ogrext.pyx":683
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 *         # TODO: we don't intend to use try/except for flow control
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "fiona/ogrext.pyx":684
 * 
 *         try:
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))             # <<<<<<<<<<<<<<
 *         # TODO: we don't intend to use try/except for flow control
 *         # this is a work around for a GDAL issue.
 */
      __pyx_t_6 = __pyx_f_5fiona_4_err_exc_wrap_pointer(OGR_L_GetSpatialRef(__pyx_v_self->cogr_layer)); if (unlikely(__pyx_t_6 == ((void *)NULL))) __PYX_ERR(0, 684, __pyx_L4_error)
      __pyx_v_cogr_crs = __pyx_t_6;

      /* "fiona/ogrext.pyx":683
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 *         # TODO: we don't intend to use try/except for flow control
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":687
 *         # TODO: we don't intend to use try/except for flow control
 *         # this is a work around for a GDAL issue.
 *         except FionaNullPointerError:             # <<<<<<<<<<<<<<
 *             log.debug("Layer has no coordinate system")
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_FionaNullPointerError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 687, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_7, __pyx_t_8);
    __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("fiona.ogrext.Session.get_crs", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 687, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "fiona/ogrext.pyx":688
 *         # this is a work around for a GDAL issue.
 *         except FionaNullPointerError:
 *             log.debug("Layer has no coordinate system")             # <<<<<<<<<<<<<<
 * 
 *         if cogr_crs is not NULL:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 688, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 688, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_kp_u_Layer_has_no_coordinate_system) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_u_Layer_has_no_coordinate_system);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 688, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "fiona/ogrext.pyx":683
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 *         # TODO: we don't intend to use try/except for flow control
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "fiona/ogrext.pyx":690
 *             log.debug("Layer has no coordinate system")
 * 
 *         if cogr_crs is not NULL:             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Got coordinate system")
 */
  __pyx_t_1 = ((__pyx_v_cogr_crs != NULL) != 0);
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":692
 *         if cogr_crs is not NULL:
 * 
 *             log.debug("Got coordinate system")             # <<<<<<<<<<<<<<
 *             crs = {}
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_log); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_kp_u_Got_coordinate_system) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u_Got_coordinate_system);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 692, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":693
 * 
 *             log.debug("Got coordinate system")
 *             crs = {}             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_crs = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":695
 *             crs = {}
 * 
 *             try:             # <<<<<<<<<<<<<<
 * 
 *                 retval = OSRAutoIdentifyEPSG(cogr_crs)
 */
    /*try:*/ {

      /* "fiona/ogrext.pyx":697
 *             try:
 * 
 *                 retval = OSRAutoIdentifyEPSG(cogr_crs)             # <<<<<<<<<<<<<<
 *                 if retval > 0:
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 */
      __pyx_v_retval = OSRAutoIdentifyEPSG(__pyx_v_cogr_crs);

      /* "fiona/ogrext.pyx":698
 * 
 *                 retval = OSRAutoIdentifyEPSG(cogr_crs)
 *                 if retval > 0:             # <<<<<<<<<<<<<<
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 * 
 */
      __pyx_t_1 = ((__pyx_v_retval > 0) != 0);
      if (__pyx_t_1) {

        /* "fiona/ogrext.pyx":699
 *                 retval = OSRAutoIdentifyEPSG(cogr_crs)
 *                 if retval > 0:
 *                     log.info("Failed to auto identify EPSG: %d", retval)             # <<<<<<<<<<<<<<
 * 
 *                 try:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_log); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 699, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_info); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 699, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_retval); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 699, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = NULL;
        __pyx_t_10 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_10 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_kp_u_Failed_to_auto_identify_EPSG_d, __pyx_t_8};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L14_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_kp_u_Failed_to_auto_identify_EPSG_d, __pyx_t_8};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L14_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        {
          __pyx_t_12 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 699, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (__pyx_t_9) {
            __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_9); __pyx_t_9 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Failed_to_auto_identify_EPSG_d);
          __Pyx_GIVEREF(__pyx_kp_u_Failed_to_auto_identify_EPSG_d);
          PyTuple_SET_ITEM(__pyx_t_12, 0+__pyx_t_10, __pyx_kp_u_Failed_to_auto_identify_EPSG_d);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_12, 1+__pyx_t_10, __pyx_t_8);
          __pyx_t_8 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_12, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 699, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":698
 * 
 *                 retval = OSRAutoIdentifyEPSG(cogr_crs)
 *                 if retval > 0:             # <<<<<<<<<<<<<<
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 * 
 */
      }

      /* "fiona/ogrext.pyx":701
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     auth_key = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityName(cogr_crs, NULL))
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        /*try:*/ {

          /* "fiona/ogrext.pyx":702
 * 
 *                 try:
 *                     auth_key = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityName(cogr_crs, NULL))             # <<<<<<<<<<<<<<
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 * 
 */
          __pyx_t_6 = __pyx_f_5fiona_4_err_exc_wrap_pointer(((void *)OSRGetAuthorityName(__pyx_v_cogr_crs, NULL))); if (unlikely(__pyx_t_6 == ((void *)NULL))) __PYX_ERR(0, 702, __pyx_L17_error)
          __pyx_v_auth_key = ((char const *)__pyx_t_6);

          /* "fiona/ogrext.pyx":703
 *                 try:
 *                     auth_key = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityName(cogr_crs, NULL))
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))             # <<<<<<<<<<<<<<
 * 
 *                 except CPLE_BaseError as exc:
 */
          __pyx_t_6 = __pyx_f_5fiona_4_err_exc_wrap_pointer(((void *)OSRGetAuthorityCode(__pyx_v_cogr_crs, NULL))); if (unlikely(__pyx_t_6 == ((void *)NULL))) __PYX_ERR(0, 703, __pyx_L17_error)
          __pyx_v_auth_val = ((char const *)__pyx_t_6);

          /* "fiona/ogrext.pyx":701
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     auth_key = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityName(cogr_crs, NULL))
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 */
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        goto __pyx_L22_try_end;
        __pyx_L17_error:;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "fiona/ogrext.pyx":705
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 * 
 *                 except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                     log.debug("{}".format(exc))
 * 
 */
        __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_7, &__pyx_t_12);
        __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_CPLE_BaseError); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L19_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_8);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_ErrRestore(__pyx_t_2, __pyx_t_7, __pyx_t_12);
        __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_12 = 0;
        if (__pyx_t_10) {
          __Pyx_AddTraceback("fiona.ogrext.Session.get_crs", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_12, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 705, __pyx_L19_except_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_7);
          __pyx_v_exc = __pyx_t_7;
          /*try:*/ {

            /* "fiona/ogrext.pyx":706
 * 
 *                 except CPLE_BaseError as exc:
 *                     log.debug("{}".format(exc))             # <<<<<<<<<<<<<<
 * 
 *                 if auth_key != NULL and auth_val != NULL:
 */
            __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_log); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 706, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_9);
            __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 706, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__13, __pyx_n_s_format); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 706, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_14 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
              __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
              if (likely(__pyx_t_14)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
                __Pyx_INCREF(__pyx_t_14);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_13, function);
              }
            }
            __pyx_t_9 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_13, __pyx_t_14, __pyx_v_exc) : __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_v_exc);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 706, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_9);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_13 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_11);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_11, function);
              }
            }
            __pyx_t_8 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_13, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_9);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 706, __pyx_L28_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }

          /* "fiona/ogrext.pyx":705
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 * 
 *                 except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                     log.debug("{}".format(exc))
 * 
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_exc);
              __pyx_v_exc = NULL;
              goto __pyx_L29;
            }
            __pyx_L28_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19) < 0)) __Pyx_ErrFetch(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __pyx_t_10 = __pyx_lineno; __pyx_t_15 = __pyx_clineno; __pyx_t_16 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_exc);
                __pyx_v_exc = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              }
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_XGIVEREF(__pyx_t_19);
              __Pyx_ErrRestore(__pyx_t_17, __pyx_t_18, __pyx_t_19);
              __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0;
              __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_15; __pyx_filename = __pyx_t_16;
              goto __pyx_L19_except_error;
            }
            __pyx_L29:;
          }
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          goto __pyx_L18_exception_handled;
        }
        goto __pyx_L19_except_error;
        __pyx_L19_except_error:;

        /* "fiona/ogrext.pyx":701
 *                     log.info("Failed to auto identify EPSG: %d", retval)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     auth_key = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityName(cogr_crs, NULL))
 *                     auth_val = <const char *>exc_wrap_pointer(<void *>OSRGetAuthorityCode(cogr_crs, NULL))
 */
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
        goto __pyx_L14_error;
        __pyx_L18_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_5);
        __Pyx_XGIVEREF(__pyx_t_4);
        __Pyx_XGIVEREF(__pyx_t_3);
        __Pyx_ExceptionReset(__pyx_t_5, __pyx_t_4, __pyx_t_3);
        __pyx_L22_try_end:;
      }

      /* "fiona/ogrext.pyx":708
 *                     log.debug("{}".format(exc))
 * 
 *                 if auth_key != NULL and auth_val != NULL:             # <<<<<<<<<<<<<<
 *                     key_b = auth_key
 *                     key = key_b.decode('utf-8')
 */
      __pyx_t_23 = ((__pyx_v_auth_key != NULL) != 0);
      if (__pyx_t_23) {
      } else {
        __pyx_t_1 = __pyx_t_23;
        goto __pyx_L35_bool_binop_done;
      }
      __pyx_t_23 = ((__pyx_v_auth_val != NULL) != 0);
      __pyx_t_1 = __pyx_t_23;
      __pyx_L35_bool_binop_done:;
      if (__pyx_t_1) {

        /* "fiona/ogrext.pyx":709
 * 
 *                 if auth_key != NULL and auth_val != NULL:
 *                     key_b = auth_key             # <<<<<<<<<<<<<<
 *                     key = key_b.decode('utf-8')
 *                     if key == 'EPSG':
 */
        __pyx_v_key_b = __pyx_v_auth_key;

        /* "fiona/ogrext.pyx":710
 *                 if auth_key != NULL and auth_val != NULL:
 *                     key_b = auth_key
 *                     key = key_b.decode('utf-8')             # <<<<<<<<<<<<<<
 *                     if key == 'EPSG':
 *                         val_b = auth_val
 */
        __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_key_b, 0, strlen(__pyx_v_key_b), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 710, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_v_key = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":711
 *                     key_b = auth_key
 *                     key = key_b.decode('utf-8')
 *                     if key == 'EPSG':             # <<<<<<<<<<<<<<
 *                         val_b = auth_val
 *                         val = val_b.decode('utf-8')
 */
        __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_key, __pyx_n_u_EPSG, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 711, __pyx_L14_error)
        if (__pyx_t_1) {

          /* "fiona/ogrext.pyx":712
 *                     key = key_b.decode('utf-8')
 *                     if key == 'EPSG':
 *                         val_b = auth_val             # <<<<<<<<<<<<<<
 *                         val = val_b.decode('utf-8')
 *                         crs['init'] = "epsg:" + val
 */
          __pyx_v_val_b = __pyx_v_auth_val;

          /* "fiona/ogrext.pyx":713
 *                     if key == 'EPSG':
 *                         val_b = auth_val
 *                         val = val_b.decode('utf-8')             # <<<<<<<<<<<<<<
 *                         crs['init'] = "epsg:" + val
 * 
 */
          __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_val_b, 0, strlen(__pyx_v_val_b), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 713, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_v_val = __pyx_t_2;
          __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":714
 *                         val_b = auth_val
 *                         val = val_b.decode('utf-8')
 *                         crs['init'] = "epsg:" + val             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
          __pyx_t_2 = PyNumber_Add(__pyx_kp_u_epsg, __pyx_v_val); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 714, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          if (unlikely(PyDict_SetItem(__pyx_v_crs, __pyx_n_u_init, __pyx_t_2) < 0)) __PYX_ERR(0, 714, __pyx_L14_error)
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":711
 *                     key_b = auth_key
 *                     key = key_b.decode('utf-8')
 *                     if key == 'EPSG':             # <<<<<<<<<<<<<<
 *                         val_b = auth_val
 *                         val = val_b.decode('utf-8')
 */
        }

        /* "fiona/ogrext.pyx":708
 *                     log.debug("{}".format(exc))
 * 
 *                 if auth_key != NULL and auth_val != NULL:             # <<<<<<<<<<<<<<
 *                     key_b = auth_key
 *                     key = key_b.decode('utf-8')
 */
        goto __pyx_L34;
      }

      /* "fiona/ogrext.pyx":717
 * 
 *                 else:
 *                     OSRExportToProj4(cogr_crs, &proj_c)             # <<<<<<<<<<<<<<
 *                     if proj_c == NULL:
 *                         raise ValueError("Null projection")
 */
      /*else*/ {
        (void)(OSRExportToProj4(__pyx_v_cogr_crs, (&__pyx_v_proj_c)));

        /* "fiona/ogrext.pyx":718
 *                 else:
 *                     OSRExportToProj4(cogr_crs, &proj_c)
 *                     if proj_c == NULL:             # <<<<<<<<<<<<<<
 *                         raise ValueError("Null projection")
 *                     proj_b = proj_c
 */
        __pyx_t_1 = ((__pyx_v_proj_c == NULL) != 0);
        if (unlikely(__pyx_t_1)) {

          /* "fiona/ogrext.pyx":719
 *                     OSRExportToProj4(cogr_crs, &proj_c)
 *                     if proj_c == NULL:
 *                         raise ValueError("Null projection")             # <<<<<<<<<<<<<<
 *                     proj_b = proj_c
 *                     log.debug("Params: %s", proj_b)
 */
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 719, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_Raise(__pyx_t_2, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __PYX_ERR(0, 719, __pyx_L14_error)

          /* "fiona/ogrext.pyx":718
 *                 else:
 *                     OSRExportToProj4(cogr_crs, &proj_c)
 *                     if proj_c == NULL:             # <<<<<<<<<<<<<<
 *                         raise ValueError("Null projection")
 *                     proj_b = proj_c
 */
        }

        /* "fiona/ogrext.pyx":720
 *                     if proj_c == NULL:
 *                         raise ValueError("Null projection")
 *                     proj_b = proj_c             # <<<<<<<<<<<<<<
 *                     log.debug("Params: %s", proj_b)
 *                     value = proj_b.decode()
 */
        __pyx_v_proj_b = __pyx_v_proj_c;

        /* "fiona/ogrext.pyx":721
 *                         raise ValueError("Null projection")
 *                     proj_b = proj_c
 *                     log.debug("Params: %s", proj_b)             # <<<<<<<<<<<<<<
 *                     value = proj_b.decode()
 *                     value = value.strip()
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_log); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 721, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_debug); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 721, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyBytes_FromString(__pyx_v_proj_b); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 721, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        __pyx_t_15 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_15 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_12)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_kp_u_Params_s, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 721, __pyx_L14_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
          PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_kp_u_Params_s, __pyx_t_7};
          __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_15, 2+__pyx_t_15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 721, __pyx_L14_error)
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        } else
        #endif
        {
          __pyx_t_11 = PyTuple_New(2+__pyx_t_15); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 721, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          if (__pyx_t_8) {
            __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Params_s);
          __Pyx_GIVEREF(__pyx_kp_u_Params_s);
          PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_15, __pyx_kp_u_Params_s);
          __Pyx_GIVEREF(__pyx_t_7);
          PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_15, __pyx_t_7);
          __pyx_t_7 = 0;
          __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 721, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":722
 *                     proj_b = proj_c
 *                     log.debug("Params: %s", proj_b)
 *                     value = proj_b.decode()             # <<<<<<<<<<<<<<
 *                     value = value.strip()
 *                     for param in value.split():
 */
        __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_proj_b, 0, strlen(__pyx_v_proj_b), NULL, NULL, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_v_value = __pyx_t_2;
        __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":723
 *                     log.debug("Params: %s", proj_b)
 *                     value = proj_b.decode()
 *                     value = value.strip()             # <<<<<<<<<<<<<<
 *                     for param in value.split():
 *                         kv = param.split("=")
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_strip); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 723, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_2 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 723, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_2);
        __pyx_t_2 = 0;

        /* "fiona/ogrext.pyx":724
 *                     value = proj_b.decode()
 *                     value = value.strip()
 *                     for param in value.split():             # <<<<<<<<<<<<<<
 *                         kv = param.split("=")
 *                         if len(kv) == 2:
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_split); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 724, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_2 = (__pyx_t_11) ? __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_t_11) : __Pyx_PyObject_CallNoArg(__pyx_t_12);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 724, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
          __pyx_t_12 = __pyx_t_2; __Pyx_INCREF(__pyx_t_12); __pyx_t_24 = 0;
          __pyx_t_25 = NULL;
        } else {
          __pyx_t_24 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 724, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_25 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 724, __pyx_L14_error)
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        for (;;) {
          if (likely(!__pyx_t_25)) {
            if (likely(PyList_CheckExact(__pyx_t_12))) {
              if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_12)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_24); __Pyx_INCREF(__pyx_t_2); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 724, __pyx_L14_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_12, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 724, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            } else {
              if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_24); __Pyx_INCREF(__pyx_t_2); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 724, __pyx_L14_error)
              #else
              __pyx_t_2 = PySequence_ITEM(__pyx_t_12, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 724, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_2);
              #endif
            }
          } else {
            __pyx_t_2 = __pyx_t_25(__pyx_t_12);
            if (unlikely(!__pyx_t_2)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 724, __pyx_L14_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_2);
          }
          __Pyx_XDECREF_SET(__pyx_v_param, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":725
 *                     value = value.strip()
 *                     for param in value.split():
 *                         kv = param.split("=")             # <<<<<<<<<<<<<<
 *                         if len(kv) == 2:
 *                             k, v = kv
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_param, __pyx_n_s_split); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 725, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_7, __pyx_kp_u__15) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_u__15);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 725, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF_SET(__pyx_v_kv, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":726
 *                     for param in value.split():
 *                         kv = param.split("=")
 *                         if len(kv) == 2:             # <<<<<<<<<<<<<<
 *                             k, v = kv
 *                             try:
 */
          __pyx_t_26 = PyObject_Length(__pyx_v_kv); if (unlikely(__pyx_t_26 == ((Py_ssize_t)-1))) __PYX_ERR(0, 726, __pyx_L14_error)
          __pyx_t_1 = ((__pyx_t_26 == 2) != 0);
          if (__pyx_t_1) {

            /* "fiona/ogrext.pyx":727
 *                         kv = param.split("=")
 *                         if len(kv) == 2:
 *                             k, v = kv             # <<<<<<<<<<<<<<
 *                             try:
 *                                 v = float(v)
 */
            if ((likely(PyTuple_CheckExact(__pyx_v_kv))) || (PyList_CheckExact(__pyx_v_kv))) {
              PyObject* sequence = __pyx_v_kv;
              Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
              if (unlikely(size != 2)) {
                if (size > 2) __Pyx_RaiseTooManyValuesError(2);
                else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
                __PYX_ERR(0, 727, __pyx_L14_error)
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              if (likely(PyTuple_CheckExact(sequence))) {
                __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
                __pyx_t_11 = PyTuple_GET_ITEM(sequence, 1); 
              } else {
                __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
                __pyx_t_11 = PyList_GET_ITEM(sequence, 1); 
              }
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_11);
              #else
              __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_2);
              __pyx_t_11 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 727, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_11);
              #endif
            } else {
              Py_ssize_t index = -1;
              __pyx_t_7 = PyObject_GetIter(__pyx_v_kv); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 727, __pyx_L14_error)
              __Pyx_GOTREF(__pyx_t_7);
              __pyx_t_27 = Py_TYPE(__pyx_t_7)->tp_iternext;
              index = 0; __pyx_t_2 = __pyx_t_27(__pyx_t_7); if (unlikely(!__pyx_t_2)) goto __pyx_L42_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_2);
              index = 1; __pyx_t_11 = __pyx_t_27(__pyx_t_7); if (unlikely(!__pyx_t_11)) goto __pyx_L42_unpacking_failed;
              __Pyx_GOTREF(__pyx_t_11);
              if (__Pyx_IternextUnpackEndCheck(__pyx_t_27(__pyx_t_7), 2) < 0) __PYX_ERR(0, 727, __pyx_L14_error)
              __pyx_t_27 = NULL;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              goto __pyx_L43_unpacking_done;
              __pyx_L42_unpacking_failed:;
              __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
              __pyx_t_27 = NULL;
              if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
              __PYX_ERR(0, 727, __pyx_L14_error)
              __pyx_L43_unpacking_done:;
            }
            __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
            __pyx_t_2 = 0;
            __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "fiona/ogrext.pyx":728
 *                         if len(kv) == 2:
 *                             k, v = kv
 *                             try:             # <<<<<<<<<<<<<<
 *                                 v = float(v)
 *                                 if v % 1 == 0:
 */
            {
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
              __Pyx_XGOTREF(__pyx_t_3);
              __Pyx_XGOTREF(__pyx_t_4);
              __Pyx_XGOTREF(__pyx_t_5);
              /*try:*/ {

                /* "fiona/ogrext.pyx":729
 *                             k, v = kv
 *                             try:
 *                                 v = float(v)             # <<<<<<<<<<<<<<
 *                                 if v % 1 == 0:
 *                                     v = int(v)
 */
                __pyx_t_11 = __Pyx_PyNumber_Float(__pyx_v_v); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 729, __pyx_L44_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_11);
                __pyx_t_11 = 0;

                /* "fiona/ogrext.pyx":730
 *                             try:
 *                                 v = float(v)
 *                                 if v % 1 == 0:             # <<<<<<<<<<<<<<
 *                                     v = int(v)
 *                             except ValueError:
 */
                __pyx_t_11 = __Pyx_PyInt_RemainderObjC(__pyx_v_v, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 730, __pyx_L44_error)
                __Pyx_GOTREF(__pyx_t_11);
                __pyx_t_2 = __Pyx_PyInt_EqObjC(__pyx_t_11, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L44_error)
                __Pyx_GOTREF(__pyx_t_2);
                __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 730, __pyx_L44_error)
                __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
                if (__pyx_t_1) {

                  /* "fiona/ogrext.pyx":731
 *                                 v = float(v)
 *                                 if v % 1 == 0:
 *                                     v = int(v)             # <<<<<<<<<<<<<<
 *                             except ValueError:
 *                                 # Leave v as a string
 */
                  __pyx_t_2 = __Pyx_PyNumber_Int(__pyx_v_v); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 731, __pyx_L44_error)
                  __Pyx_GOTREF(__pyx_t_2);
                  __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_2);
                  __pyx_t_2 = 0;

                  /* "fiona/ogrext.pyx":730
 *                             try:
 *                                 v = float(v)
 *                                 if v % 1 == 0:             # <<<<<<<<<<<<<<
 *                                     v = int(v)
 *                             except ValueError:
 */
                }

                /* "fiona/ogrext.pyx":728
 *                         if len(kv) == 2:
 *                             k, v = kv
 *                             try:             # <<<<<<<<<<<<<<
 *                                 v = float(v)
 *                                 if v % 1 == 0:
 */
              }
              __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
              goto __pyx_L51_try_end;
              __pyx_L44_error:;
              __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
              __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
              __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
              __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

              /* "fiona/ogrext.pyx":732
 *                                 if v % 1 == 0:
 *                                     v = int(v)
 *                             except ValueError:             # <<<<<<<<<<<<<<
 *                                 # Leave v as a string
 *                                 pass
 */
              __pyx_t_15 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
              if (__pyx_t_15) {
                __Pyx_ErrRestore(0,0,0);
                goto __pyx_L45_exception_handled;
              }
              goto __pyx_L46_except_error;
              __pyx_L46_except_error:;

              /* "fiona/ogrext.pyx":728
 *                         if len(kv) == 2:
 *                             k, v = kv
 *                             try:             # <<<<<<<<<<<<<<
 *                                 v = float(v)
 *                                 if v % 1 == 0:
 */
              __Pyx_XGIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_4);
              __Pyx_XGIVEREF(__pyx_t_5);
              __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
              goto __pyx_L14_error;
              __pyx_L45_exception_handled:;
              __Pyx_XGIVEREF(__pyx_t_3);
              __Pyx_XGIVEREF(__pyx_t_4);
              __Pyx_XGIVEREF(__pyx_t_5);
              __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
              __pyx_L51_try_end:;
            }

            /* "fiona/ogrext.pyx":726
 *                     for param in value.split():
 *                         kv = param.split("=")
 *                         if len(kv) == 2:             # <<<<<<<<<<<<<<
 *                             k, v = kv
 *                             try:
 */
            goto __pyx_L41;
          }

          /* "fiona/ogrext.pyx":735
 *                                 # Leave v as a string
 *                                 pass
 *                         elif len(kv) == 1:             # <<<<<<<<<<<<<<
 *                             k, v = kv[0], True
 *                         else:
 */
          __pyx_t_26 = PyObject_Length(__pyx_v_kv); if (unlikely(__pyx_t_26 == ((Py_ssize_t)-1))) __PYX_ERR(0, 735, __pyx_L14_error)
          __pyx_t_1 = ((__pyx_t_26 == 1) != 0);
          if (likely(__pyx_t_1)) {

            /* "fiona/ogrext.pyx":736
 *                                 pass
 *                         elif len(kv) == 1:
 *                             k, v = kv[0], True             # <<<<<<<<<<<<<<
 *                         else:
 *                             raise ValueError("Unexpected proj parameter %s" % param)
 */
            __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_kv, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 736, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_11 = Py_True;
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_2);
            __pyx_t_2 = 0;
            __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "fiona/ogrext.pyx":735
 *                                 # Leave v as a string
 *                                 pass
 *                         elif len(kv) == 1:             # <<<<<<<<<<<<<<
 *                             k, v = kv[0], True
 *                         else:
 */
            goto __pyx_L41;
          }

          /* "fiona/ogrext.pyx":738
 *                             k, v = kv[0], True
 *                         else:
 *                             raise ValueError("Unexpected proj parameter %s" % param)             # <<<<<<<<<<<<<<
 *                         k = k.lstrip("+")
 *                         crs[k] = v
 */
          /*else*/ {
            __pyx_t_11 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Unexpected_proj_parameter_s, __pyx_v_param); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 738, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 738, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_2);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_Raise(__pyx_t_2, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            __PYX_ERR(0, 738, __pyx_L14_error)
          }
          __pyx_L41:;

          /* "fiona/ogrext.pyx":739
 *                         else:
 *                             raise ValueError("Unexpected proj parameter %s" % param)
 *                         k = k.lstrip("+")             # <<<<<<<<<<<<<<
 *                         crs[k] = v
 * 
 */
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_k, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 739, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_7 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_7)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_7);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
            }
          }
          __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_7, __pyx_kp_u__16) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_u__16);
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 739, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF_SET(__pyx_v_k, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "fiona/ogrext.pyx":740
 *                             raise ValueError("Unexpected proj parameter %s" % param)
 *                         k = k.lstrip("+")
 *                         crs[k] = v             # <<<<<<<<<<<<<<
 * 
 *             finally:
 */
          if (unlikely(PyDict_SetItem(__pyx_v_crs, __pyx_v_k, __pyx_v_v) < 0)) __PYX_ERR(0, 740, __pyx_L14_error)

          /* "fiona/ogrext.pyx":724
 *                     value = proj_b.decode()
 *                     value = value.strip()
 *                     for param in value.split():             # <<<<<<<<<<<<<<
 *                         kv = param.split("=")
 *                         if len(kv) == 2:
 */
        }
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __pyx_L34:;
    }

    /* "fiona/ogrext.pyx":743
 * 
 *             finally:
 *                 CPLFree(proj_c)             # <<<<<<<<<<<<<<
 *                 return crs
 * 
 */
    /*finally:*/ {
      /*normal exit:*/{
        CPLFree(__pyx_v_proj_c);

        /* "fiona/ogrext.pyx":744
 *             finally:
 *                 CPLFree(proj_c)
 *                 return crs             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_crs);
        __pyx_r = __pyx_v_crs;
        goto __pyx_L0;
      }
      __pyx_L14_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_22);
        __Pyx_XGOTREF(__pyx_t_21);
        __Pyx_XGOTREF(__pyx_t_20);
        {

          /* "fiona/ogrext.pyx":743
 * 
 *             finally:
 *                 CPLFree(proj_c)             # <<<<<<<<<<<<<<
 *                 return crs
 * 
 */
          CPLFree(__pyx_v_proj_c);

          /* "fiona/ogrext.pyx":744
 *             finally:
 *                 CPLFree(proj_c)
 *                 return crs             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
          __Pyx_XDECREF(__pyx_r);
          __Pyx_INCREF(__pyx_v_crs);
          __pyx_r = __pyx_v_crs;
          goto __pyx_L53_return;
        }
        __pyx_L53_return:;
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_22);
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
        goto __pyx_L0;
      }
    }

    /* "fiona/ogrext.pyx":690
 *             log.debug("Layer has no coordinate system")
 * 
 *         if cogr_crs is not NULL:             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Got coordinate system")
 */
  }

  /* "fiona/ogrext.pyx":747
 * 
 *         else:
 *             log.debug("Projection not found (cogr_crs was NULL)")             # <<<<<<<<<<<<<<
 * 
 *         return {}
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_log); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_12 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_2, __pyx_kp_u_Projection_not_found_cogr_crs_wa) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_u_Projection_not_found_cogr_crs_wa);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 747, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  }

  /* "fiona/ogrext.pyx":749
 *             log.debug("Projection not found (cogr_crs was NULL)")
 * 
 *         return {}             # <<<<<<<<<<<<<<
 * 
 *     def get_crs_wkt(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_12 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 749, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_r = __pyx_t_12;
  __pyx_t_12 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":667
 *         return ret
 * 
 *     def get_crs(self):             # <<<<<<<<<<<<<<
 *         """Get the layer's CRS
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_crs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_crs);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_val);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_param);
  __Pyx_XDECREF(__pyx_v_kv);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":751
 *         return {}
 * 
 *     def get_crs_wkt(self):             # <<<<<<<<<<<<<<
 *         cdef char *proj_c = NULL
 *         cdef void *cogr_crs = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_23get_crs_wkt(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_23get_crs_wkt(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_crs_wkt (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_22get_crs_wkt(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_22get_crs_wkt(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  char *__pyx_v_proj_c;
  void *__pyx_v_cogr_crs;
  PyObject *__pyx_v_exc = NULL;
  char *__pyx_v_proj_b;
  PyObject *__pyx_v_crs_wkt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  void *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  __Pyx_RefNannySetupContext("get_crs_wkt", 0);

  /* "fiona/ogrext.pyx":752
 * 
 *     def get_crs_wkt(self):
 *         cdef char *proj_c = NULL             # <<<<<<<<<<<<<<
 *         cdef void *cogr_crs = NULL
 * 
 */
  __pyx_v_proj_c = NULL;

  /* "fiona/ogrext.pyx":753
 *     def get_crs_wkt(self):
 *         cdef char *proj_c = NULL
 *         cdef void *cogr_crs = NULL             # <<<<<<<<<<<<<<
 * 
 *         if self.cogr_layer == NULL:
 */
  __pyx_v_cogr_crs = NULL;

  /* "fiona/ogrext.pyx":755
 *         cdef void *cogr_crs = NULL
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fiona/ogrext.pyx":756
 * 
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 756, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 756, __pyx_L1_error)

    /* "fiona/ogrext.pyx":755
 *         cdef void *cogr_crs = NULL
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  }

  /* "fiona/ogrext.pyx":758
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 * 
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "fiona/ogrext.pyx":759
 * 
 *         try:
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))             # <<<<<<<<<<<<<<
 * 
 *         # TODO: we don't intend to use try/except for flow control
 */
      __pyx_t_6 = __pyx_f_5fiona_4_err_exc_wrap_pointer(OGR_L_GetSpatialRef(__pyx_v_self->cogr_layer)); if (unlikely(__pyx_t_6 == ((void *)NULL))) __PYX_ERR(0, 759, __pyx_L4_error)
      __pyx_v_cogr_crs = __pyx_t_6;

      /* "fiona/ogrext.pyx":758
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 * 
 */
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L4_error:;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":763
 *         # TODO: we don't intend to use try/except for flow control
 *         # this is a work around for a GDAL issue.
 *         except FionaNullPointerError:             # <<<<<<<<<<<<<<
 *             log.debug("Layer has no coordinate system")
 *         except fiona._err.CPLE_OpenFailedError as exc:
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_FionaNullPointerError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 763, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_7, __pyx_t_8);
    __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("fiona.ogrext.Session.get_crs_wkt", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 763, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);

      /* "fiona/ogrext.pyx":764
 *         # this is a work around for a GDAL issue.
 *         except FionaNullPointerError:
 *             log.debug("Layer has no coordinate system")             # <<<<<<<<<<<<<<
 *         except fiona._err.CPLE_OpenFailedError as exc:
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 764, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 764, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_kp_u_Layer_has_no_coordinate_system) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_u_Layer_has_no_coordinate_system);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 764, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5_exception_handled;
    }

    /* "fiona/ogrext.pyx":765
 *         except FionaNullPointerError:
 *             log.debug("Layer has no coordinate system")
 *         except fiona._err.CPLE_OpenFailedError as exc:             # <<<<<<<<<<<<<<
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 */
    __Pyx_ErrFetch(&__pyx_t_2, &__pyx_t_7, &__pyx_t_8);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_fiona); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 765, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_err); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 765, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_CPLE_OpenFailedError); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 765, __pyx_L6_except_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_10 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_2, __pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_ErrRestore(__pyx_t_2, __pyx_t_7, __pyx_t_8);
    __pyx_t_2 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0;
    if (__pyx_t_10) {
      __Pyx_AddTraceback("fiona.ogrext.Session.get_crs_wkt", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_2) < 0) __PYX_ERR(0, 765, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_7);
      __pyx_v_exc = __pyx_t_7;
      /*try:*/ {

        /* "fiona/ogrext.pyx":766
 *             log.debug("Layer has no coordinate system")
 *         except fiona._err.CPLE_OpenFailedError as exc:
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")             # <<<<<<<<<<<<<<
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")
 */
        __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_log); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 766, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_debug); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 766, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __pyx_t_12 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_9 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_kp_u_A_support_file_wasn_t_opened_See) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_kp_u_A_support_file_wasn_t_opened_See);
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 766, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "fiona/ogrext.pyx":767
 *         except fiona._err.CPLE_OpenFailedError as exc:
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)             # <<<<<<<<<<<<<<
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")
 *             if cogr_crs == NULL:
 */
        __pyx_v_cogr_crs = OGR_L_GetSpatialRef(__pyx_v_self->cogr_layer);

        /* "fiona/ogrext.pyx":768
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")             # <<<<<<<<<<<<<<
 *             if cogr_crs == NULL:
 *                 raise exc
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_log); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 768, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_debug); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 768, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
          }
        }
        __pyx_t_9 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_kp_u_Called_OGR_L_GetSpatialRef_again) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_kp_u_Called_OGR_L_GetSpatialRef_again);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 768, __pyx_L17_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "fiona/ogrext.pyx":769
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")
 *             if cogr_crs == NULL:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
        __pyx_t_1 = ((__pyx_v_cogr_crs == NULL) != 0);
        if (unlikely(__pyx_t_1)) {

          /* "fiona/ogrext.pyx":770
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")
 *             if cogr_crs == NULL:
 *                 raise exc             # <<<<<<<<<<<<<<
 * 
 *         if cogr_crs is not NULL:
 */
          __Pyx_Raise(__pyx_v_exc, 0, 0, 0);
          __PYX_ERR(0, 770, __pyx_L17_error)

          /* "fiona/ogrext.pyx":769
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 *             log.debug("Called OGR_L_GetSpatialRef() again without error checking.")
 *             if cogr_crs == NULL:             # <<<<<<<<<<<<<<
 *                 raise exc
 * 
 */
        }
      }

      /* "fiona/ogrext.pyx":765
 *         except FionaNullPointerError:
 *             log.debug("Layer has no coordinate system")
 *         except fiona._err.CPLE_OpenFailedError as exc:             # <<<<<<<<<<<<<<
 *             log.debug("A support file wasn't opened. See the preceding ERROR level message.")
 *             cogr_crs = OGR_L_GetSpatialRef(self.cogr_layer)
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_exc);
          __pyx_v_exc = NULL;
          goto __pyx_L18;
        }
        __pyx_L17_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __Pyx_XGOTREF(__pyx_t_20);
          __pyx_t_10 = __pyx_lineno; __pyx_t_13 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_exc);
            __pyx_v_exc = NULL;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
          }
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_XGIVEREF(__pyx_t_17);
          __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
          __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
          __pyx_lineno = __pyx_t_10; __pyx_clineno = __pyx_t_13; __pyx_filename = __pyx_t_14;
          goto __pyx_L6_except_error;
        }
        __pyx_L18:;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      goto __pyx_L5_exception_handled;
    }
    goto __pyx_L6_except_error;
    __pyx_L6_except_error:;

    /* "fiona/ogrext.pyx":758
 *             raise ValueError("Null layer")
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_crs = exc_wrap_pointer(OGR_L_GetSpatialRef(self.cogr_layer))
 * 
 */
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L5_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "fiona/ogrext.pyx":772
 *                 raise exc
 * 
 *         if cogr_crs is not NULL:             # <<<<<<<<<<<<<<
 *             log.debug("Got coordinate system")
 * 
 */
  __pyx_t_1 = ((__pyx_v_cogr_crs != NULL) != 0);
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":773
 * 
 *         if cogr_crs is not NULL:
 *             log.debug("Got coordinate system")             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_log); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_debug); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_2 = (__pyx_t_7) ? __Pyx_PyObject_Call2Args(__pyx_t_8, __pyx_t_7, __pyx_kp_u_Got_coordinate_system) : __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_kp_u_Got_coordinate_system);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":775
 *             log.debug("Got coordinate system")
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 OSRExportToWkt(cogr_crs, &proj_c)
 *                 if proj_c == NULL:
 */
    /*try:*/ {

      /* "fiona/ogrext.pyx":776
 * 
 *             try:
 *                 OSRExportToWkt(cogr_crs, &proj_c)             # <<<<<<<<<<<<<<
 *                 if proj_c == NULL:
 *                     raise ValueError("Null projection")
 */
      (void)(OSRExportToWkt(__pyx_v_cogr_crs, (&__pyx_v_proj_c)));

      /* "fiona/ogrext.pyx":777
 *             try:
 *                 OSRExportToWkt(cogr_crs, &proj_c)
 *                 if proj_c == NULL:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Null projection")
 *                 proj_b = proj_c
 */
      __pyx_t_1 = ((__pyx_v_proj_c == NULL) != 0);
      if (unlikely(__pyx_t_1)) {

        /* "fiona/ogrext.pyx":778
 *                 OSRExportToWkt(cogr_crs, &proj_c)
 *                 if proj_c == NULL:
 *                     raise ValueError("Null projection")             # <<<<<<<<<<<<<<
 *                 proj_b = proj_c
 *                 crs_wkt = proj_b.decode('utf-8')
 */
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 778, __pyx_L26_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_Raise(__pyx_t_2, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __PYX_ERR(0, 778, __pyx_L26_error)

        /* "fiona/ogrext.pyx":777
 *             try:
 *                 OSRExportToWkt(cogr_crs, &proj_c)
 *                 if proj_c == NULL:             # <<<<<<<<<<<<<<
 *                     raise ValueError("Null projection")
 *                 proj_b = proj_c
 */
      }

      /* "fiona/ogrext.pyx":779
 *                 if proj_c == NULL:
 *                     raise ValueError("Null projection")
 *                 proj_b = proj_c             # <<<<<<<<<<<<<<
 *                 crs_wkt = proj_b.decode('utf-8')
 * 
 */
      __pyx_v_proj_b = __pyx_v_proj_c;

      /* "fiona/ogrext.pyx":780
 *                     raise ValueError("Null projection")
 *                 proj_b = proj_c
 *                 crs_wkt = proj_b.decode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *             finally:
 */
      __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_proj_b, 0, strlen(__pyx_v_proj_b), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 780, __pyx_L26_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_v_crs_wkt = __pyx_t_2;
      __pyx_t_2 = 0;
    }

    /* "fiona/ogrext.pyx":783
 * 
 *             finally:
 *                 CPLFree(proj_c)             # <<<<<<<<<<<<<<
 *                 return crs_wkt
 * 
 */
    /*finally:*/ {
      /*normal exit:*/{
        CPLFree(__pyx_v_proj_c);

        /* "fiona/ogrext.pyx":784
 *             finally:
 *                 CPLFree(proj_c)
 *                 return crs_wkt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_crs_wkt);
        __pyx_r = __pyx_v_crs_wkt;
        goto __pyx_L0;
      }
      __pyx_L26_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_5 = 0; __pyx_t_4 = 0; __pyx_t_3 = 0; __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_20, &__pyx_t_19, &__pyx_t_18);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3) < 0)) __Pyx_ErrFetch(&__pyx_t_5, &__pyx_t_4, &__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_5);
        __Pyx_XGOTREF(__pyx_t_4);
        __Pyx_XGOTREF(__pyx_t_3);
        __Pyx_XGOTREF(__pyx_t_20);
        __Pyx_XGOTREF(__pyx_t_19);
        __Pyx_XGOTREF(__pyx_t_18);
        {

          /* "fiona/ogrext.pyx":783
 * 
 *             finally:
 *                 CPLFree(proj_c)             # <<<<<<<<<<<<<<
 *                 return crs_wkt
 * 
 */
          CPLFree(__pyx_v_proj_c);

          /* "fiona/ogrext.pyx":784
 *             finally:
 *                 CPLFree(proj_c)
 *                 return crs_wkt             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
          __Pyx_XDECREF(__pyx_r);
          if (unlikely(!__pyx_v_crs_wkt)) { __Pyx_RaiseUnboundLocalError("crs_wkt"); __PYX_ERR(0, 784, __pyx_L30_error) }
          __Pyx_INCREF(__pyx_v_crs_wkt);
          __pyx_r = __pyx_v_crs_wkt;
          goto __pyx_L29_return;
        }
        __pyx_L29_return:;
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
        goto __pyx_L0;
        __pyx_L30_error:;
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_XGIVEREF(__pyx_t_19);
          __Pyx_XGIVEREF(__pyx_t_18);
          __Pyx_ExceptionReset(__pyx_t_20, __pyx_t_19, __pyx_t_18);
        }
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_20 = 0; __pyx_t_19 = 0; __pyx_t_18 = 0;
        goto __pyx_L1_error;
      }
    }

    /* "fiona/ogrext.pyx":772
 *                 raise exc
 * 
 *         if cogr_crs is not NULL:             # <<<<<<<<<<<<<<
 *             log.debug("Got coordinate system")
 * 
 */
  }

  /* "fiona/ogrext.pyx":787
 * 
 *         else:
 *             log.debug("Projection not found (cogr_crs was NULL)")             # <<<<<<<<<<<<<<
 *             return ""
 * 
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_log); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_debug); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_u_Projection_not_found_cogr_crs_wa) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_Projection_not_found_cogr_crs_wa);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 787, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":788
 *         else:
 *             log.debug("Projection not found (cogr_crs was NULL)")
 *             return ""             # <<<<<<<<<<<<<<
 * 
 *     def get_extent(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u__12);
    __pyx_r = __pyx_kp_u__12;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":751
 *         return {}
 * 
 *     def get_crs_wkt(self):             # <<<<<<<<<<<<<<
 *         cdef char *proj_c = NULL
 *         cdef void *cogr_crs = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_crs_wkt", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_crs_wkt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":790
 *             return ""
 * 
 *     def get_extent(self):             # <<<<<<<<<<<<<<
 *         cdef OGREnvelope extent
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_25get_extent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_25get_extent(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_extent (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_24get_extent(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_24get_extent(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  OGREnvelope __pyx_v_extent;
  CYTHON_UNUSED OGRErr __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("get_extent", 0);

  /* "fiona/ogrext.pyx":793
 *         cdef OGREnvelope extent
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "fiona/ogrext.pyx":794
 * 
 *         if self.cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 * 
 *         result = OGR_L_GetExtent(self.cogr_layer, &extent, 1)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 794, __pyx_L1_error)

    /* "fiona/ogrext.pyx":793
 *         cdef OGREnvelope extent
 * 
 *         if self.cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  }

  /* "fiona/ogrext.pyx":796
 *             raise ValueError("Null layer")
 * 
 *         result = OGR_L_GetExtent(self.cogr_layer, &extent, 1)             # <<<<<<<<<<<<<<
 *         return (extent.MinX, extent.MinY, extent.MaxX, extent.MaxY)
 * 
 */
  __pyx_v_result = OGR_L_GetExtent(__pyx_v_self->cogr_layer, (&__pyx_v_extent), 1);

  /* "fiona/ogrext.pyx":797
 * 
 *         result = OGR_L_GetExtent(self.cogr_layer, &extent, 1)
 *         return (extent.MinX, extent.MinY, extent.MaxX, extent.MaxY)             # <<<<<<<<<<<<<<
 * 
 *     def has_feature(self, fid):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_extent.MinX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_extent.MinY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_extent.MaxX); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_extent.MaxY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_6, 3, __pyx_t_5);
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":790
 *             return ""
 * 
 *     def get_extent(self):             # <<<<<<<<<<<<<<
 *         cdef OGREnvelope extent
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_extent", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":799
 *         return (extent.MinX, extent.MinY, extent.MaxX, extent.MaxY)
 * 
 *     def has_feature(self, fid):             # <<<<<<<<<<<<<<
 *         """Provides access to feature data by FID.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_27has_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_fid); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_26has_feature[] = "Provides access to feature data by FID.\n\n        Supports Collection.__contains__().\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_27has_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_fid) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("has_feature (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_26has_feature(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self), ((PyObject *)__pyx_v_fid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_26has_feature(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_fid) {
  void *__pyx_v_cogr_feature;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("has_feature", 0);
  __Pyx_INCREF(__pyx_v_fid);

  /* "fiona/ogrext.pyx":805
 *         """
 *         cdef void * cogr_feature
 *         fid = int(fid)             # <<<<<<<<<<<<<<
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:
 */
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_fid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 805, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_fid, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":806
 *         cdef void * cogr_feature
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)             # <<<<<<<<<<<<<<
 *         if cogr_feature != NULL:
 *             _deleteOgrFeature(cogr_feature)
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_fid); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_v_cogr_feature = OGR_L_GetFeature(__pyx_v_self->cogr_layer, __pyx_t_2);

  /* "fiona/ogrext.pyx":807
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:             # <<<<<<<<<<<<<<
 *             _deleteOgrFeature(cogr_feature)
 *             return True
 */
  __pyx_t_3 = ((__pyx_v_cogr_feature != NULL) != 0);
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":808
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:
 *             _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
    __pyx_t_1 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 808, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":809
 *         if cogr_feature != NULL:
 *             _deleteOgrFeature(cogr_feature)
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":807
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:             # <<<<<<<<<<<<<<
 *             _deleteOgrFeature(cogr_feature)
 *             return True
 */
  }

  /* "fiona/ogrext.pyx":811
 *             return True
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     def get_feature(self, fid):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":799
 *         return (extent.MinX, extent.MinY, extent.MaxX, extent.MaxY)
 * 
 *     def has_feature(self, fid):             # <<<<<<<<<<<<<<
 *         """Provides access to feature data by FID.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Session.has_feature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_fid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":813
 *             return False
 * 
 *     def get_feature(self, fid):             # <<<<<<<<<<<<<<
 *         """Provides access to feature data by FID.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_29get_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_fid); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7Session_28get_feature[] = "Provides access to feature data by FID.\n\n        Supports Collection.__contains__().\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_29get_feature(PyObject *__pyx_v_self, PyObject *__pyx_v_fid) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_feature (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_28get_feature(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self), ((PyObject *)__pyx_v_fid));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_28get_feature(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_fid) {
  void *__pyx_v_cogr_feature;
  PyObject *__pyx_v_feature = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build __pyx_t_9;
  __Pyx_RefNannySetupContext("get_feature", 0);
  __Pyx_INCREF(__pyx_v_fid);

  /* "fiona/ogrext.pyx":819
 *         """
 *         cdef void * cogr_feature
 *         fid = int(fid)             # <<<<<<<<<<<<<<
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:
 */
  __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_fid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 819, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_fid, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":820
 *         cdef void * cogr_feature
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)             # <<<<<<<<<<<<<<
 *         if cogr_feature != NULL:
 *             feature = FeatureBuilder().build(
 */
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_fid); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 820, __pyx_L1_error)
  __pyx_v_cogr_feature = OGR_L_GetFeature(__pyx_v_self->cogr_layer, __pyx_t_2);

  /* "fiona/ogrext.pyx":821
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:             # <<<<<<<<<<<<<<
 *             feature = FeatureBuilder().build(
 *                 cogr_feature,
 */
  __pyx_t_3 = ((__pyx_v_cogr_feature != NULL) != 0);
  if (likely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":822
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:
 *             feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "fiona/ogrext.pyx":824
 *             feature = FeatureBuilder().build(
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),             # <<<<<<<<<<<<<<
 *                 bbox=False,
 *                 driver=self.collection.driver,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 824, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "fiona/ogrext.pyx":826
 *                 encoding=self._get_internal_encoding(),
 *                 bbox=False,
 *                 driver=self.collection.driver,             # <<<<<<<<<<<<<<
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 826, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "fiona/ogrext.pyx":827
 *                 bbox=False,
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,             # <<<<<<<<<<<<<<
 *                 ignore_geometry=self.collection.ignore_geometry,
 *             )
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 827, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "fiona/ogrext.pyx":828
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,             # <<<<<<<<<<<<<<
 *             )
 *             _deleteOgrFeature(cogr_feature)
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_geometry); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 828, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "fiona/ogrext.pyx":822
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:
 *             feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),
 */
    __pyx_t_9.__pyx_n = 5;
    __pyx_t_9.encoding = __pyx_t_4;
    __pyx_t_9.bbox = Py_False;
    __pyx_t_9.driver = __pyx_t_5;
    __pyx_t_9.ignore_fields = __pyx_t_6;
    __pyx_t_9.ignore_geometry = __pyx_t_7;
    __pyx_t_8 = ((struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1), __pyx_v_cogr_feature, &__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_v_feature = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":830
 *                 ignore_geometry=self.collection.ignore_geometry,
 *             )
 *             _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 *             return feature
 *         else:
 */
    __pyx_t_8 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 830, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":831
 *             )
 *             _deleteOgrFeature(cogr_feature)
 *             return feature             # <<<<<<<<<<<<<<
 *         else:
 *             raise KeyError("There is no feature with fid {!r}".format(fid))
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_feature);
    __pyx_r = __pyx_v_feature;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":821
 *         fid = int(fid)
 *         cogr_feature = OGR_L_GetFeature(self.cogr_layer, fid)
 *         if cogr_feature != NULL:             # <<<<<<<<<<<<<<
 *             feature = FeatureBuilder().build(
 *                 cogr_feature,
 */
  }

  /* "fiona/ogrext.pyx":833
 *             return feature
 *         else:
 *             raise KeyError("There is no feature with fid {!r}".format(fid))             # <<<<<<<<<<<<<<
 * 
 *     get = get_feature
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_There_is_no_feature_with_fid_r, __pyx_n_s_format); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_8 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_6, __pyx_v_fid) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_v_fid);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_builtin_KeyError, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 833, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 833, __pyx_L1_error)
  }

  /* "fiona/ogrext.pyx":813
 *             return False
 * 
 *     def get_feature(self, fid):             # <<<<<<<<<<<<<<
 *         """Provides access to feature data by FID.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fiona.ogrext.Session.get_feature", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XDECREF(__pyx_v_fid);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":838
 * 
 *     # TODO: Make this an alias for get_feature in a future version.
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         cdef void * cogr_feature
 *         if isinstance(item, slice):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_31__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_31__getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_30__getitem__(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_30__getitem__(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, PyObject *__pyx_v_item) {
  void *__pyx_v_cogr_feature;
  struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_itr = NULL;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_v_ftcount = NULL;
  PyObject *__pyx_v_feature = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build __pyx_t_10;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "fiona/ogrext.pyx":840
 *     def __getitem__(self, item):
 *         cdef void * cogr_feature
 *         if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *             warnings.warn("Collection slicing is deprecated and will be disabled in a future version.", FionaDeprecationWarning)
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)
 */
  __pyx_t_1 = PySlice_Check(__pyx_v_item); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":841
 *         cdef void * cogr_feature
 *         if isinstance(item, slice):
 *             warnings.warn("Collection slicing is deprecated and will be disabled in a future version.", FionaDeprecationWarning)             # <<<<<<<<<<<<<<
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)
 *             log.debug("Slice: %r", item)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_warnings); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warn); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FionaDeprecationWarning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 841, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_u_Collection_slicing_is_deprecated, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_kp_u_Collection_slicing_is_deprecated, __pyx_t_4};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_Collection_slicing_is_deprecated);
      __Pyx_GIVEREF(__pyx_kp_u_Collection_slicing_is_deprecated);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_kp_u_Collection_slicing_is_deprecated);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":842
 *         if isinstance(item, slice):
 *             warnings.warn("Collection slicing is deprecated and will be disabled in a future version.", FionaDeprecationWarning)
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)             # <<<<<<<<<<<<<<
 *             log.debug("Slice: %r", item)
 *             return list(itr)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_stop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_item, __pyx_n_s_step); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_self->collection);
    __Pyx_GIVEREF(__pyx_v_self->collection);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->collection);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_8);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5fiona_6ogrext_Iterator), __pyx_t_4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 842, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_itr = ((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_t_8);
    __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":843
 *             warnings.warn("Collection slicing is deprecated and will be disabled in a future version.", FionaDeprecationWarning)
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)
 *             log.debug("Slice: %r", item)             # <<<<<<<<<<<<<<
 *             return list(itr)
 *         elif isinstance(item, int):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 843, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_Slice_r, __pyx_v_item};
      __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 843, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_8);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_Slice_r, __pyx_v_item};
      __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 843, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_t_8);
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_Slice_r);
      __Pyx_GIVEREF(__pyx_kp_u_Slice_r);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_7, __pyx_kp_u_Slice_r);
      __Pyx_INCREF(__pyx_v_item);
      __Pyx_GIVEREF(__pyx_v_item);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_7, __pyx_v_item);
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_3, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 843, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "fiona/ogrext.pyx":844
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)
 *             log.debug("Slice: %r", item)
 *             return list(itr)             # <<<<<<<<<<<<<<
 *         elif isinstance(item, int):
 *             index = item
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_8 = PySequence_List(((PyObject *)__pyx_v_itr)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 844, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_r = __pyx_t_8;
    __pyx_t_8 = 0;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":840
 *     def __getitem__(self, item):
 *         cdef void * cogr_feature
 *         if isinstance(item, slice):             # <<<<<<<<<<<<<<
 *             warnings.warn("Collection slicing is deprecated and will be disabled in a future version.", FionaDeprecationWarning)
 *             itr = Iterator(self.collection, item.start, item.stop, item.step)
 */
  }

  /* "fiona/ogrext.pyx":845
 *             log.debug("Slice: %r", item)
 *             return list(itr)
 *         elif isinstance(item, int):             # <<<<<<<<<<<<<<
 *             index = item
 *             # from the back
 */
  __pyx_t_2 = PyInt_Check(__pyx_v_item); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":846
 *             return list(itr)
 *         elif isinstance(item, int):
 *             index = item             # <<<<<<<<<<<<<<
 *             # from the back
 *             if index < 0:
 */
    __Pyx_INCREF(__pyx_v_item);
    __pyx_v_index = __pyx_v_item;

    /* "fiona/ogrext.pyx":848
 *             index = item
 *             # from the back
 *             if index < 0:             # <<<<<<<<<<<<<<
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:
 */
    __pyx_t_8 = PyObject_RichCompare(__pyx_v_index, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 848, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 848, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_1) {

      /* "fiona/ogrext.pyx":849
 *             # from the back
 *             if index < 0:
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)             # <<<<<<<<<<<<<<
 *                 if ftcount == -1:
 *                     raise IndexError(
 */
      __pyx_t_8 = __Pyx_PyInt_From_int(OGR_L_GetFeatureCount(__pyx_v_self->cogr_layer, 0)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 849, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_v_ftcount = __pyx_t_8;
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":850
 *             if index < 0:
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:             # <<<<<<<<<<<<<<
 *                     raise IndexError(
 *                         "collection's dataset does not support negative indexes")
 */
      __pyx_t_8 = __Pyx_PyInt_EqObjC(__pyx_v_ftcount, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 850, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 850, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(__pyx_t_1)) {

        /* "fiona/ogrext.pyx":851
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:
 *                     raise IndexError(             # <<<<<<<<<<<<<<
 *                         "collection's dataset does not support negative indexes")
 *                 index += ftcount
 */
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 851, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_Raise(__pyx_t_8, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __PYX_ERR(0, 851, __pyx_L1_error)

        /* "fiona/ogrext.pyx":850
 *             if index < 0:
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:             # <<<<<<<<<<<<<<
 *                     raise IndexError(
 *                         "collection's dataset does not support negative indexes")
 */
      }

      /* "fiona/ogrext.pyx":853
 *                     raise IndexError(
 *                         "collection's dataset does not support negative indexes")
 *                 index += ftcount             # <<<<<<<<<<<<<<
 *             cogr_feature = OGR_L_GetFeature(self.cogr_layer, index)
 *             if cogr_feature == NULL:
 */
      __pyx_t_8 = PyNumber_InPlaceAdd(__pyx_v_index, __pyx_v_ftcount); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 853, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":848
 *             index = item
 *             # from the back
 *             if index < 0:             # <<<<<<<<<<<<<<
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:
 */
    }

    /* "fiona/ogrext.pyx":854
 *                         "collection's dataset does not support negative indexes")
 *                 index += ftcount
 *             cogr_feature = OGR_L_GetFeature(self.cogr_layer, index)             # <<<<<<<<<<<<<<
 *             if cogr_feature == NULL:
 *                 return None
 */
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_v_index); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 854, __pyx_L1_error)
    __pyx_v_cogr_feature = OGR_L_GetFeature(__pyx_v_self->cogr_layer, __pyx_t_7);

    /* "fiona/ogrext.pyx":855
 *                 index += ftcount
 *             cogr_feature = OGR_L_GetFeature(self.cogr_layer, index)
 *             if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *                 return None
 *             feature = FeatureBuilder().build(
 */
    __pyx_t_1 = ((__pyx_v_cogr_feature == NULL) != 0);
    if (__pyx_t_1) {

      /* "fiona/ogrext.pyx":856
 *             cogr_feature = OGR_L_GetFeature(self.cogr_layer, index)
 *             if cogr_feature == NULL:
 *                 return None             # <<<<<<<<<<<<<<
 *             feature = FeatureBuilder().build(
 *                 cogr_feature,
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "fiona/ogrext.pyx":855
 *                 index += ftcount
 *             cogr_feature = OGR_L_GetFeature(self.cogr_layer, index)
 *             if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *                 return None
 *             feature = FeatureBuilder().build(
 */
    }

    /* "fiona/ogrext.pyx":857
 *             if cogr_feature == NULL:
 *                 return None
 *             feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),
 */
    __pyx_t_8 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "fiona/ogrext.pyx":859
 *             feature = FeatureBuilder().build(
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),             # <<<<<<<<<<<<<<
 *                 bbox=False,
 *                 driver=self.collection.driver,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_5 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":861
 *                 encoding=self._get_internal_encoding(),
 *                 bbox=False,
 *                 driver=self.collection.driver,             # <<<<<<<<<<<<<<
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "fiona/ogrext.pyx":862
 *                 bbox=False,
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,             # <<<<<<<<<<<<<<
 *                 ignore_geometry=self.collection.ignore_geometry,
 *             )
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 862, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "fiona/ogrext.pyx":863
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,             # <<<<<<<<<<<<<<
 *             )
 *             _deleteOgrFeature(cogr_feature)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_geometry); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 863, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "fiona/ogrext.pyx":857
 *             if cogr_feature == NULL:
 *                 return None
 *             feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self._get_internal_encoding(),
 */
    __pyx_t_10.__pyx_n = 5;
    __pyx_t_10.encoding = __pyx_t_5;
    __pyx_t_10.bbox = Py_False;
    __pyx_t_10.driver = __pyx_t_3;
    __pyx_t_10.ignore_fields = __pyx_t_4;
    __pyx_t_10.ignore_geometry = __pyx_t_6;
    __pyx_t_9 = ((struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_8)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_8), __pyx_v_cogr_feature, &__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 857, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_feature = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "fiona/ogrext.pyx":865
 *                 ignore_geometry=self.collection.ignore_geometry,
 *             )
 *             _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 *             return feature
 * 
 */
    __pyx_t_9 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 865, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "fiona/ogrext.pyx":866
 *             )
 *             _deleteOgrFeature(cogr_feature)
 *             return feature             # <<<<<<<<<<<<<<
 * 
 *     def isactive(self):
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_feature);
    __pyx_r = __pyx_v_feature;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":845
 *             log.debug("Slice: %r", item)
 *             return list(itr)
 *         elif isinstance(item, int):             # <<<<<<<<<<<<<<
 *             index = item
 *             # from the back
 */
  }

  /* "fiona/ogrext.pyx":838
 * 
 *     # TODO: Make this an alias for get_feature in a future version.
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         cdef void * cogr_feature
 *         if isinstance(item, slice):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fiona.ogrext.Session.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_itr);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XDECREF(__pyx_v_ftcount);
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":868
 *             return feature
 * 
 *     def isactive(self):             # <<<<<<<<<<<<<<
 *         if self.cogr_layer != NULL and self.cogr_ds != NULL:
 *             return 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_33isactive(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_33isactive(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("isactive (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_32isactive(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_32isactive(struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("isactive", 0);

  /* "fiona/ogrext.pyx":869
 * 
 *     def isactive(self):
 *         if self.cogr_layer != NULL and self.cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  __pyx_t_2 = ((__pyx_v_self->cogr_layer != NULL) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = ((__pyx_v_self->cogr_ds != NULL) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":870
 *     def isactive(self):
 *         if self.cogr_layer != NULL and self.cogr_ds != NULL:
 *             return 1             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_1);
    __pyx_r = __pyx_int_1;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":869
 * 
 *     def isactive(self):
 *         if self.cogr_layer != NULL and self.cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *             return 1
 *         else:
 */
  }

  /* "fiona/ogrext.pyx":872
 *             return 1
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":868
 *             return feature
 * 
 *     def isactive(self):             # <<<<<<<<<<<<<<
 *         if self.cogr_layer != NULL and self.cogr_ds != NULL:
 *             return 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_35__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_35__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_34__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_34__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Session.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_37__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_7Session_37__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_7Session_36__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7Session_36__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Session.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":879
 *     cdef object _schema_mapping
 * 
 *     def start(self, collection, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef OGRSpatialReferenceH cogr_srs = NULL
 *         cdef char **options = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_1start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_1start(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_collection = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_collection,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "start") < 0)) __PYX_ERR(0, 879, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_collection = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 879, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_start(((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_v_self), __pyx_v_collection, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_start(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_kwargs) {
  OGRSpatialReferenceH __pyx_v_cogr_srs;
  char **__pyx_v_options;
  char const *__pyx_v_path_c;
  char const *__pyx_v_driver_c;
  char const *__pyx_v_name_c;
  char const *__pyx_v_proj_c;
  char const *__pyx_v_fileencoding_c;
  OGRFieldSubType __pyx_v_field_subtype;
  PyObject *__pyx_v_path = NULL;
  PyObject *__pyx_v_userencoding = NULL;
  PyObject *__pyx_v_path_b = NULL;
  PyObject *__pyx_v_name_b = NULL;
  PyObject *__pyx_v_exc = NULL;
  PyObject *__pyx_v_driver_b = NULL;
  void *__pyx_v_cogr_driver;
  void *__pyx_v_cogr_ds;
  int __pyx_v_capability;
  PyObject *__pyx_v_col_crs = NULL;
  PyObject *__pyx_v_proj_b = NULL;
  PyObject *__pyx_v_sysencoding = NULL;
  PyObject *__pyx_v_fileencoding_b = NULL;
  int __pyx_v_layer_count;
  PyObject *__pyx_v_layer_names = NULL;
  int __pyx_v_i;
  void *__pyx_v_cogr_layer;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_k = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_v_geometry_type = NULL;
  PyObject *__pyx_v_geometry_types = NULL;
  unsigned int __pyx_v_geometry_code;
  PyObject *__pyx_v_encoding = NULL;
  PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_width = NULL;
  PyObject *__pyx_v_precision = NULL;
  PyObject *__pyx_v_fmt = NULL;
  PyObject *__pyx_v_field_type = NULL;
  PyObject *__pyx_v_key_bytes = NULL;
  void *__pyx_v_cogr_fielddefn;
  PyObject *__pyx_v_ogr_schema = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  char const *__pyx_t_10;
  void *__pyx_t_11;
  char const *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  char const *__pyx_t_19;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  char const *__pyx_t_26;
  int __pyx_t_27;
  char const *__pyx_t_28;
  char const *__pyx_t_29;
  char const *__pyx_t_30;
  int __pyx_t_31;
  int __pyx_t_32;
  Py_ssize_t __pyx_t_33;
  Py_ssize_t __pyx_t_34;
  char const *__pyx_t_35;
  char const *__pyx_t_36;
  unsigned int __pyx_t_37;
  char const *__pyx_t_38;
  char const *__pyx_t_39;
  PyObject *(*__pyx_t_40)(PyObject *);
  char *__pyx_t_41;
  OGRFieldType __pyx_t_42;
  PyObject *__pyx_t_43 = NULL;
  char const *__pyx_t_44;
  __Pyx_RefNannySetupContext("start", 0);

  /* "fiona/ogrext.pyx":880
 * 
 *     def start(self, collection, **kwargs):
 *         cdef OGRSpatialReferenceH cogr_srs = NULL             # <<<<<<<<<<<<<<
 *         cdef char **options = NULL
 *         cdef const char *path_c = NULL
 */
  __pyx_v_cogr_srs = NULL;

  /* "fiona/ogrext.pyx":881
 *     def start(self, collection, **kwargs):
 *         cdef OGRSpatialReferenceH cogr_srs = NULL
 *         cdef char **options = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *path_c = NULL
 *         cdef const char *driver_c = NULL
 */
  __pyx_v_options = NULL;

  /* "fiona/ogrext.pyx":882
 *         cdef OGRSpatialReferenceH cogr_srs = NULL
 *         cdef char **options = NULL
 *         cdef const char *path_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *driver_c = NULL
 *         cdef const char *name_c = NULL
 */
  __pyx_v_path_c = NULL;

  /* "fiona/ogrext.pyx":883
 *         cdef char **options = NULL
 *         cdef const char *path_c = NULL
 *         cdef const char *driver_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *name_c = NULL
 *         cdef const char *proj_c = NULL
 */
  __pyx_v_driver_c = NULL;

  /* "fiona/ogrext.pyx":884
 *         cdef const char *path_c = NULL
 *         cdef const char *driver_c = NULL
 *         cdef const char *name_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *proj_c = NULL
 *         cdef const char *fileencoding_c = NULL
 */
  __pyx_v_name_c = NULL;

  /* "fiona/ogrext.pyx":885
 *         cdef const char *driver_c = NULL
 *         cdef const char *name_c = NULL
 *         cdef const char *proj_c = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *fileencoding_c = NULL
 *         cdef OGRFieldSubType field_subtype
 */
  __pyx_v_proj_c = NULL;

  /* "fiona/ogrext.pyx":886
 *         cdef const char *name_c = NULL
 *         cdef const char *proj_c = NULL
 *         cdef const char *fileencoding_c = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRFieldSubType field_subtype
 *         cdef int ret
 */
  __pyx_v_fileencoding_c = NULL;

  /* "fiona/ogrext.pyx":889
 *         cdef OGRFieldSubType field_subtype
 *         cdef int ret
 *         path = collection.path             # <<<<<<<<<<<<<<
 *         self.collection = collection
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_path = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":890
 *         cdef int ret
 *         path = collection.path
 *         self.collection = collection             # <<<<<<<<<<<<<<
 * 
 *         userencoding = kwargs.get('encoding')
 */
  __Pyx_INCREF(__pyx_v_collection);
  __Pyx_GIVEREF(__pyx_v_collection);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.collection);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.collection);
  __pyx_v_self->__pyx_base.collection = __pyx_v_collection;

  /* "fiona/ogrext.pyx":892
 *         self.collection = collection
 * 
 *         userencoding = kwargs.get('encoding')             # <<<<<<<<<<<<<<
 * 
 *         if collection.mode == 'a':
 */
  __pyx_t_1 = __Pyx_PyDict_GetItemDefault(__pyx_v_kwargs, __pyx_n_u_encoding, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 892, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_userencoding = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":894
 *         userencoding = kwargs.get('encoding')
 * 
 *         if collection.mode == 'a':             # <<<<<<<<<<<<<<
 * 
 *             if not os.path.exists(path):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_a, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":896
 *         if collection.mode == 'a':
 * 
 *             if not os.path.exists(path):             # <<<<<<<<<<<<<<
 *                 raise OSError("No such file or directory %s" % path)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_exists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 896, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = ((!__pyx_t_2) != 0);
    if (unlikely(__pyx_t_5)) {

      /* "fiona/ogrext.pyx":897
 * 
 *             if not os.path.exists(path):
 *                 raise OSError("No such file or directory %s" % path)             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
      __pyx_t_1 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_No_such_file_or_directory_s, __pyx_v_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_OSError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 897, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 897, __pyx_L1_error)

      /* "fiona/ogrext.pyx":896
 *         if collection.mode == 'a':
 * 
 *             if not os.path.exists(path):             # <<<<<<<<<<<<<<
 *                 raise OSError("No such file or directory %s" % path)
 * 
 */
    }

    /* "fiona/ogrext.pyx":899
 *                 raise OSError("No such file or directory %s" % path)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "fiona/ogrext.pyx":900
 * 
 *             try:
 *                 path_b = path.encode('utf-8')             # <<<<<<<<<<<<<<
 *             except UnicodeDecodeError:
 *                 path_b = path
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 900, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_4, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 900, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_path_b = __pyx_t_3;
        __pyx_t_3 = 0;

        /* "fiona/ogrext.pyx":899
 *                 raise OSError("No such file or directory %s" % path)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L10_try_end;
      __pyx_L5_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":901
 *             try:
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *                 path_b = path
 *             path_c = path_b
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 901, __pyx_L7_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);

        /* "fiona/ogrext.pyx":902
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 *                 path_b = path             # <<<<<<<<<<<<<<
 *             path_c = path_b
 * 
 */
        __Pyx_INCREF(__pyx_v_path);
        __Pyx_XDECREF_SET(__pyx_v_path_b, __pyx_v_path);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L6_exception_handled;
      }
      goto __pyx_L7_except_error;
      __pyx_L7_except_error:;

      /* "fiona/ogrext.pyx":899
 *                 raise OSError("No such file or directory %s" % path)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L6_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L10_try_end:;
    }

    /* "fiona/ogrext.pyx":903
 *             except UnicodeDecodeError:
 *                 path_b = path
 *             path_c = path_b             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __pyx_t_10 = __Pyx_PyObject_AsString(__pyx_v_path_b); if (unlikely((!__pyx_t_10) && PyErr_Occurred())) __PYX_ERR(0, 903, __pyx_L1_error)
    __pyx_v_path_c = __pyx_t_10;

    /* "fiona/ogrext.pyx":905
 *             path_c = path_b
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 * 
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_6);
      /*try:*/ {

        /* "fiona/ogrext.pyx":906
 * 
 *             try:
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)             # <<<<<<<<<<<<<<
 * 
 *                 if isinstance(collection.name, string_types):
 */
        __pyx_t_11 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_v_path_c, 1, Py_None, __pyx_v_kwargs); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 906, __pyx_L13_error)
        __pyx_v_self->__pyx_base.cogr_ds = __pyx_t_11;

        /* "fiona/ogrext.pyx":908
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 * 
 *                 if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *                     name_b = collection.name.encode('utf-8')
 *                     name_c = name_b
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 908, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_string_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 908, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_5 = PyObject_IsInstance(__pyx_t_4, __pyx_t_1); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(0, 908, __pyx_L13_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_2 = (__pyx_t_5 != 0);
        if (__pyx_t_2) {

          /* "fiona/ogrext.pyx":909
 * 
 *                 if isinstance(collection.name, string_types):
 *                     name_b = collection.name.encode('utf-8')             # <<<<<<<<<<<<<<
 *                     name_c = name_b
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayerByName(self.cogr_ds, name_c))
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 909, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 909, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
            }
          }
          __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_utf_8);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_v_name_b = __pyx_t_1;
          __pyx_t_1 = 0;

          /* "fiona/ogrext.pyx":910
 *                 if isinstance(collection.name, string_types):
 *                     name_b = collection.name.encode('utf-8')
 *                     name_c = name_b             # <<<<<<<<<<<<<<
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayerByName(self.cogr_ds, name_c))
 * 
 */
          __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_v_name_b); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 910, __pyx_L13_error)
          __pyx_v_name_c = __pyx_t_12;

          /* "fiona/ogrext.pyx":911
 *                     name_b = collection.name.encode('utf-8')
 *                     name_c = name_b
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayerByName(self.cogr_ds, name_c))             # <<<<<<<<<<<<<<
 * 
 *                 elif isinstance(collection.name, int):
 */
          __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(GDALDatasetGetLayerByName(__pyx_v_self->__pyx_base.cogr_ds, __pyx_v_name_c)); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 911, __pyx_L13_error)
          __pyx_v_self->__pyx_base.cogr_layer = __pyx_t_11;

          /* "fiona/ogrext.pyx":908
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 * 
 *                 if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *                     name_b = collection.name.encode('utf-8')
 *                     name_c = name_b
 */
          goto __pyx_L19;
        }

        /* "fiona/ogrext.pyx":913
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayerByName(self.cogr_ds, name_c))
 * 
 *                 elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayer(self.cogr_ds, collection.name))
 * 
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L13_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_2 = PyInt_Check(__pyx_t_1); 
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_5 = (__pyx_t_2 != 0);
        if (__pyx_t_5) {

          /* "fiona/ogrext.pyx":914
 * 
 *                 elif isinstance(collection.name, int):
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayer(self.cogr_ds, collection.name))             # <<<<<<<<<<<<<<
 * 
 *             except CPLE_BaseError as exc:
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 914, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 914, __pyx_L13_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(GDALDatasetGetLayer(__pyx_v_self->__pyx_base.cogr_ds, __pyx_t_9)); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 914, __pyx_L13_error)
          __pyx_v_self->__pyx_base.cogr_layer = __pyx_t_11;

          /* "fiona/ogrext.pyx":913
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayerByName(self.cogr_ds, name_c))
 * 
 *                 elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayer(self.cogr_ds, collection.name))
 * 
 */
        }
        __pyx_L19:;

        /* "fiona/ogrext.pyx":905
 *             path_c = path_b
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 * 
 */
      }

      /* "fiona/ogrext.pyx":923
 * 
 *             else:
 *                 self._fileencoding = userencoding or self._get_fallback_encoding()             # <<<<<<<<<<<<<<
 * 
 *         elif collection.mode == 'w':
 */
      /*else:*/ {
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_userencoding); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 923, __pyx_L15_except_error)
        if (!__pyx_t_5) {
        } else {
          __Pyx_INCREF(__pyx_v_userencoding);
          __pyx_t_1 = __pyx_v_userencoding;
          goto __pyx_L20_bool_binop_done;
        }
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_fallback_encoding); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 923, __pyx_L15_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 923, __pyx_L15_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_t_3);
        __pyx_t_1 = __pyx_t_3;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_L20_bool_binop_done:;
        __Pyx_GIVEREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_v_self->__pyx_base._fileencoding);
        __Pyx_DECREF(__pyx_v_self->__pyx_base._fileencoding);
        __pyx_v_self->__pyx_base._fileencoding = __pyx_t_1;
        __pyx_t_1 = 0;
      }
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L18_try_end;
      __pyx_L13_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":916
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayer(self.cogr_ds, collection.name))
 * 
 *             except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
      __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_CPLE_BaseError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 916, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_9 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_13);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_ErrRestore(__pyx_t_1, __pyx_t_3, __pyx_t_4);
      __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_4 = 0;
      if (__pyx_t_9) {
        __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 916, __pyx_L15_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __pyx_v_exc = __pyx_t_3;
        /*try:*/ {

          /* "fiona/ogrext.pyx":917
 * 
 *             except CPLE_BaseError as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL
 */
          (void)(OGRReleaseDataSource(__pyx_v_self->__pyx_base.cogr_ds));

          /* "fiona/ogrext.pyx":918
 *             except CPLE_BaseError as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 *                 self.cogr_layer = NULL
 *                 raise DriverError(u"{}".format(exc))
 */
          __pyx_v_self->__pyx_base.cogr_ds = NULL;

          /* "fiona/ogrext.pyx":919
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL             # <<<<<<<<<<<<<<
 *                 raise DriverError(u"{}".format(exc))
 * 
 */
          __pyx_v_self->__pyx_base.cogr_layer = NULL;

          /* "fiona/ogrext.pyx":920
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL
 *                 raise DriverError(u"{}".format(exc))             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_DriverError); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 920, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__13, __pyx_n_s_format); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 920, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
            }
          }
          __pyx_t_15 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_17, __pyx_v_exc) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_exc);
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 920, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_16 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
            }
          }
          __pyx_t_13 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_15);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 920, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_Raise(__pyx_t_13, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __PYX_ERR(0, 920, __pyx_L27_error)
        }

        /* "fiona/ogrext.pyx":916
 *                     self.cogr_layer = exc_wrap_pointer(GDALDatasetGetLayer(self.cogr_ds, collection.name))
 * 
 *             except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
        /*finally:*/ {
          __pyx_L27_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_20);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_24);
            __Pyx_XGOTREF(__pyx_t_25);
            __pyx_t_9 = __pyx_lineno; __pyx_t_18 = __pyx_clineno; __pyx_t_19 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc);
              __pyx_v_exc = NULL;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_23);
              __Pyx_XGIVEREF(__pyx_t_24);
              __Pyx_XGIVEREF(__pyx_t_25);
              __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
            }
            __Pyx_XGIVEREF(__pyx_t_20);
            __Pyx_XGIVEREF(__pyx_t_21);
            __Pyx_XGIVEREF(__pyx_t_22);
            __Pyx_ErrRestore(__pyx_t_20, __pyx_t_21, __pyx_t_22);
            __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
            __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_18; __pyx_filename = __pyx_t_19;
            goto __pyx_L15_except_error;
          }
        }
      }
      goto __pyx_L15_except_error;
      __pyx_L15_except_error:;

      /* "fiona/ogrext.pyx":905
 *             path_c = path_b
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 * 
 */
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
      goto __pyx_L1_error;
      __pyx_L18_try_end:;
    }

    /* "fiona/ogrext.pyx":894
 *         userencoding = kwargs.get('encoding')
 * 
 *         if collection.mode == 'a':             # <<<<<<<<<<<<<<
 * 
 *             if not os.path.exists(path):
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":925
 *                 self._fileencoding = userencoding or self._get_fallback_encoding()
 * 
 *         elif collection.mode == 'w':             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_w, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "fiona/ogrext.pyx":927
 *         elif collection.mode == 'w':
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "fiona/ogrext.pyx":928
 * 
 *             try:
 *                 path_b = path.encode('utf-8')             # <<<<<<<<<<<<<<
 *             except UnicodeDecodeError:
 *                 path_b = path
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 928, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_utf_8);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 928, __pyx_L33_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_v_path_b = __pyx_t_1;
        __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":927
 *         elif collection.mode == 'w':
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L38_try_end;
      __pyx_L33_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":929
 *             try:
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *                 path_b = path
 *             path_c = path_b
 */
      __pyx_t_18 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
      if (__pyx_t_18) {
        __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_1, &__pyx_t_3, &__pyx_t_4) < 0) __PYX_ERR(0, 929, __pyx_L35_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_4);

        /* "fiona/ogrext.pyx":930
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 *                 path_b = path             # <<<<<<<<<<<<<<
 *             path_c = path_b
 * 
 */
        __Pyx_INCREF(__pyx_v_path);
        __Pyx_XDECREF_SET(__pyx_v_path_b, __pyx_v_path);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        goto __pyx_L34_exception_handled;
      }
      goto __pyx_L35_except_error;
      __pyx_L35_except_error:;

      /* "fiona/ogrext.pyx":927
 *         elif collection.mode == 'w':
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 path_b = path.encode('utf-8')
 *             except UnicodeDecodeError:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L34_exception_handled:;
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      __pyx_L38_try_end:;
    }

    /* "fiona/ogrext.pyx":931
 *             except UnicodeDecodeError:
 *                 path_b = path
 *             path_c = path_b             # <<<<<<<<<<<<<<
 * 
 *             driver_b = collection.driver.encode()
 */
    __pyx_t_10 = __Pyx_PyObject_AsString(__pyx_v_path_b); if (unlikely((!__pyx_t_10) && PyErr_Occurred())) __PYX_ERR(0, 931, __pyx_L1_error)
    __pyx_v_path_c = __pyx_t_10;

    /* "fiona/ogrext.pyx":933
 *             path_c = path_b
 * 
 *             driver_b = collection.driver.encode()             # <<<<<<<<<<<<<<
 *             driver_c = driver_b
 *             cogr_driver = exc_wrap_pointer(GDALGetDriverByName(driver_c))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 933, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_driver_b = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":934
 * 
 *             driver_b = collection.driver.encode()
 *             driver_c = driver_b             # <<<<<<<<<<<<<<
 *             cogr_driver = exc_wrap_pointer(GDALGetDriverByName(driver_c))
 * 
 */
    __pyx_t_26 = __Pyx_PyObject_AsString(__pyx_v_driver_b); if (unlikely((!__pyx_t_26) && PyErr_Occurred())) __PYX_ERR(0, 934, __pyx_L1_error)
    __pyx_v_driver_c = __pyx_t_26;

    /* "fiona/ogrext.pyx":935
 *             driver_b = collection.driver.encode()
 *             driver_c = driver_b
 *             cogr_driver = exc_wrap_pointer(GDALGetDriverByName(driver_c))             # <<<<<<<<<<<<<<
 * 
 *             # Our most common use case is the creation of a new data
 */
    __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(GDALGetDriverByName(__pyx_v_driver_c)); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 935, __pyx_L1_error)
    __pyx_v_cogr_driver = __pyx_t_11;

    /* "fiona/ogrext.pyx":942
 *             #
 *             # TODO: remove the assumption.
 *             if not os.path.exists(path):             # <<<<<<<<<<<<<<
 *                 log.debug("File doesn't exist. Creating a new one...")
 *                 cogr_ds = gdal_create(cogr_driver, path_c, {})
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_os); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_exists); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 942, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = ((!__pyx_t_5) != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":943
 *             # TODO: remove the assumption.
 *             if not os.path.exists(path):
 *                 log.debug("File doesn't exist. Creating a new one...")             # <<<<<<<<<<<<<<
 *                 cogr_ds = gdal_create(cogr_driver, path_c, {})
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_kp_u_File_doesn_t_exist_Creating_a_ne) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_File_doesn_t_exist_Creating_a_ne);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 943, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":944
 *             if not os.path.exists(path):
 *                 log.debug("File doesn't exist. Creating a new one...")
 *                 cogr_ds = gdal_create(cogr_driver, path_c, {})             # <<<<<<<<<<<<<<
 * 
 *             # TODO: revisit the logic in the following blocks when we
 */
      __pyx_t_4 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 944, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_11 = __pyx_f_5fiona_5_shim_gdal_create(__pyx_v_cogr_driver, __pyx_v_path_c, __pyx_t_4); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 944, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_cogr_ds = __pyx_t_11;

      /* "fiona/ogrext.pyx":942
 *             #
 *             # TODO: remove the assumption.
 *             if not os.path.exists(path):             # <<<<<<<<<<<<<<
 *                 log.debug("File doesn't exist. Creating a new one...")
 *                 cogr_ds = gdal_create(cogr_driver, path_c, {})
 */
      goto __pyx_L41;
    }

    /* "fiona/ogrext.pyx":949
 *             # change the assumption above.
 *             else:
 *                 if collection.driver == "GeoJSON" and os.path.exists(path):             # <<<<<<<<<<<<<<
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_n_u_GeoJSON, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_5) {
      } else {
        __pyx_t_2 = __pyx_t_5;
        goto __pyx_L43_bool_binop_done;
      }
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_exists); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_path);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 949, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __pyx_t_5;
      __pyx_L43_bool_binop_done:;
      if (__pyx_t_2) {

        /* "fiona/ogrext.pyx":951
 *                 if collection.driver == "GeoJSON" and os.path.exists(path):
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)             # <<<<<<<<<<<<<<
 *                 try:
 *                     # attempt to open existing dataset in write mode
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 951, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unlink); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 951, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 951, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":949
 *             # change the assumption above.
 *             else:
 *                 if collection.driver == "GeoJSON" and os.path.exists(path):             # <<<<<<<<<<<<<<
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)
 */
      }

      /* "fiona/ogrext.pyx":952
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)
 *                 try:             # <<<<<<<<<<<<<<
 *                     # attempt to open existing dataset in write mode
 *                     cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        /*try:*/ {

          /* "fiona/ogrext.pyx":954
 *                 try:
 *                     # attempt to open existing dataset in write mode
 *                     cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)             # <<<<<<<<<<<<<<
 *                 except DriverError:
 *                     # failed, attempt to create it
 */
          __pyx_t_11 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_v_path_c, 1, Py_None, __pyx_v_kwargs); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 954, __pyx_L45_error)
          __pyx_v_cogr_ds = __pyx_t_11;

          /* "fiona/ogrext.pyx":952
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)
 *                 try:             # <<<<<<<<<<<<<<
 *                     # attempt to open existing dataset in write mode
 *                     cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 */
        }

        /* "fiona/ogrext.pyx":960
 *                 else:
 *                     # check capability of creating a new layer in the existing dataset
 *                     capability = check_capability_create_layer(cogr_ds)             # <<<<<<<<<<<<<<
 *                     if GDAL_VERSION_NUM < 2000000 and collection.driver == "GeoJSON":
 *                         # GeoJSON driver tells lies about it's capability
 */
        /*else:*/ {
          __pyx_v_capability = __pyx_f_5fiona_5_shim_check_capability_create_layer(__pyx_v_cogr_ds);

          /* "fiona/ogrext.pyx":961
 *                     # check capability of creating a new layer in the existing dataset
 *                     capability = check_capability_create_layer(cogr_ds)
 *                     if GDAL_VERSION_NUM < 2000000 and collection.driver == "GeoJSON":             # <<<<<<<<<<<<<<
 *                         # GeoJSON driver tells lies about it's capability
 *                         capability = False
 */
          __pyx_t_5 = ((__pyx_v_5fiona_6ogrext_GDAL_VERSION_NUM < 0x1E8480) != 0);
          if (__pyx_t_5) {
          } else {
            __pyx_t_2 = __pyx_t_5;
            goto __pyx_L52_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 961, __pyx_L47_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_n_u_GeoJSON, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 961, __pyx_L47_except_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_2 = __pyx_t_5;
          __pyx_L52_bool_binop_done:;
          if (__pyx_t_2) {

            /* "fiona/ogrext.pyx":963
 *                     if GDAL_VERSION_NUM < 2000000 and collection.driver == "GeoJSON":
 *                         # GeoJSON driver tells lies about it's capability
 *                         capability = False             # <<<<<<<<<<<<<<
 *                     if not capability or collection.name is None:
 *                         # unable to use existing dataset, recreate it
 */
            __pyx_v_capability = 0;

            /* "fiona/ogrext.pyx":961
 *                     # check capability of creating a new layer in the existing dataset
 *                     capability = check_capability_create_layer(cogr_ds)
 *                     if GDAL_VERSION_NUM < 2000000 and collection.driver == "GeoJSON":             # <<<<<<<<<<<<<<
 *                         # GeoJSON driver tells lies about it's capability
 *                         capability = False
 */
          }

          /* "fiona/ogrext.pyx":964
 *                         # GeoJSON driver tells lies about it's capability
 *                         capability = False
 *                     if not capability or collection.name is None:             # <<<<<<<<<<<<<<
 *                         # unable to use existing dataset, recreate it
 *                         GDALClose(cogr_ds)
 */
          __pyx_t_5 = ((!(__pyx_v_capability != 0)) != 0);
          if (!__pyx_t_5) {
          } else {
            __pyx_t_2 = __pyx_t_5;
            goto __pyx_L55_bool_binop_done;
          }
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 964, __pyx_L47_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = (__pyx_t_4 == Py_None);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_27 = (__pyx_t_5 != 0);
          __pyx_t_2 = __pyx_t_27;
          __pyx_L55_bool_binop_done:;
          if (__pyx_t_2) {

            /* "fiona/ogrext.pyx":966
 *                     if not capability or collection.name is None:
 *                         # unable to use existing dataset, recreate it
 *                         GDALClose(cogr_ds)             # <<<<<<<<<<<<<<
 *                         cogr_ds = NULL
 *                         cogr_ds = gdal_create(cogr_driver, path_c, kwargs)
 */
            GDALClose(__pyx_v_cogr_ds);

            /* "fiona/ogrext.pyx":967
 *                         # unable to use existing dataset, recreate it
 *                         GDALClose(cogr_ds)
 *                         cogr_ds = NULL             # <<<<<<<<<<<<<<
 *                         cogr_ds = gdal_create(cogr_driver, path_c, kwargs)
 * 
 */
            __pyx_v_cogr_ds = NULL;

            /* "fiona/ogrext.pyx":968
 *                         GDALClose(cogr_ds)
 *                         cogr_ds = NULL
 *                         cogr_ds = gdal_create(cogr_driver, path_c, kwargs)             # <<<<<<<<<<<<<<
 * 
 *             self.cogr_ds = cogr_ds
 */
            __pyx_t_11 = __pyx_f_5fiona_5_shim_gdal_create(__pyx_v_cogr_driver, __pyx_v_path_c, __pyx_v_kwargs); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 968, __pyx_L47_except_error)
            __pyx_v_cogr_ds = __pyx_t_11;

            /* "fiona/ogrext.pyx":964
 *                         # GeoJSON driver tells lies about it's capability
 *                         capability = False
 *                     if not capability or collection.name is None:             # <<<<<<<<<<<<<<
 *                         # unable to use existing dataset, recreate it
 *                         GDALClose(cogr_ds)
 */
          }
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L50_try_end;
        __pyx_L45_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":955
 *                     # attempt to open existing dataset in write mode
 *                     cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 *                 except DriverError:             # <<<<<<<<<<<<<<
 *                     # failed, attempt to create it
 *                     cogr_ds = gdal_create(cogr_driver, path_c, kwargs)
 */
        __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_1, &__pyx_t_3);
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_DriverError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 955, __pyx_L47_except_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_18 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_13);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_ErrRestore(__pyx_t_4, __pyx_t_1, __pyx_t_3);
        __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_3 = 0;
        if (__pyx_t_18) {
          __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 955, __pyx_L47_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_4);

          /* "fiona/ogrext.pyx":957
 *                 except DriverError:
 *                     # failed, attempt to create it
 *                     cogr_ds = gdal_create(cogr_driver, path_c, kwargs)             # <<<<<<<<<<<<<<
 *                 else:
 *                     # check capability of creating a new layer in the existing dataset
 */
          __pyx_t_11 = __pyx_f_5fiona_5_shim_gdal_create(__pyx_v_cogr_driver, __pyx_v_path_c, __pyx_v_kwargs); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 957, __pyx_L47_except_error)
          __pyx_v_cogr_ds = __pyx_t_11;
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          goto __pyx_L46_exception_handled;
        }
        goto __pyx_L47_except_error;
        __pyx_L47_except_error:;

        /* "fiona/ogrext.pyx":952
 *                     # manually remove geojson file as GDAL doesn't do this for us
 *                     os.unlink(path)
 *                 try:             # <<<<<<<<<<<<<<
 *                     # attempt to open existing dataset in write mode
 *                     cogr_ds = gdal_open_vector(path_c, 1, None, kwargs)
 */
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        goto __pyx_L1_error;
        __pyx_L46_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        __pyx_L50_try_end:;
      }
    }
    __pyx_L41:;

    /* "fiona/ogrext.pyx":970
 *                         cogr_ds = gdal_create(cogr_driver, path_c, kwargs)
 * 
 *             self.cogr_ds = cogr_ds             # <<<<<<<<<<<<<<
 * 
 *             # Set the spatial reference system from the crs given to the
 */
    __pyx_v_self->__pyx_base.cogr_ds = __pyx_v_cogr_ds;

    /* "fiona/ogrext.pyx":976
 *             # properties because they aren't accessible until the layer
 *             # is constructed (later).
 *             try:             # <<<<<<<<<<<<<<
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      /*try:*/ {

        /* "fiona/ogrext.pyx":977
 *             # is constructed (later).
 *             try:
 *                 col_crs = collection._crs_wkt             # <<<<<<<<<<<<<<
 *                 if col_crs:
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_crs_wkt); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 977, __pyx_L59_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_v_col_crs = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":978
 *             try:
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:             # <<<<<<<<<<<<<<
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))
 *                     proj_b = col_crs.encode('utf-8')
 */
        __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_col_crs); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 978, __pyx_L59_error)
        if (__pyx_t_2) {

          /* "fiona/ogrext.pyx":979
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))             # <<<<<<<<<<<<<<
 *                     proj_b = col_crs.encode('utf-8')
 *                     proj_c = proj_b
 */
          __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(OSRNewSpatialReference(NULL)); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 979, __pyx_L59_error)
          __pyx_v_cogr_srs = __pyx_t_11;

          /* "fiona/ogrext.pyx":980
 *                 if col_crs:
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))
 *                     proj_b = col_crs.encode('utf-8')             # <<<<<<<<<<<<<<
 *                     proj_c = proj_b
 *                     OSRSetFromUserInput(cogr_srs, proj_c)
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_col_crs, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 980, __pyx_L59_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
            }
          }
          __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 980, __pyx_L59_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_v_proj_b = __pyx_t_4;
          __pyx_t_4 = 0;

          /* "fiona/ogrext.pyx":981
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))
 *                     proj_b = col_crs.encode('utf-8')
 *                     proj_c = proj_b             # <<<<<<<<<<<<<<
 *                     OSRSetFromUserInput(cogr_srs, proj_c)
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)
 */
          __pyx_t_28 = __Pyx_PyObject_AsString(__pyx_v_proj_b); if (unlikely((!__pyx_t_28) && PyErr_Occurred())) __PYX_ERR(0, 981, __pyx_L59_error)
          __pyx_v_proj_c = __pyx_t_28;

          /* "fiona/ogrext.pyx":982
 *                     proj_b = col_crs.encode('utf-8')
 *                     proj_c = proj_b
 *                     OSRSetFromUserInput(cogr_srs, proj_c)             # <<<<<<<<<<<<<<
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)
 *             except CPLE_BaseError as exc:
 */
          (void)(OSRSetFromUserInput(__pyx_v_cogr_srs, __pyx_v_proj_c));

          /* "fiona/ogrext.pyx":983
 *                     proj_c = proj_b
 *                     OSRSetFromUserInput(cogr_srs, proj_c)
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)             # <<<<<<<<<<<<<<
 *             except CPLE_BaseError as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)
 */
          __pyx_f_5fiona_5_shim_osr_set_traditional_axis_mapping_strategy(__pyx_v_cogr_srs);

          /* "fiona/ogrext.pyx":978
 *             try:
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:             # <<<<<<<<<<<<<<
 *                     cogr_srs = exc_wrap_pointer(OSRNewSpatialReference(NULL))
 *                     proj_b = col_crs.encode('utf-8')
 */
        }

        /* "fiona/ogrext.pyx":976
 *             # properties because they aren't accessible until the layer
 *             # is constructed (later).
 *             try:             # <<<<<<<<<<<<<<
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:
 */
      }
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      goto __pyx_L64_try_end;
      __pyx_L59_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":984
 *                     OSRSetFromUserInput(cogr_srs, proj_c)
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)
 *             except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
      __Pyx_ErrFetch(&__pyx_t_4, &__pyx_t_1, &__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_CPLE_BaseError); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 984, __pyx_L61_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_18 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_4, __pyx_t_13);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_ErrRestore(__pyx_t_4, __pyx_t_1, __pyx_t_3);
      __pyx_t_4 = 0; __pyx_t_1 = 0; __pyx_t_3 = 0;
      if (__pyx_t_18) {
        __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_4) < 0) __PYX_ERR(0, 984, __pyx_L61_except_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_1);
        __pyx_v_exc = __pyx_t_1;
        /*try:*/ {

          /* "fiona/ogrext.pyx":985
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)
 *             except CPLE_BaseError as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL
 */
          (void)(OGRReleaseDataSource(__pyx_v_self->__pyx_base.cogr_ds));

          /* "fiona/ogrext.pyx":986
 *             except CPLE_BaseError as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 *                 self.cogr_layer = NULL
 *                 raise CRSError(u"{}".format(exc))
 */
          __pyx_v_self->__pyx_base.cogr_ds = NULL;

          /* "fiona/ogrext.pyx":987
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL             # <<<<<<<<<<<<<<
 *                 raise CRSError(u"{}".format(exc))
 * 
 */
          __pyx_v_self->__pyx_base.cogr_layer = NULL;

          /* "fiona/ogrext.pyx":988
 *                 self.cogr_ds = NULL
 *                 self.cogr_layer = NULL
 *                 raise CRSError(u"{}".format(exc))             # <<<<<<<<<<<<<<
 * 
 *             # Determine which encoding to use. The encoding parameter given to
 */
          __Pyx_GetModuleGlobalName(__pyx_t_14, __pyx_n_s_CRSError); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 988, __pyx_L71_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__13, __pyx_n_s_format); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 988, __pyx_L71_error)
          __Pyx_GOTREF(__pyx_t_16);
          __pyx_t_17 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
            }
          }
          __pyx_t_15 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_17, __pyx_v_exc) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_exc);
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 988, __pyx_L71_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_16 = NULL;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_16)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_16);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
            }
          }
          __pyx_t_13 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_t_15);
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 988, __pyx_L71_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __Pyx_Raise(__pyx_t_13, 0, 0, 0);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __PYX_ERR(0, 988, __pyx_L71_error)
        }

        /* "fiona/ogrext.pyx":984
 *                     OSRSetFromUserInput(cogr_srs, proj_c)
 *                     osr_set_traditional_axis_mapping_strategy(cogr_srs)
 *             except CPLE_BaseError as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
        /*finally:*/ {
          __pyx_L71_error:;
          /*exception exit:*/{
            __Pyx_PyThreadState_declare
            __Pyx_PyThreadState_assign
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
            if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_25);
            __Pyx_XGOTREF(__pyx_t_24);
            __Pyx_XGOTREF(__pyx_t_23);
            __Pyx_XGOTREF(__pyx_t_22);
            __Pyx_XGOTREF(__pyx_t_21);
            __Pyx_XGOTREF(__pyx_t_20);
            __pyx_t_18 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_29 = __pyx_filename;
            {
              __Pyx_DECREF(__pyx_v_exc);
              __pyx_v_exc = NULL;
            }
            if (PY_MAJOR_VERSION >= 3) {
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
            }
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_23);
            __Pyx_ErrRestore(__pyx_t_25, __pyx_t_24, __pyx_t_23);
            __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
            __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_29;
            goto __pyx_L61_except_error;
          }
        }
      }
      goto __pyx_L61_except_error;
      __pyx_L61_except_error:;

      /* "fiona/ogrext.pyx":976
 *             # properties because they aren't accessible until the layer
 *             # is constructed (later).
 *             try:             # <<<<<<<<<<<<<<
 *                 col_crs = collection._crs_wkt
 *                 if col_crs:
 */
      __Pyx_XGIVEREF(__pyx_t_6);
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_ExceptionReset(__pyx_t_6, __pyx_t_7, __pyx_t_8);
      goto __pyx_L1_error;
      __pyx_L64_try_end:;
    }

    /* "fiona/ogrext.pyx":994
 *             # 'iso-8859-1' (for shapefiles), then the system's default encoding
 *             # as last resort.
 *             sysencoding = locale.getpreferredencoding()             # <<<<<<<<<<<<<<
 *             self._fileencoding = userencoding or ("Shapefile" in collection.driver and 'iso-8859-1') or sysencoding
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_locale); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getpreferredencoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_1) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_1) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_sysencoding = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":995
 *             # as last resort.
 *             sysencoding = locale.getpreferredencoding()
 *             self._fileencoding = userencoding or ("Shapefile" in collection.driver and 'iso-8859-1') or sysencoding             # <<<<<<<<<<<<<<
 * 
 *             if "Shapefile" in collection.driver:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_userencoding); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 995, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __Pyx_INCREF(__pyx_v_userencoding);
      __pyx_t_4 = __pyx_v_userencoding;
      goto __pyx_L77_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_Shapefile, __pyx_t_3, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_2) {
      goto __pyx_L79_next_or;
    } else {
    }
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_kp_u_iso_8859_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 995, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __Pyx_INCREF(__pyx_kp_u_iso_8859_1);
      __pyx_t_4 = __pyx_kp_u_iso_8859_1;
      goto __pyx_L77_bool_binop_done;
    }
    __pyx_L79_next_or:;
    __Pyx_INCREF(__pyx_v_sysencoding);
    __pyx_t_4 = __pyx_v_sysencoding;
    __pyx_L77_bool_binop_done:;
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->__pyx_base._fileencoding);
    __Pyx_DECREF(__pyx_v_self->__pyx_base._fileencoding);
    __pyx_v_self->__pyx_base._fileencoding = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":997
 *             self._fileencoding = userencoding or ("Shapefile" in collection.driver and 'iso-8859-1') or sysencoding
 * 
 *             if "Shapefile" in collection.driver:             # <<<<<<<<<<<<<<
 *                 if self._fileencoding:
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_Shapefile, __pyx_t_4, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 997, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_27 = (__pyx_t_2 != 0);
    if (__pyx_t_27) {

      /* "fiona/ogrext.pyx":998
 * 
 *             if "Shapefile" in collection.driver:
 *                 if self._fileencoding:             # <<<<<<<<<<<<<<
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 *                     fileencoding_c = fileencoding_b
 */
      __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_v_self->__pyx_base._fileencoding); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 998, __pyx_L1_error)
      if (__pyx_t_27) {

        /* "fiona/ogrext.pyx":999
 *             if "Shapefile" in collection.driver:
 *                 if self._fileencoding:
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 *                     fileencoding_c = fileencoding_b
 *                     options = CSLSetNameValue(options, "ENCODING", fileencoding_c)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._fileencoding, __pyx_n_s_upper); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_3 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_v_fileencoding_b = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":1000
 *                 if self._fileencoding:
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 *                     fileencoding_c = fileencoding_b             # <<<<<<<<<<<<<<
 *                     options = CSLSetNameValue(options, "ENCODING", fileencoding_c)
 * 
 */
        __pyx_t_30 = __Pyx_PyObject_AsString(__pyx_v_fileencoding_b); if (unlikely((!__pyx_t_30) && PyErr_Occurred())) __PYX_ERR(0, 1000, __pyx_L1_error)
        __pyx_v_fileencoding_c = __pyx_t_30;

        /* "fiona/ogrext.pyx":1001
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 *                     fileencoding_c = fileencoding_b
 *                     options = CSLSetNameValue(options, "ENCODING", fileencoding_c)             # <<<<<<<<<<<<<<
 * 
 *             # Does the layer exist already? If so, we delete it.
 */
        __pyx_v_options = CSLSetNameValue(__pyx_v_options, ((char const *)"ENCODING"), __pyx_v_fileencoding_c);

        /* "fiona/ogrext.pyx":998
 * 
 *             if "Shapefile" in collection.driver:
 *                 if self._fileencoding:             # <<<<<<<<<<<<<<
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 *                     fileencoding_c = fileencoding_b
 */
      }

      /* "fiona/ogrext.pyx":997
 *             self._fileencoding = userencoding or ("Shapefile" in collection.driver and 'iso-8859-1') or sysencoding
 * 
 *             if "Shapefile" in collection.driver:             # <<<<<<<<<<<<<<
 *                 if self._fileencoding:
 *                     fileencoding_b = self._fileencoding.upper().encode('utf-8')
 */
    }

    /* "fiona/ogrext.pyx":1004
 * 
 *             # Does the layer exist already? If so, we delete it.
 *             layer_count = GDALDatasetGetLayerCount(self.cogr_ds)             # <<<<<<<<<<<<<<
 *             layer_names = []
 *             for i in range(layer_count):
 */
    __pyx_v_layer_count = GDALDatasetGetLayerCount(__pyx_v_self->__pyx_base.cogr_ds);

    /* "fiona/ogrext.pyx":1005
 *             # Does the layer exist already? If so, we delete it.
 *             layer_count = GDALDatasetGetLayerCount(self.cogr_ds)
 *             layer_names = []             # <<<<<<<<<<<<<<
 *             for i in range(layer_count):
 *                 cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 */
    __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1005, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_layer_names = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":1006
 *             layer_count = GDALDatasetGetLayerCount(self.cogr_ds)
 *             layer_names = []
 *             for i in range(layer_count):             # <<<<<<<<<<<<<<
 *                 cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *                 name_c = OGR_L_GetName(cogr_layer)
 */
    __pyx_t_9 = __pyx_v_layer_count;
    __pyx_t_18 = __pyx_t_9;
    for (__pyx_t_31 = 0; __pyx_t_31 < __pyx_t_18; __pyx_t_31+=1) {
      __pyx_v_i = __pyx_t_31;

      /* "fiona/ogrext.pyx":1007
 *             layer_names = []
 *             for i in range(layer_count):
 *                 cogr_layer = GDALDatasetGetLayer(cogr_ds, i)             # <<<<<<<<<<<<<<
 *                 name_c = OGR_L_GetName(cogr_layer)
 *                 name_b = name_c
 */
      __pyx_v_cogr_layer = GDALDatasetGetLayer(__pyx_v_cogr_ds, __pyx_v_i);

      /* "fiona/ogrext.pyx":1008
 *             for i in range(layer_count):
 *                 cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *                 name_c = OGR_L_GetName(cogr_layer)             # <<<<<<<<<<<<<<
 *                 name_b = name_c
 *                 layer_names.append(name_b.decode('utf-8'))
 */
      __pyx_v_name_c = OGR_L_GetName(__pyx_v_cogr_layer);

      /* "fiona/ogrext.pyx":1009
 *                 cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *                 name_c = OGR_L_GetName(cogr_layer)
 *                 name_b = name_c             # <<<<<<<<<<<<<<
 *                 layer_names.append(name_b.decode('utf-8'))
 * 
 */
      __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_name_c); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1009, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_name_b, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":1010
 *                 name_c = OGR_L_GetName(cogr_layer)
 *                 name_b = name_c
 *                 layer_names.append(name_b.decode('utf-8'))             # <<<<<<<<<<<<<<
 * 
 *             idx = -1
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_name_b, __pyx_n_s_decode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_3, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_32 = __Pyx_PyList_Append(__pyx_v_layer_names, __pyx_t_4); if (unlikely(__pyx_t_32 == ((int)-1))) __PYX_ERR(0, 1010, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }

    /* "fiona/ogrext.pyx":1012
 *                 layer_names.append(name_b.decode('utf-8'))
 * 
 *             idx = -1             # <<<<<<<<<<<<<<
 *             if isinstance(collection.name, string_types):
 *                 if collection.name in layer_names:
 */
    __Pyx_INCREF(__pyx_int_neg_1);
    __pyx_v_idx = __pyx_int_neg_1;

    /* "fiona/ogrext.pyx":1013
 * 
 *             idx = -1
 *             if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *                 if collection.name in layer_names:
 *                     idx = layer_names.index(collection.name)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_string_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_27 = PyObject_IsInstance(__pyx_t_4, __pyx_t_1); if (unlikely(__pyx_t_27 == ((int)-1))) __PYX_ERR(0, 1013, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_27 != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":1014
 *             idx = -1
 *             if isinstance(collection.name, string_types):
 *                 if collection.name in layer_names:             # <<<<<<<<<<<<<<
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_v_layer_names, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1014, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_27 = (__pyx_t_2 != 0);
      if (__pyx_t_27) {

        /* "fiona/ogrext.pyx":1015
 *             if isinstance(collection.name, string_types):
 *                 if collection.name in layer_names:
 *                     idx = layer_names.index(collection.name)             # <<<<<<<<<<<<<<
 *             elif isinstance(collection.name, int):
 *                 if collection.name >= 0 and collection.name < layer_count:
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyList_Type_index, __pyx_v_layer_names, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1015, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":1014
 *             idx = -1
 *             if isinstance(collection.name, string_types):
 *                 if collection.name in layer_names:             # <<<<<<<<<<<<<<
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):
 */
      }

      /* "fiona/ogrext.pyx":1013
 * 
 *             idx = -1
 *             if isinstance(collection.name, string_types):             # <<<<<<<<<<<<<<
 *                 if collection.name in layer_names:
 *                     idx = layer_names.index(collection.name)
 */
      goto __pyx_L85;
    }

    /* "fiona/ogrext.pyx":1016
 *                 if collection.name in layer_names:
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *                 if collection.name >= 0 and collection.name < layer_count:
 *                     idx = collection.name
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1016, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_27 = PyInt_Check(__pyx_t_4); 
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = (__pyx_t_27 != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":1017
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):
 *                 if collection.name >= 0 and collection.name < layer_count:             # <<<<<<<<<<<<<<
 *                     idx = collection.name
 *             if idx >= 0:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_27) {
      } else {
        __pyx_t_2 = __pyx_t_27;
        goto __pyx_L88_bool_binop_done;
      }
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_layer_count); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_2 = __pyx_t_27;
      __pyx_L88_bool_binop_done:;
      if (__pyx_t_2) {

        /* "fiona/ogrext.pyx":1018
 *             elif isinstance(collection.name, int):
 *                 if collection.name >= 0 and collection.name < layer_count:
 *                     idx = collection.name             # <<<<<<<<<<<<<<
 *             if idx >= 0:
 *                 log.debug("Deleted pre-existing layer at %s", collection.name)
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF_SET(__pyx_v_idx, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "fiona/ogrext.pyx":1017
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):
 *                 if collection.name >= 0 and collection.name < layer_count:             # <<<<<<<<<<<<<<
 *                     idx = collection.name
 *             if idx >= 0:
 */
      }

      /* "fiona/ogrext.pyx":1016
 *                 if collection.name in layer_names:
 *                     idx = layer_names.index(collection.name)
 *             elif isinstance(collection.name, int):             # <<<<<<<<<<<<<<
 *                 if collection.name >= 0 and collection.name < layer_count:
 *                     idx = collection.name
 */
    }
    __pyx_L85:;

    /* "fiona/ogrext.pyx":1019
 *                 if collection.name >= 0 and collection.name < layer_count:
 *                     idx = collection.name
 *             if idx >= 0:             # <<<<<<<<<<<<<<
 *                 log.debug("Deleted pre-existing layer at %s", collection.name)
 *                 GDALDatasetDeleteLayer(self.cogr_ds, idx)
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_idx, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1019, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":1020
 *                     idx = collection.name
 *             if idx >= 0:
 *                 log.debug("Deleted pre-existing layer at %s", collection.name)             # <<<<<<<<<<<<<<
 *                 GDALDatasetDeleteLayer(self.cogr_ds, idx)
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = NULL;
      __pyx_t_9 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_9 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_kp_u_Deleted_pre_existing_layer_at_s, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1020, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_kp_u_Deleted_pre_existing_layer_at_s, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1020, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else
      #endif
      {
        __pyx_t_14 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1020, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        if (__pyx_t_13) {
          __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13); __pyx_t_13 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Deleted_pre_existing_layer_at_s);
        __Pyx_GIVEREF(__pyx_kp_u_Deleted_pre_existing_layer_at_s);
        PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_9, __pyx_kp_u_Deleted_pre_existing_layer_at_s);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_9, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1020, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":1021
 *             if idx >= 0:
 *                 log.debug("Deleted pre-existing layer at %s", collection.name)
 *                 GDALDatasetDeleteLayer(self.cogr_ds, idx)             # <<<<<<<<<<<<<<
 * 
 *             # Create the named layer in the datasource.
 */
      __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_v_idx); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1021, __pyx_L1_error)
      (void)(GDALDatasetDeleteLayer(__pyx_v_self->__pyx_base.cogr_ds, __pyx_t_9));

      /* "fiona/ogrext.pyx":1019
 *                 if collection.name >= 0 and collection.name < layer_count:
 *                     idx = collection.name
 *             if idx >= 0:             # <<<<<<<<<<<<<<
 *                 log.debug("Deleted pre-existing layer at %s", collection.name)
 *                 GDALDatasetDeleteLayer(self.cogr_ds, idx)
 */
    }

    /* "fiona/ogrext.pyx":1024
 * 
 *             # Create the named layer in the datasource.
 *             name_b = collection.name.encode('utf-8')             # <<<<<<<<<<<<<<
 *             name_c = name_b
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_14))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_14, function);
      }
    }
    __pyx_t_3 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_1, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1024, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_XDECREF_SET(__pyx_v_name_b, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1025
 *             # Create the named layer in the datasource.
 *             name_b = collection.name.encode('utf-8')
 *             name_c = name_b             # <<<<<<<<<<<<<<
 * 
 *             for k, v in kwargs.items():
 */
    __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_v_name_b); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1025, __pyx_L1_error)
    __pyx_v_name_c = __pyx_t_12;

    /* "fiona/ogrext.pyx":1027
 *             name_c = name_b
 * 
 *             for k, v in kwargs.items():             # <<<<<<<<<<<<<<
 * 
 *                 if v is None:
 */
    __pyx_t_33 = 0;
    __pyx_t_14 = __Pyx_dict_iterator(__pyx_v_kwargs, 1, __pyx_n_s_items, (&__pyx_t_34), (&__pyx_t_9)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_XDECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_14;
    __pyx_t_14 = 0;
    while (1) {
      __pyx_t_18 = __Pyx_dict_iter_next(__pyx_t_3, __pyx_t_34, &__pyx_t_33, &__pyx_t_14, &__pyx_t_1, NULL, __pyx_t_9);
      if (unlikely(__pyx_t_18 == 0)) break;
      if (unlikely(__pyx_t_18 == -1)) __PYX_ERR(0, 1027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_k, __pyx_t_14);
      __pyx_t_14 = 0;
      __Pyx_XDECREF_SET(__pyx_v_v, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":1029
 *             for k, v in kwargs.items():
 * 
 *                 if v is None:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      __pyx_t_2 = (__pyx_v_v == Py_None);
      __pyx_t_27 = (__pyx_t_2 != 0);
      if (__pyx_t_27) {

        /* "fiona/ogrext.pyx":1030
 * 
 *                 if v is None:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 # We need to remove encoding from the layer creation
 */
        goto __pyx_L91_continue;

        /* "fiona/ogrext.pyx":1029
 *             for k, v in kwargs.items():
 * 
 *                 if v is None:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      }

      /* "fiona/ogrext.pyx":1034
 *                 # We need to remove encoding from the layer creation
 *                 # options if we're not creating a shapefile.
 *                 if k == 'encoding' and "Shapefile" not in collection.driver:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_k, __pyx_n_u_encoding, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1034, __pyx_L1_error)
      if (__pyx_t_2) {
      } else {
        __pyx_t_27 = __pyx_t_2;
        goto __pyx_L95_bool_binop_done;
      }
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_n_u_Shapefile, __pyx_t_1, Py_NE)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1034, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = (__pyx_t_2 != 0);
      __pyx_t_27 = __pyx_t_5;
      __pyx_L95_bool_binop_done:;
      if (__pyx_t_27) {

        /* "fiona/ogrext.pyx":1035
 *                 # options if we're not creating a shapefile.
 *                 if k == 'encoding' and "Shapefile" not in collection.driver:
 *                     continue             # <<<<<<<<<<<<<<
 * 
 *                 k = k.upper().encode('utf-8')
 */
        goto __pyx_L91_continue;

        /* "fiona/ogrext.pyx":1034
 *                 # We need to remove encoding from the layer creation
 *                 # options if we're not creating a shapefile.
 *                 if k == 'encoding' and "Shapefile" not in collection.driver:             # <<<<<<<<<<<<<<
 *                     continue
 * 
 */
      }

      /* "fiona/ogrext.pyx":1037
 *                     continue
 * 
 *                 k = k.upper().encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *                 if isinstance(v, bool):
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_k, __pyx_n_s_upper); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_14 = (__pyx_t_13) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_13) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_14)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_14);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_14) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_14, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_utf_8);
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF_SET(__pyx_v_k, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":1039
 *                 k = k.upper().encode('utf-8')
 * 
 *                 if isinstance(v, bool):             # <<<<<<<<<<<<<<
 *                     v = ('ON' if v else 'OFF').encode('utf-8')
 *                 else:
 */
      __pyx_t_1 = ((PyObject*)&PyBool_Type);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_27 = PyObject_IsInstance(__pyx_v_v, __pyx_t_1); if (unlikely(__pyx_t_27 == ((int)-1))) __PYX_ERR(0, 1039, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_5 = (__pyx_t_27 != 0);
      if (__pyx_t_5) {

        /* "fiona/ogrext.pyx":1040
 * 
 *                 if isinstance(v, bool):
 *                     v = ('ON' if v else 'OFF').encode('utf-8')             # <<<<<<<<<<<<<<
 *                 else:
 *                     v = str(v).encode('utf-8')
 */
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_v); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1040, __pyx_L1_error)
        if (__pyx_t_5) {
          __Pyx_INCREF(__pyx_n_u_ON);
          __pyx_t_1 = __pyx_n_u_ON;
        } else {
          __Pyx_INCREF(__pyx_n_u_OFF);
          __pyx_t_1 = __pyx_n_u_OFF;
        }
        if (unlikely(__pyx_t_1 == Py_None)) {
          PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
          __PYX_ERR(0, 1040, __pyx_L1_error)
        }
        __pyx_t_4 = PyUnicode_AsUTF8String(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1040, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":1039
 *                 k = k.upper().encode('utf-8')
 * 
 *                 if isinstance(v, bool):             # <<<<<<<<<<<<<<
 *                     v = ('ON' if v else 'OFF').encode('utf-8')
 *                 else:
 */
        goto __pyx_L97;
      }

      /* "fiona/ogrext.pyx":1042
 *                     v = ('ON' if v else 'OFF').encode('utf-8')
 *                 else:
 *                     v = str(v).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 log.debug("Set option %r: %r", k, v)
 *                 options = CSLAddNameValue(options, <const char *>k, <const char *>v)
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_v_v); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = PyUnicode_AsUTF8String(((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1042, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF_SET(__pyx_v_v, __pyx_t_1);
        __pyx_t_1 = 0;
      }
      __pyx_L97:;

      /* "fiona/ogrext.pyx":1043
 *                 else:
 *                     v = str(v).encode('utf-8')
 *                 log.debug("Set option %r: %r", k, v)             # <<<<<<<<<<<<<<
 *                 options = CSLAddNameValue(options, <const char *>k, <const char *>v)
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      __pyx_t_18 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_14);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_14, function);
          __pyx_t_18 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_kp_u_Set_option_r_r, __pyx_v_k, __pyx_v_v};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
        PyObject *__pyx_temp[4] = {__pyx_t_4, __pyx_kp_u_Set_option_r_r, __pyx_v_k, __pyx_v_v};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(3+__pyx_t_18); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Set_option_r_r);
        __Pyx_GIVEREF(__pyx_kp_u_Set_option_r_r);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_18, __pyx_kp_u_Set_option_r_r);
        __Pyx_INCREF(__pyx_v_k);
        __Pyx_GIVEREF(__pyx_v_k);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_18, __pyx_v_k);
        __Pyx_INCREF(__pyx_v_v);
        __Pyx_GIVEREF(__pyx_v_v);
        PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_18, __pyx_v_v);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1043, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":1044
 *                     v = str(v).encode('utf-8')
 *                 log.debug("Set option %r: %r", k, v)
 *                 options = CSLAddNameValue(options, <const char *>k, <const char *>v)             # <<<<<<<<<<<<<<
 * 
 *             geometry_type = collection.schema.get("geometry", "Unknown")
 */
      __pyx_t_35 = __Pyx_PyObject_AsString(__pyx_v_k); if (unlikely((!__pyx_t_35) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L1_error)
      __pyx_t_36 = __Pyx_PyObject_AsString(__pyx_v_v); if (unlikely((!__pyx_t_36) && PyErr_Occurred())) __PYX_ERR(0, 1044, __pyx_L1_error)
      __pyx_v_options = CSLAddNameValue(__pyx_v_options, ((char const *)__pyx_t_35), ((char const *)__pyx_t_36));
      __pyx_L91_continue:;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1046
 *                 options = CSLAddNameValue(options, <const char *>k, <const char *>v)
 * 
 *             geometry_type = collection.schema.get("geometry", "Unknown")             # <<<<<<<<<<<<<<
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:
 *                 geometry_types = set(geometry_type)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_geometry_type = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1047
 * 
 *             geometry_type = collection.schema.get("geometry", "Unknown")
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:             # <<<<<<<<<<<<<<
 *                 geometry_types = set(geometry_type)
 *                 if len(geometry_types) > 1:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_string_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_27 = PyObject_IsInstance(__pyx_v_geometry_type, __pyx_t_3); if (unlikely(__pyx_t_27 == ((int)-1))) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = ((!(__pyx_t_27 != 0)) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_5 = __pyx_t_2;
      goto __pyx_L99_bool_binop_done;
    }
    __pyx_t_2 = (__pyx_v_geometry_type != Py_None);
    __pyx_t_27 = (__pyx_t_2 != 0);
    __pyx_t_5 = __pyx_t_27;
    __pyx_L99_bool_binop_done:;
    if (__pyx_t_5) {

      /* "fiona/ogrext.pyx":1048
 *             geometry_type = collection.schema.get("geometry", "Unknown")
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:
 *                 geometry_types = set(geometry_type)             # <<<<<<<<<<<<<<
 *                 if len(geometry_types) > 1:
 *                     geometry_type = "Unknown"
 */
      __pyx_t_3 = PySet_New(__pyx_v_geometry_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1048, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_v_geometry_types = ((PyObject*)__pyx_t_3);
      __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":1049
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:
 *                 geometry_types = set(geometry_type)
 *                 if len(geometry_types) > 1:             # <<<<<<<<<<<<<<
 *                     geometry_type = "Unknown"
 *                 else:
 */
      __pyx_t_34 = PySet_GET_SIZE(__pyx_v_geometry_types); if (unlikely(__pyx_t_34 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1049, __pyx_L1_error)
      __pyx_t_5 = ((__pyx_t_34 > 1) != 0);
      if (__pyx_t_5) {

        /* "fiona/ogrext.pyx":1050
 *                 geometry_types = set(geometry_type)
 *                 if len(geometry_types) > 1:
 *                     geometry_type = "Unknown"             # <<<<<<<<<<<<<<
 *                 else:
 *                     geometry_type = geometry_types.pop()
 */
        __Pyx_INCREF(__pyx_n_u_Unknown);
        __Pyx_DECREF_SET(__pyx_v_geometry_type, __pyx_n_u_Unknown);

        /* "fiona/ogrext.pyx":1049
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:
 *                 geometry_types = set(geometry_type)
 *                 if len(geometry_types) > 1:             # <<<<<<<<<<<<<<
 *                     geometry_type = "Unknown"
 *                 else:
 */
        goto __pyx_L101;
      }

      /* "fiona/ogrext.pyx":1052
 *                     geometry_type = "Unknown"
 *                 else:
 *                     geometry_type = geometry_types.pop()             # <<<<<<<<<<<<<<
 *             if geometry_type == "Any" or geometry_type is None:
 *                 geometry_type = "Unknown"
 */
      /*else*/ {
        __pyx_t_3 = PySet_Pop(__pyx_v_geometry_types); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1052, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF_SET(__pyx_v_geometry_type, __pyx_t_3);
        __pyx_t_3 = 0;
      }
      __pyx_L101:;

      /* "fiona/ogrext.pyx":1047
 * 
 *             geometry_type = collection.schema.get("geometry", "Unknown")
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:             # <<<<<<<<<<<<<<
 *                 geometry_types = set(geometry_type)
 *                 if len(geometry_types) > 1:
 */
    }

    /* "fiona/ogrext.pyx":1053
 *                 else:
 *                     geometry_type = geometry_types.pop()
 *             if geometry_type == "Any" or geometry_type is None:             # <<<<<<<<<<<<<<
 *                 geometry_type = "Unknown"
 *             geometry_code = geometry_type_code(geometry_type)
 */
    __pyx_t_27 = (__Pyx_PyUnicode_Equals(__pyx_v_geometry_type, __pyx_n_u_Any, Py_EQ)); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1053, __pyx_L1_error)
    if (!__pyx_t_27) {
    } else {
      __pyx_t_5 = __pyx_t_27;
      goto __pyx_L103_bool_binop_done;
    }
    __pyx_t_27 = (__pyx_v_geometry_type == Py_None);
    __pyx_t_2 = (__pyx_t_27 != 0);
    __pyx_t_5 = __pyx_t_2;
    __pyx_L103_bool_binop_done:;
    if (__pyx_t_5) {

      /* "fiona/ogrext.pyx":1054
 *                     geometry_type = geometry_types.pop()
 *             if geometry_type == "Any" or geometry_type is None:
 *                 geometry_type = "Unknown"             # <<<<<<<<<<<<<<
 *             geometry_code = geometry_type_code(geometry_type)
 * 
 */
      __Pyx_INCREF(__pyx_n_u_Unknown);
      __Pyx_DECREF_SET(__pyx_v_geometry_type, __pyx_n_u_Unknown);

      /* "fiona/ogrext.pyx":1053
 *                 else:
 *                     geometry_type = geometry_types.pop()
 *             if geometry_type == "Any" or geometry_type is None:             # <<<<<<<<<<<<<<
 *                 geometry_type = "Unknown"
 *             geometry_code = geometry_type_code(geometry_type)
 */
    }

    /* "fiona/ogrext.pyx":1055
 *             if geometry_type == "Any" or geometry_type is None:
 *                 geometry_type = "Unknown"
 *             geometry_code = geometry_type_code(geometry_type)             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
    __pyx_t_37 = __pyx_f_5fiona_9_geometry_geometry_type_code(__pyx_v_geometry_type); if (unlikely(__pyx_t_37 == ((unsigned int)0x270F) && PyErr_Occurred())) __PYX_ERR(0, 1055, __pyx_L1_error)
    __pyx_v_geometry_code = __pyx_t_37;

    /* "fiona/ogrext.pyx":1057
 *             geometry_code = geometry_type_code(geometry_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_layer = exc_wrap_pointer(
 *                     GDALDatasetCreateLayer(
 */
    /*try:*/ {
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_6);
        /*try:*/ {

          /* "fiona/ogrext.pyx":1058
 * 
 *             try:
 *                 self.cogr_layer = exc_wrap_pointer(             # <<<<<<<<<<<<<<
 *                     GDALDatasetCreateLayer(
 *                         self.cogr_ds, name_c, cogr_srs,
 */
          __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(GDALDatasetCreateLayer(__pyx_v_self->__pyx_base.cogr_ds, __pyx_v_name_c, __pyx_v_cogr_srs, ((OGRwkbGeometryType)__pyx_v_geometry_code), __pyx_v_options)); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 1058, __pyx_L108_error)
          __pyx_v_self->__pyx_base.cogr_layer = __pyx_t_11;

          /* "fiona/ogrext.pyx":1057
 *             geometry_code = geometry_type_code(geometry_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_layer = exc_wrap_pointer(
 *                     GDALDatasetCreateLayer(
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        goto __pyx_L113_try_end;
        __pyx_L108_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":1063
 *                         <OGRwkbGeometryType>geometry_code, options))
 * 
 *             except Exception as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
        __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_9) {
          __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_3, &__pyx_t_1, &__pyx_t_14) < 0) __PYX_ERR(0, 1063, __pyx_L110_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_INCREF(__pyx_t_1);
          __pyx_v_exc = __pyx_t_1;
          /*try:*/ {

            /* "fiona/ogrext.pyx":1064
 * 
 *             except Exception as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)             # <<<<<<<<<<<<<<
 *                 self.cogr_ds = NULL
 *                 raise DriverIOError(u"{}".format(exc))
 */
            (void)(OGRReleaseDataSource(__pyx_v_self->__pyx_base.cogr_ds));

            /* "fiona/ogrext.pyx":1065
 *             except Exception as exc:
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 *                 raise DriverIOError(u"{}".format(exc))
 * 
 */
            __pyx_v_self->__pyx_base.cogr_ds = NULL;

            /* "fiona/ogrext.pyx":1066
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 *                 raise DriverIOError(u"{}".format(exc))             # <<<<<<<<<<<<<<
 * 
 *             finally:
 */
            __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_DriverIOError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1066, __pyx_L119_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__13, __pyx_n_s_format); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1066, __pyx_L119_error)
            __Pyx_GOTREF(__pyx_t_16);
            __pyx_t_17 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_16))) {
              __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
              if (likely(__pyx_t_17)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
                __Pyx_INCREF(__pyx_t_17);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_16, function);
              }
            }
            __pyx_t_15 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_16, __pyx_t_17, __pyx_v_exc) : __Pyx_PyObject_CallOneArg(__pyx_t_16, __pyx_v_exc);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1066, __pyx_L119_error)
            __Pyx_GOTREF(__pyx_t_15);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __pyx_t_16 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_4);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_4, function);
              }
            }
            __pyx_t_13 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_16, __pyx_t_15) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_15);
            __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1066, __pyx_L119_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_Raise(__pyx_t_13, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __PYX_ERR(0, 1066, __pyx_L119_error)
          }

          /* "fiona/ogrext.pyx":1063
 *                         <OGRwkbGeometryType>geometry_code, options))
 * 
 *             except Exception as exc:             # <<<<<<<<<<<<<<
 *                 OGRReleaseDataSource(self.cogr_ds)
 *                 self.cogr_ds = NULL
 */
          /*finally:*/ {
            __pyx_L119_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22) < 0)) __Pyx_ErrFetch(&__pyx_t_20, &__pyx_t_21, &__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_23);
              __Pyx_XGOTREF(__pyx_t_24);
              __Pyx_XGOTREF(__pyx_t_25);
              __pyx_t_9 = __pyx_lineno; __pyx_t_18 = __pyx_clineno; __pyx_t_38 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_exc);
                __pyx_v_exc = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_23);
                __Pyx_XGIVEREF(__pyx_t_24);
                __Pyx_XGIVEREF(__pyx_t_25);
                __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
              }
              __Pyx_XGIVEREF(__pyx_t_20);
              __Pyx_XGIVEREF(__pyx_t_21);
              __Pyx_XGIVEREF(__pyx_t_22);
              __Pyx_ErrRestore(__pyx_t_20, __pyx_t_21, __pyx_t_22);
              __pyx_t_20 = 0; __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0;
              __pyx_lineno = __pyx_t_9; __pyx_clineno = __pyx_t_18; __pyx_filename = __pyx_t_38;
              goto __pyx_L110_except_error;
            }
          }
        }
        goto __pyx_L110_except_error;
        __pyx_L110_except_error:;

        /* "fiona/ogrext.pyx":1057
 *             geometry_code = geometry_type_code(geometry_type)
 * 
 *             try:             # <<<<<<<<<<<<<<
 *                 self.cogr_layer = exc_wrap_pointer(
 *                     GDALDatasetCreateLayer(
 */
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_7, __pyx_t_6);
        goto __pyx_L106_error;
        __pyx_L113_try_end:;
      }
    }

    /* "fiona/ogrext.pyx":1069
 * 
 *             finally:
 *                 if options != NULL:             # <<<<<<<<<<<<<<
 *                     CSLDestroy(options)
 * 
 */
    /*finally:*/ {
      /*normal exit:*/{
        __pyx_t_5 = ((__pyx_v_options != NULL) != 0);
        if (__pyx_t_5) {

          /* "fiona/ogrext.pyx":1070
 *             finally:
 *                 if options != NULL:
 *                     CSLDestroy(options)             # <<<<<<<<<<<<<<
 * 
 *                 # Shapefile layers make a copy of the passed srs. GPKG
 */
          CSLDestroy(__pyx_v_options);

          /* "fiona/ogrext.pyx":1069
 * 
 *             finally:
 *                 if options != NULL:             # <<<<<<<<<<<<<<
 *                     CSLDestroy(options)
 * 
 */
        }

        /* "fiona/ogrext.pyx":1076
 *                 # count. OSRRelease() is the safe way to release
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:             # <<<<<<<<<<<<<<
 *                     OSRRelease(cogr_srs)
 * 
 */
        __pyx_t_5 = ((__pyx_v_cogr_srs != NULL) != 0);
        if (__pyx_t_5) {

          /* "fiona/ogrext.pyx":1077
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:
 *                     OSRRelease(cogr_srs)             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Created layer %s", collection.name)
 */
          OSRRelease(__pyx_v_cogr_srs);

          /* "fiona/ogrext.pyx":1076
 *                 # count. OSRRelease() is the safe way to release
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:             # <<<<<<<<<<<<<<
 *                     OSRRelease(cogr_srs)
 * 
 */
        }
        goto __pyx_L107;
      }
      __pyx_L106_error:;
      /*exception exit:*/{
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_25, &__pyx_t_24, &__pyx_t_23);
        if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8) < 0)) __Pyx_ErrFetch(&__pyx_t_6, &__pyx_t_7, &__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_6);
        __Pyx_XGOTREF(__pyx_t_7);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_25);
        __Pyx_XGOTREF(__pyx_t_24);
        __Pyx_XGOTREF(__pyx_t_23);
        __pyx_t_18 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_39 = __pyx_filename;
        {

          /* "fiona/ogrext.pyx":1069
 * 
 *             finally:
 *                 if options != NULL:             # <<<<<<<<<<<<<<
 *                     CSLDestroy(options)
 * 
 */
          __pyx_t_5 = ((__pyx_v_options != NULL) != 0);
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1070
 *             finally:
 *                 if options != NULL:
 *                     CSLDestroy(options)             # <<<<<<<<<<<<<<
 * 
 *                 # Shapefile layers make a copy of the passed srs. GPKG
 */
            CSLDestroy(__pyx_v_options);

            /* "fiona/ogrext.pyx":1069
 * 
 *             finally:
 *                 if options != NULL:             # <<<<<<<<<<<<<<
 *                     CSLDestroy(options)
 * 
 */
          }

          /* "fiona/ogrext.pyx":1076
 *                 # count. OSRRelease() is the safe way to release
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:             # <<<<<<<<<<<<<<
 *                     OSRRelease(cogr_srs)
 * 
 */
          __pyx_t_5 = ((__pyx_v_cogr_srs != NULL) != 0);
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1077
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:
 *                     OSRRelease(cogr_srs)             # <<<<<<<<<<<<<<
 * 
 *             log.debug("Created layer %s", collection.name)
 */
            OSRRelease(__pyx_v_cogr_srs);

            /* "fiona/ogrext.pyx":1076
 *                 # count. OSRRelease() is the safe way to release
 *                 # OGRSpatialReferenceH.
 *                 if cogr_srs != NULL:             # <<<<<<<<<<<<<<
 *                     OSRRelease(cogr_srs)
 * 
 */
          }
        }
        if (PY_MAJOR_VERSION >= 3) {
          __Pyx_XGIVEREF(__pyx_t_25);
          __Pyx_XGIVEREF(__pyx_t_24);
          __Pyx_XGIVEREF(__pyx_t_23);
          __Pyx_ExceptionReset(__pyx_t_25, __pyx_t_24, __pyx_t_23);
        }
        __Pyx_XGIVEREF(__pyx_t_6);
        __Pyx_XGIVEREF(__pyx_t_7);
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_ErrRestore(__pyx_t_6, __pyx_t_7, __pyx_t_8);
        __pyx_t_6 = 0; __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_25 = 0; __pyx_t_24 = 0; __pyx_t_23 = 0;
        __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_39;
        goto __pyx_L1_error;
      }
      __pyx_L107:;
    }

    /* "fiona/ogrext.pyx":1079
 *                     OSRRelease(cogr_srs)
 * 
 *             log.debug("Created layer %s", collection.name)             # <<<<<<<<<<<<<<
 * 
 *             # Next, make a layer definition from the given schema properties,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_13 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_13)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_13);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_kp_u_Created_layer_s, __pyx_t_1};
      __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_13, __pyx_kp_u_Created_layer_s, __pyx_t_1};
      __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_4 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_13) {
        __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_13); __pyx_t_13 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_u_Created_layer_s);
      __Pyx_GIVEREF(__pyx_kp_u_Created_layer_s);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_9, __pyx_kp_u_Created_layer_s);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_9, __pyx_t_1);
      __pyx_t_1 = 0;
      __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1079, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "fiona/ogrext.pyx":1084
 *             # which are an ordered dict since Fiona 1.0.1.
 * 
 *             encoding = self._get_internal_encoding()             # <<<<<<<<<<<<<<
 * 
 *             for key, value in collection.schema['properties'].items():
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_14 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1084, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_encoding = __pyx_t_14;
    __pyx_t_14 = 0;

    /* "fiona/ogrext.pyx":1086
 *             encoding = self._get_internal_encoding()
 * 
 *             for key, value in collection.schema['properties'].items():             # <<<<<<<<<<<<<<
 * 
 *                 log.debug("Begin creating field: %r value: %r", key, value)
 */
    __pyx_t_34 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_properties); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_4 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 1086, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_t_4, 0, __pyx_n_s_items, (&__pyx_t_33), (&__pyx_t_9)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_14);
    __pyx_t_14 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_18 = __Pyx_dict_iter_next(__pyx_t_14, __pyx_t_33, &__pyx_t_34, &__pyx_t_3, &__pyx_t_4, NULL, __pyx_t_9);
      if (unlikely(__pyx_t_18 == 0)) break;
      if (unlikely(__pyx_t_18 == -1)) __PYX_ERR(0, 1086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":1088
 *             for key, value in collection.schema['properties'].items():
 * 
 *                 log.debug("Begin creating field: %r value: %r", key, value)             # <<<<<<<<<<<<<<
 * 
 *                 field_subtype = OFSTNone
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1088, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_18 = 0;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_18 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_u_Begin_creating_field_r_value_r, __pyx_v_key, __pyx_v_value};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1088, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[4] = {__pyx_t_3, __pyx_kp_u_Begin_creating_field_r_value_r, __pyx_v_key, __pyx_v_value};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1088, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_GOTREF(__pyx_t_4);
      } else
      #endif
      {
        __pyx_t_13 = PyTuple_New(3+__pyx_t_18); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1088, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (__pyx_t_3) {
          __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_3); __pyx_t_3 = NULL;
        }
        __Pyx_INCREF(__pyx_kp_u_Begin_creating_field_r_value_r);
        __Pyx_GIVEREF(__pyx_kp_u_Begin_creating_field_r_value_r);
        PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_18, __pyx_kp_u_Begin_creating_field_r_value_r);
        __Pyx_INCREF(__pyx_v_key);
        __Pyx_GIVEREF(__pyx_v_key);
        PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_18, __pyx_v_key);
        __Pyx_INCREF(__pyx_v_value);
        __Pyx_GIVEREF(__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_18, __pyx_v_value);
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_13, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1088, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":1090
 *                 log.debug("Begin creating field: %r value: %r", key, value)
 * 
 *                 field_subtype = OFSTNone             # <<<<<<<<<<<<<<
 * 
 *                 # Convert 'long' to 'int'. See
 */
      __pyx_v_field_subtype = OFSTNone;

      /* "fiona/ogrext.pyx":1094
 *                 # Convert 'long' to 'int'. See
 *                 # https://github.com/Toblerity/Fiona/issues/101.
 *                 if fiona.gdal_version.major >= 2 and value in ('int', 'long'):             # <<<<<<<<<<<<<<
 *                     value = 'int64'
 *                 elif value == 'int':
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_fiona); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_gdal_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_major); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyObject_RichCompare(__pyx_t_4, __pyx_int_2, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {
      } else {
        __pyx_t_5 = __pyx_t_2;
        goto __pyx_L134_bool_binop_done;
      }
      __Pyx_INCREF(__pyx_v_value);
      __pyx_t_1 = __pyx_v_value;
      __pyx_t_27 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1094, __pyx_L1_error)
      if (!__pyx_t_27) {
      } else {
        __pyx_t_2 = __pyx_t_27;
        goto __pyx_L136_bool_binop_done;
      }
      __pyx_t_27 = (__Pyx_PyUnicode_Equals(__pyx_t_1, __pyx_n_u_long, Py_EQ)); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1094, __pyx_L1_error)
      __pyx_t_2 = __pyx_t_27;
      __pyx_L136_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_27 = (__pyx_t_2 != 0);
      __pyx_t_5 = __pyx_t_27;
      __pyx_L134_bool_binop_done:;
      if (__pyx_t_5) {

        /* "fiona/ogrext.pyx":1095
 *                 # https://github.com/Toblerity/Fiona/issues/101.
 *                 if fiona.gdal_version.major >= 2 and value in ('int', 'long'):
 *                     value = 'int64'             # <<<<<<<<<<<<<<
 *                 elif value == 'int':
 *                     value = 'int32'
 */
        __Pyx_INCREF(__pyx_n_u_int64);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_n_u_int64);

        /* "fiona/ogrext.pyx":1094
 *                 # Convert 'long' to 'int'. See
 *                 # https://github.com/Toblerity/Fiona/issues/101.
 *                 if fiona.gdal_version.major >= 2 and value in ('int', 'long'):             # <<<<<<<<<<<<<<
 *                     value = 'int64'
 *                 elif value == 'int':
 */
        goto __pyx_L133;
      }

      /* "fiona/ogrext.pyx":1096
 *                 if fiona.gdal_version.major >= 2 and value in ('int', 'long'):
 *                     value = 'int64'
 *                 elif value == 'int':             # <<<<<<<<<<<<<<
 *                     value = 'int32'
 * 
 */
      __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_value, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1096, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "fiona/ogrext.pyx":1097
 *                     value = 'int64'
 *                 elif value == 'int':
 *                     value = 'int32'             # <<<<<<<<<<<<<<
 * 
 *                 if value == 'bool':
 */
        __Pyx_INCREF(__pyx_n_u_int32);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_n_u_int32);

        /* "fiona/ogrext.pyx":1096
 *                 if fiona.gdal_version.major >= 2 and value in ('int', 'long'):
 *                     value = 'int64'
 *                 elif value == 'int':             # <<<<<<<<<<<<<<
 *                     value = 'int32'
 * 
 */
      }
      __pyx_L133:;

      /* "fiona/ogrext.pyx":1099
 *                     value = 'int32'
 * 
 *                 if value == 'bool':             # <<<<<<<<<<<<<<
 *                     value = 'int32'
 *                     field_subtype = OFSTBoolean
 */
      __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_value, __pyx_n_u_bool, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1099, __pyx_L1_error)
      if (__pyx_t_5) {

        /* "fiona/ogrext.pyx":1100
 * 
 *                 if value == 'bool':
 *                     value = 'int32'             # <<<<<<<<<<<<<<
 *                     field_subtype = OFSTBoolean
 * 
 */
        __Pyx_INCREF(__pyx_n_u_int32);
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_n_u_int32);

        /* "fiona/ogrext.pyx":1101
 *                 if value == 'bool':
 *                     value = 'int32'
 *                     field_subtype = OFSTBoolean             # <<<<<<<<<<<<<<
 * 
 *                 # Is there a field width/precision?
 */
        __pyx_v_field_subtype = OFSTBoolean;

        /* "fiona/ogrext.pyx":1099
 *                     value = 'int32'
 * 
 *                 if value == 'bool':             # <<<<<<<<<<<<<<
 *                     value = 'int32'
 *                     field_subtype = OFSTBoolean
 */
      }

      /* "fiona/ogrext.pyx":1104
 * 
 *                 # Is there a field width/precision?
 *                 width = precision = None             # <<<<<<<<<<<<<<
 *                 if ':' in value:
 *                     value, fmt = value.split(':')
 */
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_width, Py_None);
      __Pyx_INCREF(Py_None);
      __Pyx_XDECREF_SET(__pyx_v_precision, Py_None);

      /* "fiona/ogrext.pyx":1105
 *                 # Is there a field width/precision?
 *                 width = precision = None
 *                 if ':' in value:             # <<<<<<<<<<<<<<
 *                     value, fmt = value.split(':')
 * 
 */
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_kp_u__21, __pyx_v_value, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1105, __pyx_L1_error)
      __pyx_t_27 = (__pyx_t_5 != 0);
      if (__pyx_t_27) {

        /* "fiona/ogrext.pyx":1106
 *                 width = precision = None
 *                 if ':' in value:
 *                     value, fmt = value.split(':')             # <<<<<<<<<<<<<<
 * 
 *                     log.debug("Field format parsing, value: %r, fmt: %r", value, fmt)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_split); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_13 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_1 = (__pyx_t_13) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_13, __pyx_kp_u__21) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u__21);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
          PyObject* sequence = __pyx_t_1;
          Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
          if (unlikely(size != 2)) {
            if (size > 2) __Pyx_RaiseTooManyValuesError(2);
            else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
            __PYX_ERR(0, 1106, __pyx_L1_error)
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          if (likely(PyTuple_CheckExact(sequence))) {
            __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
            __pyx_t_13 = PyTuple_GET_ITEM(sequence, 1); 
          } else {
            __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
            __pyx_t_13 = PyList_GET_ITEM(sequence, 1); 
          }
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_13);
          #else
          __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1106, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_13 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1106, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          #endif
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        } else {
          Py_ssize_t index = -1;
          __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1106, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_40 = Py_TYPE(__pyx_t_3)->tp_iternext;
          index = 0; __pyx_t_4 = __pyx_t_40(__pyx_t_3); if (unlikely(!__pyx_t_4)) goto __pyx_L140_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_4);
          index = 1; __pyx_t_13 = __pyx_t_40(__pyx_t_3); if (unlikely(!__pyx_t_13)) goto __pyx_L140_unpacking_failed;
          __Pyx_GOTREF(__pyx_t_13);
          if (__Pyx_IternextUnpackEndCheck(__pyx_t_40(__pyx_t_3), 2) < 0) __PYX_ERR(0, 1106, __pyx_L1_error)
          __pyx_t_40 = NULL;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          goto __pyx_L141_unpacking_done;
          __pyx_L140_unpacking_failed:;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_40 = NULL;
          if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
          __PYX_ERR(0, 1106, __pyx_L1_error)
          __pyx_L141_unpacking_done:;
        }
        __Pyx_DECREF_SET(__pyx_v_value, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_XDECREF_SET(__pyx_v_fmt, __pyx_t_13);
        __pyx_t_13 = 0;

        /* "fiona/ogrext.pyx":1108
 *                     value, fmt = value.split(':')
 * 
 *                     log.debug("Field format parsing, value: %r, fmt: %r", value, fmt)             # <<<<<<<<<<<<<<
 * 
 *                     if '.' in fmt:
 */
        __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_log); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1108, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_13 = NULL;
        __pyx_t_18 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_13)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_13);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
            __pyx_t_18 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_kp_u_Field_format_parsing_value_r_fmt, __pyx_v_value, __pyx_v_fmt};
          __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
          PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_kp_u_Field_format_parsing_value_r_fmt, __pyx_v_value, __pyx_v_fmt};
          __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_18, 3+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_GOTREF(__pyx_t_1);
        } else
        #endif
        {
          __pyx_t_3 = PyTuple_New(3+__pyx_t_18); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          if (__pyx_t_13) {
            __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_13); __pyx_t_13 = NULL;
          }
          __Pyx_INCREF(__pyx_kp_u_Field_format_parsing_value_r_fmt);
          __Pyx_GIVEREF(__pyx_kp_u_Field_format_parsing_value_r_fmt);
          PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_18, __pyx_kp_u_Field_format_parsing_value_r_fmt);
          __Pyx_INCREF(__pyx_v_value);
          __Pyx_GIVEREF(__pyx_v_value);
          PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_18, __pyx_v_value);
          __Pyx_INCREF(__pyx_v_fmt);
          __Pyx_GIVEREF(__pyx_v_fmt);
          PyTuple_SET_ITEM(__pyx_t_3, 2+__pyx_t_18, __pyx_v_fmt);
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "fiona/ogrext.pyx":1110
 *                     log.debug("Field format parsing, value: %r, fmt: %r", value, fmt)
 * 
 *                     if '.' in fmt:             # <<<<<<<<<<<<<<
 *                         width, precision = map(int, fmt.split('.'))
 *                     else:
 */
        __pyx_t_27 = (__Pyx_PySequence_ContainsTF(__pyx_kp_u__22, __pyx_v_fmt, Py_EQ)); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1110, __pyx_L1_error)
        __pyx_t_5 = (__pyx_t_27 != 0);
        if (__pyx_t_5) {

          /* "fiona/ogrext.pyx":1111
 * 
 *                     if '.' in fmt:
 *                         width, precision = map(int, fmt.split('.'))             # <<<<<<<<<<<<<<
 *                     else:
 *                         width = int(fmt)
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_fmt, __pyx_n_s_split); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_kp_u__22) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u__22);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_INCREF(((PyObject *)(&PyInt_Type)));
          __Pyx_GIVEREF(((PyObject *)(&PyInt_Type)));
          PyTuple_SET_ITEM(__pyx_t_4, 0, ((PyObject *)(&PyInt_Type)));
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_map, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
            PyObject* sequence = __pyx_t_1;
            Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
            if (unlikely(size != 2)) {
              if (size > 2) __Pyx_RaiseTooManyValuesError(2);
              else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
              __PYX_ERR(0, 1111, __pyx_L1_error)
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            if (likely(PyTuple_CheckExact(sequence))) {
              __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
            } else {
              __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
              __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
            }
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            #else
            __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            #endif
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          } else {
            Py_ssize_t index = -1;
            __pyx_t_13 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __pyx_t_40 = Py_TYPE(__pyx_t_13)->tp_iternext;
            index = 0; __pyx_t_4 = __pyx_t_40(__pyx_t_13); if (unlikely(!__pyx_t_4)) goto __pyx_L143_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_4);
            index = 1; __pyx_t_3 = __pyx_t_40(__pyx_t_13); if (unlikely(!__pyx_t_3)) goto __pyx_L143_unpacking_failed;
            __Pyx_GOTREF(__pyx_t_3);
            if (__Pyx_IternextUnpackEndCheck(__pyx_t_40(__pyx_t_13), 2) < 0) __PYX_ERR(0, 1111, __pyx_L1_error)
            __pyx_t_40 = NULL;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            goto __pyx_L144_unpacking_done;
            __pyx_L143_unpacking_failed:;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            __pyx_t_40 = NULL;
            if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
            __PYX_ERR(0, 1111, __pyx_L1_error)
            __pyx_L144_unpacking_done:;
          }
          __Pyx_DECREF_SET(__pyx_v_width, __pyx_t_4);
          __pyx_t_4 = 0;
          __Pyx_DECREF_SET(__pyx_v_precision, __pyx_t_3);
          __pyx_t_3 = 0;

          /* "fiona/ogrext.pyx":1110
 *                     log.debug("Field format parsing, value: %r, fmt: %r", value, fmt)
 * 
 *                     if '.' in fmt:             # <<<<<<<<<<<<<<
 *                         width, precision = map(int, fmt.split('.'))
 *                     else:
 */
          goto __pyx_L142;
        }

        /* "fiona/ogrext.pyx":1113
 *                         width, precision = map(int, fmt.split('.'))
 *                     else:
 *                         width = int(fmt)             # <<<<<<<<<<<<<<
 * 
 *                     if value == 'int':
 */
        /*else*/ {
          __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_v_fmt); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF_SET(__pyx_v_width, __pyx_t_1);
          __pyx_t_1 = 0;
        }
        __pyx_L142:;

        /* "fiona/ogrext.pyx":1115
 *                         width = int(fmt)
 * 
 *                     if value == 'int':             # <<<<<<<<<<<<<<
 *                         if GDAL_VERSION_NUM >= 2000000 and (width == 0 or width >= 10):
 *                             value = 'int64'
 */
        __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_v_value, __pyx_n_u_int, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1115, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "fiona/ogrext.pyx":1116
 * 
 *                     if value == 'int':
 *                         if GDAL_VERSION_NUM >= 2000000 and (width == 0 or width >= 10):             # <<<<<<<<<<<<<<
 *                             value = 'int64'
 *                         else:
 */
          __pyx_t_27 = ((__pyx_v_5fiona_6ogrext_GDAL_VERSION_NUM >= 0x1E8480) != 0);
          if (__pyx_t_27) {
          } else {
            __pyx_t_5 = __pyx_t_27;
            goto __pyx_L147_bool_binop_done;
          }
          __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_width, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1116, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (!__pyx_t_27) {
          } else {
            __pyx_t_5 = __pyx_t_27;
            goto __pyx_L147_bool_binop_done;
          }
          __pyx_t_1 = PyObject_RichCompare(__pyx_v_width, __pyx_int_10, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1116, __pyx_L1_error)
          __pyx_t_27 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_27 < 0)) __PYX_ERR(0, 1116, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_5 = __pyx_t_27;
          __pyx_L147_bool_binop_done:;
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1117
 *                     if value == 'int':
 *                         if GDAL_VERSION_NUM >= 2000000 and (width == 0 or width >= 10):
 *                             value = 'int64'             # <<<<<<<<<<<<<<
 *                         else:
 *                             value = 'int32'
 */
            __Pyx_INCREF(__pyx_n_u_int64);
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_n_u_int64);

            /* "fiona/ogrext.pyx":1116
 * 
 *                     if value == 'int':
 *                         if GDAL_VERSION_NUM >= 2000000 and (width == 0 or width >= 10):             # <<<<<<<<<<<<<<
 *                             value = 'int64'
 *                         else:
 */
            goto __pyx_L146;
          }

          /* "fiona/ogrext.pyx":1119
 *                             value = 'int64'
 *                         else:
 *                             value = 'int32'             # <<<<<<<<<<<<<<
 * 
 *                 field_type = FIELD_TYPES.index(value)
 */
          /*else*/ {
            __Pyx_INCREF(__pyx_n_u_int32);
            __Pyx_DECREF_SET(__pyx_v_value, __pyx_n_u_int32);
          }
          __pyx_L146:;

          /* "fiona/ogrext.pyx":1115
 *                         width = int(fmt)
 * 
 *                     if value == 'int':             # <<<<<<<<<<<<<<
 *                         if GDAL_VERSION_NUM >= 2000000 and (width == 0 or width >= 10):
 *                             value = 'int64'
 */
        }

        /* "fiona/ogrext.pyx":1105
 *                 # Is there a field width/precision?
 *                 width = precision = None
 *                 if ':' in value:             # <<<<<<<<<<<<<<
 *                     value, fmt = value.split(':')
 * 
 */
      }

      /* "fiona/ogrext.pyx":1121
 *                             value = 'int32'
 * 
 *                 field_type = FIELD_TYPES.index(value)             # <<<<<<<<<<<<<<
 * 
 *                 try:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_FIELD_TYPES); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_value) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_value);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1121, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF_SET(__pyx_v_field_type, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":1123
 *                 field_type = FIELD_TYPES.index(value)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_23, &__pyx_t_24, &__pyx_t_25);
        __Pyx_XGOTREF(__pyx_t_23);
        __Pyx_XGOTREF(__pyx_t_24);
        __Pyx_XGOTREF(__pyx_t_25);
        /*try:*/ {

          /* "fiona/ogrext.pyx":1124
 * 
 *                 try:
 *                     key_bytes = key.encode(encoding)             # <<<<<<<<<<<<<<
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 *                     if width:
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_key, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1124, __pyx_L150_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = NULL;
          if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
            }
          }
          __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_v_encoding) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_encoding);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1124, __pyx_L150_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_XDECREF_SET(__pyx_v_key_bytes, __pyx_t_1);
          __pyx_t_1 = 0;

          /* "fiona/ogrext.pyx":1125
 *                 try:
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))             # <<<<<<<<<<<<<<
 *                     if width:
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 */
          __pyx_t_41 = __Pyx_PyObject_AsWritableString(__pyx_v_key_bytes); if (unlikely((!__pyx_t_41) && PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L150_error)
          __pyx_t_42 = ((OGRFieldType)__Pyx_PyInt_As_OGRFieldType(__pyx_v_field_type)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L150_error)
          __pyx_t_11 = __pyx_f_5fiona_4_err_exc_wrap_pointer(OGR_Fld_Create(__pyx_t_41, ((OGRFieldType)__pyx_t_42))); if (unlikely(__pyx_t_11 == ((void *)NULL))) __PYX_ERR(0, 1125, __pyx_L150_error)
          __pyx_v_cogr_fielddefn = __pyx_t_11;

          /* "fiona/ogrext.pyx":1126
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 *                     if width:             # <<<<<<<<<<<<<<
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 *                     if precision:
 */
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_width); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1126, __pyx_L150_error)
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1127
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 *                     if width:
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)             # <<<<<<<<<<<<<<
 *                     if precision:
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)
 */
            __pyx_t_18 = __Pyx_PyInt_As_int(__pyx_v_width); if (unlikely((__pyx_t_18 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1127, __pyx_L150_error)
            OGR_Fld_SetWidth(__pyx_v_cogr_fielddefn, __pyx_t_18);

            /* "fiona/ogrext.pyx":1126
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 *                     if width:             # <<<<<<<<<<<<<<
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 *                     if precision:
 */
          }

          /* "fiona/ogrext.pyx":1128
 *                     if width:
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 *                     if precision:             # <<<<<<<<<<<<<<
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)
 *                     if field_subtype != OFSTNone:
 */
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_precision); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1128, __pyx_L150_error)
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1129
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 *                     if precision:
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)             # <<<<<<<<<<<<<<
 *                     if field_subtype != OFSTNone:
 *                         # subtypes are new in GDAL 2.x, ignored in 1.x
 */
            __pyx_t_18 = __Pyx_PyInt_As_int(__pyx_v_precision); if (unlikely((__pyx_t_18 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1129, __pyx_L150_error)
            OGR_Fld_SetPrecision(__pyx_v_cogr_fielddefn, __pyx_t_18);

            /* "fiona/ogrext.pyx":1128
 *                     if width:
 *                         OGR_Fld_SetWidth(cogr_fielddefn, width)
 *                     if precision:             # <<<<<<<<<<<<<<
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)
 *                     if field_subtype != OFSTNone:
 */
          }

          /* "fiona/ogrext.pyx":1130
 *                     if precision:
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)
 *                     if field_subtype != OFSTNone:             # <<<<<<<<<<<<<<
 *                         # subtypes are new in GDAL 2.x, ignored in 1.x
 *                         set_field_subtype(cogr_fielddefn, field_subtype)
 */
          __pyx_t_5 = ((__pyx_v_field_subtype != OFSTNone) != 0);
          if (__pyx_t_5) {

            /* "fiona/ogrext.pyx":1132
 *                     if field_subtype != OFSTNone:
 *                         # subtypes are new in GDAL 2.x, ignored in 1.x
 *                         set_field_subtype(cogr_fielddefn, field_subtype)             # <<<<<<<<<<<<<<
 *                     exc_wrap_int(OGR_L_CreateField(self.cogr_layer, cogr_fielddefn, 1))
 * 
 */
            __pyx_f_5fiona_5_shim_set_field_subtype(__pyx_v_cogr_fielddefn, __pyx_v_field_subtype);

            /* "fiona/ogrext.pyx":1130
 *                     if precision:
 *                         OGR_Fld_SetPrecision(cogr_fielddefn, precision)
 *                     if field_subtype != OFSTNone:             # <<<<<<<<<<<<<<
 *                         # subtypes are new in GDAL 2.x, ignored in 1.x
 *                         set_field_subtype(cogr_fielddefn, field_subtype)
 */
          }

          /* "fiona/ogrext.pyx":1133
 *                         # subtypes are new in GDAL 2.x, ignored in 1.x
 *                         set_field_subtype(cogr_fielddefn, field_subtype)
 *                     exc_wrap_int(OGR_L_CreateField(self.cogr_layer, cogr_fielddefn, 1))             # <<<<<<<<<<<<<<
 * 
 *                 except (UnicodeEncodeError, CPLE_BaseError) as exc:
 */
          __pyx_t_18 = __pyx_f_5fiona_4_err_exc_wrap_int(OGR_L_CreateField(__pyx_v_self->__pyx_base.cogr_layer, __pyx_v_cogr_fielddefn, 1)); if (unlikely(__pyx_t_18 == ((int)-1))) __PYX_ERR(0, 1133, __pyx_L150_error)

          /* "fiona/ogrext.pyx":1123
 *                 field_type = FIELD_TYPES.index(value)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 */
        }

        /* "fiona/ogrext.pyx":1142
 * 
 *                 else:
 *                     OGR_Fld_Destroy(cogr_fielddefn)             # <<<<<<<<<<<<<<
 *                     log.debug("End creating field %r", key)
 * 
 */
        /*else:*/ {
          OGR_Fld_Destroy(__pyx_v_cogr_fielddefn);

          /* "fiona/ogrext.pyx":1143
 *                 else:
 *                     OGR_Fld_Destroy(cogr_fielddefn)
 *                     log.debug("End creating field %r", key)             # <<<<<<<<<<<<<<
 * 
 *         # Mapping of the Python collection schema to the munged
 */
          __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __pyx_t_4 = NULL;
          __pyx_t_18 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_4)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_4);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_18 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_End_creating_field_r, __pyx_v_key};
            __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_18, 2+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
            PyObject *__pyx_temp[3] = {__pyx_t_4, __pyx_kp_u_End_creating_field_r, __pyx_v_key};
            __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_18, 2+__pyx_t_18); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
            __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_GOTREF(__pyx_t_1);
          } else
          #endif
          {
            __pyx_t_13 = PyTuple_New(2+__pyx_t_18); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
            __Pyx_GOTREF(__pyx_t_13);
            if (__pyx_t_4) {
              __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_4); __pyx_t_4 = NULL;
            }
            __Pyx_INCREF(__pyx_kp_u_End_creating_field_r);
            __Pyx_GIVEREF(__pyx_kp_u_End_creating_field_r);
            PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_18, __pyx_kp_u_End_creating_field_r);
            __Pyx_INCREF(__pyx_v_key);
            __Pyx_GIVEREF(__pyx_v_key);
            PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_18, __pyx_v_key);
            __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1143, __pyx_L152_except_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_XDECREF(__pyx_t_23); __pyx_t_23 = 0;
        __Pyx_XDECREF(__pyx_t_24); __pyx_t_24 = 0;
        __Pyx_XDECREF(__pyx_t_25); __pyx_t_25 = 0;
        goto __pyx_L157_try_end;
        __pyx_L150_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fiona/ogrext.pyx":1135
 *                     exc_wrap_int(OGR_L_CreateField(self.cogr_layer, cogr_fielddefn, 1))
 * 
 *                 except (UnicodeEncodeError, CPLE_BaseError) as exc:             # <<<<<<<<<<<<<<
 *                     OGRReleaseDataSource(self.cogr_ds)
 *                     self.cogr_ds = NULL
 */
        __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_3, &__pyx_t_13);
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_CPLE_BaseError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1135, __pyx_L152_except_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_18 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_builtin_UnicodeEncodeError) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_ErrRestore(__pyx_t_1, __pyx_t_3, __pyx_t_13);
        __pyx_t_1 = 0; __pyx_t_3 = 0; __pyx_t_13 = 0;
        if (__pyx_t_18) {
          __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_13, &__pyx_t_3, &__pyx_t_1) < 0) __PYX_ERR(0, 1135, __pyx_L152_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __pyx_v_exc = __pyx_t_3;
          /*try:*/ {

            /* "fiona/ogrext.pyx":1136
 * 
 *                 except (UnicodeEncodeError, CPLE_BaseError) as exc:
 *                     OGRReleaseDataSource(self.cogr_ds)             # <<<<<<<<<<<<<<
 *                     self.cogr_ds = NULL
 *                     self.cogr_layer = NULL
 */
            (void)(OGRReleaseDataSource(__pyx_v_self->__pyx_base.cogr_ds));

            /* "fiona/ogrext.pyx":1137
 *                 except (UnicodeEncodeError, CPLE_BaseError) as exc:
 *                     OGRReleaseDataSource(self.cogr_ds)
 *                     self.cogr_ds = NULL             # <<<<<<<<<<<<<<
 *                     self.cogr_layer = NULL
 *                     raise SchemaError(u"{}".format(exc))
 */
            __pyx_v_self->__pyx_base.cogr_ds = NULL;

            /* "fiona/ogrext.pyx":1138
 *                     OGRReleaseDataSource(self.cogr_ds)
 *                     self.cogr_ds = NULL
 *                     self.cogr_layer = NULL             # <<<<<<<<<<<<<<
 *                     raise SchemaError(u"{}".format(exc))
 * 
 */
            __pyx_v_self->__pyx_base.cogr_layer = NULL;

            /* "fiona/ogrext.pyx":1139
 *                     self.cogr_ds = NULL
 *                     self.cogr_layer = NULL
 *                     raise SchemaError(u"{}".format(exc))             # <<<<<<<<<<<<<<
 * 
 *                 else:
 */
            __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1139, __pyx_L166_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__13, __pyx_n_s_format); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 1139, __pyx_L166_error)
            __Pyx_GOTREF(__pyx_t_17);
            __pyx_t_43 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_17))) {
              __pyx_t_43 = PyMethod_GET_SELF(__pyx_t_17);
              if (likely(__pyx_t_43)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
                __Pyx_INCREF(__pyx_t_43);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_17, function);
              }
            }
            __pyx_t_16 = (__pyx_t_43) ? __Pyx_PyObject_Call2Args(__pyx_t_17, __pyx_t_43, __pyx_v_exc) : __Pyx_PyObject_CallOneArg(__pyx_t_17, __pyx_v_exc);
            __Pyx_XDECREF(__pyx_t_43); __pyx_t_43 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1139, __pyx_L166_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
            __pyx_t_17 = NULL;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_15))) {
              __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_15);
              if (likely(__pyx_t_17)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
                __Pyx_INCREF(__pyx_t_17);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_15, function);
              }
            }
            __pyx_t_4 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_17, __pyx_t_16) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_t_16);
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1139, __pyx_L166_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            __Pyx_Raise(__pyx_t_4, 0, 0, 0);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __PYX_ERR(0, 1139, __pyx_L166_error)
          }

          /* "fiona/ogrext.pyx":1135
 *                     exc_wrap_int(OGR_L_CreateField(self.cogr_layer, cogr_fielddefn, 1))
 * 
 *                 except (UnicodeEncodeError, CPLE_BaseError) as exc:             # <<<<<<<<<<<<<<
 *                     OGRReleaseDataSource(self.cogr_ds)
 *                     self.cogr_ds = NULL
 */
          /*finally:*/ {
            __pyx_L166_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
              __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              __Pyx_XDECREF(__pyx_t_43); __pyx_t_43 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_22, &__pyx_t_21, &__pyx_t_20);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6) < 0)) __Pyx_ErrFetch(&__pyx_t_8, &__pyx_t_7, &__pyx_t_6);
              __Pyx_XGOTREF(__pyx_t_8);
              __Pyx_XGOTREF(__pyx_t_7);
              __Pyx_XGOTREF(__pyx_t_6);
              __Pyx_XGOTREF(__pyx_t_22);
              __Pyx_XGOTREF(__pyx_t_21);
              __Pyx_XGOTREF(__pyx_t_20);
              __pyx_t_18 = __pyx_lineno; __pyx_t_31 = __pyx_clineno; __pyx_t_44 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_exc);
                __pyx_v_exc = NULL;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_22);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_ExceptionReset(__pyx_t_22, __pyx_t_21, __pyx_t_20);
              }
              __Pyx_XGIVEREF(__pyx_t_8);
              __Pyx_XGIVEREF(__pyx_t_7);
              __Pyx_XGIVEREF(__pyx_t_6);
              __Pyx_ErrRestore(__pyx_t_8, __pyx_t_7, __pyx_t_6);
              __pyx_t_8 = 0; __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_22 = 0; __pyx_t_21 = 0; __pyx_t_20 = 0;
              __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_31; __pyx_filename = __pyx_t_44;
              goto __pyx_L152_except_error;
            }
          }
        }
        goto __pyx_L152_except_error;
        __pyx_L152_except_error:;

        /* "fiona/ogrext.pyx":1123
 *                 field_type = FIELD_TYPES.index(value)
 * 
 *                 try:             # <<<<<<<<<<<<<<
 *                     key_bytes = key.encode(encoding)
 *                     cogr_fielddefn = exc_wrap_pointer(OGR_Fld_Create(key_bytes, <OGRFieldType>field_type))
 */
        __Pyx_XGIVEREF(__pyx_t_23);
        __Pyx_XGIVEREF(__pyx_t_24);
        __Pyx_XGIVEREF(__pyx_t_25);
        __Pyx_ExceptionReset(__pyx_t_23, __pyx_t_24, __pyx_t_25);
        goto __pyx_L1_error;
        __pyx_L157_try_end:;
      }
    }
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "fiona/ogrext.pyx":925
 *                 self._fileencoding = userencoding or self._get_fallback_encoding()
 * 
 *         elif collection.mode == 'w':             # <<<<<<<<<<<<<<
 * 
 *             try:
 */
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1147
 *         # Mapping of the Python collection schema to the munged
 *         # OGR schema.
 *         ogr_schema = self.get_schema()             # <<<<<<<<<<<<<<
 *         self._schema_mapping = dict(zip(
 *             collection.schema['properties'].keys(),
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_ogr_schema = __pyx_t_14;
  __pyx_t_14 = 0;

  /* "fiona/ogrext.pyx":1149
 *         ogr_schema = self.get_schema()
 *         self._schema_mapping = dict(zip(
 *             collection.schema['properties'].keys(),             # <<<<<<<<<<<<<<
 *             ogr_schema['properties'].keys() ))
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_n_u_properties); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_keys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  __pyx_t_14 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1150
 *         self._schema_mapping = dict(zip(
 *             collection.schema['properties'].keys(),
 *             ogr_schema['properties'].keys() ))             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Writing started")
 */
  __pyx_t_3 = __Pyx_PyObject_Dict_GetItem(__pyx_v_ogr_schema, __pyx_n_u_properties); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_keys); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_13))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_13, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_13, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "fiona/ogrext.pyx":1148
 *         # OGR schema.
 *         ogr_schema = self.get_schema()
 *         self._schema_mapping = dict(zip(             # <<<<<<<<<<<<<<
 *             collection.schema['properties'].keys(),
 *             ogr_schema['properties'].keys() ))
 */
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_t_1);
  __pyx_t_14 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyDict_Type)), __pyx_t_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_13);
  __Pyx_GOTREF(__pyx_v_self->_schema_mapping);
  __Pyx_DECREF(__pyx_v_self->_schema_mapping);
  __pyx_v_self->_schema_mapping = __pyx_t_13;
  __pyx_t_13 = 0;

  /* "fiona/ogrext.pyx":1152
 *             ogr_schema['properties'].keys() ))
 * 
 *         log.debug("Writing started")             # <<<<<<<<<<<<<<
 * 
 *     def writerecs(self, records, collection):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
    }
  }
  __pyx_t_13 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_14, __pyx_t_1, __pyx_kp_u_Writing_started) : __Pyx_PyObject_CallOneArg(__pyx_t_14, __pyx_kp_u_Writing_started);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;

  /* "fiona/ogrext.pyx":879
 *     cdef object _schema_mapping
 * 
 *     def start(self, collection, **kwargs):             # <<<<<<<<<<<<<<
 *         cdef OGRSpatialReferenceH cogr_srs = NULL
 *         cdef char **options = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_43);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path);
  __Pyx_XDECREF(__pyx_v_userencoding);
  __Pyx_XDECREF(__pyx_v_path_b);
  __Pyx_XDECREF(__pyx_v_name_b);
  __Pyx_XDECREF(__pyx_v_exc);
  __Pyx_XDECREF(__pyx_v_driver_b);
  __Pyx_XDECREF(__pyx_v_col_crs);
  __Pyx_XDECREF(__pyx_v_proj_b);
  __Pyx_XDECREF(__pyx_v_sysencoding);
  __Pyx_XDECREF(__pyx_v_fileencoding_b);
  __Pyx_XDECREF(__pyx_v_layer_names);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_k);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XDECREF(__pyx_v_geometry_type);
  __Pyx_XDECREF(__pyx_v_geometry_types);
  __Pyx_XDECREF(__pyx_v_encoding);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_width);
  __Pyx_XDECREF(__pyx_v_precision);
  __Pyx_XDECREF(__pyx_v_fmt);
  __Pyx_XDECREF(__pyx_v_field_type);
  __Pyx_XDECREF(__pyx_v_key_bytes);
  __Pyx_XDECREF(__pyx_v_ogr_schema);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1154
 *         log.debug("Writing started")
 * 
 *     def writerecs(self, records, collection):             # <<<<<<<<<<<<<<
 *         """Writes buffered records to OGR."""
 *         cdef void *cogr_driver
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_3writerecs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14WritingSession_2writerecs[] = "Writes buffered records to OGR.";
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_3writerecs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_records = 0;
  PyObject *__pyx_v_collection = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("writerecs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_records,&__pyx_n_s_collection,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_records)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collection)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("writerecs", 1, 2, 2, 1); __PYX_ERR(0, 1154, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "writerecs") < 0)) __PYX_ERR(0, 1154, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_records = values[0];
    __pyx_v_collection = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("writerecs", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1154, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.writerecs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_2writerecs(((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_v_self), __pyx_v_records, __pyx_v_collection);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1169
 * 
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *             if record["geometry"] is None:
 *                 return True
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_9writerecs_1validate_geometry_type(PyObject *__pyx_self, PyObject *__pyx_v_record); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_14WritingSession_9writerecs_1validate_geometry_type = {"validate_geometry_type", (PyCFunction)__pyx_pw_5fiona_6ogrext_14WritingSession_9writerecs_1validate_geometry_type, METH_O, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_9writerecs_1validate_geometry_type(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("validate_geometry_type (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_9writerecs_validate_geometry_type(__pyx_self, ((PyObject *)__pyx_v_record));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_9writerecs_validate_geometry_type(PyObject *__pyx_self, PyObject *__pyx_v_record) {
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *__pyx_cur_scope;
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("validate_geometry_type", 0);
  __pyx_outer_scope = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;

  /* "fiona/ogrext.pyx":1170
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):
 *             if record["geometry"] is None:             # <<<<<<<<<<<<<<
 *                 return True
 *             return record["geometry"]["type"].lstrip("3D ") in valid_geom_types
 */
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_v_record, __pyx_n_u_geometry); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1171
 *         def validate_geometry_type(record):
 *             if record["geometry"] is None:
 *                 return True             # <<<<<<<<<<<<<<
 *             return record["geometry"]["type"].lstrip("3D ") in valid_geom_types
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":1170
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):
 *             if record["geometry"] is None:             # <<<<<<<<<<<<<<
 *                 return True
 *             return record["geometry"]["type"].lstrip("3D ") in valid_geom_types
 */
  }

  /* "fiona/ogrext.pyx":1172
 *             if record["geometry"] is None:
 *                 return True
 *             return record["geometry"]["type"].lstrip("3D ") in valid_geom_types             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Starting transaction (initial)")
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_v_record, __pyx_n_u_geometry); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_3D) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_3D);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_cur_scope->__pyx_v_valid_geom_types)) { __Pyx_RaiseClosureNameError("valid_geom_types"); __PYX_ERR(0, 1172, __pyx_L1_error) }
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_cur_scope->__pyx_v_valid_geom_types, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1169
 * 
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *             if record["geometry"] is None:
 *                 return True
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.writerecs.validate_geometry_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1154
 *         log.debug("Writing started")
 * 
 *     def writerecs(self, records, collection):             # <<<<<<<<<<<<<<
 *         """Writes buffered records to OGR."""
 *         cdef void *cogr_driver
 */

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_2writerecs(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, PyObject *__pyx_v_records, PyObject *__pyx_v_collection) {
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *__pyx_cur_scope;
  void *__pyx_v_cogr_driver;
  void *__pyx_v_cogr_feature;
  int __pyx_v_features_in_transaction;
  void *__pyx_v_cogr_layer;
  CYTHON_UNUSED PyObject *__pyx_v_schema_geom_type = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_driver_name = NULL;
  PyObject *__pyx_v_validate_geometry_type = 0;
  OGRErr __pyx_v_result;
  PyObject *__pyx_v_schema_props_keys = NULL;
  PyObject *__pyx_v_record = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  char const *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("writerecs", 0);
  __pyx_cur_scope = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct_1_writerecs(__pyx_ptype_5fiona_6ogrext___pyx_scope_struct_1_writerecs, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1154, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "fiona/ogrext.pyx":1158
 *         cdef void *cogr_driver
 *         cdef void *cogr_feature
 *         cdef int features_in_transaction = 0             # <<<<<<<<<<<<<<
 * 
 *         cdef void *cogr_layer = self.cogr_layer
 */
  __pyx_v_features_in_transaction = 0;

  /* "fiona/ogrext.pyx":1160
 *         cdef int features_in_transaction = 0
 * 
 *         cdef void *cogr_layer = self.cogr_layer             # <<<<<<<<<<<<<<
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.cogr_layer;
  __pyx_v_cogr_layer = __pyx_t_1;

  /* "fiona/ogrext.pyx":1161
 * 
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  __pyx_t_2 = ((__pyx_v_cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1162
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 * 
 *         schema_geom_type = collection.schema['geometry']
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1162, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1161
 * 
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1164
 *             raise ValueError("Null layer")
 * 
 *         schema_geom_type = collection.schema['geometry']             # <<<<<<<<<<<<<<
 *         cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         driver_name = OGR_Dr_GetName(cogr_driver).decode("utf-8")
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_geometry); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_schema_geom_type = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1165
 * 
 *         schema_geom_type = collection.schema['geometry']
 *         cogr_driver = GDALGetDatasetDriver(self.cogr_ds)             # <<<<<<<<<<<<<<
 *         driver_name = OGR_Dr_GetName(cogr_driver).decode("utf-8")
 * 
 */
  __pyx_v_cogr_driver = GDALGetDatasetDriver(__pyx_v_self->__pyx_base.cogr_ds);

  /* "fiona/ogrext.pyx":1166
 *         schema_geom_type = collection.schema['geometry']
 *         cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         driver_name = OGR_Dr_GetName(cogr_driver).decode("utf-8")             # <<<<<<<<<<<<<<
 * 
 *         valid_geom_types = collection._valid_geom_types
 */
  __pyx_t_5 = OGR_Dr_GetName(__pyx_v_cogr_driver);
  __pyx_t_4 = __Pyx_decode_c_string(__pyx_t_5, 0, strlen(__pyx_t_5), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_t_4);
  __pyx_v_driver_name = __pyx_t_4;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1168
 *         driver_name = OGR_Dr_GetName(cogr_driver).decode("utf-8")
 * 
 *         valid_geom_types = collection._valid_geom_types             # <<<<<<<<<<<<<<
 *         def validate_geometry_type(record):
 *             if record["geometry"] is None:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_valid_geom_types); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_valid_geom_types = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1169
 * 
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *             if record["geometry"] is None:
 *                 return True
 */
  __pyx_t_4 = __Pyx_CyFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_14WritingSession_9writerecs_1validate_geometry_type, 0, __pyx_n_s_writerecs_locals_validate_geomet, ((PyObject*)__pyx_cur_scope), __pyx_n_s_fiona_ogrext, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_validate_geometry_type = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1174
 *             return record["geometry"]["type"].lstrip("3D ") in valid_geom_types
 * 
 *         log.debug("Starting transaction (initial)")             # <<<<<<<<<<<<<<
 *         result = gdal_start_transaction(self.cogr_ds, 0)
 *         if result == OGRERR_FAILURE:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_log); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_4 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_3, __pyx_kp_u_Starting_transaction_initial) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_Starting_transaction_initial);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1175
 * 
 *         log.debug("Starting transaction (initial)")
 *         result = gdal_start_transaction(self.cogr_ds, 0)             # <<<<<<<<<<<<<<
 *         if result == OGRERR_FAILURE:
 *             raise TransactionError("Failed to start transaction")
 */
  __pyx_v_result = __pyx_f_5fiona_5_shim_gdal_start_transaction(__pyx_v_self->__pyx_base.cogr_ds, 0);

  /* "fiona/ogrext.pyx":1176
 *         log.debug("Starting transaction (initial)")
 *         result = gdal_start_transaction(self.cogr_ds, 0)
 *         if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *             raise TransactionError("Failed to start transaction")
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_From_OGRErr(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OGRERR_FAILURE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1177
 *         result = gdal_start_transaction(self.cogr_ds, 0)
 *         if result == OGRERR_FAILURE:
 *             raise TransactionError("Failed to start transaction")             # <<<<<<<<<<<<<<
 * 
 *         schema_props_keys = set(collection.schema['properties'].keys())
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_TransactionError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_4, __pyx_kp_u_Failed_to_start_transaction) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_Failed_to_start_transaction);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1177, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1176
 *         log.debug("Starting transaction (initial)")
 *         result = gdal_start_transaction(self.cogr_ds, 0)
 *         if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *             raise TransactionError("Failed to start transaction")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1179
 *             raise TransactionError("Failed to start transaction")
 * 
 *         schema_props_keys = set(collection.schema['properties'].keys())             # <<<<<<<<<<<<<<
 *         for record in records:
 *             log.debug("Creating feature in layer: %s" % record)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_4 = __Pyx_PyObject_Dict_GetItem(__pyx_t_6, __pyx_n_u_properties); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_keys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_schema_props_keys = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "fiona/ogrext.pyx":1180
 * 
 *         schema_props_keys = set(collection.schema['properties'].keys())
 *         for record in records:             # <<<<<<<<<<<<<<
 *             log.debug("Creating feature in layer: %s" % record)
 *             # Validate against collection's schema.
 */
  if (likely(PyList_CheckExact(__pyx_v_records)) || PyTuple_CheckExact(__pyx_v_records)) {
    __pyx_t_6 = __pyx_v_records; __Pyx_INCREF(__pyx_t_6); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_records); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1180, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_6))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 1180, __pyx_L1_error)
        #else
        __pyx_t_3 = PySequence_ITEM(__pyx_t_6, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1180, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_6);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1180, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_record, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1181
 *         schema_props_keys = set(collection.schema['properties'].keys())
 *         for record in records:
 *             log.debug("Creating feature in layer: %s" % record)             # <<<<<<<<<<<<<<
 *             # Validate against collection's schema.
 *             if set(record['properties'].keys()) != schema_props_keys:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Creating_feature_in_layer_s, __pyx_v_record); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_10, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1183
 *             log.debug("Creating feature in layer: %s" % record)
 *             # Validate against collection's schema.
 *             if set(record['properties'].keys()) != schema_props_keys:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     "Record does not match collection schema: %r != %r" % (
 */
    __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_record, __pyx_n_u_properties); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_keys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySet_New(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_v_schema_props_keys, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(__pyx_t_2)) {

      /* "fiona/ogrext.pyx":1185
 *             if set(record['properties'].keys()) != schema_props_keys:
 *                 raise ValueError(
 *                     "Record does not match collection schema: %r != %r" % (             # <<<<<<<<<<<<<<
 *                         record['properties'].keys(),
 *                         list(schema_props_keys) ))
 */
      __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = 0;
      __pyx_t_12 = 127;
      __Pyx_INCREF(__pyx_kp_u_Record_does_not_match_collection);
      __pyx_t_11 += 41;
      __Pyx_GIVEREF(__pyx_kp_u_Record_does_not_match_collection);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Record_does_not_match_collection);

      /* "fiona/ogrext.pyx":1186
 *                 raise ValueError(
 *                     "Record does not match collection schema: %r != %r" % (
 *                         record['properties'].keys(),             # <<<<<<<<<<<<<<
 *                         list(schema_props_keys) ))
 *             if not validate_geometry_type(record):
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_record, __pyx_n_u_properties); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallNoArg(__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_4), __pyx_empty_unicode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_10) : __pyx_t_12;
      __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_INCREF(__pyx_kp_u__25);
      __pyx_t_11 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__25);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__25);

      /* "fiona/ogrext.pyx":1187
 *                     "Record does not match collection schema: %r != %r" % (
 *                         record['properties'].keys(),
 *                         list(schema_props_keys) ))             # <<<<<<<<<<<<<<
 *             if not validate_geometry_type(record):
 *                 raise GeometryTypeValidationError(
 */
      __pyx_t_10 = PySequence_List(__pyx_v_schema_props_keys); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_4 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_10), __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1187, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_12;
      __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":1185
 *             if set(record['properties'].keys()) != schema_props_keys:
 *                 raise ValueError(
 *                     "Record does not match collection schema: %r != %r" % (             # <<<<<<<<<<<<<<
 *                         record['properties'].keys(),
 *                         list(schema_props_keys) ))
 */
      __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 4, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":1184
 *             # Validate against collection's schema.
 *             if set(record['properties'].keys()) != schema_props_keys:
 *                 raise ValueError(             # <<<<<<<<<<<<<<
 *                     "Record does not match collection schema: %r != %r" % (
 *                         record['properties'].keys(),
 */
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1184, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1184, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1183
 *             log.debug("Creating feature in layer: %s" % record)
 *             # Validate against collection's schema.
 *             if set(record['properties'].keys()) != schema_props_keys:             # <<<<<<<<<<<<<<
 *                 raise ValueError(
 *                     "Record does not match collection schema: %r != %r" % (
 */
    }

    /* "fiona/ogrext.pyx":1188
 *                         record['properties'].keys(),
 *                         list(schema_props_keys) ))
 *             if not validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *                 raise GeometryTypeValidationError(
 *                     "Record's geometry type does not match "
 */
    __pyx_t_3 = __pyx_pf_5fiona_6ogrext_14WritingSession_9writerecs_validate_geometry_type(__pyx_v_validate_geometry_type, __pyx_v_record); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1188, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_13 = ((!__pyx_t_2) != 0);
    if (unlikely(__pyx_t_13)) {

      /* "fiona/ogrext.pyx":1189
 *                         list(schema_props_keys) ))
 *             if not validate_geometry_type(record):
 *                 raise GeometryTypeValidationError(             # <<<<<<<<<<<<<<
 *                     "Record's geometry type does not match "
 *                     "collection schema's geometry type: %r != %r" % (
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_GeometryTypeValidationError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "fiona/ogrext.pyx":1190
 *             if not validate_geometry_type(record):
 *                 raise GeometryTypeValidationError(
 *                     "Record's geometry type does not match "             # <<<<<<<<<<<<<<
 *                     "collection schema's geometry type: %r != %r" % (
 *                          record['geometry']['type'],
 */
      __pyx_t_10 = PyTuple_New(4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = 0;
      __pyx_t_12 = 127;
      __Pyx_INCREF(__pyx_kp_u_Record_s_geometry_type_does_not);
      __pyx_t_11 += 73;
      __Pyx_GIVEREF(__pyx_kp_u_Record_s_geometry_type_does_not);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_kp_u_Record_s_geometry_type_does_not);

      /* "fiona/ogrext.pyx":1192
 *                     "Record's geometry type does not match "
 *                     "collection schema's geometry type: %r != %r" % (
 *                          record['geometry']['type'],             # <<<<<<<<<<<<<<
 *                          collection.schema['geometry'] ))
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_Dict_GetItem(__pyx_v_record, __pyx_n_u_geometry); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_14 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_u_type); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_14), __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_12;
      __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_9);
      __pyx_t_9 = 0;
      __Pyx_INCREF(__pyx_kp_u__25);
      __pyx_t_11 += 4;
      __Pyx_GIVEREF(__pyx_kp_u__25);
      PyTuple_SET_ITEM(__pyx_t_10, 2, __pyx_kp_u__25);

      /* "fiona/ogrext.pyx":1193
 *                     "collection schema's geometry type: %r != %r" % (
 *                          record['geometry']['type'],
 *                          collection.schema['geometry'] ))             # <<<<<<<<<<<<<<
 * 
 *             cogr_feature = OGRFeatureBuilder().build(record, collection)
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_schema); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_14 = __Pyx_PyObject_Dict_GetItem(__pyx_t_9, __pyx_n_u_geometry); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = __Pyx_PyObject_FormatSimpleAndDecref(PyObject_Repr(__pyx_t_14), __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_12;
      __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
      __Pyx_GIVEREF(__pyx_t_9);
      PyTuple_SET_ITEM(__pyx_t_10, 3, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "fiona/ogrext.pyx":1190
 *             if not validate_geometry_type(record):
 *                 raise GeometryTypeValidationError(
 *                     "Record's geometry type does not match "             # <<<<<<<<<<<<<<
 *                     "collection schema's geometry type: %r != %r" % (
 *                          record['geometry']['type'],
 */
      __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_10, 4, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1190, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_10, __pyx_t_9) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1189, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1188
 *                         record['properties'].keys(),
 *                         list(schema_props_keys) ))
 *             if not validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *                 raise GeometryTypeValidationError(
 *                     "Record's geometry type does not match "
 */
    }

    /* "fiona/ogrext.pyx":1195
 *                          collection.schema['geometry'] ))
 * 
 *             cogr_feature = OGRFeatureBuilder().build(record, collection)             # <<<<<<<<<<<<<<
 *             result = OGR_L_CreateFeature(cogr_layer, cogr_feature)
 *             if result != OGRERR_NONE:
 */
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_OGRFeatureBuilder)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = ((struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_t_3)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_t_3), __pyx_v_record, __pyx_v_collection); if (unlikely(__pyx_t_1 == ((void *)NULL))) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cogr_feature = __pyx_t_1;

    /* "fiona/ogrext.pyx":1196
 * 
 *             cogr_feature = OGRFeatureBuilder().build(record, collection)
 *             result = OGR_L_CreateFeature(cogr_layer, cogr_feature)             # <<<<<<<<<<<<<<
 *             if result != OGRERR_NONE:
 *                 raise RuntimeError("Failed to write record: %s" % record)
 */
    __pyx_v_result = OGR_L_CreateFeature(__pyx_v_cogr_layer, __pyx_v_cogr_feature);

    /* "fiona/ogrext.pyx":1197
 *             cogr_feature = OGRFeatureBuilder().build(record, collection)
 *             result = OGR_L_CreateFeature(cogr_layer, cogr_feature)
 *             if result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Failed to write record: %s" % record)
 *             _deleteOgrFeature(cogr_feature)
 */
    __pyx_t_3 = __Pyx_PyInt_From_OGRErr(__pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OGRERR_NONE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_NE); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(__pyx_t_13)) {

      /* "fiona/ogrext.pyx":1198
 *             result = OGR_L_CreateFeature(cogr_layer, cogr_feature)
 *             if result != OGRERR_NONE:
 *                 raise RuntimeError("Failed to write record: %s" % record)             # <<<<<<<<<<<<<<
 *             _deleteOgrFeature(cogr_feature)
 * 
 */
      __pyx_t_9 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_Failed_to_write_record_s, __pyx_v_record); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_RuntimeError, __pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1198, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1197
 *             cogr_feature = OGRFeatureBuilder().build(record, collection)
 *             result = OGR_L_CreateFeature(cogr_layer, cogr_feature)
 *             if result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError("Failed to write record: %s" % record)
 *             _deleteOgrFeature(cogr_feature)
 */
    }

    /* "fiona/ogrext.pyx":1199
 *             if result != OGRERR_NONE:
 *                 raise RuntimeError("Failed to write record: %s" % record)
 *             _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 * 
 *             features_in_transaction += 1
 */
    __pyx_t_4 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1199, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "fiona/ogrext.pyx":1201
 *             _deleteOgrFeature(cogr_feature)
 * 
 *             features_in_transaction += 1             # <<<<<<<<<<<<<<
 *             if features_in_transaction == DEFAULT_TRANSACTION_SIZE:
 *                 log.debug("Comitting transaction (intermediate)")
 */
    __pyx_v_features_in_transaction = (__pyx_v_features_in_transaction + 1);

    /* "fiona/ogrext.pyx":1202
 * 
 *             features_in_transaction += 1
 *             if features_in_transaction == DEFAULT_TRANSACTION_SIZE:             # <<<<<<<<<<<<<<
 *                 log.debug("Comitting transaction (intermediate)")
 *                 result = gdal_commit_transaction(self.cogr_ds)
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_features_in_transaction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_DEFAULT_TRANSACTION_SIZE); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_13) {

      /* "fiona/ogrext.pyx":1203
 *             features_in_transaction += 1
 *             if features_in_transaction == DEFAULT_TRANSACTION_SIZE:
 *                 log.debug("Comitting transaction (intermediate)")             # <<<<<<<<<<<<<<
 *                 result = gdal_commit_transaction(self.cogr_ds)
 *                 if result == OGRERR_FAILURE:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_log); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_debug); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_9, __pyx_kp_u_Comitting_transaction_intermedia) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_Comitting_transaction_intermedia);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":1204
 *             if features_in_transaction == DEFAULT_TRANSACTION_SIZE:
 *                 log.debug("Comitting transaction (intermediate)")
 *                 result = gdal_commit_transaction(self.cogr_ds)             # <<<<<<<<<<<<<<
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to commit transaction")
 */
      __pyx_v_result = __pyx_f_5fiona_5_shim_gdal_commit_transaction(__pyx_v_self->__pyx_base.cogr_ds);

      /* "fiona/ogrext.pyx":1205
 *                 log.debug("Comitting transaction (intermediate)")
 *                 result = gdal_commit_transaction(self.cogr_ds)
 *                 if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *                     raise TransactionError("Failed to commit transaction")
 *                 log.debug("Starting transaction (intermediate)")
 */
      __pyx_t_3 = __Pyx_PyInt_From_OGRErr(__pyx_v_result); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OGRERR_FAILURE); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1205, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1205, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1205, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(__pyx_t_13)) {

        /* "fiona/ogrext.pyx":1206
 *                 result = gdal_commit_transaction(self.cogr_ds)
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to commit transaction")             # <<<<<<<<<<<<<<
 *                 log.debug("Starting transaction (intermediate)")
 *                 result = gdal_start_transaction(self.cogr_ds, 0)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_TransactionError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_3 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_4, function);
          }
        }
        __pyx_t_9 = (__pyx_t_3) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_3, __pyx_kp_u_Failed_to_commit_transaction) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_Failed_to_commit_transaction);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_9, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __PYX_ERR(0, 1206, __pyx_L1_error)

        /* "fiona/ogrext.pyx":1205
 *                 log.debug("Comitting transaction (intermediate)")
 *                 result = gdal_commit_transaction(self.cogr_ds)
 *                 if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *                     raise TransactionError("Failed to commit transaction")
 *                 log.debug("Starting transaction (intermediate)")
 */
      }

      /* "fiona/ogrext.pyx":1207
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to commit transaction")
 *                 log.debug("Starting transaction (intermediate)")             # <<<<<<<<<<<<<<
 *                 result = gdal_start_transaction(self.cogr_ds, 0)
 *                 if result == OGRERR_FAILURE:
 */
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      __pyx_t_9 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_u_Starting_transaction_intermediat) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_Starting_transaction_intermediat);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1207, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "fiona/ogrext.pyx":1208
 *                     raise TransactionError("Failed to commit transaction")
 *                 log.debug("Starting transaction (intermediate)")
 *                 result = gdal_start_transaction(self.cogr_ds, 0)             # <<<<<<<<<<<<<<
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to start transaction")
 */
      __pyx_v_result = __pyx_f_5fiona_5_shim_gdal_start_transaction(__pyx_v_self->__pyx_base.cogr_ds, 0);

      /* "fiona/ogrext.pyx":1209
 *                 log.debug("Starting transaction (intermediate)")
 *                 result = gdal_start_transaction(self.cogr_ds, 0)
 *                 if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *                     raise TransactionError("Failed to start transaction")
 *                 features_in_transaction = 0
 */
      __pyx_t_9 = __Pyx_PyInt_From_OGRErr(__pyx_v_result); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OGRERR_FAILURE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_9, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1209, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(__pyx_t_13)) {

        /* "fiona/ogrext.pyx":1210
 *                 result = gdal_start_transaction(self.cogr_ds, 0)
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to start transaction")             # <<<<<<<<<<<<<<
 *                 features_in_transaction = 0
 * 
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TransactionError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_9 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_9)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_9);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        __pyx_t_4 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_9, __pyx_kp_u_Failed_to_start_transaction) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_Failed_to_start_transaction);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(0, 1210, __pyx_L1_error)

        /* "fiona/ogrext.pyx":1209
 *                 log.debug("Starting transaction (intermediate)")
 *                 result = gdal_start_transaction(self.cogr_ds, 0)
 *                 if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *                     raise TransactionError("Failed to start transaction")
 *                 features_in_transaction = 0
 */
      }

      /* "fiona/ogrext.pyx":1211
 *                 if result == OGRERR_FAILURE:
 *                     raise TransactionError("Failed to start transaction")
 *                 features_in_transaction = 0             # <<<<<<<<<<<<<<
 * 
 *         log.debug("Comitting transaction (final)")
 */
      __pyx_v_features_in_transaction = 0;

      /* "fiona/ogrext.pyx":1202
 * 
 *             features_in_transaction += 1
 *             if features_in_transaction == DEFAULT_TRANSACTION_SIZE:             # <<<<<<<<<<<<<<
 *                 log.debug("Comitting transaction (intermediate)")
 *                 result = gdal_commit_transaction(self.cogr_ds)
 */
    }

    /* "fiona/ogrext.pyx":1180
 * 
 *         schema_props_keys = set(collection.schema['properties'].keys())
 *         for record in records:             # <<<<<<<<<<<<<<
 *             log.debug("Creating feature in layer: %s" % record)
 *             # Validate against collection's schema.
 */
  }
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "fiona/ogrext.pyx":1213
 *                 features_in_transaction = 0
 * 
 *         log.debug("Comitting transaction (final)")             # <<<<<<<<<<<<<<
 *         result = gdal_commit_transaction(self.cogr_ds)
 *         if result == OGRERR_FAILURE:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_6 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_kp_u_Comitting_transaction_final) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_Comitting_transaction_final);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

  /* "fiona/ogrext.pyx":1214
 * 
 *         log.debug("Comitting transaction (final)")
 *         result = gdal_commit_transaction(self.cogr_ds)             # <<<<<<<<<<<<<<
 *         if result == OGRERR_FAILURE:
 *             raise TransactionError("Failed to commit transaction")
 */
  __pyx_v_result = __pyx_f_5fiona_5_shim_gdal_commit_transaction(__pyx_v_self->__pyx_base.cogr_ds);

  /* "fiona/ogrext.pyx":1215
 *         log.debug("Comitting transaction (final)")
 *         result = gdal_commit_transaction(self.cogr_ds)
 *         if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *             raise TransactionError("Failed to commit transaction")
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_From_OGRErr(__pyx_v_result); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OGRERR_FAILURE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_6, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 1215, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(__pyx_t_13)) {

    /* "fiona/ogrext.pyx":1216
 *         result = gdal_commit_transaction(self.cogr_ds)
 *         if result == OGRERR_FAILURE:
 *             raise TransactionError("Failed to commit transaction")             # <<<<<<<<<<<<<<
 * 
 *     def sync(self, collection):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TransactionError); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_6, __pyx_kp_u_Failed_to_commit_transaction) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_Failed_to_commit_transaction);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1216, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(0, 1216, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1215
 *         log.debug("Comitting transaction (final)")
 *         result = gdal_commit_transaction(self.cogr_ds)
 *         if result == OGRERR_FAILURE:             # <<<<<<<<<<<<<<
 *             raise TransactionError("Failed to commit transaction")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1154
 *         log.debug("Writing started")
 * 
 *     def writerecs(self, records, collection):             # <<<<<<<<<<<<<<
 *         """Writes buffered records to OGR."""
 *         cdef void *cogr_driver
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.writerecs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_schema_geom_type);
  __Pyx_XDECREF(__pyx_v_driver_name);
  __Pyx_XDECREF(__pyx_v_validate_geometry_type);
  __Pyx_XDECREF(__pyx_v_schema_props_keys);
  __Pyx_XDECREF(__pyx_v_record);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1218
 *             raise TransactionError("Failed to commit transaction")
 * 
 *     def sync(self, collection):             # <<<<<<<<<<<<<<
 *         """Syncs OGR to disk."""
 *         cdef void *cogr_ds = self.cogr_ds
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_5sync(PyObject *__pyx_v_self, PyObject *__pyx_v_collection); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14WritingSession_4sync[] = "Syncs OGR to disk.";
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_5sync(PyObject *__pyx_v_self, PyObject *__pyx_v_collection) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sync (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_4sync(((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_v_self), ((PyObject *)__pyx_v_collection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_4sync(struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_collection) {
  void *__pyx_v_cogr_ds;
  CYTHON_UNUSED void *__pyx_v_cogr_layer;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  void *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("sync", 0);

  /* "fiona/ogrext.pyx":1220
 *     def sync(self, collection):
 *         """Syncs OGR to disk."""
 *         cdef void *cogr_ds = self.cogr_ds             # <<<<<<<<<<<<<<
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_ds == NULL:
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.cogr_ds;
  __pyx_v_cogr_ds = __pyx_t_1;

  /* "fiona/ogrext.pyx":1221
 *         """Syncs OGR to disk."""
 *         cdef void *cogr_ds = self.cogr_ds
 *         cdef void *cogr_layer = self.cogr_layer             # <<<<<<<<<<<<<<
 *         if cogr_ds == NULL:
 *             raise ValueError("Null data source")
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base.cogr_layer;
  __pyx_v_cogr_layer = __pyx_t_1;

  /* "fiona/ogrext.pyx":1222
 *         cdef void *cogr_ds = self.cogr_ds
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_ds == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null data source")
 * 
 */
  __pyx_t_2 = ((__pyx_v_cogr_ds == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1223
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_ds == NULL:
 *             raise ValueError("Null data source")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1223, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1222
 *         cdef void *cogr_ds = self.cogr_ds
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_ds == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null data source")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1226
 * 
 * 
 *         gdal_flush_cache(cogr_ds)             # <<<<<<<<<<<<<<
 *         log.debug("Flushed data source cache")
 * 
 */
  __pyx_f_5fiona_5_shim_gdal_flush_cache(__pyx_v_cogr_ds);

  /* "fiona/ogrext.pyx":1227
 * 
 *         gdal_flush_cache(cogr_ds)
 *         log.debug("Flushed data source cache")             # <<<<<<<<<<<<<<
 * 
 * cdef class Iterator:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_log); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_debug); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_kp_u_Flushed_data_source_cache) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_Flushed_data_source_cache);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":1218
 *             raise TransactionError("Failed to commit transaction")
 * 
 *     def sync(self, collection):             # <<<<<<<<<<<<<<
 *         """Syncs OGR to disk."""
 *         cdef void *cogr_ds = self.cogr_ds
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.sync", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_7__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_6__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_6__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14WritingSession_9__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14WritingSession_8__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_WritingSession *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14WritingSession_8__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_WritingSession *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.WritingSession.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1244
 *     cdef stepsign
 * 
 *     def __cinit__(self, collection, start=None, stop=None, step=None,             # <<<<<<<<<<<<<<
 *                   bbox=None, mask=None):
 *         if collection.session is None:
 */

/* Python wrapper */
static int __pyx_pw_5fiona_6ogrext_8Iterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_5fiona_6ogrext_8Iterator_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_collection = 0;
  PyObject *__pyx_v_start = 0;
  PyObject *__pyx_v_stop = 0;
  PyObject *__pyx_v_step = 0;
  PyObject *__pyx_v_bbox = 0;
  PyObject *__pyx_v_mask = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_collection,&__pyx_n_s_start,&__pyx_n_s_stop,&__pyx_n_s_step,&__pyx_n_s_bbox,&__pyx_n_s_mask,0};
    PyObject* values[6] = {0,0,0,0,0,0};
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)Py_None);
    values[3] = ((PyObject *)Py_None);

    /* "fiona/ogrext.pyx":1245
 * 
 *     def __cinit__(self, collection, start=None, stop=None, step=None,
 *                   bbox=None, mask=None):             # <<<<<<<<<<<<<<
 *         if collection.session is None:
 *             raise ValueError("I/O operation on closed collection")
 */
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_collection)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_start);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stop);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_step);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bbox);
          if (value) { values[4] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mask);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 1244, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_collection = values[0];
    __pyx_v_start = values[1];
    __pyx_v_stop = values[2];
    __pyx_v_step = values[3];
    __pyx_v_bbox = values[4];
    __pyx_v_mask = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 1, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1244, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.Iterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator___cinit__(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self), __pyx_v_collection, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_bbox, __pyx_v_mask);

  /* "fiona/ogrext.pyx":1244
 *     cdef stepsign
 * 
 *     def __cinit__(self, collection, start=None, stop=None, step=None,             # <<<<<<<<<<<<<<
 *                   bbox=None, mask=None):
 *         if collection.session is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5fiona_6ogrext_8Iterator___cinit__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self, PyObject *__pyx_v_collection, PyObject *__pyx_v_start, PyObject *__pyx_v_stop, PyObject *__pyx_v_step, PyObject *__pyx_v_bbox, PyObject *__pyx_v_mask) {
  struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_session = 0;
  void *__pyx_v_cogr_geometry;
  void *__pyx_v_cogr_layer;
  PyObject *__pyx_v_ftcount = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  void *__pyx_t_4;
  double __pyx_t_5;
  double __pyx_t_6;
  double __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_start);
  __Pyx_INCREF(__pyx_v_stop);
  __Pyx_INCREF(__pyx_v_step);

  /* "fiona/ogrext.pyx":1246
 *     def __cinit__(self, collection, start=None, stop=None, step=None,
 *                   bbox=None, mask=None):
 *         if collection.session is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("I/O operation on closed collection")
 *         self.collection = collection
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1247
 *                   bbox=None, mask=None):
 *         if collection.session is None:
 *             raise ValueError("I/O operation on closed collection")             # <<<<<<<<<<<<<<
 *         self.collection = collection
 *         cdef Session session
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1247, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1246
 *     def __cinit__(self, collection, start=None, stop=None, step=None,
 *                   bbox=None, mask=None):
 *         if collection.session is None:             # <<<<<<<<<<<<<<
 *             raise ValueError("I/O operation on closed collection")
 *         self.collection = collection
 */
  }

  /* "fiona/ogrext.pyx":1248
 *         if collection.session is None:
 *             raise ValueError("I/O operation on closed collection")
 *         self.collection = collection             # <<<<<<<<<<<<<<
 *         cdef Session session
 *         cdef void *cogr_geometry
 */
  __Pyx_INCREF(__pyx_v_collection);
  __Pyx_GIVEREF(__pyx_v_collection);
  __Pyx_GOTREF(__pyx_v_self->collection);
  __Pyx_DECREF(__pyx_v_self->collection);
  __pyx_v_self->collection = __pyx_v_collection;

  /* "fiona/ogrext.pyx":1251
 *         cdef Session session
 *         cdef void *cogr_geometry
 *         session = self.collection.session             # <<<<<<<<<<<<<<
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1251, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_Session))))) __PYX_ERR(0, 1251, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1252
 *         cdef void *cogr_geometry
 *         session = self.collection.session
 *         cdef void *cogr_layer = session.cogr_layer             # <<<<<<<<<<<<<<
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")
 */
  __pyx_t_4 = __pyx_v_session->cogr_layer;
  __pyx_v_cogr_layer = __pyx_t_4;

  /* "fiona/ogrext.pyx":1253
 *         session = self.collection.session
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         OGR_L_ResetReading(cogr_layer)
 */
  __pyx_t_3 = ((__pyx_v_cogr_layer == NULL) != 0);
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1254
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 *         OGR_L_ResetReading(cogr_layer)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1254, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1254, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1253
 *         session = self.collection.session
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null layer")
 *         OGR_L_ResetReading(cogr_layer)
 */
  }

  /* "fiona/ogrext.pyx":1255
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")
 *         OGR_L_ResetReading(cogr_layer)             # <<<<<<<<<<<<<<
 * 
 *         if bbox and mask:
 */
  OGR_L_ResetReading(__pyx_v_cogr_layer);

  /* "fiona/ogrext.pyx":1257
 *         OGR_L_ResetReading(cogr_layer)
 * 
 *         if bbox and mask:             # <<<<<<<<<<<<<<
 *             raise ValueError("mask and bbox can not be set together")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_bbox); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1257, __pyx_L1_error)
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_mask); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1257, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1258
 * 
 *         if bbox and mask:
 *             raise ValueError("mask and bbox can not be set together")             # <<<<<<<<<<<<<<
 * 
 *         if bbox:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1258, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1257
 *         OGR_L_ResetReading(cogr_layer)
 * 
 *         if bbox and mask:             # <<<<<<<<<<<<<<
 *             raise ValueError("mask and bbox can not be set together")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1260
 *             raise ValueError("mask and bbox can not be set together")
 * 
 *         if bbox:             # <<<<<<<<<<<<<<
 *             OGR_L_SetSpatialFilterRect(
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_bbox); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1260, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1262
 *         if bbox:
 *             OGR_L_SetSpatialFilterRect(
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])             # <<<<<<<<<<<<<<
 *         elif mask:
 *             cogr_geometry = OGRGeomBuilder().build(mask)
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_bbox, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_5 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_bbox, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_6 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_bbox, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_bbox, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1262, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1261
 * 
 *         if bbox:
 *             OGR_L_SetSpatialFilterRect(             # <<<<<<<<<<<<<<
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 *         elif mask:
 */
    OGR_L_SetSpatialFilterRect(__pyx_v_cogr_layer, __pyx_t_5, __pyx_t_6, __pyx_t_7, __pyx_t_8);

    /* "fiona/ogrext.pyx":1260
 *             raise ValueError("mask and bbox can not be set together")
 * 
 *         if bbox:             # <<<<<<<<<<<<<<
 *             OGR_L_SetSpatialFilterRect(
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 */
    goto __pyx_L8;
  }

  /* "fiona/ogrext.pyx":1263
 *             OGR_L_SetSpatialFilterRect(
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 *         elif mask:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGRGeomBuilder().build(mask)
 *             OGR_L_SetSpatialFilter(cogr_layer, cogr_geometry)
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_mask); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1263, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1264
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 *         elif mask:
 *             cogr_geometry = OGRGeomBuilder().build(mask)             # <<<<<<<<<<<<<<
 *             OGR_L_SetSpatialFilter(cogr_layer, cogr_geometry)
 *             OGR_G_DestroyGeometry(cogr_geometry)
 */
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_9_geometry_OGRGeomBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = ((struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder *)((struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder *)__pyx_t_1), __pyx_v_mask); if (unlikely(__pyx_t_4 == ((void *)NULL))) __PYX_ERR(0, 1264, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_cogr_geometry = __pyx_t_4;

    /* "fiona/ogrext.pyx":1265
 *         elif mask:
 *             cogr_geometry = OGRGeomBuilder().build(mask)
 *             OGR_L_SetSpatialFilter(cogr_layer, cogr_geometry)             # <<<<<<<<<<<<<<
 *             OGR_G_DestroyGeometry(cogr_geometry)
 * 
 */
    OGR_L_SetSpatialFilter(__pyx_v_cogr_layer, __pyx_v_cogr_geometry);

    /* "fiona/ogrext.pyx":1266
 *             cogr_geometry = OGRGeomBuilder().build(mask)
 *             OGR_L_SetSpatialFilter(cogr_layer, cogr_geometry)
 *             OGR_G_DestroyGeometry(cogr_geometry)             # <<<<<<<<<<<<<<
 * 
 *         else:
 */
    OGR_G_DestroyGeometry(__pyx_v_cogr_geometry);

    /* "fiona/ogrext.pyx":1263
 *             OGR_L_SetSpatialFilterRect(
 *                 cogr_layer, bbox[0], bbox[1], bbox[2], bbox[3])
 *         elif mask:             # <<<<<<<<<<<<<<
 *             cogr_geometry = OGRGeomBuilder().build(mask)
 *             OGR_L_SetSpatialFilter(cogr_layer, cogr_geometry)
 */
    goto __pyx_L8;
  }

  /* "fiona/ogrext.pyx":1269
 * 
 *         else:
 *             OGR_L_SetSpatialFilter(cogr_layer, NULL)             # <<<<<<<<<<<<<<
 * 
 *         self.encoding = session._get_internal_encoding()
 */
  /*else*/ {
    OGR_L_SetSpatialFilter(__pyx_v_cogr_layer, NULL);
  }
  __pyx_L8:;

  /* "fiona/ogrext.pyx":1271
 *             OGR_L_SetSpatialFilter(cogr_layer, NULL)
 * 
 *         self.encoding = session._get_internal_encoding()             # <<<<<<<<<<<<<<
 * 
 *         self.fastindex = OGR_L_TestCapability(
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_session), __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_10)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_10) : __Pyx_PyObject_CallNoArg(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->encoding);
  __Pyx_DECREF(__pyx_v_self->encoding);
  __pyx_v_self->encoding = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1273
 *         self.encoding = session._get_internal_encoding()
 * 
 *         self.fastindex = OGR_L_TestCapability(             # <<<<<<<<<<<<<<
 *             session.cogr_layer, OLC_FASTSETNEXTBYINDEX)
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(OGR_L_TestCapability(__pyx_v_session->cogr_layer, __pyx_v_5fiona_6ogrext_OLC_FASTSETNEXTBYINDEX)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->fastindex);
  __Pyx_DECREF(__pyx_v_self->fastindex);
  __pyx_v_self->fastindex = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1276
 *             session.cogr_layer, OLC_FASTSETNEXTBYINDEX)
 * 
 *         ftcount = OGR_L_GetFeatureCount(session.cogr_layer, 0)             # <<<<<<<<<<<<<<
 *         if ftcount == -1 and ((start is not None and start < 0) or
 *                               (stop is not None and stop < 0)):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(OGR_L_GetFeatureCount(__pyx_v_session->cogr_layer, 0)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1276, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ftcount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1277
 * 
 *         ftcount = OGR_L_GetFeatureCount(session.cogr_layer, 0)
 *         if ftcount == -1 and ((start is not None and start < 0) or             # <<<<<<<<<<<<<<
 *                               (stop is not None and stop < 0)):
 *             raise IndexError(
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_ftcount, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_start != Py_None);
  __pyx_t_11 = (__pyx_t_2 != 0);
  if (!__pyx_t_11) {
    goto __pyx_L12_next_or;
  } else {
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_start, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_11) {
  } else {
    __pyx_t_3 = __pyx_t_11;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_L12_next_or:;

  /* "fiona/ogrext.pyx":1278
 *         ftcount = OGR_L_GetFeatureCount(session.cogr_layer, 0)
 *         if ftcount == -1 and ((start is not None and start < 0) or
 *                               (stop is not None and stop < 0)):             # <<<<<<<<<<<<<<
 *             raise IndexError(
 *                 "collection's dataset does not support negative slice indexes")
 */
  __pyx_t_11 = (__pyx_v_stop != Py_None);
  __pyx_t_2 = (__pyx_t_11 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_3 = __pyx_t_2;
    goto __pyx_L10_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_stop, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1278, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_2;
  __pyx_L10_bool_binop_done:;

  /* "fiona/ogrext.pyx":1277
 * 
 *         ftcount = OGR_L_GetFeatureCount(session.cogr_layer, 0)
 *         if ftcount == -1 and ((start is not None and start < 0) or             # <<<<<<<<<<<<<<
 *                               (stop is not None and stop < 0)):
 *             raise IndexError(
 */
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1279
 *         if ftcount == -1 and ((start is not None and start < 0) or
 *                               (stop is not None and stop < 0)):
 *             raise IndexError(             # <<<<<<<<<<<<<<
 *                 "collection's dataset does not support negative slice indexes")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_IndexError, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1279, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1277
 * 
 *         ftcount = OGR_L_GetFeatureCount(session.cogr_layer, 0)
 *         if ftcount == -1 and ((start is not None and start < 0) or             # <<<<<<<<<<<<<<
 *                               (stop is not None and stop < 0)):
 *             raise IndexError(
 */
  }

  /* "fiona/ogrext.pyx":1282
 *                 "collection's dataset does not support negative slice indexes")
 * 
 *         if stop is not None and stop < 0:             # <<<<<<<<<<<<<<
 *             stop += ftcount
 * 
 */
  __pyx_t_2 = (__pyx_v_stop != Py_None);
  __pyx_t_11 = (__pyx_t_2 != 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_3 = __pyx_t_11;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_stop, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __pyx_t_11;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1283
 * 
 *         if stop is not None and stop < 0:
 *             stop += ftcount             # <<<<<<<<<<<<<<
 * 
 *         if start is None:
 */
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_stop, __pyx_v_ftcount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_stop, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1282
 *                 "collection's dataset does not support negative slice indexes")
 * 
 *         if stop is not None and stop < 0:             # <<<<<<<<<<<<<<
 *             stop += ftcount
 * 
 */
  }

  /* "fiona/ogrext.pyx":1285
 *             stop += ftcount
 * 
 *         if start is None:             # <<<<<<<<<<<<<<
 *             start = 0
 *         if start is not None and start < 0:
 */
  __pyx_t_3 = (__pyx_v_start == Py_None);
  __pyx_t_11 = (__pyx_t_3 != 0);
  if (__pyx_t_11) {

    /* "fiona/ogrext.pyx":1286
 * 
 *         if start is None:
 *             start = 0             # <<<<<<<<<<<<<<
 *         if start is not None and start < 0:
 *             start += ftcount
 */
    __Pyx_INCREF(__pyx_int_0);
    __Pyx_DECREF_SET(__pyx_v_start, __pyx_int_0);

    /* "fiona/ogrext.pyx":1285
 *             stop += ftcount
 * 
 *         if start is None:             # <<<<<<<<<<<<<<
 *             start = 0
 *         if start is not None and start < 0:
 */
  }

  /* "fiona/ogrext.pyx":1287
 *         if start is None:
 *             start = 0
 *         if start is not None and start < 0:             # <<<<<<<<<<<<<<
 *             start += ftcount
 * 
 */
  __pyx_t_3 = (__pyx_v_start != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_11 = __pyx_t_2;
    goto __pyx_L20_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_start, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1287, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __pyx_t_2;
  __pyx_L20_bool_binop_done:;
  if (__pyx_t_11) {

    /* "fiona/ogrext.pyx":1288
 *             start = 0
 *         if start is not None and start < 0:
 *             start += ftcount             # <<<<<<<<<<<<<<
 * 
 *         # step size
 */
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_start, __pyx_v_ftcount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_start, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1287
 *         if start is None:
 *             start = 0
 *         if start is not None and start < 0:             # <<<<<<<<<<<<<<
 *             start += ftcount
 * 
 */
  }

  /* "fiona/ogrext.pyx":1291
 * 
 *         # step size
 *         if step is None:             # <<<<<<<<<<<<<<
 *             step = 1
 *         if step == 0:
 */
  __pyx_t_11 = (__pyx_v_step == Py_None);
  __pyx_t_2 = (__pyx_t_11 != 0);
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1292
 *         # step size
 *         if step is None:
 *             step = 1             # <<<<<<<<<<<<<<
 *         if step == 0:
 *             raise ValueError("slice step cannot be zero")
 */
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_DECREF_SET(__pyx_v_step, __pyx_int_1);

    /* "fiona/ogrext.pyx":1291
 * 
 *         # step size
 *         if step is None:             # <<<<<<<<<<<<<<
 *             step = 1
 *         if step == 0:
 */
  }

  /* "fiona/ogrext.pyx":1293
 *         if step is None:
 *             step = 1
 *         if step == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_step, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1294
 *             step = 1
 *         if step == 0:
 *             raise ValueError("slice step cannot be zero")             # <<<<<<<<<<<<<<
 *         if step < 0 and not self.fastindex:
 *             warnings.warn("Layer does not support" \
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1294, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1293
 *         if step is None:
 *             step = 1
 *         if step == 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:
 */
  }

  /* "fiona/ogrext.pyx":1295
 *         if step == 0:
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:             # <<<<<<<<<<<<<<
 *             warnings.warn("Layer does not support" \
 *                     "OLCFastSetNextByIndex, negative step size may" \
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_step, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {
  } else {
    __pyx_t_2 = __pyx_t_11;
    goto __pyx_L25_bool_binop_done;
  }
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_self->fastindex); if (unlikely(__pyx_t_11 < 0)) __PYX_ERR(0, 1295, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_11) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L25_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1296
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:
 *             warnings.warn("Layer does not support" \             # <<<<<<<<<<<<<<
 *                     "OLCFastSetNextByIndex, negative step size may" \
 *                     " be slow", RuntimeWarning)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_warnings); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_warn); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1298
 *             warnings.warn("Layer does not support" \
 *                     "OLCFastSetNextByIndex, negative step size may" \
 *                     " be slow", RuntimeWarning)             # <<<<<<<<<<<<<<
 *         self.stepsign = int(math.copysign(1, step))
 *         self.stop = stop
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1296, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1295
 *         if step == 0:
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:             # <<<<<<<<<<<<<<
 *             warnings.warn("Layer does not support" \
 *                     "OLCFastSetNextByIndex, negative step size may" \
 */
  }

  /* "fiona/ogrext.pyx":1299
 *                     "OLCFastSetNextByIndex, negative step size may" \
 *                     " be slow", RuntimeWarning)
 *         self.stepsign = int(math.copysign(1, step))             # <<<<<<<<<<<<<<
 *         self.stop = stop
 *         self.start = start
 */
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_math); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_copysign); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_10)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_int_1, __pyx_v_step};
    __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_int_1, __pyx_v_step};
    __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_1);
  } else
  #endif
  {
    __pyx_t_13 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_int_1);
    __Pyx_GIVEREF(__pyx_int_1);
    PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_12, __pyx_int_1);
    __Pyx_INCREF(__pyx_v_step);
    __Pyx_GIVEREF(__pyx_v_step);
    PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_12, __pyx_v_step);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_10);
  __Pyx_GOTREF(__pyx_v_self->stepsign);
  __Pyx_DECREF(__pyx_v_self->stepsign);
  __pyx_v_self->stepsign = __pyx_t_10;
  __pyx_t_10 = 0;

  /* "fiona/ogrext.pyx":1300
 *                     " be slow", RuntimeWarning)
 *         self.stepsign = int(math.copysign(1, step))
 *         self.stop = stop             # <<<<<<<<<<<<<<
 *         self.start = start
 *         self.step = step
 */
  __Pyx_INCREF(__pyx_v_stop);
  __Pyx_GIVEREF(__pyx_v_stop);
  __Pyx_GOTREF(__pyx_v_self->stop);
  __Pyx_DECREF(__pyx_v_self->stop);
  __pyx_v_self->stop = __pyx_v_stop;

  /* "fiona/ogrext.pyx":1301
 *         self.stepsign = int(math.copysign(1, step))
 *         self.stop = stop
 *         self.start = start             # <<<<<<<<<<<<<<
 *         self.step = step
 * 
 */
  __Pyx_INCREF(__pyx_v_start);
  __Pyx_GIVEREF(__pyx_v_start);
  __Pyx_GOTREF(__pyx_v_self->start);
  __Pyx_DECREF(__pyx_v_self->start);
  __pyx_v_self->start = __pyx_v_start;

  /* "fiona/ogrext.pyx":1302
 *         self.stop = stop
 *         self.start = start
 *         self.step = step             # <<<<<<<<<<<<<<
 * 
 *         self.next_index = start
 */
  __Pyx_INCREF(__pyx_v_step);
  __Pyx_GIVEREF(__pyx_v_step);
  __Pyx_GOTREF(__pyx_v_self->step);
  __Pyx_DECREF(__pyx_v_self->step);
  __pyx_v_self->step = __pyx_v_step;

  /* "fiona/ogrext.pyx":1304
 *         self.step = step
 * 
 *         self.next_index = start             # <<<<<<<<<<<<<<
 *         log.debug("Index: %d", self.next_index)
 *         OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 */
  __pyx_t_12 = __Pyx_PyInt_As_int(__pyx_v_start); if (unlikely((__pyx_t_12 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1304, __pyx_L1_error)
  __pyx_v_self->next_index = __pyx_t_12;

  /* "fiona/ogrext.pyx":1305
 * 
 *         self.next_index = start
 *         log.debug("Index: %d", self.next_index)             # <<<<<<<<<<<<<<
 *         OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_log); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_debug); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = NULL;
  __pyx_t_12 = 0;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_13))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_13);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_13, function);
      __pyx_t_12 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_13)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_kp_u_Index_d, __pyx_t_1};
    __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_13)) {
    PyObject *__pyx_temp[3] = {__pyx_t_9, __pyx_kp_u_Index_d, __pyx_t_1};
    __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_13, __pyx_temp+1-__pyx_t_12, 2+__pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else
  #endif
  {
    __pyx_t_14 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_u_Index_d);
    __Pyx_GIVEREF(__pyx_kp_u_Index_d);
    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_12, __pyx_kp_u_Index_d);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_12, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_13, __pyx_t_14, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1305, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "fiona/ogrext.pyx":1306
 *         self.next_index = start
 *         log.debug("Index: %d", self.next_index)
 *         OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)             # <<<<<<<<<<<<<<
 * 
 *     def __iter__(self):
 */
  (void)(OGR_L_SetNextByIndex(__pyx_v_session->cogr_layer, __pyx_v_self->next_index));

  /* "fiona/ogrext.pyx":1244
 *     cdef stepsign
 * 
 *     def __cinit__(self, collection, start=None, stop=None, step=None,             # <<<<<<<<<<<<<<
 *                   bbox=None, mask=None):
 *         if collection.session is None:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("fiona.ogrext.Iterator.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XDECREF(__pyx_v_ftcount);
  __Pyx_XDECREF(__pyx_v_start);
  __Pyx_XDECREF(__pyx_v_stop);
  __Pyx_XDECREF(__pyx_v_step);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1308
 *         OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_3__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_3__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator_2__iter__(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_2__iter__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "fiona/ogrext.pyx":1309
 * 
 *     def __iter__(self):
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def _next(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1308
 *         OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         return self
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1311
 *         return self
 * 
 *     def _next(self):             # <<<<<<<<<<<<<<
 *         """Internal method to set read cursor to next item"""
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_5_next(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_8Iterator_4_next[] = "Internal method to set read cursor to next item";
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_5_next(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_next (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator_4_next(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_4_next(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self) {
  struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_session = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  void *__pyx_v_cogr_feature;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  __Pyx_RefNannySetupContext("_next", 0);

  /* "fiona/ogrext.pyx":1315
 * 
 *         cdef Session session
 *         session = self.collection.session             # <<<<<<<<<<<<<<
 * 
 *         # Check if next_index is valid
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_Session))))) __PYX_ERR(0, 1315, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1318
 * 
 *         # Check if next_index is valid
 *         if self.next_index < 0:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  __pyx_t_2 = ((__pyx_v_self->next_index < 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1319
 *         # Check if next_index is valid
 *         if self.next_index < 0:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *         if self.stepsign == 1:
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 1319, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1318
 * 
 *         # Check if next_index is valid
 *         if self.next_index < 0:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  }

  /* "fiona/ogrext.pyx":1321
 *             raise StopIteration
 * 
 *         if self.stepsign == 1:             # <<<<<<<<<<<<<<
 *             if self.next_index < self.start or (self.stop is not None and self.next_index >= self.stop):
 *                 raise StopIteration
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_self->stepsign, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1322
 * 
 *         if self.stepsign == 1:
 *             if self.next_index < self.start or (self.stop is not None and self.next_index >= self.stop):             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 *         else:
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_self->start, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_4 = (__pyx_v_self->stop != Py_None);
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_v_self->stop, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_5;
    __pyx_L6_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "fiona/ogrext.pyx":1323
 *         if self.stepsign == 1:
 *             if self.next_index < self.start or (self.stop is not None and self.next_index >= self.stop):
 *                 raise StopIteration             # <<<<<<<<<<<<<<
 *         else:
 *             if self.next_index > self.start or (self.stop is not None and self.next_index <= self.stop):
 */
      __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
      __PYX_ERR(0, 1323, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1322
 * 
 *         if self.stepsign == 1:
 *             if self.next_index < self.start or (self.stop is not None and self.next_index >= self.stop):             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 *         else:
 */
    }

    /* "fiona/ogrext.pyx":1321
 *             raise StopIteration
 * 
 *         if self.stepsign == 1:             # <<<<<<<<<<<<<<
 *             if self.next_index < self.start or (self.stop is not None and self.next_index >= self.stop):
 *                 raise StopIteration
 */
    goto __pyx_L4;
  }

  /* "fiona/ogrext.pyx":1325
 *                 raise StopIteration
 *         else:
 *             if self.next_index > self.start or (self.stop is not None and self.next_index <= self.stop):             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_self->start, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (!__pyx_t_5) {
    } else {
      __pyx_t_2 = __pyx_t_5;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->stop != Py_None);
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {
    } else {
      __pyx_t_2 = __pyx_t_4;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_v_self->stop, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_4;
    __pyx_L10_bool_binop_done:;
    if (unlikely(__pyx_t_2)) {

      /* "fiona/ogrext.pyx":1326
 *         else:
 *             if self.next_index > self.start or (self.stop is not None and self.next_index <= self.stop):
 *                 raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *         # Set read cursor to next_item position
 */
      __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
      __PYX_ERR(0, 1326, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1325
 *                 raise StopIteration
 *         else:
 *             if self.next_index > self.start or (self.stop is not None and self.next_index <= self.stop):             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 * 
 */
    }
  }
  __pyx_L4:;

  /* "fiona/ogrext.pyx":1329
 * 
 *         # Set read cursor to next_item position
 *         if self.step > 1 and self.fastindex:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->step, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_self->fastindex); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1329, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L14_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1330
 *         # Set read cursor to next_item position
 *         if self.step > 1 and self.fastindex:
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)             # <<<<<<<<<<<<<<
 * 
 *         elif self.step > 1 and not self.fastindex and not self.next_index == self.start:
 */
    (void)(OGR_L_SetNextByIndex(__pyx_v_session->cogr_layer, __pyx_v_self->next_index));

    /* "fiona/ogrext.pyx":1329
 * 
 *         # Set read cursor to next_item position
 *         if self.step > 1 and self.fastindex:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
    goto __pyx_L13;
  }

  /* "fiona/ogrext.pyx":1332
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 *         elif self.step > 1 and not self.fastindex and not self.next_index == self.start:             # <<<<<<<<<<<<<<
 *             for _ in range(self.step - 1):
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->step, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_self->fastindex); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L16_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_self->start, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1332, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  __pyx_t_2 = __pyx_t_4;
  __pyx_L16_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1333
 * 
 *         elif self.step > 1 and not self.fastindex and not self.next_index == self.start:
 *             for _ in range(self.step - 1):             # <<<<<<<<<<<<<<
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 */
    __pyx_t_3 = __Pyx_PyInt_SubtractObjC(__pyx_v_self->step, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
      __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3); __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1333, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_7 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1333, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          if (__pyx_t_6 >= PyList_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1333, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          if (__pyx_t_6 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely(0 < 0)) __PYX_ERR(0, 1333, __pyx_L1_error)
          #else
          __pyx_t_1 = PySequence_ITEM(__pyx_t_3, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1333, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_7(__pyx_t_3);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1333, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fiona/ogrext.pyx":1335
 *             for _ in range(self.step - 1):
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)             # <<<<<<<<<<<<<<
 *                 if cogr_feature == NULL:
 *                     raise StopIteration
 */
      __pyx_v_cogr_feature = OGR_L_GetNextFeature(__pyx_v_session->cogr_layer);

      /* "fiona/ogrext.pyx":1336
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *                 if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *                     raise StopIteration
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:
 */
      __pyx_t_2 = ((__pyx_v_cogr_feature == NULL) != 0);
      if (unlikely(__pyx_t_2)) {

        /* "fiona/ogrext.pyx":1337
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *                 if cogr_feature == NULL:
 *                     raise StopIteration             # <<<<<<<<<<<<<<
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 */
        __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
        __PYX_ERR(0, 1337, __pyx_L1_error)

        /* "fiona/ogrext.pyx":1336
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *                 if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *                     raise StopIteration
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:
 */
      }

      /* "fiona/ogrext.pyx":1333
 * 
 *         elif self.step > 1 and not self.fastindex and not self.next_index == self.start:
 *             for _ in range(self.step - 1):             # <<<<<<<<<<<<<<
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 *                 cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 */
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1332
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 *         elif self.step > 1 and not self.fastindex and not self.next_index == self.start:             # <<<<<<<<<<<<<<
 *             for _ in range(self.step - 1):
 *                 # TODO rbuffat add test -> OGR_L_GetNextFeature increments cursor by 1, therefore self.step - 1 as one increment was performed when feature is read
 */
    goto __pyx_L13;
  }

  /* "fiona/ogrext.pyx":1338
 *                 if cogr_feature == NULL:
 *                     raise StopIteration
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
  __pyx_t_3 = PyObject_RichCompare(__pyx_v_self->step, __pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_self->fastindex); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L22_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_v_self->start, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __pyx_t_5;
  __pyx_L22_bool_binop_done:;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1339
 *                     raise StopIteration
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)             # <<<<<<<<<<<<<<
 * 
 *         elif self.step == 0:
 */
    (void)(OGR_L_SetNextByIndex(__pyx_v_session->cogr_layer, __pyx_v_self->next_index));

    /* "fiona/ogrext.pyx":1338
 *                 if cogr_feature == NULL:
 *                     raise StopIteration
 *         elif self.step > 1 and not self.fastindex and self.next_index == self.start:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
    goto __pyx_L13;
  }

  /* "fiona/ogrext.pyx":1341
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 *         elif self.step == 0:             # <<<<<<<<<<<<<<
 *             # OGR_L_GetNextFeature increments read cursor by one
 *             pass
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_self->step, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1341, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {
    goto __pyx_L13;
  }

  /* "fiona/ogrext.pyx":1344
 *             # OGR_L_GetNextFeature increments read cursor by one
 *             pass
 *         elif self.step < 0:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->step, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1345
 *             pass
 *         elif self.step < 0:
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)             # <<<<<<<<<<<<<<
 * 
 *         # set the next index
 */
    (void)(OGR_L_SetNextByIndex(__pyx_v_session->cogr_layer, __pyx_v_self->next_index));

    /* "fiona/ogrext.pyx":1344
 *             # OGR_L_GetNextFeature increments read cursor by one
 *             pass
 *         elif self.step < 0:             # <<<<<<<<<<<<<<
 *             OGR_L_SetNextByIndex(session.cogr_layer, self.next_index)
 * 
 */
  }
  __pyx_L13:;

  /* "fiona/ogrext.pyx":1348
 * 
 *         # set the next index
 *         self.next_index += self.step             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->next_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_v_self->step); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1348, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_self->next_index = __pyx_t_8;

  /* "fiona/ogrext.pyx":1311
 *         return self
 * 
 *     def _next(self):             # <<<<<<<<<<<<<<
 *         """Internal method to set read cursor to next item"""
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fiona.ogrext.Iterator._next", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1350
 *         self.next_index += self.step
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef OGRFeatureH cogr_feature = NULL
 *         cdef OGRLayerH cogr_layer = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_7__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_7__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator_6__next__(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_6__next__(struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self) {
  OGRFeatureH __pyx_v_cogr_feature;
  CYTHON_UNUSED OGRLayerH __pyx_v_cogr_layer;
  struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_session = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "fiona/ogrext.pyx":1351
 * 
 *     def __next__(self):
 *         cdef OGRFeatureH cogr_feature = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRLayerH cogr_layer = NULL
 *         cdef Session session
 */
  __pyx_v_cogr_feature = NULL;

  /* "fiona/ogrext.pyx":1352
 *     def __next__(self):
 *         cdef OGRFeatureH cogr_feature = NULL
 *         cdef OGRLayerH cogr_layer = NULL             # <<<<<<<<<<<<<<
 *         cdef Session session
 * 
 */
  __pyx_v_cogr_layer = NULL;

  /* "fiona/ogrext.pyx":1355
 *         cdef Session session
 * 
 *         session = self.collection.session             # <<<<<<<<<<<<<<
 * 
 *         if not session or not session.isactive:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_Session))))) __PYX_ERR(0, 1355, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1357
 *         session = self.collection.session
 * 
 *         if not session or not session.isactive:             # <<<<<<<<<<<<<<
 *             raise FionaValueError("Session is inactive, dataset is closed or layer is unavailable.")
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_session)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_3) != 0);
  if (!__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_session), __pyx_n_s_isactive); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!__pyx_t_4) != 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1358
 * 
 *         if not session or not session.isactive:
 *             raise FionaValueError("Session is inactive, dataset is closed or layer is unavailable.")             # <<<<<<<<<<<<<<
 * 
 *         # Update read cursor
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_FionaValueError); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_Session_is_inactive_dataset_is_c) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_Session_is_inactive_dataset_is_c);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1358, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1357
 *         session = self.collection.session
 * 
 *         if not session or not session.isactive:             # <<<<<<<<<<<<<<
 *             raise FionaValueError("Session is inactive, dataset is closed or layer is unavailable.")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1361
 * 
 *         # Update read cursor
 *         self._next()             # <<<<<<<<<<<<<<
 * 
 *         # Get the next feature.
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_next); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_1 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1361, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1364
 * 
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)             # <<<<<<<<<<<<<<
 *         if cogr_feature == NULL:
 *             raise StopIteration
 */
  __pyx_v_cogr_feature = OGR_L_GetNextFeature(__pyx_v_session->cogr_layer);

  /* "fiona/ogrext.pyx":1365
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  __pyx_t_2 = ((__pyx_v_cogr_feature == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1366
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 1366, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1365
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  }

  /* "fiona/ogrext.pyx":1368
 *             raise StopIteration
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             return FeatureBuilder().build(
 *                 cogr_feature,
 */
  /*try:*/ {

    /* "fiona/ogrext.pyx":1369
 * 
 *         try:
 *             return FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self.collection.session._get_internal_encoding(),
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1369, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "fiona/ogrext.pyx":1371
 *             return FeatureBuilder().build(
 *                 cogr_feature,
 *                 encoding=self.collection.session._get_internal_encoding(),             # <<<<<<<<<<<<<<
 *                 bbox=False,
 *                 driver=self.collection.driver,
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_session); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1371, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1371, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6) : __Pyx_PyObject_CallNoArg(__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1371, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "fiona/ogrext.pyx":1373
 *                 encoding=self.collection.session._get_internal_encoding(),
 *                 bbox=False,
 *                 driver=self.collection.driver,             # <<<<<<<<<<<<<<
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1373, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "fiona/ogrext.pyx":1374
 *                 bbox=False,
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,             # <<<<<<<<<<<<<<
 *                 ignore_geometry=self.collection.ignore_geometry,
 *             )
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1374, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "fiona/ogrext.pyx":1375
 *                 driver=self.collection.driver,
 *                 ignore_fields=self.collection.ignore_fields,
 *                 ignore_geometry=self.collection.ignore_geometry,             # <<<<<<<<<<<<<<
 *             )
 *         finally:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->collection, __pyx_n_s_ignore_geometry); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1375, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_8);

    /* "fiona/ogrext.pyx":1369
 * 
 *         try:
 *             return FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *                 cogr_feature,
 *                 encoding=self.collection.session._get_internal_encoding(),
 */
    __pyx_t_10.__pyx_n = 5;
    __pyx_t_10.encoding = __pyx_t_5;
    __pyx_t_10.bbox = Py_False;
    __pyx_t_10.driver = __pyx_t_7;
    __pyx_t_10.ignore_fields = __pyx_t_6;
    __pyx_t_10.ignore_geometry = __pyx_t_8;
    __pyx_t_9 = ((struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1), __pyx_v_cogr_feature, &__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1369, __pyx_L8_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L7_return;
  }

  /* "fiona/ogrext.pyx":1378
 *             )
 *         finally:
 *             _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*finally:*/ {
    __pyx_L8_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
      {
        __pyx_t_9 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1378, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
      goto __pyx_L1_error;
      __pyx_L11_error:;
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L7_return: {
      __pyx_t_19 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_9 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1378, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_r = __pyx_t_19;
      __pyx_t_19 = 0;
      goto __pyx_L0;
    }
  }

  /* "fiona/ogrext.pyx":1350
 *         self.next_index += self.step
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef OGRFeatureH cogr_feature = NULL
 *         cdef OGRLayerH cogr_layer = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fiona.ogrext.Iterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_9__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator_8__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_8__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Iterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_8Iterator_11__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_8Iterator_10__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_Iterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_8Iterator_10__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_Iterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.Iterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1383
 * cdef class ItemsIterator(Iterator):
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 * 
 *         cdef long fid
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_13ItemsIterator___next__(((struct __pyx_obj_5fiona_6ogrext_ItemsIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator___next__(struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self) {
  long __pyx_v_fid;
  void *__pyx_v_cogr_feature;
  struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_session = 0;
  PyObject *__pyx_v_feature = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build __pyx_t_8;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "fiona/ogrext.pyx":1388
 *         cdef void * cogr_feature
 *         cdef Session session
 *         session = self.collection.session             # <<<<<<<<<<<<<<
 * 
 *         #Update read cursor
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_Session))))) __PYX_ERR(0, 1388, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1391
 * 
 *         #Update read cursor
 *         self._next()             # <<<<<<<<<<<<<<
 * 
 *         # Get the next feature.
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_next); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1391, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1394
 * 
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)             # <<<<<<<<<<<<<<
 *         if cogr_feature == NULL:
 *             raise StopIteration
 */
  __pyx_v_cogr_feature = OGR_L_GetNextFeature(__pyx_v_session->cogr_layer);

  /* "fiona/ogrext.pyx":1395
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  __pyx_t_4 = ((__pyx_v_cogr_feature == NULL) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "fiona/ogrext.pyx":1396
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *         fid = OGR_F_GetFID(cogr_feature)
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 1396, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1395
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  }

  /* "fiona/ogrext.pyx":1398
 *             raise StopIteration
 * 
 *         fid = OGR_F_GetFID(cogr_feature)             # <<<<<<<<<<<<<<
 *         feature = FeatureBuilder().build(
 *             cogr_feature,
 */
  __pyx_v_fid = OGR_F_GetFID(__pyx_v_cogr_feature);

  /* "fiona/ogrext.pyx":1399
 * 
 *         fid = OGR_F_GetFID(cogr_feature)
 *         feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *             cogr_feature,
 *             encoding=self.collection.session._get_internal_encoding(),
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "fiona/ogrext.pyx":1401
 *         feature = FeatureBuilder().build(
 *             cogr_feature,
 *             encoding=self.collection.session._get_internal_encoding(),             # <<<<<<<<<<<<<<
 *             bbox=False,
 *             driver=self.collection.driver,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_session); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_get_internal_encoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_2 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1401, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":1403
 *             encoding=self.collection.session._get_internal_encoding(),
 *             bbox=False,
 *             driver=self.collection.driver,             # <<<<<<<<<<<<<<
 *             ignore_fields=self.collection.ignore_fields,
 *             ignore_geometry=self.collection.ignore_geometry,
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_driver); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "fiona/ogrext.pyx":1404
 *             bbox=False,
 *             driver=self.collection.driver,
 *             ignore_fields=self.collection.ignore_fields,             # <<<<<<<<<<<<<<
 *             ignore_geometry=self.collection.ignore_geometry,
 *         )
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_ignore_fields); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1404, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "fiona/ogrext.pyx":1405
 *             driver=self.collection.driver,
 *             ignore_fields=self.collection.ignore_fields,
 *             ignore_geometry=self.collection.ignore_geometry,             # <<<<<<<<<<<<<<
 *         )
 *         _deleteOgrFeature(cogr_feature)
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_ignore_geometry); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "fiona/ogrext.pyx":1399
 * 
 *         fid = OGR_F_GetFID(cogr_feature)
 *         feature = FeatureBuilder().build(             # <<<<<<<<<<<<<<
 *             cogr_feature,
 *             encoding=self.collection.session._get_internal_encoding(),
 */
  __pyx_t_8.__pyx_n = 5;
  __pyx_t_8.encoding = __pyx_t_2;
  __pyx_t_8.bbox = Py_False;
  __pyx_t_8.driver = __pyx_t_5;
  __pyx_t_8.ignore_fields = __pyx_t_3;
  __pyx_t_8.ignore_geometry = __pyx_t_6;
  __pyx_t_7 = ((struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder *)((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_t_1), __pyx_v_cogr_feature, &__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_feature = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "fiona/ogrext.pyx":1407
 *             ignore_geometry=self.collection.ignore_geometry,
 *         )
 *         _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 * 
 *         return fid, feature
 */
  __pyx_t_7 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "fiona/ogrext.pyx":1409
 *         _deleteOgrFeature(cogr_feature)
 * 
 *         return fid, feature             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_7 = __Pyx_PyInt_From_long(__pyx_v_fid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_7);
  __Pyx_INCREF(__pyx_v_feature);
  __Pyx_GIVEREF(__pyx_v_feature);
  PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_v_feature);
  __pyx_t_7 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1383
 * cdef class ItemsIterator(Iterator):
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 * 
 *         cdef long fid
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("fiona.ogrext.ItemsIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_13ItemsIterator_2__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_ItemsIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.ItemsIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_13ItemsIterator_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_13ItemsIterator_4__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_ItemsIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_13ItemsIterator_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_ItemsIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.ItemsIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1414
 * cdef class KeysIterator(Iterator):
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef long fid
 *         cdef void * cogr_feature
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_12KeysIterator___next__(((struct __pyx_obj_5fiona_6ogrext_KeysIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator___next__(struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self) {
  long __pyx_v_fid;
  void *__pyx_v_cogr_feature;
  struct __pyx_obj_5fiona_6ogrext_Session *__pyx_v_session = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "fiona/ogrext.pyx":1418
 *         cdef void * cogr_feature
 *         cdef Session session
 *         session = self.collection.session             # <<<<<<<<<<<<<<
 * 
 *         #Update read cursor
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base.collection, __pyx_n_s_session); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_5fiona_6ogrext_Session))))) __PYX_ERR(0, 1418, __pyx_L1_error)
  __pyx_v_session = ((struct __pyx_obj_5fiona_6ogrext_Session *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1421
 * 
 *         #Update read cursor
 *         self._next()             # <<<<<<<<<<<<<<
 * 
 *         # Get the next feature.
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_next); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1424
 * 
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)             # <<<<<<<<<<<<<<
 *         if cogr_feature == NULL:
 *             raise StopIteration
 */
  __pyx_v_cogr_feature = OGR_L_GetNextFeature(__pyx_v_session->cogr_layer);

  /* "fiona/ogrext.pyx":1425
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  __pyx_t_4 = ((__pyx_v_cogr_feature == NULL) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "fiona/ogrext.pyx":1426
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:
 *             raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *         fid = OGR_F_GetFID(cogr_feature)
 */
    __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
    __PYX_ERR(0, 1426, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1425
 *         # Get the next feature.
 *         cogr_feature = OGR_L_GetNextFeature(session.cogr_layer)
 *         if cogr_feature == NULL:             # <<<<<<<<<<<<<<
 *             raise StopIteration
 * 
 */
  }

  /* "fiona/ogrext.pyx":1428
 *             raise StopIteration
 * 
 *         fid = OGR_F_GetFID(cogr_feature)             # <<<<<<<<<<<<<<
 *         _deleteOgrFeature(cogr_feature)
 * 
 */
  __pyx_v_fid = OGR_F_GetFID(__pyx_v_cogr_feature);

  /* "fiona/ogrext.pyx":1429
 * 
 *         fid = OGR_F_GetFID(cogr_feature)
 *         _deleteOgrFeature(cogr_feature)             # <<<<<<<<<<<<<<
 * 
 *         return fid
 */
  __pyx_t_1 = __pyx_f_5fiona_6ogrext__deleteOgrFeature(__pyx_v_cogr_feature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1431
 *         _deleteOgrFeature(cogr_feature)
 * 
 *         return fid             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v_fid); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1431, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1414
 * cdef class KeysIterator(Iterator):
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef long fid
 *         cdef void * cogr_feature
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("fiona.ogrext.KeysIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_session);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_3__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_12KeysIterator_2__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_KeysIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator_2__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.KeysIterator.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_12KeysIterator_5__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_12KeysIterator_4__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_KeysIterator *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_12KeysIterator_4__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_5fiona_6ogrext_KeysIterator *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.KeysIterator.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1434
 * 
 * 
 * def _remove(path, driver=None):             # <<<<<<<<<<<<<<
 *     """Deletes an OGR data source
 *     """
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_8_remove(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_7_remove[] = "Deletes an OGR data source\n    ";
static PyMethodDef __pyx_mdef_5fiona_6ogrext_8_remove = {"_remove", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_8_remove, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_7_remove};
static PyObject *__pyx_pw_5fiona_6ogrext_8_remove(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_driver = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_remove (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_driver,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_driver);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_remove") < 0)) __PYX_ERR(0, 1434, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_path = values[0];
    __pyx_v_driver = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_remove", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1434, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext._remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_7_remove(__pyx_self, __pyx_v_path, __pyx_v_driver);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_7_remove(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_driver) {
  void *__pyx_v_cogr_driver;
  void *__pyx_v_cogr_ds;
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  char const *__pyx_t_9;
  void *__pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  char *__pyx_t_17;
  char const *__pyx_t_18;
  __Pyx_RefNannySetupContext("_remove", 0);

  /* "fiona/ogrext.pyx":1442
 *     cdef char *driver_c
 * 
 *     if driver is None:             # <<<<<<<<<<<<<<
 *         # attempt to identify the driver by opening the dataset
 *         try:
 */
  __pyx_t_1 = (__pyx_v_driver == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1444
 *     if driver is None:
 *         # attempt to identify the driver by opening the dataset
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})
 *         except (DriverError, FionaNullPointerError):
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "fiona/ogrext.pyx":1445
 *         # attempt to identify the driver by opening the dataset
 *         try:
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})             # <<<<<<<<<<<<<<
 *         except (DriverError, FionaNullPointerError):
 *             raise DatasetDeleteError("Failed to remove data source {}".format(path))
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_8 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_8)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_8);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
          }
        }
        __pyx_t_6 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u_utf_8);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_t_9 = __Pyx_PyObject_AsString(__pyx_t_6); if (unlikely((!__pyx_t_9) && PyErr_Occurred())) __PYX_ERR(0, 1445, __pyx_L4_error)
        __pyx_t_7 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1445, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_10 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_t_9, 0, Py_None, __pyx_t_7); if (unlikely(__pyx_t_10 == ((void *)NULL))) __PYX_ERR(0, 1445, __pyx_L4_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_cogr_ds = __pyx_t_10;

        /* "fiona/ogrext.pyx":1444
 *     if driver is None:
 *         # attempt to identify the driver by opening the dataset
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})
 *         except (DriverError, FionaNullPointerError):
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "fiona/ogrext.pyx":1446
 *         try:
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})
 *         except (DriverError, FionaNullPointerError):             # <<<<<<<<<<<<<<
 *             raise DatasetDeleteError("Failed to remove data source {}".format(path))
 *         cogr_driver = GDALGetDatasetDriver(cogr_ds)
 */
      __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_6, &__pyx_t_8);
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_DriverError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1446, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_FionaNullPointerError); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1446, __pyx_L6_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_11) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_7, __pyx_t_12);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_6, __pyx_t_8);
      __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_8 = 0;
      if (__pyx_t_13) {
        __Pyx_AddTraceback("fiona.ogrext._remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_8, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 1446, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GOTREF(__pyx_t_7);

        /* "fiona/ogrext.pyx":1447
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})
 *         except (DriverError, FionaNullPointerError):
 *             raise DatasetDeleteError("Failed to remove data source {}".format(path))             # <<<<<<<<<<<<<<
 *         cogr_driver = GDALGetDatasetDriver(cogr_ds)
 *         GDALClose(cogr_ds)
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1447, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Failed_to_remove_data_source, __pyx_n_s_format); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1447, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_15);
        __pyx_t_16 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_15))) {
          __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_15);
          if (likely(__pyx_t_16)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_15);
            __Pyx_INCREF(__pyx_t_16);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_15, function);
          }
        }
        __pyx_t_14 = (__pyx_t_16) ? __Pyx_PyObject_Call2Args(__pyx_t_15, __pyx_t_16, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_15, __pyx_v_path);
        __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
        if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1447, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = NULL;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
          }
        }
        __pyx_t_12 = (__pyx_t_15) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_15, __pyx_t_14) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_14);
        __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1447, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_Raise(__pyx_t_12, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __PYX_ERR(0, 1447, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "fiona/ogrext.pyx":1444
 *     if driver is None:
 *         # attempt to identify the driver by opening the dataset
 *         try:             # <<<<<<<<<<<<<<
 *             cogr_ds = gdal_open_vector(path.encode("utf-8"), 0, None, {})
 *         except (DriverError, FionaNullPointerError):
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "fiona/ogrext.pyx":1448
 *         except (DriverError, FionaNullPointerError):
 *             raise DatasetDeleteError("Failed to remove data source {}".format(path))
 *         cogr_driver = GDALGetDatasetDriver(cogr_ds)             # <<<<<<<<<<<<<<
 *         GDALClose(cogr_ds)
 *     else:
 */
    __pyx_v_cogr_driver = GDALGetDatasetDriver(__pyx_v_cogr_ds);

    /* "fiona/ogrext.pyx":1449
 *             raise DatasetDeleteError("Failed to remove data source {}".format(path))
 *         cogr_driver = GDALGetDatasetDriver(cogr_ds)
 *         GDALClose(cogr_ds)             # <<<<<<<<<<<<<<
 *     else:
 *         cogr_driver = OGRGetDriverByName(driver.encode("utf-8"))
 */
    GDALClose(__pyx_v_cogr_ds);

    /* "fiona/ogrext.pyx":1442
 *     cdef char *driver_c
 * 
 *     if driver is None:             # <<<<<<<<<<<<<<
 *         # attempt to identify the driver by opening the dataset
 *         try:
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1451
 *         GDALClose(cogr_ds)
 *     else:
 *         cogr_driver = OGRGetDriverByName(driver.encode("utf-8"))             # <<<<<<<<<<<<<<
 * 
 *     if cogr_driver == NULL:
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_driver, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_17 = __Pyx_PyObject_AsWritableString(__pyx_t_7); if (unlikely((!__pyx_t_17) && PyErr_Occurred())) __PYX_ERR(0, 1451, __pyx_L1_error)
    __pyx_v_cogr_driver = OGRGetDriverByName(__pyx_t_17);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1453
 *         cogr_driver = OGRGetDriverByName(driver.encode("utf-8"))
 * 
 *     if cogr_driver == NULL:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Null driver when attempting to delete {}".format(path))
 * 
 */
  __pyx_t_2 = ((__pyx_v_cogr_driver == NULL) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1454
 * 
 *     if cogr_driver == NULL:
 *         raise DatasetDeleteError("Null driver when attempting to delete {}".format(path))             # <<<<<<<<<<<<<<
 * 
 *     if not OGR_Dr_TestCapability(cogr_driver, ODrCDeleteDataSource):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Null_driver_when_attempting_to_d, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_8 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_12, __pyx_t_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_8);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1454, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1454, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1453
 *         cogr_driver = OGRGetDriverByName(driver.encode("utf-8"))
 * 
 *     if cogr_driver == NULL:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Null driver when attempting to delete {}".format(path))
 * 
 */
  }

  /* "fiona/ogrext.pyx":1456
 *         raise DatasetDeleteError("Null driver when attempting to delete {}".format(path))
 * 
 *     if not OGR_Dr_TestCapability(cogr_driver, ODrCDeleteDataSource):             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Driver does not support dataset removal operation")
 * 
 */
  __pyx_t_2 = ((!(OGR_Dr_TestCapability(__pyx_v_cogr_driver, __pyx_v_5fiona_6ogrext_ODrCDeleteDataSource) != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1457
 * 
 *     if not OGR_Dr_TestCapability(cogr_driver, ODrCDeleteDataSource):
 *         raise DatasetDeleteError("Driver does not support dataset removal operation")             # <<<<<<<<<<<<<<
 * 
 *     result = GDALDeleteDataset(cogr_driver, path.encode('utf-8'))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_kp_u_Driver_does_not_support_dataset) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_Driver_does_not_support_dataset);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 1457, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1456
 *         raise DatasetDeleteError("Null driver when attempting to delete {}".format(path))
 * 
 *     if not OGR_Dr_TestCapability(cogr_driver, ODrCDeleteDataSource):             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Driver does not support dataset removal operation")
 * 
 */
  }

  /* "fiona/ogrext.pyx":1459
 *         raise DatasetDeleteError("Driver does not support dataset removal operation")
 * 
 *     result = GDALDeleteDataset(cogr_driver, path.encode('utf-8'))             # <<<<<<<<<<<<<<
 *     if result != OGRERR_NONE:
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
    }
  }
  __pyx_t_7 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_8, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_18 = __Pyx_PyObject_AsString(__pyx_t_7); if (unlikely((!__pyx_t_18) && PyErr_Occurred())) __PYX_ERR(0, 1459, __pyx_L1_error)
  __pyx_v_result = GDALDeleteDataset(__pyx_v_cogr_driver, __pyx_t_18);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "fiona/ogrext.pyx":1460
 * 
 *     result = GDALDeleteDataset(cogr_driver, path.encode('utf-8'))
 *     if result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OGRERR_NONE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_8 = PyObject_RichCompare(__pyx_t_7, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1461
 *     result = GDALDeleteDataset(cogr_driver, path.encode('utf-8'))
 *     if result != OGRERR_NONE:
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Failed_to_remove_data_source, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
      }
    }
    __pyx_t_7 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_11, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_12)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_12);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_8 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_12, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 1461, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1460
 * 
 *     result = GDALDeleteDataset(cogr_driver, path.encode('utf-8'))
 *     if result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 * 
 */
  }

  /* "fiona/ogrext.pyx":1434
 * 
 * 
 * def _remove(path, driver=None):             # <<<<<<<<<<<<<<
 *     """Deletes an OGR data source
 *     """
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("fiona.ogrext._remove", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1464
 * 
 * 
 * def _remove_layer(path, layer, driver=None):             # <<<<<<<<<<<<<<
 *     cdef void *cogr_ds
 *     cdef int layer_index
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_10_remove_layer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_10_remove_layer = {"_remove_layer", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_10_remove_layer, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_10_remove_layer(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_layer = 0;
  CYTHON_UNUSED PyObject *__pyx_v_driver = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_remove_layer (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,&__pyx_n_s_layer,&__pyx_n_s_driver,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_layer)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_remove_layer", 0, 2, 3, 1); __PYX_ERR(0, 1464, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_driver);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_remove_layer") < 0)) __PYX_ERR(0, 1464, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_path = values[0];
    __pyx_v_layer = values[1];
    __pyx_v_driver = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_remove_layer", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1464, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext._remove_layer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_9_remove_layer(__pyx_self, __pyx_v_path, __pyx_v_layer, __pyx_v_driver);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_9_remove_layer(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_layer, CYTHON_UNUSED PyObject *__pyx_v_driver) {
  void *__pyx_v_cogr_ds;
  int __pyx_v_layer_index;
  PyObject *__pyx_v_layer_str = NULL;
  PyObject *__pyx_v_layer_names = NULL;
  int __pyx_v_result;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  Py_ssize_t __pyx_t_14;
  char const *__pyx_t_15;
  void *__pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  __Pyx_RefNannySetupContext("_remove_layer", 0);

  /* "fiona/ogrext.pyx":1468
 *     cdef int layer_index
 * 
 *     if isinstance(layer, integer_types):             # <<<<<<<<<<<<<<
 *         layer_index = layer
 *         layer_str = str(layer_index)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_integer_types); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_layer, __pyx_t_1); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 1468, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1469
 * 
 *     if isinstance(layer, integer_types):
 *         layer_index = layer             # <<<<<<<<<<<<<<
 *         layer_str = str(layer_index)
 *     else:
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_layer); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1469, __pyx_L1_error)
    __pyx_v_layer_index = __pyx_t_4;

    /* "fiona/ogrext.pyx":1470
 *     if isinstance(layer, integer_types):
 *         layer_index = layer
 *         layer_str = str(layer_index)             # <<<<<<<<<<<<<<
 *     else:
 *         layer_names = _listlayers(path)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_layer_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_CallOneArg(((PyObject *)(&PyUnicode_Type)), __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1470, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_layer_str = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "fiona/ogrext.pyx":1468
 *     cdef int layer_index
 * 
 *     if isinstance(layer, integer_types):             # <<<<<<<<<<<<<<
 *         layer_index = layer
 *         layer_str = str(layer_index)
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1472
 *         layer_str = str(layer_index)
 *     else:
 *         layer_names = _listlayers(path)             # <<<<<<<<<<<<<<
 *         try:
 *             layer_index = layer_names.index(layer)
 */
  /*else*/ {
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_listlayers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_layer_names = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "fiona/ogrext.pyx":1473
 *     else:
 *         layer_names = _listlayers(path)
 *         try:             # <<<<<<<<<<<<<<
 *             layer_index = layer_names.index(layer)
 *         except ValueError:
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      /*try:*/ {

        /* "fiona/ogrext.pyx":1474
 *         layer_names = _listlayers(path)
 *         try:
 *             layer_index = layer_names.index(layer)             # <<<<<<<<<<<<<<
 *         except ValueError:
 *             raise ValueError("Layer \"{}\" does not exist in datasource: {}".format(layer, path))
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_layer_names, __pyx_n_s_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1474, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_6 = NULL;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
          }
        }
        __pyx_t_5 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_6, __pyx_v_layer) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_layer);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1474, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1474, __pyx_L4_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_layer_index = __pyx_t_4;

        /* "fiona/ogrext.pyx":1473
 *     else:
 *         layer_names = _listlayers(path)
 *         try:             # <<<<<<<<<<<<<<
 *             layer_index = layer_names.index(layer)
 *         except ValueError:
 */
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "fiona/ogrext.pyx":1475
 *         try:
 *             layer_index = layer_names.index(layer)
 *         except ValueError:             # <<<<<<<<<<<<<<
 *             raise ValueError("Layer \"{}\" does not exist in datasource: {}".format(layer, path))
 *         layer_str = '"{}"'.format(layer)
 */
      __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ValueError);
      if (__pyx_t_4) {
        __Pyx_AddTraceback("fiona.ogrext._remove_layer", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_1, &__pyx_t_6) < 0) __PYX_ERR(0, 1475, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GOTREF(__pyx_t_6);

        /* "fiona/ogrext.pyx":1476
 *             layer_index = layer_names.index(layer)
 *         except ValueError:
 *             raise ValueError("Layer \"{}\" does not exist in datasource: {}".format(layer, path))             # <<<<<<<<<<<<<<
 *         layer_str = '"{}"'.format(layer)
 * 
 */
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Layer_does_not_exist_in_datasour, __pyx_n_s_format); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_12 = NULL;
        __pyx_t_4 = 0;
        if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_4 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_layer, __pyx_v_path};
          __pyx_t_10 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_10);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
          PyObject *__pyx_temp[3] = {__pyx_t_12, __pyx_v_layer, __pyx_v_path};
          __pyx_t_10 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_GOTREF(__pyx_t_10);
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_12) {
            __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_12); __pyx_t_12 = NULL;
          }
          __Pyx_INCREF(__pyx_v_layer);
          __Pyx_GIVEREF(__pyx_v_layer);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_4, __pyx_v_layer);
          __Pyx_INCREF(__pyx_v_path);
          __Pyx_GIVEREF(__pyx_v_path);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_4, __pyx_v_path);
          __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_13, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1476, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 1476, __pyx_L6_except_error)
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "fiona/ogrext.pyx":1473
 *     else:
 *         layer_names = _listlayers(path)
 *         try:             # <<<<<<<<<<<<<<
 *             layer_index = layer_names.index(layer)
 *         except ValueError:
 */
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ExceptionReset(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      goto __pyx_L1_error;
      __pyx_L9_try_end:;
    }

    /* "fiona/ogrext.pyx":1477
 *         except ValueError:
 *             raise ValueError("Layer \"{}\" does not exist in datasource: {}".format(layer, path))
 *         layer_str = '"{}"'.format(layer)             # <<<<<<<<<<<<<<
 * 
 *     if layer_index < 0:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__40, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_layer) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_layer);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1477, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_layer_str = __pyx_t_6;
    __pyx_t_6 = 0;
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1479
 *         layer_str = '"{}"'.format(layer)
 * 
 *     if layer_index < 0:             # <<<<<<<<<<<<<<
 *         layer_names = _listlayers(path)
 *         layer_index = len(layer_names) + layer_index
 */
  __pyx_t_3 = ((__pyx_v_layer_index < 0) != 0);
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1480
 * 
 *     if layer_index < 0:
 *         layer_names = _listlayers(path)             # <<<<<<<<<<<<<<
 *         layer_index = len(layer_names) + layer_index
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_listlayers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_path);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1480, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_layer_names, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "fiona/ogrext.pyx":1481
 *     if layer_index < 0:
 *         layer_names = _listlayers(path)
 *         layer_index = len(layer_names) + layer_index             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
    __pyx_t_14 = PyObject_Length(__pyx_v_layer_names); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1481, __pyx_L1_error)
    __pyx_v_layer_index = (__pyx_t_14 + __pyx_v_layer_index);

    /* "fiona/ogrext.pyx":1479
 *         layer_str = '"{}"'.format(layer)
 * 
 *     if layer_index < 0:             # <<<<<<<<<<<<<<
 *         layer_names = _listlayers(path)
 *         layer_index = len(layer_names) + layer_index
 */
  }

  /* "fiona/ogrext.pyx":1483
 *         layer_index = len(layer_names) + layer_index
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})
 *     except (DriverError, FionaNullPointerError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_9, &__pyx_t_8, &__pyx_t_7);
    __Pyx_XGOTREF(__pyx_t_9);
    __Pyx_XGOTREF(__pyx_t_8);
    __Pyx_XGOTREF(__pyx_t_7);
    /*try:*/ {

      /* "fiona/ogrext.pyx":1484
 * 
 *     try:
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})             # <<<<<<<<<<<<<<
 *     except (DriverError, FionaNullPointerError):
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
        }
      }
      __pyx_t_6 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_kp_u_utf_8);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1484, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_15 = __Pyx_PyObject_AsString(__pyx_t_6); if (unlikely((!__pyx_t_15) && PyErr_Occurred())) __PYX_ERR(0, 1484, __pyx_L13_error)
      __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1484, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_16 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_t_15, 1, Py_None, __pyx_t_1); if (unlikely(__pyx_t_16 == ((void *)NULL))) __PYX_ERR(0, 1484, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_cogr_ds = __pyx_t_16;

      /* "fiona/ogrext.pyx":1483
 *         layer_index = len(layer_names) + layer_index
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})
 *     except (DriverError, FionaNullPointerError):
 */
    }
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    goto __pyx_L18_try_end;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fiona/ogrext.pyx":1485
 *     try:
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})
 *     except (DriverError, FionaNullPointerError):             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_6, &__pyx_t_5);
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_DriverError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1485, __pyx_L15_except_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_FionaNullPointerError); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1485, __pyx_L15_except_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_4 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_11) || __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_6, __pyx_t_5);
    __pyx_t_1 = 0; __pyx_t_6 = 0; __pyx_t_5 = 0;
    if (__pyx_t_4) {
      __Pyx_AddTraceback("fiona.ogrext._remove_layer", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_1) < 0) __PYX_ERR(0, 1485, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_1);

      /* "fiona/ogrext.pyx":1486
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})
 *     except (DriverError, FionaNullPointerError):
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))             # <<<<<<<<<<<<<<
 * 
 *     result = OGR_DS_DeleteLayer(cogr_ds, layer_index)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1486, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Failed_to_remove_data_source, __pyx_n_s_format); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1486, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_17 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_17)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_17);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
        }
      }
      __pyx_t_13 = (__pyx_t_17) ? __Pyx_PyObject_Call2Args(__pyx_t_12, __pyx_t_17, __pyx_v_path) : __Pyx_PyObject_CallOneArg(__pyx_t_12, __pyx_v_path);
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1486, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_12 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_11, function);
        }
      }
      __pyx_t_10 = (__pyx_t_12) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_12, __pyx_t_13) : __Pyx_PyObject_CallOneArg(__pyx_t_11, __pyx_t_13);
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1486, __pyx_L15_except_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(0, 1486, __pyx_L15_except_error)
    }
    goto __pyx_L15_except_error;
    __pyx_L15_except_error:;

    /* "fiona/ogrext.pyx":1483
 *         layer_index = len(layer_names) + layer_index
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         cogr_ds = gdal_open_vector(path.encode("utf-8"), 1, None, {})
 *     except (DriverError, FionaNullPointerError):
 */
    __Pyx_XGIVEREF(__pyx_t_9);
    __Pyx_XGIVEREF(__pyx_t_8);
    __Pyx_XGIVEREF(__pyx_t_7);
    __Pyx_ExceptionReset(__pyx_t_9, __pyx_t_8, __pyx_t_7);
    goto __pyx_L1_error;
    __pyx_L18_try_end:;
  }

  /* "fiona/ogrext.pyx":1488
 *         raise DatasetDeleteError("Failed to remove data source {}".format(path))
 * 
 *     result = OGR_DS_DeleteLayer(cogr_ds, layer_index)             # <<<<<<<<<<<<<<
 *     GDALClose(cogr_ds)
 *     if result == OGRERR_UNSUPPORTED_OPERATION:
 */
  __pyx_v_result = OGR_DS_DeleteLayer(__pyx_v_cogr_ds, __pyx_v_layer_index);

  /* "fiona/ogrext.pyx":1489
 * 
 *     result = OGR_DS_DeleteLayer(cogr_ds, layer_index)
 *     GDALClose(cogr_ds)             # <<<<<<<<<<<<<<
 *     if result == OGRERR_UNSUPPORTED_OPERATION:
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 */
  GDALClose(__pyx_v_cogr_ds);

  /* "fiona/ogrext.pyx":1490
 *     result = OGR_DS_DeleteLayer(cogr_ds, layer_index)
 *     GDALClose(cogr_ds)
 *     if result == OGRERR_UNSUPPORTED_OPERATION:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 *     elif result != OGRERR_NONE:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OGRERR_UNSUPPORTED_OPERATION); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_6, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1490, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1491
 *     GDALClose(cogr_ds)
 *     if result == OGRERR_UNSUPPORTED_OPERATION:
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))             # <<<<<<<<<<<<<<
 *     elif result != OGRERR_NONE:
 *         raise DatasetDeleteError("Failed to remove layer {} from datasource: {}".format(layer_str, path))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Removal_of_layer_not_supported_b, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    __pyx_t_1 = (__pyx_t_11) ? __Pyx_PyObject_Call2Args(__pyx_t_10, __pyx_t_11, __pyx_v_layer_str) : __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_v_layer_str);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_t_1) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_1);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_5, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __PYX_ERR(0, 1491, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1490
 *     result = OGR_DS_DeleteLayer(cogr_ds, layer_index)
 *     GDALClose(cogr_ds)
 *     if result == OGRERR_UNSUPPORTED_OPERATION:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 *     elif result != OGRERR_NONE:
 */
  }

  /* "fiona/ogrext.pyx":1492
 *     if result == OGRERR_UNSUPPORTED_OPERATION:
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 *     elif result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Failed to remove layer {} from datasource: {}".format(layer_str, path))
 * 
 */
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OGRERR_NONE); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 1492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_3)) {

    /* "fiona/ogrext.pyx":1493
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 *     elif result != OGRERR_NONE:
 *         raise DatasetDeleteError("Failed to remove layer {} from datasource: {}".format(layer_str, path))             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_Failed_to_remove_layer_from_data, __pyx_n_s_format); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = NULL;
    __pyx_t_4 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
        __pyx_t_4 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_layer_str, __pyx_v_path};
      __pyx_t_5 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1493, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
      PyObject *__pyx_temp[3] = {__pyx_t_11, __pyx_v_layer_str, __pyx_v_path};
      __pyx_t_5 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-__pyx_t_4, 2+__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1493, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_5);
    } else
    #endif
    {
      __pyx_t_13 = PyTuple_New(2+__pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      if (__pyx_t_11) {
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_11); __pyx_t_11 = NULL;
      }
      __Pyx_INCREF(__pyx_v_layer_str);
      __Pyx_GIVEREF(__pyx_v_layer_str);
      PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_4, __pyx_v_layer_str);
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_GIVEREF(__pyx_v_path);
      PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_4, __pyx_v_path);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_13, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1493, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    __pyx_t_1 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_6, __pyx_t_10, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1493, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1492
 *     if result == OGRERR_UNSUPPORTED_OPERATION:
 *         raise DatasetDeleteError("Removal of layer {} not supported by driver".format(layer_str))
 *     elif result != OGRERR_NONE:             # <<<<<<<<<<<<<<
 *         raise DatasetDeleteError("Failed to remove layer {} from datasource: {}".format(layer_str, path))
 * 
 */
  }

  /* "fiona/ogrext.pyx":1464
 * 
 * 
 * def _remove_layer(path, layer, driver=None):             # <<<<<<<<<<<<<<
 *     cdef void *cogr_ds
 *     cdef int layer_index
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("fiona.ogrext._remove_layer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_layer_str);
  __Pyx_XDECREF(__pyx_v_layer_names);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1496
 * 
 * 
 * def _listlayers(path, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *     """Provides a list of the layers in an OGR data source.
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_12_listlayers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_11_listlayers[] = "Provides a list of the layers in an OGR data source.\n    ";
static PyMethodDef __pyx_mdef_5fiona_6ogrext_12_listlayers = {"_listlayers", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_12_listlayers, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_11_listlayers};
static PyObject *__pyx_pw_5fiona_6ogrext_12_listlayers(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_path = 0;
  PyObject *__pyx_v_kwargs = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_listlayers (wrapper)", 0);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_path,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_path)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, __pyx_v_kwargs, values, pos_args, "_listlayers") < 0)) __PYX_ERR(0, 1496, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_path = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_listlayers", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1496, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("fiona.ogrext._listlayers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_11_listlayers(__pyx_self, __pyx_v_path, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_kwargs);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_11_listlayers(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_path, PyObject *__pyx_v_kwargs) {
  void *__pyx_v_cogr_ds;
  void *__pyx_v_cogr_layer;
  char const *__pyx_v_path_c;
  char const *__pyx_v_name_c;
  PyObject *__pyx_v_path_b = NULL;
  int __pyx_v_layer_count;
  PyObject *__pyx_v_layer_names = NULL;
  int __pyx_v_i;
  char const *__pyx_v_name_b;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  char const *__pyx_t_8;
  void *__pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("_listlayers", 0);

  /* "fiona/ogrext.pyx":1501
 *     """
 * 
 *     cdef void *cogr_ds = NULL             # <<<<<<<<<<<<<<
 *     cdef void *cogr_layer = NULL
 *     cdef const char *path_c
 */
  __pyx_v_cogr_ds = NULL;

  /* "fiona/ogrext.pyx":1502
 * 
 *     cdef void *cogr_ds = NULL
 *     cdef void *cogr_layer = NULL             # <<<<<<<<<<<<<<
 *     cdef const char *path_c
 *     cdef const char *name_c
 */
  __pyx_v_cogr_layer = NULL;

  /* "fiona/ogrext.pyx":1507
 * 
 *     # Open OGR data source.
 *     try:             # <<<<<<<<<<<<<<
 *         path_b = path.encode('utf-8')
 *     except UnicodeDecodeError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "fiona/ogrext.pyx":1508
 *     # Open OGR data source.
 *     try:
 *         path_b = path.encode('utf-8')             # <<<<<<<<<<<<<<
 *     except UnicodeDecodeError:
 *         path_b = path
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_path, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1508, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      __pyx_t_4 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_kp_u_utf_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1508, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_path_b = __pyx_t_4;
      __pyx_t_4 = 0;

      /* "fiona/ogrext.pyx":1507
 * 
 *     # Open OGR data source.
 *     try:             # <<<<<<<<<<<<<<
 *         path_b = path.encode('utf-8')
 *     except UnicodeDecodeError:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "fiona/ogrext.pyx":1509
 *     try:
 *         path_b = path.encode('utf-8')
 *     except UnicodeDecodeError:             # <<<<<<<<<<<<<<
 *         path_b = path
 *     path_c = path_b
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_UnicodeDecodeError);
    if (__pyx_t_7) {
      __Pyx_AddTraceback("fiona.ogrext._listlayers", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 1509, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);

      /* "fiona/ogrext.pyx":1510
 *         path_b = path.encode('utf-8')
 *     except UnicodeDecodeError:
 *         path_b = path             # <<<<<<<<<<<<<<
 *     path_c = path_b
 *     cogr_ds = gdal_open_vector(path_c, 0, None, kwargs)
 */
      __Pyx_INCREF(__pyx_v_path);
      __Pyx_XDECREF_SET(__pyx_v_path_b, __pyx_v_path);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "fiona/ogrext.pyx":1507
 * 
 *     # Open OGR data source.
 *     try:             # <<<<<<<<<<<<<<
 *         path_b = path.encode('utf-8')
 *     except UnicodeDecodeError:
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "fiona/ogrext.pyx":1511
 *     except UnicodeDecodeError:
 *         path_b = path
 *     path_c = path_b             # <<<<<<<<<<<<<<
 *     cogr_ds = gdal_open_vector(path_c, 0, None, kwargs)
 * 
 */
  __pyx_t_8 = __Pyx_PyObject_AsString(__pyx_v_path_b); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1511, __pyx_L1_error)
  __pyx_v_path_c = __pyx_t_8;

  /* "fiona/ogrext.pyx":1512
 *         path_b = path
 *     path_c = path_b
 *     cogr_ds = gdal_open_vector(path_c, 0, None, kwargs)             # <<<<<<<<<<<<<<
 * 
 *     # Loop over the layers to get their names.
 */
  __pyx_t_9 = __pyx_f_5fiona_5_shim_gdal_open_vector(__pyx_v_path_c, 0, Py_None, __pyx_v_kwargs); if (unlikely(__pyx_t_9 == ((void *)NULL))) __PYX_ERR(0, 1512, __pyx_L1_error)
  __pyx_v_cogr_ds = __pyx_t_9;

  /* "fiona/ogrext.pyx":1515
 * 
 *     # Loop over the layers to get their names.
 *     layer_count = GDALDatasetGetLayerCount(cogr_ds)             # <<<<<<<<<<<<<<
 *     layer_names = []
 *     for i in range(layer_count):
 */
  __pyx_v_layer_count = GDALDatasetGetLayerCount(__pyx_v_cogr_ds);

  /* "fiona/ogrext.pyx":1516
 *     # Loop over the layers to get their names.
 *     layer_count = GDALDatasetGetLayerCount(cogr_ds)
 *     layer_names = []             # <<<<<<<<<<<<<<
 *     for i in range(layer_count):
 *         cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 */
  __pyx_t_6 = PyList_New(0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_v_layer_names = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "fiona/ogrext.pyx":1517
 *     layer_count = GDALDatasetGetLayerCount(cogr_ds)
 *     layer_names = []
 *     for i in range(layer_count):             # <<<<<<<<<<<<<<
 *         cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *         name_c = OGR_L_GetName(cogr_layer)
 */
  __pyx_t_7 = __pyx_v_layer_count;
  __pyx_t_10 = __pyx_t_7;
  for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_10; __pyx_t_11+=1) {
    __pyx_v_i = __pyx_t_11;

    /* "fiona/ogrext.pyx":1518
 *     layer_names = []
 *     for i in range(layer_count):
 *         cogr_layer = GDALDatasetGetLayer(cogr_ds, i)             # <<<<<<<<<<<<<<
 *         name_c = OGR_L_GetName(cogr_layer)
 *         name_b = name_c
 */
    __pyx_v_cogr_layer = GDALDatasetGetLayer(__pyx_v_cogr_ds, __pyx_v_i);

    /* "fiona/ogrext.pyx":1519
 *     for i in range(layer_count):
 *         cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *         name_c = OGR_L_GetName(cogr_layer)             # <<<<<<<<<<<<<<
 *         name_b = name_c
 *         layer_names.append(name_b.decode('utf-8'))
 */
    __pyx_v_name_c = OGR_L_GetName(__pyx_v_cogr_layer);

    /* "fiona/ogrext.pyx":1520
 *         cogr_layer = GDALDatasetGetLayer(cogr_ds, i)
 *         name_c = OGR_L_GetName(cogr_layer)
 *         name_b = name_c             # <<<<<<<<<<<<<<
 *         layer_names.append(name_b.decode('utf-8'))
 * 
 */
    __pyx_v_name_b = __pyx_v_name_c;

    /* "fiona/ogrext.pyx":1521
 *         name_c = OGR_L_GetName(cogr_layer)
 *         name_b = name_c
 *         layer_names.append(name_b.decode('utf-8'))             # <<<<<<<<<<<<<<
 * 
 *     # Close up data source.
 */
    __pyx_t_6 = __Pyx_decode_c_string(__pyx_v_name_b, 0, strlen(__pyx_v_name_b), NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_layer_names, __pyx_t_6); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 1521, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "fiona/ogrext.pyx":1524
 * 
 *     # Close up data source.
 *     if cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *         GDALClose(cogr_ds)
 *     cogr_ds = NULL
 */
  __pyx_t_13 = ((__pyx_v_cogr_ds != NULL) != 0);
  if (__pyx_t_13) {

    /* "fiona/ogrext.pyx":1525
 *     # Close up data source.
 *     if cogr_ds != NULL:
 *         GDALClose(cogr_ds)             # <<<<<<<<<<<<<<
 *     cogr_ds = NULL
 * 
 */
    GDALClose(__pyx_v_cogr_ds);

    /* "fiona/ogrext.pyx":1524
 * 
 *     # Close up data source.
 *     if cogr_ds != NULL:             # <<<<<<<<<<<<<<
 *         GDALClose(cogr_ds)
 *     cogr_ds = NULL
 */
  }

  /* "fiona/ogrext.pyx":1526
 *     if cogr_ds != NULL:
 *         GDALClose(cogr_ds)
 *     cogr_ds = NULL             # <<<<<<<<<<<<<<
 * 
 *     return layer_names
 */
  __pyx_v_cogr_ds = NULL;

  /* "fiona/ogrext.pyx":1528
 *     cogr_ds = NULL
 * 
 *     return layer_names             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_layer_names);
  __pyx_r = __pyx_v_layer_names;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1496
 * 
 * 
 * def _listlayers(path, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *     """Provides a list of the layers in an OGR data source.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fiona.ogrext._listlayers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_path_b);
  __Pyx_XDECREF(__pyx_v_layer_names);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1531
 * 
 * 
 * def buffer_to_virtual_file(bytesbuf, ext=''):             # <<<<<<<<<<<<<<
 *     """Maps a bytes buffer to a virtual file.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14buffer_to_virtual_file(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_13buffer_to_virtual_file[] = "Maps a bytes buffer to a virtual file.\n\n    `ext` is empty or begins with a period and contains at most one period.\n    ";
static PyMethodDef __pyx_mdef_5fiona_6ogrext_14buffer_to_virtual_file = {"buffer_to_virtual_file", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_14buffer_to_virtual_file, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_13buffer_to_virtual_file};
static PyObject *__pyx_pw_5fiona_6ogrext_14buffer_to_virtual_file(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_bytesbuf = 0;
  PyObject *__pyx_v_ext = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer_to_virtual_file (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_bytesbuf,&__pyx_n_s_ext,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_kp_u__12);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_bytesbuf)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ext);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "buffer_to_virtual_file") < 0)) __PYX_ERR(0, 1531, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_bytesbuf = values[0];
    __pyx_v_ext = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("buffer_to_virtual_file", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1531, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.buffer_to_virtual_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_13buffer_to_virtual_file(__pyx_self, __pyx_v_bytesbuf, __pyx_v_ext);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_13buffer_to_virtual_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_bytesbuf, PyObject *__pyx_v_ext) {
  PyObject *__pyx_v_vsi_filename = NULL;
  PyObject *__pyx_v_vsi_cfilename = NULL;
  VSILFILE *__pyx_v_vsi_handle;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  char const *__pyx_t_7;
  unsigned char *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  __Pyx_RefNannySetupContext("buffer_to_virtual_file", 0);

  /* "fiona/ogrext.pyx":1537
 *     """
 * 
 *     vsi_filename = '/vsimem/{}'.format(uuid.uuid4().hex + ext)             # <<<<<<<<<<<<<<
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_vsimem, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_uuid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_hex); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_5, __pyx_v_ext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1537, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_vsi_filename = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1538
 * 
 *     vsi_filename = '/vsimem/{}'.format(uuid.uuid4().hex + ext)
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     vsi_handle = VSIFileFromMemBuffer(vsi_cfilename, <unsigned char *>bytesbuf, len(bytesbuf), 0)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyObject_IsInstance(__pyx_v_vsi_filename, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (((!(__pyx_t_6 != 0)) != 0)) {
    __Pyx_INCREF(__pyx_v_vsi_filename);
    __pyx_t_1 = __pyx_v_vsi_filename;
  } else {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_vsi_filename, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_v_vsi_cfilename = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1540
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 * 
 *     vsi_handle = VSIFileFromMemBuffer(vsi_cfilename, <unsigned char *>bytesbuf, len(bytesbuf), 0)             # <<<<<<<<<<<<<<
 * 
 *     if vsi_handle == NULL:
 */
  __pyx_t_7 = __Pyx_PyObject_AsString(__pyx_v_vsi_cfilename); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(0, 1540, __pyx_L1_error)
  __pyx_t_8 = __Pyx_PyObject_AsWritableUString(__pyx_v_bytesbuf); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1540, __pyx_L1_error)
  __pyx_t_9 = PyObject_Length(__pyx_v_bytesbuf); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1540, __pyx_L1_error)
  __pyx_v_vsi_handle = VSIFileFromMemBuffer(__pyx_t_7, ((unsigned char *)__pyx_t_8), __pyx_t_9, 0);

  /* "fiona/ogrext.pyx":1542
 *     vsi_handle = VSIFileFromMemBuffer(vsi_cfilename, <unsigned char *>bytesbuf, len(bytesbuf), 0)
 * 
 *     if vsi_handle == NULL:             # <<<<<<<<<<<<<<
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:
 */
  __pyx_t_6 = ((__pyx_v_vsi_handle == NULL) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "fiona/ogrext.pyx":1543
 * 
 *     if vsi_handle == NULL:
 *         raise OSError('failed to map buffer to file')             # <<<<<<<<<<<<<<
 *     if VSIFCloseL(vsi_handle) != 0:
 *         raise OSError('failed to close mapped file handle')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__41, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1543, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1543, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1542
 *     vsi_handle = VSIFileFromMemBuffer(vsi_cfilename, <unsigned char *>bytesbuf, len(bytesbuf), 0)
 * 
 *     if vsi_handle == NULL:             # <<<<<<<<<<<<<<
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:
 */
  }

  /* "fiona/ogrext.pyx":1544
 *     if vsi_handle == NULL:
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:             # <<<<<<<<<<<<<<
 *         raise OSError('failed to close mapped file handle')
 * 
 */
  __pyx_t_6 = ((VSIFCloseL(__pyx_v_vsi_handle) != 0) != 0);
  if (unlikely(__pyx_t_6)) {

    /* "fiona/ogrext.pyx":1545
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:
 *         raise OSError('failed to close mapped file handle')             # <<<<<<<<<<<<<<
 * 
 *     return vsi_filename
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_OSError, __pyx_tuple__42, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 1545, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1544
 *     if vsi_handle == NULL:
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:             # <<<<<<<<<<<<<<
 *         raise OSError('failed to close mapped file handle')
 * 
 */
  }

  /* "fiona/ogrext.pyx":1547
 *         raise OSError('failed to close mapped file handle')
 * 
 *     return vsi_filename             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_vsi_filename);
  __pyx_r = __pyx_v_vsi_filename;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1531
 * 
 * 
 * def buffer_to_virtual_file(bytesbuf, ext=''):             # <<<<<<<<<<<<<<
 *     """Maps a bytes buffer to a virtual file.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.buffer_to_virtual_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vsi_filename);
  __Pyx_XDECREF(__pyx_v_vsi_cfilename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1550
 * 
 * 
 * def remove_virtual_file(vsi_filename):             # <<<<<<<<<<<<<<
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 *     return VSIUnlink(vsi_cfilename)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_16remove_virtual_file(PyObject *__pyx_self, PyObject *__pyx_v_vsi_filename); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_16remove_virtual_file = {"remove_virtual_file", (PyCFunction)__pyx_pw_5fiona_6ogrext_16remove_virtual_file, METH_O, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_16remove_virtual_file(PyObject *__pyx_self, PyObject *__pyx_v_vsi_filename) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("remove_virtual_file (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_15remove_virtual_file(__pyx_self, ((PyObject *)__pyx_v_vsi_filename));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_15remove_virtual_file(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_vsi_filename) {
  PyObject *__pyx_v_vsi_cfilename = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char const *__pyx_t_6;
  __Pyx_RefNannySetupContext("remove_virtual_file", 0);

  /* "fiona/ogrext.pyx":1551
 * 
 * def remove_virtual_file(vsi_filename):
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')             # <<<<<<<<<<<<<<
 *     return VSIUnlink(vsi_cfilename)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_vsi_filename, __pyx_t_2); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 1551, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (((!(__pyx_t_3 != 0)) != 0)) {
    __Pyx_INCREF(__pyx_v_vsi_filename);
    __pyx_t_1 = __pyx_v_vsi_filename;
  } else {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_vsi_filename, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_kp_u_utf_8);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
  }
  __pyx_v_vsi_cfilename = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1552
 * def remove_virtual_file(vsi_filename):
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 *     return VSIUnlink(vsi_cfilename)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = __Pyx_PyObject_AsString(__pyx_v_vsi_cfilename); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 1552, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyInt_From_int(VSIUnlink(__pyx_t_6)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1550
 * 
 * 
 * def remove_virtual_file(vsi_filename):             # <<<<<<<<<<<<<<
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 *     return VSIUnlink(vsi_cfilename)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.remove_virtual_file", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vsi_cfilename);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1558
 *     """Base for a BytesIO-like class backed by an in-memory file."""
 * 
 *     def __init__(self, file_or_bytes=None, filename=None, ext=''):             # <<<<<<<<<<<<<<
 *         """A file in an in-memory filesystem.
 * 
 */

/* Python wrapper */
static int __pyx_pw_5fiona_6ogrext_14MemoryFileBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase___init__[] = "A file in an in-memory filesystem.\n\n        Parameters\n        ----------\n        file_or_bytes : file or bytes\n            A file opened in binary mode or bytes or a bytearray\n        filename : str\n            A filename for the in-memory file under /vsimem\n        ext : str\n            A file extension for the in-memory file under /vsimem. Ignored if\n            filename was provided.\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase___init__;
#endif
static int __pyx_pw_5fiona_6ogrext_14MemoryFileBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_file_or_bytes = 0;
  PyObject *__pyx_v_filename = 0;
  PyObject *__pyx_v_ext = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_file_or_bytes,&__pyx_n_s_filename,&__pyx_n_s_ext,0};
    PyObject* values[3] = {0,0,0};
    values[0] = ((PyObject *)Py_None);
    values[1] = ((PyObject *)Py_None);
    values[2] = ((PyObject *)__pyx_kp_u__12);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_file_or_bytes);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_filename);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ext);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 1558, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_file_or_bytes = values[0];
    __pyx_v_filename = values[1];
    __pyx_v_ext = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 0, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1558, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase___init__(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self), __pyx_v_file_or_bytes, __pyx_v_filename, __pyx_v_ext);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5fiona_6ogrext_14MemoryFileBase___init__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_file_or_bytes, PyObject *__pyx_v_filename, PyObject *__pyx_v_ext) {
  VSILFILE *__pyx_v_vsi_handle;
  PyObject *__pyx_v_initial_bytes = NULL;
  unsigned char *__pyx_v_buffer;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  unsigned char *__pyx_t_11;
  char const *__pyx_t_12;
  Py_ssize_t __pyx_t_13;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "fiona/ogrext.pyx":1571
 *             filename was provided.
 *         """
 *         cdef VSILFILE *vsi_handle = NULL             # <<<<<<<<<<<<<<
 * 
 *         if file_or_bytes:
 */
  __pyx_v_vsi_handle = NULL;

  /* "fiona/ogrext.pyx":1573
 *         cdef VSILFILE *vsi_handle = NULL
 * 
 *         if file_or_bytes:             # <<<<<<<<<<<<<<
 *             if hasattr(file_or_bytes, 'read'):
 *                 initial_bytes = file_or_bytes.read()
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_file_or_bytes); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1573, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":1574
 * 
 *         if file_or_bytes:
 *             if hasattr(file_or_bytes, 'read'):             # <<<<<<<<<<<<<<
 *                 initial_bytes = file_or_bytes.read()
 *             else:
 */
    __pyx_t_1 = __Pyx_HasAttr(__pyx_v_file_or_bytes, __pyx_n_u_read); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(0, 1574, __pyx_L1_error)
    __pyx_t_2 = (__pyx_t_1 != 0);
    if (__pyx_t_2) {

      /* "fiona/ogrext.pyx":1575
 *         if file_or_bytes:
 *             if hasattr(file_or_bytes, 'read'):
 *                 initial_bytes = file_or_bytes.read()             # <<<<<<<<<<<<<<
 *             else:
 *                 initial_bytes = file_or_bytes
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_file_or_bytes, __pyx_n_s_read); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_4, function);
        }
      }
      __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1575, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_initial_bytes = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "fiona/ogrext.pyx":1574
 * 
 *         if file_or_bytes:
 *             if hasattr(file_or_bytes, 'read'):             # <<<<<<<<<<<<<<
 *                 initial_bytes = file_or_bytes.read()
 *             else:
 */
      goto __pyx_L4;
    }

    /* "fiona/ogrext.pyx":1577
 *                 initial_bytes = file_or_bytes.read()
 *             else:
 *                 initial_bytes = file_or_bytes             # <<<<<<<<<<<<<<
 *             if not isinstance(initial_bytes, (bytearray, bytes)):
 *                 raise TypeError(
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_file_or_bytes);
      __pyx_v_initial_bytes = __pyx_v_file_or_bytes;
    }
    __pyx_L4:;

    /* "fiona/ogrext.pyx":1578
 *             else:
 *                 initial_bytes = file_or_bytes
 *             if not isinstance(initial_bytes, (bytearray, bytes)):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     "Constructor argument must be a file opened in binary "
 */
    __pyx_t_1 = PyByteArray_Check(__pyx_v_initial_bytes); 
    __pyx_t_6 = (__pyx_t_1 != 0);
    if (!__pyx_t_6) {
    } else {
      __pyx_t_2 = __pyx_t_6;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_6 = PyBytes_Check(__pyx_v_initial_bytes); 
    __pyx_t_1 = (__pyx_t_6 != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_t_1 = ((!(__pyx_t_2 != 0)) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "fiona/ogrext.pyx":1579
 *                 initial_bytes = file_or_bytes
 *             if not isinstance(initial_bytes, (bytearray, bytes)):
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Constructor argument must be a file opened in binary "
 *                     "mode or bytes/bytearray.")
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__43, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1579, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1578
 *             else:
 *                 initial_bytes = file_or_bytes
 *             if not isinstance(initial_bytes, (bytearray, bytes)):             # <<<<<<<<<<<<<<
 *                 raise TypeError(
 *                     "Constructor argument must be a file opened in binary "
 */
    }

    /* "fiona/ogrext.pyx":1573
 *         cdef VSILFILE *vsi_handle = NULL
 * 
 *         if file_or_bytes:             # <<<<<<<<<<<<<<
 *             if hasattr(file_or_bytes, 'read'):
 *                 initial_bytes = file_or_bytes.read()
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1583
 *                     "mode or bytes/bytearray.")
 *         else:
 *             initial_bytes = b''             # <<<<<<<<<<<<<<
 * 
 *         if filename:
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_b__12);
    __pyx_v_initial_bytes = __pyx_kp_b__12;
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1585
 *             initial_bytes = b''
 * 
 *         if filename:             # <<<<<<<<<<<<<<
 *             # GDAL's SRTMHGT driver requires the filename to be "correct" (match
 *             # the bounds being written)
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_filename); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1585, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":1588
 *             # GDAL's SRTMHGT driver requires the filename to be "correct" (match
 *             # the bounds being written)
 *             self.name = '/vsimem/{0}'.format(filename)             # <<<<<<<<<<<<<<
 *         else:
 *             # GDAL 2.1 requires a .zip extension for zipped files.
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_vsimem_0, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_v_filename) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_filename);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 1588, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1585
 *             initial_bytes = b''
 * 
 *         if filename:             # <<<<<<<<<<<<<<
 *             # GDAL's SRTMHGT driver requires the filename to be "correct" (match
 *             # the bounds being written)
 */
    goto __pyx_L8;
  }

  /* "fiona/ogrext.pyx":1591
 *         else:
 *             # GDAL 2.1 requires a .zip extension for zipped files.
 *             self.name = '/vsimem/{0}.{1}'.format(uuid.uuid4(), ext.lstrip('.'))             # <<<<<<<<<<<<<<
 * 
 *         self.path = self.name.encode('utf-8')
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_vsimem_0_1, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_uuid); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_uuid4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    __pyx_t_5 = (__pyx_t_7) ? __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_t_7) : __Pyx_PyObject_CallNoArg(__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_ext, __pyx_n_s_lstrip); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_8 = (__pyx_t_9) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_9, __pyx_kp_u__22) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_kp_u__22);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_5, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
      PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_5, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_7) {
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_t_8);
      __pyx_t_5 = 0;
      __pyx_t_8 = 0;
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1591, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name, __pyx_t_3) < 0) __PYX_ERR(0, 1591, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L8:;

  /* "fiona/ogrext.pyx":1593
 *             self.name = '/vsimem/{0}.{1}'.format(uuid.uuid4(), ext.lstrip('.'))
 * 
 *         self.path = self.name.encode('utf-8')             # <<<<<<<<<<<<<<
 *         self._len = 0
 *         self._pos = 0
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_encode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_9, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_4, __pyx_kp_u_utf_8) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_kp_u_utf_8);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path, __pyx_t_3) < 0) __PYX_ERR(0, 1593, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":1594
 * 
 *         self.path = self.name.encode('utf-8')
 *         self._len = 0             # <<<<<<<<<<<<<<
 *         self._pos = 0
 *         self.closed = False
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_len, __pyx_int_0) < 0) __PYX_ERR(0, 1594, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1595
 *         self.path = self.name.encode('utf-8')
 *         self._len = 0
 *         self._pos = 0             # <<<<<<<<<<<<<<
 *         self.closed = False
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_int_0) < 0) __PYX_ERR(0, 1595, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1596
 *         self._len = 0
 *         self._pos = 0
 *         self.closed = False             # <<<<<<<<<<<<<<
 * 
 *         self._initial_bytes = initial_bytes
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed, Py_False) < 0) __PYX_ERR(0, 1596, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1598
 *         self.closed = False
 * 
 *         self._initial_bytes = initial_bytes             # <<<<<<<<<<<<<<
 *         cdef unsigned char *buffer = self._initial_bytes
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes, __pyx_v_initial_bytes) < 0) __PYX_ERR(0, 1598, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1599
 * 
 *         self._initial_bytes = initial_bytes
 *         cdef unsigned char *buffer = self._initial_bytes             # <<<<<<<<<<<<<<
 * 
 *         if self._initial_bytes:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = __Pyx_PyObject_AsWritableUString(__pyx_t_3); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 1599, __pyx_L1_error)
  __pyx_v_buffer = __pyx_t_11;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":1601
 *         cdef unsigned char *buffer = self._initial_bytes
 * 
 *         if self._initial_bytes:             # <<<<<<<<<<<<<<
 * 
 *             vsi_handle = VSIFileFromMemBuffer(
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "fiona/ogrext.pyx":1604
 * 
 *             vsi_handle = VSIFileFromMemBuffer(
 *                 self.path, buffer, len(self._initial_bytes), 0)             # <<<<<<<<<<<<<<
 *             self._len = len(self._initial_bytes)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_t_3); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(0, 1604, __pyx_L1_error)
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_13 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1604, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "fiona/ogrext.pyx":1603
 *         if self._initial_bytes:
 * 
 *             vsi_handle = VSIFileFromMemBuffer(             # <<<<<<<<<<<<<<
 *                 self.path, buffer, len(self._initial_bytes), 0)
 *             self._len = len(self._initial_bytes)
 */
    __pyx_v_vsi_handle = VSIFileFromMemBuffer(__pyx_t_12, __pyx_v_buffer, __pyx_t_13, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1605
 *             vsi_handle = VSIFileFromMemBuffer(
 *                 self.path, buffer, len(self._initial_bytes), 0)
 *             self._len = len(self._initial_bytes)             # <<<<<<<<<<<<<<
 * 
 *             if vsi_handle == NULL:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_13 = PyObject_Length(__pyx_t_3); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_len, __pyx_t_3) < 0) __PYX_ERR(0, 1605, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1607
 *             self._len = len(self._initial_bytes)
 * 
 *             if vsi_handle == NULL:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to create in-memory file using initial bytes.")
 */
    __pyx_t_1 = ((__pyx_v_vsi_handle == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "fiona/ogrext.pyx":1608
 * 
 *             if vsi_handle == NULL:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to create in-memory file using initial bytes.")
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IOError, __pyx_tuple__44, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1608, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1608, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1607
 *             self._len = len(self._initial_bytes)
 * 
 *             if vsi_handle == NULL:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to create in-memory file using initial bytes.")
 */
    }

    /* "fiona/ogrext.pyx":1611
 *                     "Failed to create in-memory file using initial bytes.")
 * 
 *             if VSIFCloseL(vsi_handle) != 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to properly close in-memory file.")
 */
    __pyx_t_1 = ((VSIFCloseL(__pyx_v_vsi_handle) != 0) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "fiona/ogrext.pyx":1612
 * 
 *             if VSIFCloseL(vsi_handle) != 0:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to properly close in-memory file.")
 * 
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_IOError, __pyx_tuple__45, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(0, 1612, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1611
 *                     "Failed to create in-memory file using initial bytes.")
 * 
 *             if VSIFCloseL(vsi_handle) != 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to properly close in-memory file.")
 */
    }

    /* "fiona/ogrext.pyx":1601
 *         cdef unsigned char *buffer = self._initial_bytes
 * 
 *         if self._initial_bytes:             # <<<<<<<<<<<<<<
 * 
 *             vsi_handle = VSIFileFromMemBuffer(
 */
  }

  /* "fiona/ogrext.pyx":1558
 *     """Base for a BytesIO-like class backed by an in-memory file."""
 * 
 *     def __init__(self, file_or_bytes=None, filename=None, ext=''):             # <<<<<<<<<<<<<<
 *         """A file in an in-memory filesystem.
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_initial_bytes);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1615
 *                     "Failed to properly close in-memory file.")
 * 
 *     def exists(self):             # <<<<<<<<<<<<<<
 *         """Test if the in-memory file exists.
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_3exists(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_2exists[] = "Test if the in-memory file exists.\n\n        Returns\n        -------\n        bool\n            True if the in-memory file exists.\n        ";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_3exists(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exists (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_2exists(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_2exists(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self) {
  VSILFILE *__pyx_v_fp;
  char const *__pyx_v_cypath;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  int __pyx_t_3;
  __Pyx_RefNannySetupContext("exists", 0);

  /* "fiona/ogrext.pyx":1623
 *             True if the in-memory file exists.
 *         """
 *         cdef VSILFILE *fp = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *cypath = self.path
 * 
 */
  __pyx_v_fp = NULL;

  /* "fiona/ogrext.pyx":1624
 *         """
 *         cdef VSILFILE *fp = NULL
 *         cdef const char *cypath = self.path             # <<<<<<<<<<<<<<
 * 
 *         with nogil:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1624, __pyx_L1_error)
  __pyx_v_cypath = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1626
 *         cdef const char *cypath = self.path
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             fp = VSIFOpenL(cypath, 'r')
 * 
 */
  {
      #ifdef WITH_THREAD
      PyThreadState *_save;
      Py_UNBLOCK_THREADS
      __Pyx_FastGIL_Remember();
      #endif
      /*try:*/ {

        /* "fiona/ogrext.pyx":1627
 * 
 *         with nogil:
 *             fp = VSIFOpenL(cypath, 'r')             # <<<<<<<<<<<<<<
 * 
 *         if fp != NULL:
 */
        __pyx_v_fp = VSIFOpenL(__pyx_v_cypath, ((char const *)"r"));
      }

      /* "fiona/ogrext.pyx":1626
 *         cdef const char *cypath = self.path
 * 
 *         with nogil:             # <<<<<<<<<<<<<<
 *             fp = VSIFOpenL(cypath, 'r')
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          #ifdef WITH_THREAD
          __Pyx_FastGIL_Forget();
          Py_BLOCK_THREADS
          #endif
          goto __pyx_L5;
        }
        __pyx_L5:;
      }
  }

  /* "fiona/ogrext.pyx":1629
 *             fp = VSIFOpenL(cypath, 'r')
 * 
 *         if fp != NULL:             # <<<<<<<<<<<<<<
 *             VSIFCloseL(fp)
 *             return True
 */
  __pyx_t_3 = ((__pyx_v_fp != NULL) != 0);
  if (__pyx_t_3) {

    /* "fiona/ogrext.pyx":1630
 * 
 *         if fp != NULL:
 *             VSIFCloseL(fp)             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
    (void)(VSIFCloseL(__pyx_v_fp));

    /* "fiona/ogrext.pyx":1631
 *         if fp != NULL:
 *             VSIFCloseL(fp)
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":1629
 *             fp = VSIFOpenL(cypath, 'r')
 * 
 *         if fp != NULL:             # <<<<<<<<<<<<<<
 *             VSIFCloseL(fp)
 *             return True
 */
  }

  /* "fiona/ogrext.pyx":1633
 *             return True
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "fiona/ogrext.pyx":1615
 *                     "Failed to properly close in-memory file.")
 * 
 *     def exists(self):             # <<<<<<<<<<<<<<
 *         """Test if the in-memory file exists.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.exists", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1635
 *             return False
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         """Length of the file's buffer in number of bytes.
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_pw_5fiona_6ogrext_14MemoryFileBase_5__len__(PyObject *__pyx_v_self); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_4__len__[] = "Length of the file's buffer in number of bytes.\n\n        Returns\n        -------\n        int\n        ";
#if CYTHON_COMPILING_IN_CPYTHON
struct wrapperbase __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase_4__len__;
#endif
static Py_ssize_t __pyx_pw_5fiona_6ogrext_14MemoryFileBase_5__len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_4__len__(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_pf_5fiona_6ogrext_14MemoryFileBase_4__len__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self) {
  CYTHON_UNUSED unsigned char *__pyx_v_buff;
  CYTHON_UNUSED char const *__pyx_v_cfilename;
  vsi_l_offset __pyx_v_buff_len;
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  char const *__pyx_t_3;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "fiona/ogrext.pyx":1642
 *         int
 *         """
 *         cdef unsigned char *buff = NULL             # <<<<<<<<<<<<<<
 *         cdef const char *cfilename = self.path
 *         cdef vsi_l_offset buff_len = 0
 */
  __pyx_v_buff = NULL;

  /* "fiona/ogrext.pyx":1643
 *         """
 *         cdef unsigned char *buff = NULL
 *         cdef const char *cfilename = self.path             # <<<<<<<<<<<<<<
 *         cdef vsi_l_offset buff_len = 0
 *         buff = VSIGetMemFileBuffer(self.path, &buff_len, 0)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1643, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1643, __pyx_L1_error)
  __pyx_v_cfilename = __pyx_t_2;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1644
 *         cdef unsigned char *buff = NULL
 *         cdef const char *cfilename = self.path
 *         cdef vsi_l_offset buff_len = 0             # <<<<<<<<<<<<<<
 *         buff = VSIGetMemFileBuffer(self.path, &buff_len, 0)
 *         return int(buff_len)
 */
  __pyx_v_buff_len = 0;

  /* "fiona/ogrext.pyx":1645
 *         cdef const char *cfilename = self.path
 *         cdef vsi_l_offset buff_len = 0
 *         buff = VSIGetMemFileBuffer(self.path, &buff_len, 0)             # <<<<<<<<<<<<<<
 *         return int(buff_len)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_AsString(__pyx_t_1); if (unlikely((!__pyx_t_3) && PyErr_Occurred())) __PYX_ERR(0, 1645, __pyx_L1_error)
  __pyx_v_buff = VSIGetMemFileBuffer(__pyx_t_3, (&__pyx_v_buff_len), 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1646
 *         cdef vsi_l_offset buff_len = 0
 *         buff = VSIGetMemFileBuffer(self.path, &buff_len, 0)
 *         return int(buff_len)             # <<<<<<<<<<<<<<
 * 
 *     def close(self):
 */
  __pyx_r = ((Py_ssize_t)__pyx_v_buff_len);
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1635
 *             return False
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         """Length of the file's buffer in number of bytes.
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.__len__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1648
 *         return int(buff_len)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close MemoryFile and release allocated memory."""
 *         VSIUnlink(self.path)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_7close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_6close[] = "Close MemoryFile and release allocated memory.";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_7close(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_6close(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_6close(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  char const *__pyx_t_2;
  __Pyx_RefNannySetupContext("close", 0);

  /* "fiona/ogrext.pyx":1650
 *     def close(self):
 *         """Close MemoryFile and release allocated memory."""
 *         VSIUnlink(self.path)             # <<<<<<<<<<<<<<
 *         self._pos = 0
 *         self._initial_bytes = None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1650, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_AsString(__pyx_t_1); if (unlikely((!__pyx_t_2) && PyErr_Occurred())) __PYX_ERR(0, 1650, __pyx_L1_error)
  (void)(VSIUnlink(__pyx_t_2));
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1651
 *         """Close MemoryFile and release allocated memory."""
 *         VSIUnlink(self.path)
 *         self._pos = 0             # <<<<<<<<<<<<<<
 *         self._initial_bytes = None
 *         self.closed = True
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_int_0) < 0) __PYX_ERR(0, 1651, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1652
 *         VSIUnlink(self.path)
 *         self._pos = 0
 *         self._initial_bytes = None             # <<<<<<<<<<<<<<
 *         self.closed = True
 * 
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_initial_bytes, Py_None) < 0) __PYX_ERR(0, 1652, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1653
 *         self._pos = 0
 *         self._initial_bytes = None
 *         self.closed = True             # <<<<<<<<<<<<<<
 * 
 *     def read(self, size=-1):
 */
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_closed, Py_True) < 0) __PYX_ERR(0, 1653, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1648
 *         return int(buff_len)
 * 
 *     def close(self):             # <<<<<<<<<<<<<<
 *         """Close MemoryFile and release allocated memory."""
 *         VSIUnlink(self.path)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1655
 *         self.closed = True
 * 
 *     def read(self, size=-1):             # <<<<<<<<<<<<<<
 *         """Read size bytes from MemoryFile."""
 *         cdef VSILFILE *fp = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_9read(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_8read[] = "Read size bytes from MemoryFile.";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_9read(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("read (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_size,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)__pyx_int_neg_1);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_size);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "read") < 0)) __PYX_ERR(0, 1655, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_size = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("read", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1655, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_8read(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self), __pyx_v_size);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_8read(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_size) {
  VSILFILE *__pyx_v_fp;
  PyObject *__pyx_v_length = NULL;
  unsigned char *__pyx_v_buffer;
  PyObject *__pyx_v_result = 0;
  size_t __pyx_v_objects_read;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  char const *__pyx_t_8;
  VSILFILE *__pyx_t_9;
  vsi_l_offset __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  __Pyx_RefNannySetupContext("read", 0);
  __Pyx_INCREF(__pyx_v_size);

  /* "fiona/ogrext.pyx":1657
 *     def read(self, size=-1):
 *         """Read size bytes from MemoryFile."""
 *         cdef VSILFILE *fp = NULL             # <<<<<<<<<<<<<<
 *         # Return no bytes immediately if the position is at or past the
 *         # end of the file.
 */
  __pyx_v_fp = NULL;

  /* "fiona/ogrext.pyx":1660
 *         # Return no bytes immediately if the position is at or past the
 *         # end of the file.
 *         length = len(self)             # <<<<<<<<<<<<<<
 * 
 *         if self._pos >= length:
 */
  __pyx_t_1 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1660, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_length = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":1662
 *         length = len(self)
 * 
 *         if self._pos >= length:             # <<<<<<<<<<<<<<
 *             self._pos = length
 *             return b''
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_v_length, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1662, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "fiona/ogrext.pyx":1663
 * 
 *         if self._pos >= length:
 *             self._pos = length             # <<<<<<<<<<<<<<
 *             return b''
 * 
 */
    if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_v_length) < 0) __PYX_ERR(0, 1663, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1664
 *         if self._pos >= length:
 *             self._pos = length
 *             return b''             # <<<<<<<<<<<<<<
 * 
 *         if size == -1:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_b__12);
    __pyx_r = __pyx_kp_b__12;
    goto __pyx_L0;

    /* "fiona/ogrext.pyx":1662
 *         length = len(self)
 * 
 *         if self._pos >= length:             # <<<<<<<<<<<<<<
 *             self._pos = length
 *             return b''
 */
  }

  /* "fiona/ogrext.pyx":1666
 *             return b''
 * 
 *         if size == -1:             # <<<<<<<<<<<<<<
 *             size = length - self._pos
 *         else:
 */
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_size, __pyx_int_neg_1, -1L, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1666, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "fiona/ogrext.pyx":1667
 * 
 *         if size == -1:
 *             size = length - self._pos             # <<<<<<<<<<<<<<
 *         else:
 *             size = min(size, length - self._pos)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_length, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1667, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "fiona/ogrext.pyx":1666
 *             return b''
 * 
 *         if size == -1:             # <<<<<<<<<<<<<<
 *             size = length - self._pos
 *         else:
 */
    goto __pyx_L4;
  }

  /* "fiona/ogrext.pyx":1669
 *             size = length - self._pos
 *         else:
 *             size = min(size, length - self._pos)             # <<<<<<<<<<<<<<
 * 
 *         cdef unsigned char *buffer = <unsigned char *>CPLMalloc(size)
 */
  /*else*/ {
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_length, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v_size);
    __pyx_t_2 = __pyx_v_size;
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 1669, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_5 = __pyx_t_3;
    } else {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_5 = __pyx_t_2;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __pyx_t_5;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_3);
    __pyx_t_3 = 0;
  }
  __pyx_L4:;

  /* "fiona/ogrext.pyx":1671
 *             size = min(size, length - self._pos)
 * 
 *         cdef unsigned char *buffer = <unsigned char *>CPLMalloc(size)             # <<<<<<<<<<<<<<
 *         cdef bytes result
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_size); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1671, __pyx_L1_error)
  __pyx_v_buffer = ((unsigned char *)CPLMalloc(__pyx_t_7));

  /* "fiona/ogrext.pyx":1674
 *         cdef bytes result
 * 
 *         fp = VSIFOpenL(self.path, 'r')             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_8 = __Pyx_PyObject_AsString(__pyx_t_3); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1674, __pyx_L1_error)
  __pyx_v_fp = VSIFOpenL(__pyx_t_8, ((char const *)"r"));
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":1676
 *         fp = VSIFOpenL(self.path, 'r')
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             fp = exc_wrap_vsilfile(fp)
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 */
  /*try:*/ {

    /* "fiona/ogrext.pyx":1677
 * 
 *         try:
 *             fp = exc_wrap_vsilfile(fp)             # <<<<<<<<<<<<<<
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 *                 raise IOError(
 */
    __pyx_t_9 = __pyx_f_5fiona_4_err_exc_wrap_vsilfile(__pyx_v_fp); if (unlikely(__pyx_t_9 == ((VSILFILE *)NULL))) __PYX_ERR(0, 1677, __pyx_L6_error)
    __pyx_v_fp = __pyx_t_9;

    /* "fiona/ogrext.pyx":1678
 *         try:
 *             fp = exc_wrap_vsilfile(fp)
 *             if VSIFSeekL(fp, self._pos, 0) < 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.",
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1678, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_As_vsi_l_offset(__pyx_t_3); if (unlikely((__pyx_t_10 == ((vsi_l_offset)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1678, __pyx_L6_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = ((VSIFSeekL(__pyx_v_fp, __pyx_t_10, 0) < 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "fiona/ogrext.pyx":1681
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.",
 *                     self._pos, self.name)             # <<<<<<<<<<<<<<
 * 
 *             objects_read = VSIFReadL(buffer, 1, size, fp)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1681, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1681, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);

      /* "fiona/ogrext.pyx":1679
 *             fp = exc_wrap_vsilfile(fp)
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to seek to offset %s in %s.",
 *                     self._pos, self.name)
 */
      __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1679, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      __Pyx_GIVEREF(__pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
      __pyx_t_3 = 0;
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_IOError, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1679, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 1679, __pyx_L6_error)

      /* "fiona/ogrext.pyx":1678
 *         try:
 *             fp = exc_wrap_vsilfile(fp)
 *             if VSIFSeekL(fp, self._pos, 0) < 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.",
 */
    }

    /* "fiona/ogrext.pyx":1683
 *                     self._pos, self.name)
 * 
 *             objects_read = VSIFReadL(buffer, 1, size, fp)             # <<<<<<<<<<<<<<
 *             result = <bytes>buffer[:objects_read]
 * 
 */
    __pyx_t_7 = __Pyx_PyInt_As_size_t(__pyx_v_size); if (unlikely((__pyx_t_7 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 1683, __pyx_L6_error)
    __pyx_v_objects_read = VSIFReadL(__pyx_v_buffer, 1, __pyx_t_7, __pyx_v_fp);

    /* "fiona/ogrext.pyx":1684
 * 
 *             objects_read = VSIFReadL(buffer, 1, size, fp)
 *             result = <bytes>buffer[:objects_read]             # <<<<<<<<<<<<<<
 * 
 *         finally:
 */
    __pyx_t_5 = __Pyx_PyBytes_FromStringAndSize(((const char*)__pyx_v_buffer) + 0, __pyx_v_objects_read - 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1684, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;
  }

  /* "fiona/ogrext.pyx":1687
 * 
 *         finally:
 *             VSIFCloseL(fp)             # <<<<<<<<<<<<<<
 *             CPLFree(buffer)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      (void)(VSIFCloseL(__pyx_v_fp));

      /* "fiona/ogrext.pyx":1688
 *         finally:
 *             VSIFCloseL(fp)
 *             CPLFree(buffer)             # <<<<<<<<<<<<<<
 * 
 *         self._pos += len(result)
 */
      CPLFree(__pyx_v_buffer);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_14);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __pyx_t_11 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
      {

        /* "fiona/ogrext.pyx":1687
 * 
 *         finally:
 *             VSIFCloseL(fp)             # <<<<<<<<<<<<<<
 *             CPLFree(buffer)
 * 
 */
        (void)(VSIFCloseL(__pyx_v_fp));

        /* "fiona/ogrext.pyx":1688
 *         finally:
 *             VSIFCloseL(fp)
 *             CPLFree(buffer)             # <<<<<<<<<<<<<<
 * 
 *         self._pos += len(result)
 */
        CPLFree(__pyx_v_buffer);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_17);
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
      }
      __Pyx_XGIVEREF(__pyx_t_14);
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
      __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
      __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "fiona/ogrext.pyx":1690
 *             CPLFree(buffer)
 * 
 *         self._pos += len(result)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_v_result == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1690, __pyx_L1_error)
  }
  __pyx_t_1 = PyBytes_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1690, __pyx_L1_error)
  __pyx_t_5 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_t_3) < 0) __PYX_ERR(0, 1690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":1691
 * 
 *         self._pos += len(result)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def seek(self, offset, whence=0):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1655
 *         self.closed = True
 * 
 *     def read(self, size=-1):             # <<<<<<<<<<<<<<
 *         """Read size bytes from MemoryFile."""
 *         cdef VSILFILE *fp = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.read", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1693
 *         return result
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         """Seek to position in MemoryFile."""
 *         if whence == 0:
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_11seek(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_10seek[] = "Seek to position in MemoryFile.";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_11seek(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_offset = 0;
  PyObject *__pyx_v_whence = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("seek (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_offset,&__pyx_n_s_whence,0};
    PyObject* values[2] = {0,0};
    values[1] = ((PyObject *)__pyx_int_0);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_offset)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_whence);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "seek") < 0)) __PYX_ERR(0, 1693, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_offset = values[0];
    __pyx_v_whence = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("seek", 0, 1, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1693, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_10seek(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self), __pyx_v_offset, __pyx_v_whence);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_10seek(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_offset, PyObject *__pyx_v_whence) {
  PyObject *__pyx_v_pos = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("seek", 0);

  /* "fiona/ogrext.pyx":1695
 *     def seek(self, offset, whence=0):
 *         """Seek to position in MemoryFile."""
 *         if whence == 0:             # <<<<<<<<<<<<<<
 *             pos = offset
 *         elif whence == 1:
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_whence, __pyx_int_0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1696
 *         """Seek to position in MemoryFile."""
 *         if whence == 0:
 *             pos = offset             # <<<<<<<<<<<<<<
 *         elif whence == 1:
 *             pos = self._pos + offset
 */
    __Pyx_INCREF(__pyx_v_offset);
    __pyx_v_pos = __pyx_v_offset;

    /* "fiona/ogrext.pyx":1695
 *     def seek(self, offset, whence=0):
 *         """Seek to position in MemoryFile."""
 *         if whence == 0:             # <<<<<<<<<<<<<<
 *             pos = offset
 *         elif whence == 1:
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1697
 *         if whence == 0:
 *             pos = offset
 *         elif whence == 1:             # <<<<<<<<<<<<<<
 *             pos = self._pos + offset
 *         elif whence == 2:
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_v_whence, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1697, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1698
 *             pos = offset
 *         elif whence == 1:
 *             pos = self._pos + offset             # <<<<<<<<<<<<<<
 *         elif whence == 2:
 *             pos = len(self) - offset
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_offset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_pos = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "fiona/ogrext.pyx":1697
 *         if whence == 0:
 *             pos = offset
 *         elif whence == 1:             # <<<<<<<<<<<<<<
 *             pos = self._pos + offset
 *         elif whence == 2:
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1699
 *         elif whence == 1:
 *             pos = self._pos + offset
 *         elif whence == 2:             # <<<<<<<<<<<<<<
 *             pos = len(self) - offset
 *         if pos < 0:
 */
  __pyx_t_3 = __Pyx_PyInt_EqObjC(__pyx_v_whence, __pyx_int_2, 2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1699, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_2) {

    /* "fiona/ogrext.pyx":1700
 *             pos = self._pos + offset
 *         elif whence == 2:
 *             pos = len(self) - offset             # <<<<<<<<<<<<<<
 *         if pos < 0:
 *             raise ValueError("negative seek position: {}".format(pos))
 */
    __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1700, __pyx_L1_error)
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyNumber_Subtract(__pyx_t_3, __pyx_v_offset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_pos = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "fiona/ogrext.pyx":1699
 *         elif whence == 1:
 *             pos = self._pos + offset
 *         elif whence == 2:             # <<<<<<<<<<<<<<
 *             pos = len(self) - offset
 *         if pos < 0:
 */
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1701
 *         elif whence == 2:
 *             pos = len(self) - offset
 *         if pos < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("negative seek position: {}".format(pos))
 *         if pos > len(self):
 */
  if (unlikely(!__pyx_v_pos)) { __Pyx_RaiseUnboundLocalError("pos"); __PYX_ERR(0, 1701, __pyx_L1_error) }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_pos, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1701, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1702
 *             pos = len(self) - offset
 *         if pos < 0:
 *             raise ValueError("negative seek position: {}".format(pos))             # <<<<<<<<<<<<<<
 *         if pos > len(self):
 *             raise ValueError("seek position past end of file: {}".format(pos))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_negative_seek_position, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(!__pyx_v_pos)) { __Pyx_RaiseUnboundLocalError("pos"); __PYX_ERR(0, 1702, __pyx_L1_error) }
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_pos) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_pos);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1702, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1702, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1701
 *         elif whence == 2:
 *             pos = len(self) - offset
 *         if pos < 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("negative seek position: {}".format(pos))
 *         if pos > len(self):
 */
  }

  /* "fiona/ogrext.pyx":1703
 *         if pos < 0:
 *             raise ValueError("negative seek position: {}".format(pos))
 *         if pos > len(self):             # <<<<<<<<<<<<<<
 *             raise ValueError("seek position past end of file: {}".format(pos))
 *         self._pos = pos
 */
  if (unlikely(!__pyx_v_pos)) { __Pyx_RaiseUnboundLocalError("pos"); __PYX_ERR(0, 1703, __pyx_L1_error) }
  __pyx_t_4 = PyObject_Length(((PyObject *)__pyx_v_self)); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1703, __pyx_L1_error)
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_pos, __pyx_t_3, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1703, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(__pyx_t_2)) {

    /* "fiona/ogrext.pyx":1704
 *             raise ValueError("negative seek position: {}".format(pos))
 *         if pos > len(self):
 *             raise ValueError("seek position past end of file: {}".format(pos))             # <<<<<<<<<<<<<<
 *         self._pos = pos
 *         return self._pos
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_seek_position_past_end_of_file, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(!__pyx_v_pos)) { __Pyx_RaiseUnboundLocalError("pos"); __PYX_ERR(0, 1704, __pyx_L1_error) }
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_pos) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_pos);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 1704, __pyx_L1_error)

    /* "fiona/ogrext.pyx":1703
 *         if pos < 0:
 *             raise ValueError("negative seek position: {}".format(pos))
 *         if pos > len(self):             # <<<<<<<<<<<<<<
 *             raise ValueError("seek position past end of file: {}".format(pos))
 *         self._pos = pos
 */
  }

  /* "fiona/ogrext.pyx":1705
 *         if pos > len(self):
 *             raise ValueError("seek position past end of file: {}".format(pos))
 *         self._pos = pos             # <<<<<<<<<<<<<<
 *         return self._pos
 * 
 */
  if (unlikely(!__pyx_v_pos)) { __Pyx_RaiseUnboundLocalError("pos"); __PYX_ERR(0, 1705, __pyx_L1_error) }
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_v_pos) < 0) __PYX_ERR(0, 1705, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1706
 *             raise ValueError("seek position past end of file: {}".format(pos))
 *         self._pos = pos
 *         return self._pos             # <<<<<<<<<<<<<<
 * 
 *     def tell(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1693
 *         return result
 * 
 *     def seek(self, offset, whence=0):             # <<<<<<<<<<<<<<
 *         """Seek to position in MemoryFile."""
 *         if whence == 0:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.seek", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pos);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1708
 *         return self._pos
 * 
 *     def tell(self):             # <<<<<<<<<<<<<<
 *         """Tell current position in MemoryFile."""
 *         return self._pos
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_13tell(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_12tell[] = "Tell current position in MemoryFile.";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_13tell(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("tell (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_12tell(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_12tell(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("tell", 0);

  /* "fiona/ogrext.pyx":1710
 *     def tell(self):
 *         """Tell current position in MemoryFile."""
 *         return self._pos             # <<<<<<<<<<<<<<
 * 
 *     def write(self, data):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1708
 *         return self._pos
 * 
 *     def tell(self):             # <<<<<<<<<<<<<<
 *         """Tell current position in MemoryFile."""
 *         return self._pos
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.tell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fiona/ogrext.pyx":1712
 *         return self._pos
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write data bytes to MemoryFile"""
 *         cdef VSILFILE *fp = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_15write(PyObject *__pyx_v_self, PyObject *__pyx_v_data); /*proto*/
static char __pyx_doc_5fiona_6ogrext_14MemoryFileBase_14write[] = "Write data bytes to MemoryFile";
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_15write(PyObject *__pyx_v_self, PyObject *__pyx_v_data) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("write (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_14write(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self), ((PyObject *)__pyx_v_data));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_14write(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v_data) {
  VSILFILE *__pyx_v_fp;
  unsigned char const *__pyx_v_view;
  Py_ssize_t __pyx_v_n;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned char const *__pyx_t_1;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  char const *__pyx_t_8;
  VSILFILE *__pyx_t_9;
  vsi_l_offset __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  __Pyx_RefNannySetupContext("write", 0);

  /* "fiona/ogrext.pyx":1714
 *     def write(self, data):
 *         """Write data bytes to MemoryFile"""
 *         cdef VSILFILE *fp = NULL             # <<<<<<<<<<<<<<
 *         cdef const unsigned char *view = <bytes>data
 *         n = len(data)
 */
  __pyx_v_fp = NULL;

  /* "fiona/ogrext.pyx":1715
 *         """Write data bytes to MemoryFile"""
 *         cdef VSILFILE *fp = NULL
 *         cdef const unsigned char *view = <bytes>data             # <<<<<<<<<<<<<<
 *         n = len(data)
 * 
 */
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(0, 1715, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyBytes_AsUString(__pyx_v_data); if (unlikely((!__pyx_t_1) && PyErr_Occurred())) __PYX_ERR(0, 1715, __pyx_L1_error)
  __pyx_v_view = __pyx_t_1;

  /* "fiona/ogrext.pyx":1716
 *         cdef VSILFILE *fp = NULL
 *         cdef const unsigned char *view = <bytes>data
 *         n = len(data)             # <<<<<<<<<<<<<<
 * 
 *         if not self.exists():
 */
  __pyx_t_2 = PyObject_Length(__pyx_v_data); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1716, __pyx_L1_error)
  __pyx_v_n = __pyx_t_2;

  /* "fiona/ogrext.pyx":1718
 *         n = len(data)
 * 
 *         if not self.exists():             # <<<<<<<<<<<<<<
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'w'))
 *         else:
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_exists); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
    }
  }
  __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallNoArg(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 1718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = ((!__pyx_t_6) != 0);
  if (__pyx_t_7) {

    /* "fiona/ogrext.pyx":1719
 * 
 *         if not self.exists():
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'w'))             # <<<<<<<<<<<<<<
 *         else:
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'r+'))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_AsString(__pyx_t_3); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1719, __pyx_L1_error)
    __pyx_t_9 = __pyx_f_5fiona_4_err_exc_wrap_vsilfile(VSIFOpenL(__pyx_t_8, ((char const *)"w"))); if (unlikely(__pyx_t_9 == ((VSILFILE *)NULL))) __PYX_ERR(0, 1719, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_fp = __pyx_t_9;

    /* "fiona/ogrext.pyx":1718
 *         n = len(data)
 * 
 *         if not self.exists():             # <<<<<<<<<<<<<<
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'w'))
 *         else:
 */
    goto __pyx_L3;
  }

  /* "fiona/ogrext.pyx":1721
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'w'))
 *         else:
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'r+'))             # <<<<<<<<<<<<<<
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 *                 raise IOError(
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_path); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_AsString(__pyx_t_3); if (unlikely((!__pyx_t_8) && PyErr_Occurred())) __PYX_ERR(0, 1721, __pyx_L1_error)
    __pyx_t_9 = __pyx_f_5fiona_4_err_exc_wrap_vsilfile(VSIFOpenL(__pyx_t_8, ((char const *)"r+"))); if (unlikely(__pyx_t_9 == ((VSILFILE *)NULL))) __PYX_ERR(0, 1721, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_fp = __pyx_t_9;

    /* "fiona/ogrext.pyx":1722
 *         else:
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'r+'))
 *             if VSIFSeekL(fp, self._pos, 0) < 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.", self._pos, self.name)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_As_vsi_l_offset(__pyx_t_3); if (unlikely((__pyx_t_10 == ((vsi_l_offset)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1722, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = ((VSIFSeekL(__pyx_v_fp, __pyx_t_10, 0) < 0) != 0);
    if (unlikely(__pyx_t_7)) {

      /* "fiona/ogrext.pyx":1724
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.", self._pos, self.name)             # <<<<<<<<<<<<<<
 * 
 *         result = VSIFWriteL(view, 1, n, fp)
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1724, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1724, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);

      /* "fiona/ogrext.pyx":1723
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'r+'))
 *             if VSIFSeekL(fp, self._pos, 0) < 0:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to seek to offset %s in %s.", self._pos, self.name)
 * 
 */
      __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1723, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      __Pyx_GIVEREF(__pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Failed_to_seek_to_offset_s_in_s);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_IOError, __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1723, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(0, 1723, __pyx_L1_error)

      /* "fiona/ogrext.pyx":1722
 *         else:
 *             fp = exc_wrap_vsilfile(VSIFOpenL(self.path, 'r+'))
 *             if VSIFSeekL(fp, self._pos, 0) < 0:             # <<<<<<<<<<<<<<
 *                 raise IOError(
 *                     "Failed to seek to offset %s in %s.", self._pos, self.name)
 */
    }
  }
  __pyx_L3:;

  /* "fiona/ogrext.pyx":1726
 *                     "Failed to seek to offset %s in %s.", self._pos, self.name)
 * 
 *         result = VSIFWriteL(view, 1, n, fp)             # <<<<<<<<<<<<<<
 *         VSIFFlushL(fp)
 *         VSIFCloseL(fp)
 */
  __pyx_t_4 = __Pyx_PyInt_FromSize_t(VSIFWriteL(__pyx_v_view, 1, __pyx_v_n, __pyx_v_fp)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_result = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "fiona/ogrext.pyx":1727
 * 
 *         result = VSIFWriteL(view, 1, n, fp)
 *         VSIFFlushL(fp)             # <<<<<<<<<<<<<<
 *         VSIFCloseL(fp)
 * 
 */
  (void)(VSIFFlushL(__pyx_v_fp));

  /* "fiona/ogrext.pyx":1728
 *         result = VSIFWriteL(view, 1, n, fp)
 *         VSIFFlushL(fp)
 *         VSIFCloseL(fp)             # <<<<<<<<<<<<<<
 * 
 *         self._pos += result
 */
  (void)(VSIFCloseL(__pyx_v_fp));

  /* "fiona/ogrext.pyx":1730
 *         VSIFCloseL(fp)
 * 
 *         self._pos += result             # <<<<<<<<<<<<<<
 *         self._len = max(self._len, self._pos)
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_4, __pyx_v_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos, __pyx_t_5) < 0) __PYX_ERR(0, 1730, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":1731
 * 
 *         self._pos += result
 *         self._len = max(self._len, self._pos)             # <<<<<<<<<<<<<<
 * 
 *         return result
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pos); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_len); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyObject_RichCompare(__pyx_t_5, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  if (__pyx_t_7) {
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_3 = __pyx_t_5;
  } else {
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_len, __pyx_t_5) < 0) __PYX_ERR(0, 1731, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fiona/ogrext.pyx":1733
 *         self._len = max(self._len, self._pos)
 * 
 *         return result             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "fiona/ogrext.pyx":1712
 *         return self._pos
 * 
 *     def write(self, data):             # <<<<<<<<<<<<<<
 *         """Write data bytes to MemoryFile"""
 *         cdef VSILFILE *fp = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.write", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_17__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_16__reduce_cython__(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_16__reduce_cython__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __Pyx_INCREF(__pyx_empty_tuple);
  __pyx_v_state = __pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, None), state
 */
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_MemoryFileBase); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, None), state
 *     else:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryFileBase__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_MemoryFileBase); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_222419149);
    __Pyx_GIVEREF(__pyx_int_222419149);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_222419149);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryFileBase__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_5fiona_6ogrext_14MemoryFileBase_19__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5fiona_6ogrext_14MemoryFileBase_18__setstate_cython__(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_14MemoryFileBase_18__setstate_cython__(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MemoryFileBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_5fiona_6ogrext___pyx_unpickle_MemoryFileBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MemoryFileBase, (type(self), 0xd41d8cd, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MemoryFileBase__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("fiona.ogrext.MemoryFileBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_18__pyx_unpickle_FeatureBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_18__pyx_unpickle_FeatureBuilder = {"__pyx_unpickle_FeatureBuilder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_18__pyx_unpickle_FeatureBuilder, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_18__pyx_unpickle_FeatureBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBuilder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBuilder", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBuilder", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_FeatureBuilder") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_FeatureBuilder", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_FeatureBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_17__pyx_unpickle_FeatureBuilder(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_17__pyx_unpickle_FeatureBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBuilder", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5fiona_6ogrext_FeatureBuilder), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_5fiona_6ogrext___pyx_unpickle_FeatureBuilder__set_state(((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = FeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_FeatureBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_FeatureBuilder__set_state(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_FeatureBuilder__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_FeatureBuilder__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_OGRFeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_20__pyx_unpickle_OGRFeatureBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_20__pyx_unpickle_OGRFeatureBuilder = {"__pyx_unpickle_OGRFeatureBuilder", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_20__pyx_unpickle_OGRFeatureBuilder, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_20__pyx_unpickle_OGRFeatureBuilder(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_OGRFeatureBuilder (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OGRFeatureBuilder", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OGRFeatureBuilder", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_OGRFeatureBuilder") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OGRFeatureBuilder", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_OGRFeatureBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_19__pyx_unpickle_OGRFeatureBuilder(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_19__pyx_unpickle_OGRFeatureBuilder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OGRFeatureBuilder", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5fiona_6ogrext_OGRFeatureBuilder), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_5fiona_6ogrext___pyx_unpickle_OGRFeatureBuilder__set_state(((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = OGRFeatureBuilder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_OGRFeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_OGRFeatureBuilder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_OGRFeatureBuilder__set_state(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OGRFeatureBuilder__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_OGRFeatureBuilder__set_state(<OGRFeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OGRFeatureBuilder__set_state(OGRFeatureBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_OGRFeatureBuilder__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MemoryFileBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_5fiona_6ogrext_22__pyx_unpickle_MemoryFileBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_5fiona_6ogrext_22__pyx_unpickle_MemoryFileBase = {"__pyx_unpickle_MemoryFileBase", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_22__pyx_unpickle_MemoryFileBase, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_5fiona_6ogrext_22__pyx_unpickle_MemoryFileBase(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryFileBase (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryFileBase", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryFileBase", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_MemoryFileBase") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MemoryFileBase", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_MemoryFileBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_5fiona_6ogrext_21__pyx_unpickle_MemoryFileBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_5fiona_6ogrext_21__pyx_unpickle_MemoryFileBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryFileBase", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xd41d8cd) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xd41d8cd:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xd41d8cd:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_5fiona_6ogrext_MemoryFileBase), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_5fiona_6ogrext___pyx_unpickle_MemoryFileBase__set_state(((struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xd41d8cd = ())" % __pyx_checksum)
 *     __pyx_result = MemoryFileBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MemoryFileBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_MemoryFileBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

static PyObject *__pyx_f_5fiona_6ogrext___pyx_unpickle_MemoryFileBase__set_state(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MemoryFileBase__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = ((__pyx_t_2 > 0) != 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_4 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_5 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MemoryFileBase__set_state(<MemoryFileBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MemoryFileBase__set_state(MemoryFileBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fiona.ogrext.__pyx_unpickle_MemoryFileBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_5fiona_6ogrext_FeatureBuilder __pyx_vtable_5fiona_6ogrext_FeatureBuilder;

static PyObject *__pyx_tp_new_5fiona_6ogrext_FeatureBuilder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *)o);
  p->__pyx_vtab = __pyx_vtabptr_5fiona_6ogrext_FeatureBuilder;
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_FeatureBuilder(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_5fiona_6ogrext_FeatureBuilder[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14FeatureBuilder_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14FeatureBuilder_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_FeatureBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.FeatureBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_FeatureBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Build Fiona features from OGR feature pointers.\n\n    No OGR objects are allocated by this function and the feature\n    argument is not destroyed.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_FeatureBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_FeatureBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_5fiona_6ogrext_OGRFeatureBuilder __pyx_vtable_5fiona_6ogrext_OGRFeatureBuilder;

static PyObject *__pyx_tp_new_5fiona_6ogrext_OGRFeatureBuilder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *)o);
  p->__pyx_vtab = __pyx_vtabptr_5fiona_6ogrext_OGRFeatureBuilder;
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_OGRFeatureBuilder(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_5fiona_6ogrext_OGRFeatureBuilder[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_17OGRFeatureBuilder_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_OGRFeatureBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.OGRFeatureBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_OGRFeatureBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Builds an OGR Feature from a Fiona feature mapping.\n\n    Allocates one OGR Feature which should be destroyed by the caller.\n    Borrows a layer definition from the collection.\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_OGRFeatureBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_OGRFeatureBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_5fiona_6ogrext_Session __pyx_vtable_5fiona_6ogrext_Session;

static PyObject *__pyx_tp_new_5fiona_6ogrext_Session(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_5fiona_6ogrext_Session *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5fiona_6ogrext_Session *)o);
  p->__pyx_vtab = __pyx_vtabptr_5fiona_6ogrext_Session;
  p->_fileencoding = Py_None; Py_INCREF(Py_None);
  p->_encoding = Py_None; Py_INCREF(Py_None);
  p->collection = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_Session(PyObject *o) {
  struct __pyx_obj_5fiona_6ogrext_Session *p = (struct __pyx_obj_5fiona_6ogrext_Session *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    ++Py_REFCNT(o);
    __pyx_pw_5fiona_6ogrext_7Session_3__dealloc__(o);
    --Py_REFCNT(o);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_fileencoding);
  Py_CLEAR(p->_encoding);
  Py_CLEAR(p->collection);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5fiona_6ogrext_Session(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5fiona_6ogrext_Session *p = (struct __pyx_obj_5fiona_6ogrext_Session *)o;
  if (p->_fileencoding) {
    e = (*v)(p->_fileencoding, a); if (e) return e;
  }
  if (p->_encoding) {
    e = (*v)(p->_encoding, a); if (e) return e;
  }
  if (p->collection) {
    e = (*v)(p->collection, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5fiona_6ogrext_Session(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5fiona_6ogrext_Session *p = (struct __pyx_obj_5fiona_6ogrext_Session *)o;
  tmp = ((PyObject*)p->_fileencoding);
  p->_fileencoding = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_encoding);
  p->_encoding = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->collection);
  p->collection = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
static PyObject *__pyx_sq_item_5fiona_6ogrext_Session(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static PyMethodDef __pyx_methods_5fiona_6ogrext_Session[] = {
  {"start", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_7Session_5start, METH_VARARGS|METH_KEYWORDS, 0},
  {"stop", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_7stop, METH_NOARGS, 0},
  {"get_fileencoding", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_9get_fileencoding, METH_NOARGS, __pyx_doc_5fiona_6ogrext_7Session_8get_fileencoding},
  {"_get_fallback_encoding", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_11_get_fallback_encoding, METH_NOARGS, __pyx_doc_5fiona_6ogrext_7Session_10_get_fallback_encoding},
  {"_get_internal_encoding", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_13_get_internal_encoding, METH_NOARGS, __pyx_doc_5fiona_6ogrext_7Session_12_get_internal_encoding},
  {"get_length", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_15get_length, METH_NOARGS, 0},
  {"get_driver", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_17get_driver, METH_NOARGS, 0},
  {"get_schema", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_19get_schema, METH_NOARGS, 0},
  {"get_crs", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_21get_crs, METH_NOARGS, __pyx_doc_5fiona_6ogrext_7Session_20get_crs},
  {"get_crs_wkt", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_23get_crs_wkt, METH_NOARGS, 0},
  {"get_extent", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_25get_extent, METH_NOARGS, 0},
  {"has_feature", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_27has_feature, METH_O, __pyx_doc_5fiona_6ogrext_7Session_26has_feature},
  {"get_feature", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_29get_feature, METH_O, __pyx_doc_5fiona_6ogrext_7Session_28get_feature},
  {"isactive", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_33isactive, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_35__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_7Session_37__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_Session = {
  0, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_5fiona_6ogrext_Session, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_Session = {
  0, /*mp_length*/
  __pyx_pw_5fiona_6ogrext_7Session_31__getitem__, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_5fiona_6ogrext_Session = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.Session", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_Session), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_Session, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_Session, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_Session, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext_Session, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext_Session, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_Session, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5fiona_6ogrext_7Session_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_Session, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_5fiona_6ogrext_WritingSession __pyx_vtable_5fiona_6ogrext_WritingSession;

static PyObject *__pyx_tp_new_5fiona_6ogrext_WritingSession(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5fiona_6ogrext_WritingSession *p;
  PyObject *o = __pyx_tp_new_5fiona_6ogrext_Session(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5fiona_6ogrext_WritingSession *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_5fiona_6ogrext_Session*)__pyx_vtabptr_5fiona_6ogrext_WritingSession;
  p->_schema_mapping = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_WritingSession(PyObject *o) {
  struct __pyx_obj_5fiona_6ogrext_WritingSession *p = (struct __pyx_obj_5fiona_6ogrext_WritingSession *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_schema_mapping);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_5fiona_6ogrext_Session(o);
}

static int __pyx_tp_traverse_5fiona_6ogrext_WritingSession(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5fiona_6ogrext_WritingSession *p = (struct __pyx_obj_5fiona_6ogrext_WritingSession *)o;
  e = __pyx_tp_traverse_5fiona_6ogrext_Session(o, v, a); if (e) return e;
  if (p->_schema_mapping) {
    e = (*v)(p->_schema_mapping, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5fiona_6ogrext_WritingSession(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5fiona_6ogrext_WritingSession *p = (struct __pyx_obj_5fiona_6ogrext_WritingSession *)o;
  __pyx_tp_clear_5fiona_6ogrext_Session(o);
  tmp = ((PyObject*)p->_schema_mapping);
  p->_schema_mapping = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_5fiona_6ogrext_WritingSession[] = {
  {"start", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_14WritingSession_1start, METH_VARARGS|METH_KEYWORDS, 0},
  {"writerecs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_14WritingSession_3writerecs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_14WritingSession_2writerecs},
  {"sync", (PyCFunction)__pyx_pw_5fiona_6ogrext_14WritingSession_5sync, METH_O, __pyx_doc_5fiona_6ogrext_14WritingSession_4sync},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14WritingSession_7__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14WritingSession_9__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_WritingSession = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.WritingSession", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_WritingSession), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_WritingSession, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext_WritingSession, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext_WritingSession, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_WritingSession, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_5fiona_6ogrext_7Session_1__init__, /*tp_init*/
  #else
  0, /*tp_init*/
  #endif
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_WritingSession, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_5fiona_6ogrext_Iterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_5fiona_6ogrext_Iterator *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_5fiona_6ogrext_Iterator *)o);
  p->collection = Py_None; Py_INCREF(Py_None);
  p->encoding = Py_None; Py_INCREF(Py_None);
  p->stop = Py_None; Py_INCREF(Py_None);
  p->start = Py_None; Py_INCREF(Py_None);
  p->step = Py_None; Py_INCREF(Py_None);
  p->fastindex = Py_None; Py_INCREF(Py_None);
  p->stepsign = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_5fiona_6ogrext_8Iterator_1__cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_Iterator(PyObject *o) {
  struct __pyx_obj_5fiona_6ogrext_Iterator *p = (struct __pyx_obj_5fiona_6ogrext_Iterator *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->collection);
  Py_CLEAR(p->encoding);
  Py_CLEAR(p->stop);
  Py_CLEAR(p->start);
  Py_CLEAR(p->step);
  Py_CLEAR(p->fastindex);
  Py_CLEAR(p->stepsign);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_5fiona_6ogrext_Iterator(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5fiona_6ogrext_Iterator *p = (struct __pyx_obj_5fiona_6ogrext_Iterator *)o;
  if (p->collection) {
    e = (*v)(p->collection, a); if (e) return e;
  }
  if (p->encoding) {
    e = (*v)(p->encoding, a); if (e) return e;
  }
  if (p->stop) {
    e = (*v)(p->stop, a); if (e) return e;
  }
  if (p->start) {
    e = (*v)(p->start, a); if (e) return e;
  }
  if (p->step) {
    e = (*v)(p->step, a); if (e) return e;
  }
  if (p->fastindex) {
    e = (*v)(p->fastindex, a); if (e) return e;
  }
  if (p->stepsign) {
    e = (*v)(p->stepsign, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5fiona_6ogrext_Iterator(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5fiona_6ogrext_Iterator *p = (struct __pyx_obj_5fiona_6ogrext_Iterator *)o;
  tmp = ((PyObject*)p->collection);
  p->collection = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->encoding);
  p->encoding = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->stop);
  p->stop = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->start);
  p->start = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->step);
  p->step = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->fastindex);
  p->fastindex = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->stepsign);
  p->stepsign = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_specialmethod___pyx_pw_5fiona_6ogrext_8Iterator_7__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_5fiona_6ogrext_8Iterator_7__next__(self);}

static PyMethodDef __pyx_methods_5fiona_6ogrext_Iterator[] = {
  {"_next", (PyCFunction)__pyx_pw_5fiona_6ogrext_8Iterator_5_next, METH_NOARGS, __pyx_doc_5fiona_6ogrext_8Iterator_4_next},
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_5fiona_6ogrext_8Iterator_7__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_8Iterator_9__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_8Iterator_11__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_Iterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.Iterator", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_Iterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_Iterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Provides iterated access to feature data.\n    ", /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext_Iterator, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext_Iterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_5fiona_6ogrext_8Iterator_3__iter__, /*tp_iter*/
  __pyx_pw_5fiona_6ogrext_8Iterator_7__next__, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_Iterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_Iterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_5fiona_6ogrext_ItemsIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __pyx_tp_new_5fiona_6ogrext_Iterator(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static PyObject *__pyx_specialmethod___pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__(self);}

static PyMethodDef __pyx_methods_5fiona_6ogrext_ItemsIterator[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_13ItemsIterator_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_13ItemsIterator_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_ItemsIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.ItemsIterator", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_ItemsIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_Iterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext_Iterator, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext_Iterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_5fiona_6ogrext_8Iterator_3__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  __pyx_pw_5fiona_6ogrext_13ItemsIterator_1__next__, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_ItemsIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_ItemsIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_5fiona_6ogrext_KeysIterator(PyTypeObject *t, PyObject *a, PyObject *k) {
  PyObject *o = __pyx_tp_new_5fiona_6ogrext_Iterator(t, a, k);
  if (unlikely(!o)) return 0;
  return o;
}

static PyObject *__pyx_specialmethod___pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__(PyObject *self, CYTHON_UNUSED PyObject *arg) {return __pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__(self);}

static PyMethodDef __pyx_methods_5fiona_6ogrext_KeysIterator[] = {
  {"__next__", (PyCFunction)__pyx_specialmethod___pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_12KeysIterator_3__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_12KeysIterator_5__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_5fiona_6ogrext_KeysIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.KeysIterator", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_KeysIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_Iterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext_Iterator, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext_Iterator, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_pw_5fiona_6ogrext_8Iterator_3__iter__, /*tp_iter*/
  #else
  0, /*tp_iter*/
  #endif
  __pyx_pw_5fiona_6ogrext_12KeysIterator_1__next__, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_KeysIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_KeysIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_5fiona_6ogrext_MemoryFileBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext_MemoryFileBase(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_5fiona_6ogrext_MemoryFileBase[] = {
  {"exists", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_3exists, METH_NOARGS, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_2exists},
  {"close", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_7close, METH_NOARGS, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_6close},
  {"read", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_9read, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_8read},
  {"seek", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_11seek, METH_VARARGS|METH_KEYWORDS, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_10seek},
  {"tell", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_13tell, METH_NOARGS, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_12tell},
  {"write", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_15write, METH_O, __pyx_doc_5fiona_6ogrext_14MemoryFileBase_14write},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_17__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_5fiona_6ogrext_14MemoryFileBase_19__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_MemoryFileBase = {
  __pyx_pw_5fiona_6ogrext_14MemoryFileBase_5__len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  0, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_MemoryFileBase = {
  __pyx_pw_5fiona_6ogrext_14MemoryFileBase_5__len__, /*mp_length*/
  0, /*mp_subscript*/
  0, /*mp_ass_subscript*/
};

static PyTypeObject __pyx_type_5fiona_6ogrext_MemoryFileBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.MemoryFileBase", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext_MemoryFileBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext_MemoryFileBase, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_MemoryFileBase, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_MemoryFileBase, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Base for a BytesIO-like class backed by an in-memory file.", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_5fiona_6ogrext_MemoryFileBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_5fiona_6ogrext_14MemoryFileBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext_MemoryFileBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *__pyx_freelist_5fiona_6ogrext___pyx_scope_struct___explode[8];
static int __pyx_freecount_5fiona_6ogrext___pyx_scope_struct___explode = 0;

static PyObject *__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct___explode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5fiona_6ogrext___pyx_scope_struct___explode > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode)))) {
    o = (PyObject*)__pyx_freelist_5fiona_6ogrext___pyx_scope_struct___explode[--__pyx_freecount_5fiona_6ogrext___pyx_scope_struct___explode];
    memset(o, 0, sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext___pyx_scope_struct___explode(PyObject *o) {
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *p = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_coords);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_3);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5fiona_6ogrext___pyx_scope_struct___explode < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode)))) {
    __pyx_freelist_5fiona_6ogrext___pyx_scope_struct___explode[__pyx_freecount_5fiona_6ogrext___pyx_scope_struct___explode++] = ((struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5fiona_6ogrext___pyx_scope_struct___explode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *p = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode *)o;
  if (p->__pyx_v_coords) {
    e = (*v)(p->__pyx_v_coords, a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_3) {
    e = (*v)(p->__pyx_t_3, a); if (e) return e;
  }
  return 0;
}

static PyTypeObject __pyx_type_5fiona_6ogrext___pyx_scope_struct___explode = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.__pyx_scope_struct___explode", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct___explode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext___pyx_scope_struct___explode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext___pyx_scope_struct___explode, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext___pyx_scope_struct___explode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *__pyx_freelist_5fiona_6ogrext___pyx_scope_struct_1_writerecs[8];
static int __pyx_freecount_5fiona_6ogrext___pyx_scope_struct_1_writerecs = 0;

static PyObject *__pyx_tp_new_5fiona_6ogrext___pyx_scope_struct_1_writerecs(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_5fiona_6ogrext___pyx_scope_struct_1_writerecs > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs)))) {
    o = (PyObject*)__pyx_freelist_5fiona_6ogrext___pyx_scope_struct_1_writerecs[--__pyx_freecount_5fiona_6ogrext___pyx_scope_struct_1_writerecs];
    memset(o, 0, sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_5fiona_6ogrext___pyx_scope_struct_1_writerecs(PyObject *o) {
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *p = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_valid_geom_types);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_5fiona_6ogrext___pyx_scope_struct_1_writerecs < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs)))) {
    __pyx_freelist_5fiona_6ogrext___pyx_scope_struct_1_writerecs[__pyx_freecount_5fiona_6ogrext___pyx_scope_struct_1_writerecs++] = ((struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_5fiona_6ogrext___pyx_scope_struct_1_writerecs(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *p = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)o;
  if (p->__pyx_v_valid_geom_types) {
    e = (*v)(p->__pyx_v_valid_geom_types, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_5fiona_6ogrext___pyx_scope_struct_1_writerecs(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *p = (struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs *)o;
  tmp = ((PyObject*)p->__pyx_v_valid_geom_types);
  p->__pyx_v_valid_geom_types = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs = {
  PyVarObject_HEAD_INIT(0, 0)
  "fiona.ogrext.__pyx_scope_struct_1_writerecs", /*tp_name*/
  sizeof(struct __pyx_obj_5fiona_6ogrext___pyx_scope_struct_1_writerecs), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_5fiona_6ogrext___pyx_scope_struct_1_writerecs, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_5fiona_6ogrext___pyx_scope_struct_1_writerecs, /*tp_traverse*/
  __pyx_tp_clear_5fiona_6ogrext___pyx_scope_struct_1_writerecs, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_5fiona_6ogrext___pyx_scope_struct_1_writerecs, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_ogrext(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_ogrext},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "ogrext",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_3D, __pyx_k_3D, sizeof(__pyx_k_3D), 0, 1, 0, 0},
  {&__pyx_kp_u_A_support_file_wasn_t_opened_See, __pyx_k_A_support_file_wasn_t_opened_See, sizeof(__pyx_k_A_support_file_wasn_t_opened_See), 0, 1, 0, 0},
  {&__pyx_n_u_Any, __pyx_k_Any, sizeof(__pyx_k_Any), 0, 1, 0, 1},
  {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
  {&__pyx_kp_u_Begin_creating_field_r_value_r, __pyx_k_Begin_creating_field_r_value_r, sizeof(__pyx_k_Begin_creating_field_r_value_r), 0, 1, 0, 0},
  {&__pyx_kp_u_By_request_ignoring_field_r, __pyx_k_By_request_ignoring_field_r, sizeof(__pyx_k_By_request_ignoring_field_r), 0, 1, 0, 0},
  {&__pyx_n_s_CPLE_BaseError, __pyx_k_CPLE_BaseError, sizeof(__pyx_k_CPLE_BaseError), 0, 0, 1, 1},
  {&__pyx_n_s_CPLE_OpenFailedError, __pyx_k_CPLE_OpenFailedError, sizeof(__pyx_k_CPLE_OpenFailedError), 0, 0, 1, 1},
  {&__pyx_n_s_CRSError, __pyx_k_CRSError, sizeof(__pyx_k_CRSError), 0, 0, 1, 1},
  {&__pyx_kp_u_Called_OGR_L_GetSpatialRef_again, __pyx_k_Called_OGR_L_GetSpatialRef_again, sizeof(__pyx_k_Called_OGR_L_GetSpatialRef_again), 0, 1, 0, 0},
  {&__pyx_kp_u_Collection_slicing_is_deprecated, __pyx_k_Collection_slicing_is_deprecated, sizeof(__pyx_k_Collection_slicing_is_deprecated), 0, 1, 0, 0},
  {&__pyx_kp_u_Comitting_transaction_final, __pyx_k_Comitting_transaction_final, sizeof(__pyx_k_Comitting_transaction_final), 0, 1, 0, 0},
  {&__pyx_kp_u_Comitting_transaction_intermedia, __pyx_k_Comitting_transaction_intermedia, sizeof(__pyx_k_Comitting_transaction_intermedia), 0, 1, 0, 0},
  {&__pyx_kp_u_Constructor_argument_must_be_a_f, __pyx_k_Constructor_argument_must_be_a_f, sizeof(__pyx_k_Constructor_argument_must_be_a_f), 0, 1, 0, 0},
  {&__pyx_kp_u_Created_layer_s, __pyx_k_Created_layer_s, sizeof(__pyx_k_Created_layer_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Creating_feature_in_layer_s, __pyx_k_Creating_feature_in_layer_s, sizeof(__pyx_k_Creating_feature_in_layer_s), 0, 1, 0, 0},
  {&__pyx_n_s_DEFAULT_TRANSACTION_SIZE, __pyx_k_DEFAULT_TRANSACTION_SIZE, sizeof(__pyx_k_DEFAULT_TRANSACTION_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_DatasetDeleteError, __pyx_k_DatasetDeleteError, sizeof(__pyx_k_DatasetDeleteError), 0, 0, 1, 1},
  {&__pyx_kp_u_Deleted_pre_existing_layer_at_s, __pyx_k_Deleted_pre_existing_layer_at_s, sizeof(__pyx_k_Deleted_pre_existing_layer_at_s), 0, 1, 0, 0},
  {&__pyx_n_s_DriverError, __pyx_k_DriverError, sizeof(__pyx_k_DriverError), 0, 0, 1, 1},
  {&__pyx_n_s_DriverIOError, __pyx_k_DriverIOError, sizeof(__pyx_k_DriverIOError), 0, 0, 1, 1},
  {&__pyx_kp_u_Driver_does_not_support_dataset, __pyx_k_Driver_does_not_support_dataset, sizeof(__pyx_k_Driver_does_not_support_dataset), 0, 1, 0, 0},
  {&__pyx_n_u_EPSG, __pyx_k_EPSG, sizeof(__pyx_k_EPSG), 0, 1, 0, 1},
  {&__pyx_kp_u_End_creating_field_r, __pyx_k_End_creating_field_r, sizeof(__pyx_k_End_creating_field_r), 0, 1, 0, 0},
  {&__pyx_n_s_FIELD_TYPES, __pyx_k_FIELD_TYPES, sizeof(__pyx_k_FIELD_TYPES), 0, 0, 1, 1},
  {&__pyx_n_s_FIELD_TYPES_MAP, __pyx_k_FIELD_TYPES_MAP, sizeof(__pyx_k_FIELD_TYPES_MAP), 0, 0, 1, 1},
  {&__pyx_kp_u_Failed_to_auto_identify_EPSG_d, __pyx_k_Failed_to_auto_identify_EPSG_d, sizeof(__pyx_k_Failed_to_auto_identify_EPSG_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_commit_transaction, __pyx_k_Failed_to_commit_transaction, sizeof(__pyx_k_Failed_to_commit_transaction), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_create_in_memory_file, __pyx_k_Failed_to_create_in_memory_file, sizeof(__pyx_k_Failed_to_create_in_memory_file), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_decode_s_using_s_codec, __pyx_k_Failed_to_decode_s_using_s_codec, sizeof(__pyx_k_Failed_to_decode_s_using_s_codec), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_encode_s_using_s_codec, __pyx_k_Failed_to_encode_s_using_s_codec, sizeof(__pyx_k_Failed_to_encode_s_using_s_codec), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_properly_close_in_memo, __pyx_k_Failed_to_properly_close_in_memo, sizeof(__pyx_k_Failed_to_properly_close_in_memo), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_remove_data_source, __pyx_k_Failed_to_remove_data_source, sizeof(__pyx_k_Failed_to_remove_data_source), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_remove_layer_from_data, __pyx_k_Failed_to_remove_layer_from_data, sizeof(__pyx_k_Failed_to_remove_layer_from_data), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_seek_to_offset_s_in_s, __pyx_k_Failed_to_seek_to_offset_s_in_s, sizeof(__pyx_k_Failed_to_seek_to_offset_s_in_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_start_transaction, __pyx_k_Failed_to_start_transaction, sizeof(__pyx_k_Failed_to_start_transaction), 0, 1, 0, 0},
  {&__pyx_kp_u_Failed_to_write_record_s, __pyx_k_Failed_to_write_record_s, sizeof(__pyx_k_Failed_to_write_record_s), 0, 1, 0, 0},
  {&__pyx_n_u_False, __pyx_k_False, sizeof(__pyx_k_False), 0, 1, 0, 1},
  {&__pyx_n_u_Feature, __pyx_k_Feature, sizeof(__pyx_k_Feature), 0, 1, 0, 1},
  {&__pyx_n_s_FeatureBuilder, __pyx_k_FeatureBuilder, sizeof(__pyx_k_FeatureBuilder), 0, 0, 1, 1},
  {&__pyx_kp_u_Field_format_parsing_value_r_fmt, __pyx_k_Field_format_parsing_value_r_fmt, sizeof(__pyx_k_Field_format_parsing_value_r_fmt), 0, 1, 0, 0},
  {&__pyx_kp_u_File_doesn_t_exist_Creating_a_ne, __pyx_k_File_doesn_t_exist_Creating_a_ne, sizeof(__pyx_k_File_doesn_t_exist_Creating_a_ne), 0, 1, 0, 0},
  {&__pyx_n_s_FionaDateTimeType, __pyx_k_FionaDateTimeType, sizeof(__pyx_k_FionaDateTimeType), 0, 0, 1, 1},
  {&__pyx_n_s_FionaDateType, __pyx_k_FionaDateType, sizeof(__pyx_k_FionaDateType), 0, 0, 1, 1},
  {&__pyx_n_s_FionaDeprecationWarning, __pyx_k_FionaDeprecationWarning, sizeof(__pyx_k_FionaDeprecationWarning), 0, 0, 1, 1},
  {&__pyx_n_s_FionaNullPointerError, __pyx_k_FionaNullPointerError, sizeof(__pyx_k_FionaNullPointerError), 0, 0, 1, 1},
  {&__pyx_n_s_FionaTimeType, __pyx_k_FionaTimeType, sizeof(__pyx_k_FionaTimeType), 0, 0, 1, 1},
  {&__pyx_n_s_FionaValueError, __pyx_k_FionaValueError, sizeof(__pyx_k_FionaValueError), 0, 0, 1, 1},
  {&__pyx_kp_u_Flushed_data_source_cache, __pyx_k_Flushed_data_source_cache, sizeof(__pyx_k_Flushed_data_source_cache), 0, 1, 0, 0},
  {&__pyx_n_s_GDALVersion, __pyx_k_GDALVersion, sizeof(__pyx_k_GDALVersion), 0, 0, 1, 1},
  {&__pyx_n_s_GEOMETRY_TYPES, __pyx_k_GEOMETRY_TYPES, sizeof(__pyx_k_GEOMETRY_TYPES), 0, 0, 1, 1},
  {&__pyx_n_u_GeoJSON, __pyx_k_GeoJSON, sizeof(__pyx_k_GeoJSON), 0, 1, 0, 1},
  {&__pyx_n_s_GeometryTypeValidationError, __pyx_k_GeometryTypeValidationError, sizeof(__pyx_k_GeometryTypeValidationError), 0, 0, 1, 1},
  {&__pyx_kp_u_Got_coordinate_system, __pyx_k_Got_coordinate_system, sizeof(__pyx_k_Got_coordinate_system), 0, 1, 0, 0},
  {&__pyx_n_s_IOError, __pyx_k_IOError, sizeof(__pyx_k_IOError), 0, 0, 1, 1},
  {&__pyx_kp_u_I_O_operation_on_closed_collecti, __pyx_k_I_O_operation_on_closed_collecti, sizeof(__pyx_k_I_O_operation_on_closed_collecti), 0, 1, 0, 0},
  {&__pyx_kp_u_Ignored_field_has_type_expected, __pyx_k_Ignored_field_has_type_expected, sizeof(__pyx_k_Ignored_field_has_type_expected), 0, 1, 0, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xd4, __pyx_k_Incompatible_checksums_s_vs_0xd4, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xd4), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_u_Index_d, __pyx_k_Index_d, sizeof(__pyx_k_Index_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Invalid_field_name_ref_s, __pyx_k_Invalid_field_name_ref_s, sizeof(__pyx_k_Invalid_field_name_ref_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Invalid_field_type_s, __pyx_k_Invalid_field_type_s, sizeof(__pyx_k_Invalid_field_type_s), 0, 1, 0, 0},
  {&__pyx_n_s_ItemsIterator, __pyx_k_ItemsIterator, sizeof(__pyx_k_ItemsIterator), 0, 0, 1, 1},
  {&__pyx_n_s_Iterator, __pyx_k_Iterator, sizeof(__pyx_k_Iterator), 0, 0, 1, 1},
  {&__pyx_n_s_KeyError, __pyx_k_KeyError, sizeof(__pyx_k_KeyError), 0, 0, 1, 1},
  {&__pyx_n_s_KeysIterator, __pyx_k_KeysIterator, sizeof(__pyx_k_KeysIterator), 0, 0, 1, 1},
  {&__pyx_kp_u_Layer_does_not_exist_in_datasour, __pyx_k_Layer_does_not_exist_in_datasour, sizeof(__pyx_k_Layer_does_not_exist_in_datasour), 0, 1, 0, 0},
  {&__pyx_kp_u_Layer_does_not_supportOLCFastSet, __pyx_k_Layer_does_not_supportOLCFastSet, sizeof(__pyx_k_Layer_does_not_supportOLCFastSet), 0, 1, 0, 0},
  {&__pyx_kp_u_Layer_has_no_coordinate_system, __pyx_k_Layer_has_no_coordinate_system, sizeof(__pyx_k_Layer_has_no_coordinate_system), 0, 1, 0, 0},
  {&__pyx_kp_u_Looking_up_s_in_s, __pyx_k_Looking_up_s_in_s, sizeof(__pyx_k_Looking_up_s_in_s), 0, 1, 0, 0},
  {&__pyx_n_s_MemoryFileBase, __pyx_k_MemoryFileBase, sizeof(__pyx_k_MemoryFileBase), 0, 0, 1, 1},
  {&__pyx_kp_u_No_such_file_or_directory_s, __pyx_k_No_such_file_or_directory_s, sizeof(__pyx_k_No_such_file_or_directory_s), 0, 1, 0, 0},
  {&__pyx_kp_u_None_fieldtype, __pyx_k_None_fieldtype, sizeof(__pyx_k_None_fieldtype), 0, 1, 0, 0},
  {&__pyx_kp_u_Normalizing_schema_type_for_key, __pyx_k_Normalizing_schema_type_for_key, sizeof(__pyx_k_Normalizing_schema_type_for_key), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_data_source, __pyx_k_Null_data_source, sizeof(__pyx_k_Null_data_source), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_driver, __pyx_k_Null_driver, sizeof(__pyx_k_Null_driver), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_driver_when_attempting_to_d, __pyx_k_Null_driver_when_attempting_to_d, sizeof(__pyx_k_Null_driver_when_attempting_to_d), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_feature, __pyx_k_Null_feature, sizeof(__pyx_k_Null_feature), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_feature_definition, __pyx_k_Null_feature_definition, sizeof(__pyx_k_Null_feature_definition), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_field_definition, __pyx_k_Null_field_definition, sizeof(__pyx_k_Null_field_definition), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_field_name_reference, __pyx_k_Null_field_name_reference, sizeof(__pyx_k_Null_field_name_reference), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_geometry, __pyx_k_Null_geometry, sizeof(__pyx_k_Null_geometry), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_layer, __pyx_k_Null_layer, sizeof(__pyx_k_Null_layer), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_layer_2, __pyx_k_Null_layer_2, sizeof(__pyx_k_Null_layer_2), 0, 1, 0, 0},
  {&__pyx_kp_u_Null_projection, __pyx_k_Null_projection, sizeof(__pyx_k_Null_projection), 0, 1, 0, 0},
  {&__pyx_n_u_OFF, __pyx_k_OFF, sizeof(__pyx_k_OFF), 0, 1, 0, 1},
  {&__pyx_n_s_OGRERR_CORRUPT_DATA, __pyx_k_OGRERR_CORRUPT_DATA, sizeof(__pyx_k_OGRERR_CORRUPT_DATA), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_FAILURE, __pyx_k_OGRERR_FAILURE, sizeof(__pyx_k_OGRERR_FAILURE), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_INVALID_HANDLE, __pyx_k_OGRERR_INVALID_HANDLE, sizeof(__pyx_k_OGRERR_INVALID_HANDLE), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_NONE, __pyx_k_OGRERR_NONE, sizeof(__pyx_k_OGRERR_NONE), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_NOT_ENOUGH_DATA, __pyx_k_OGRERR_NOT_ENOUGH_DATA, sizeof(__pyx_k_OGRERR_NOT_ENOUGH_DATA), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_NOT_ENOUGH_MEMORY, __pyx_k_OGRERR_NOT_ENOUGH_MEMORY, sizeof(__pyx_k_OGRERR_NOT_ENOUGH_MEMORY), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_UNSUPPORTED_GEOMETRY_TYPE, __pyx_k_OGRERR_UNSUPPORTED_GEOMETRY_TYPE, sizeof(__pyx_k_OGRERR_UNSUPPORTED_GEOMETRY_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_UNSUPPORTED_OPERATION, __pyx_k_OGRERR_UNSUPPORTED_OPERATION, sizeof(__pyx_k_OGRERR_UNSUPPORTED_OPERATION), 0, 0, 1, 1},
  {&__pyx_n_s_OGRERR_UNSUPPORTED_SRS, __pyx_k_OGRERR_UNSUPPORTED_SRS, sizeof(__pyx_k_OGRERR_UNSUPPORTED_SRS), 0, 0, 1, 1},
  {&__pyx_n_s_OGRFeatureBuilder, __pyx_k_OGRFeatureBuilder, sizeof(__pyx_k_OGRFeatureBuilder), 0, 0, 1, 1},
  {&__pyx_n_u_ON, __pyx_k_ON, sizeof(__pyx_k_ON), 0, 1, 0, 1},
  {&__pyx_n_s_OSError, __pyx_k_OSError, sizeof(__pyx_k_OSError), 0, 0, 1, 1},
  {&__pyx_n_s_OrderedDict, __pyx_k_OrderedDict, sizeof(__pyx_k_OrderedDict), 0, 0, 1, 1},
  {&__pyx_kp_u_Params_s, __pyx_k_Params_s, sizeof(__pyx_k_Params_s), 0, 1, 0, 0},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_kp_u_Projection_not_found_cogr_crs_wa, __pyx_k_Projection_not_found_cogr_crs_wa, sizeof(__pyx_k_Projection_not_found_cogr_crs_wa), 0, 1, 0, 0},
  {&__pyx_kp_u_Record_does_not_match_collection, __pyx_k_Record_does_not_match_collection, sizeof(__pyx_k_Record_does_not_match_collection), 0, 1, 0, 0},
  {&__pyx_kp_u_Record_s_geometry_type_does_not, __pyx_k_Record_s_geometry_type_does_not, sizeof(__pyx_k_Record_s_geometry_type_does_not), 0, 1, 0, 0},
  {&__pyx_kp_u_Removal_of_layer_not_supported_b, __pyx_k_Removal_of_layer_not_supported_b, sizeof(__pyx_k_Removal_of_layer_not_supported_b), 0, 1, 0, 0},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeWarning, __pyx_k_RuntimeWarning, sizeof(__pyx_k_RuntimeWarning), 0, 0, 1, 1},
  {&__pyx_n_s_SchemaError, __pyx_k_SchemaError, sizeof(__pyx_k_SchemaError), 0, 0, 1, 1},
  {&__pyx_n_s_Session, __pyx_k_Session, sizeof(__pyx_k_Session), 0, 0, 1, 1},
  {&__pyx_kp_u_Session_is_inactive_dataset_is_c, __pyx_k_Session_is_inactive_dataset_is_c, sizeof(__pyx_k_Session_is_inactive_dataset_is_c), 0, 1, 0, 0},
  {&__pyx_kp_u_Set_field, __pyx_k_Set_field, sizeof(__pyx_k_Set_field), 0, 1, 0, 0},
  {&__pyx_kp_u_Set_option_r_r, __pyx_k_Set_option_r_r, sizeof(__pyx_k_Set_option_r_r), 0, 1, 0, 0},
  {&__pyx_kp_u_Setting_field_r_type_r_to_value, __pyx_k_Setting_field_r_type_r_to_value, sizeof(__pyx_k_Setting_field_r_type_r_to_value), 0, 1, 0, 0},
  {&__pyx_n_u_Shapefile, __pyx_k_Shapefile, sizeof(__pyx_k_Shapefile), 0, 1, 0, 1},
  {&__pyx_kp_u_Skipping_field_s_invalid_type_s, __pyx_k_Skipping_field_s_invalid_type_s, sizeof(__pyx_k_Skipping_field_s_invalid_type_s), 0, 1, 0, 0},
  {&__pyx_kp_u_Slice_r, __pyx_k_Slice_r, sizeof(__pyx_k_Slice_r), 0, 1, 0, 0},
  {&__pyx_kp_u_Starting_transaction_initial, __pyx_k_Starting_transaction_initial, sizeof(__pyx_k_Starting_transaction_initial), 0, 1, 0, 0},
  {&__pyx_kp_u_Starting_transaction_intermediat, __pyx_k_Starting_transaction_intermediat, sizeof(__pyx_k_Starting_transaction_intermediat), 0, 1, 0, 0},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_kp_u_There_is_no_feature_with_fid_r, __pyx_k_There_is_no_feature_with_fid_r, sizeof(__pyx_k_There_is_no_feature_with_fid_r), 0, 1, 0, 0},
  {&__pyx_n_s_TransactionError, __pyx_k_TransactionError, sizeof(__pyx_k_TransactionError), 0, 0, 1, 1},
  {&__pyx_n_u_True, __pyx_k_True, sizeof(__pyx_k_True), 0, 1, 0, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_u_Unexpected_proj_parameter_s, __pyx_k_Unexpected_proj_parameter_s, sizeof(__pyx_k_Unexpected_proj_parameter_s), 0, 1, 0, 0},
  {&__pyx_n_s_UnicodeDecodeError, __pyx_k_UnicodeDecodeError, sizeof(__pyx_k_UnicodeDecodeError), 0, 0, 1, 1},
  {&__pyx_n_s_UnicodeEncodeError, __pyx_k_UnicodeEncodeError, sizeof(__pyx_k_UnicodeEncodeError), 0, 0, 1, 1},
  {&__pyx_n_u_Unknown, __pyx_k_Unknown, sizeof(__pyx_k_Unknown), 0, 1, 0, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_WritingSession, __pyx_k_WritingSession, sizeof(__pyx_k_WritingSession), 0, 0, 1, 1},
  {&__pyx_kp_u_Writing_started, __pyx_k_Writing_started, sizeof(__pyx_k_Writing_started), 0, 1, 0, 0},
  {&__pyx_kp_b__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 0, 0, 0},
  {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
  {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
  {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
  {&__pyx_kp_u__16, __pyx_k__16, sizeof(__pyx_k__16), 0, 1, 0, 0},
  {&__pyx_kp_u__21, __pyx_k__21, sizeof(__pyx_k__21), 0, 1, 0, 0},
  {&__pyx_kp_u__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 1, 0, 0},
  {&__pyx_kp_u__25, __pyx_k__25, sizeof(__pyx_k__25), 0, 1, 0, 0},
  {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
  {&__pyx_kp_u__40, __pyx_k__40, sizeof(__pyx_k__40), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
  {&__pyx_n_u_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 1, 0, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_bbox, __pyx_k_bbox, sizeof(__pyx_k_bbox), 0, 0, 1, 1},
  {&__pyx_n_u_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 1, 0, 1},
  {&__pyx_n_s_bounds, __pyx_k_bounds, sizeof(__pyx_k_bounds), 0, 0, 1, 1},
  {&__pyx_n_s_buffer_to_virtual_file, __pyx_k_buffer_to_virtual_file, sizeof(__pyx_k_buffer_to_virtual_file), 0, 0, 1, 1},
  {&__pyx_n_u_bytes, __pyx_k_bytes, sizeof(__pyx_k_bytes), 0, 1, 0, 1},
  {&__pyx_n_s_bytesbuf, __pyx_k_bytesbuf, sizeof(__pyx_k_bytesbuf), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_closed, __pyx_k_closed, sizeof(__pyx_k_closed), 0, 0, 1, 1},
  {&__pyx_n_s_cogr_driver, __pyx_k_cogr_driver, sizeof(__pyx_k_cogr_driver), 0, 0, 1, 1},
  {&__pyx_n_s_cogr_ds, __pyx_k_cogr_ds, sizeof(__pyx_k_cogr_ds), 0, 0, 1, 1},
  {&__pyx_n_s_cogr_feature, __pyx_k_cogr_feature, sizeof(__pyx_k_cogr_feature), 0, 0, 1, 1},
  {&__pyx_n_s_cogr_geometry, __pyx_k_cogr_geometry, sizeof(__pyx_k_cogr_geometry), 0, 0, 1, 1},
  {&__pyx_n_s_cogr_layer, __pyx_k_cogr_layer, sizeof(__pyx_k_cogr_layer), 0, 0, 1, 1},
  {&__pyx_n_s_collection, __pyx_k_collection, sizeof(__pyx_k_collection), 0, 0, 1, 1},
  {&__pyx_kp_u_collection_s_dataset_does_not_su, __pyx_k_collection_s_dataset_does_not_su, sizeof(__pyx_k_collection_s_dataset_does_not_su), 0, 1, 0, 0},
  {&__pyx_kp_u_collection_s_dataset_does_not_su_2, __pyx_k_collection_s_dataset_does_not_su_2, sizeof(__pyx_k_collection_s_dataset_does_not_su_2), 0, 1, 0, 0},
  {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
  {&__pyx_n_s_compat, __pyx_k_compat, sizeof(__pyx_k_compat), 0, 0, 1, 1},
  {&__pyx_n_u_coordinates, __pyx_k_coordinates, sizeof(__pyx_k_coordinates), 0, 1, 0, 1},
  {&__pyx_n_s_coords, __pyx_k_coords, sizeof(__pyx_k_coords), 0, 0, 1, 1},
  {&__pyx_n_s_copysign, __pyx_k_copysign, sizeof(__pyx_k_copysign), 0, 0, 1, 1},
  {&__pyx_n_s_cpl_errs, __pyx_k_cpl_errs, sizeof(__pyx_k_cpl_errs), 0, 0, 1, 1},
  {&__pyx_n_s_crs_wkt, __pyx_k_crs_wkt, sizeof(__pyx_k_crs_wkt), 0, 0, 1, 1},
  {&__pyx_kp_u_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 1, 0, 0},
  {&__pyx_kp_u_d_2, __pyx_k_d_2, sizeof(__pyx_k_d_2), 0, 1, 0, 0},
  {&__pyx_n_s_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 0, 1, 1},
  {&__pyx_n_u_date, __pyx_k_date, sizeof(__pyx_k_date), 0, 1, 0, 1},
  {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
  {&__pyx_n_u_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 1, 0, 1},
  {&__pyx_n_s_day, __pyx_k_day, sizeof(__pyx_k_day), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_driver, __pyx_k_driver, sizeof(__pyx_k_driver), 0, 0, 1, 1},
  {&__pyx_n_s_driver_2, __pyx_k_driver_2, sizeof(__pyx_k_driver_2), 0, 0, 1, 1},
  {&__pyx_n_s_driver_c, __pyx_k_driver_c, sizeof(__pyx_k_driver_c), 0, 0, 1, 1},
  {&__pyx_n_s_dumps, __pyx_k_dumps, sizeof(__pyx_k_dumps), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_enabled_drivers, __pyx_k_enabled_drivers, sizeof(__pyx_k_enabled_drivers), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_u_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 1, 0, 1},
  {&__pyx_kp_u_epsg, __pyx_k_epsg, sizeof(__pyx_k_epsg), 0, 1, 0, 0},
  {&__pyx_n_s_err, __pyx_k_err, sizeof(__pyx_k_err), 0, 0, 1, 1},
  {&__pyx_n_s_exception, __pyx_k_exception, sizeof(__pyx_k_exception), 0, 0, 1, 1},
  {&__pyx_n_s_exists, __pyx_k_exists, sizeof(__pyx_k_exists), 0, 0, 1, 1},
  {&__pyx_n_s_explode, __pyx_k_explode, sizeof(__pyx_k_explode), 0, 0, 1, 1},
  {&__pyx_n_s_ext, __pyx_k_ext, sizeof(__pyx_k_ext), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_kp_u_failed_to_close_mapped_file_hand, __pyx_k_failed_to_close_mapped_file_hand, sizeof(__pyx_k_failed_to_close_mapped_file_hand), 0, 1, 0, 0},
  {&__pyx_kp_u_failed_to_map_buffer_to_file, __pyx_k_failed_to_map_buffer_to_file, sizeof(__pyx_k_failed_to_map_buffer_to_file), 0, 1, 0, 0},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_s_featureRT, __pyx_k_featureRT, sizeof(__pyx_k_featureRT), 0, 0, 1, 1},
  {&__pyx_n_s_file_or_bytes, __pyx_k_file_or_bytes, sizeof(__pyx_k_file_or_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_filename, __pyx_k_filename, sizeof(__pyx_k_filename), 0, 0, 1, 1},
  {&__pyx_n_s_fiona, __pyx_k_fiona, sizeof(__pyx_k_fiona), 0, 0, 1, 1},
  {&__pyx_n_s_fiona__env, __pyx_k_fiona__env, sizeof(__pyx_k_fiona__env), 0, 0, 1, 1},
  {&__pyx_n_s_fiona__err, __pyx_k_fiona__err, sizeof(__pyx_k_fiona__err), 0, 0, 1, 1},
  {&__pyx_n_s_fiona__geometry, __pyx_k_fiona__geometry, sizeof(__pyx_k_fiona__geometry), 0, 0, 1, 1},
  {&__pyx_n_s_fiona_compat, __pyx_k_fiona_compat, sizeof(__pyx_k_fiona_compat), 0, 0, 1, 1},
  {&__pyx_n_s_fiona_errors, __pyx_k_fiona_errors, sizeof(__pyx_k_fiona_errors), 0, 0, 1, 1},
  {&__pyx_n_s_fiona_ogrext, __pyx_k_fiona_ogrext, sizeof(__pyx_k_fiona_ogrext), 0, 0, 1, 1},
  {&__pyx_kp_s_fiona_ogrext_pyx, __pyx_k_fiona_ogrext_pyx, sizeof(__pyx_k_fiona_ogrext_pyx), 0, 0, 1, 0},
  {&__pyx_n_s_fiona_path, __pyx_k_fiona_path, sizeof(__pyx_k_fiona_path), 0, 0, 1, 1},
  {&__pyx_n_s_fiona_rfc3339, __pyx_k_fiona_rfc3339, sizeof(__pyx_k_fiona_rfc3339), 0, 0, 1, 1},
  {&__pyx_n_s_fiona_schema, __pyx_k_fiona_schema, sizeof(__pyx_k_fiona_schema), 0, 0, 1, 1},
  {&__pyx_n_u_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 1, 0, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_gdal_version, __pyx_k_gdal_version, sizeof(__pyx_k_gdal_version), 0, 0, 1, 1},
  {&__pyx_n_s_geometry, __pyx_k_geometry, sizeof(__pyx_k_geometry), 0, 0, 1, 1},
  {&__pyx_n_u_geometry, __pyx_k_geometry, sizeof(__pyx_k_geometry), 0, 1, 0, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_get_driver, __pyx_k_get_driver, sizeof(__pyx_k_get_driver), 0, 0, 1, 1},
  {&__pyx_n_s_get_fallback_encoding, __pyx_k_get_fallback_encoding, sizeof(__pyx_k_get_fallback_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_get_feature, __pyx_k_get_feature, sizeof(__pyx_k_get_feature), 0, 0, 1, 1},
  {&__pyx_kp_u_get_fileencoding_is_deprecated_a, __pyx_k_get_fileencoding_is_deprecated_a, sizeof(__pyx_k_get_fileencoding_is_deprecated_a), 0, 1, 0, 0},
  {&__pyx_n_s_get_gdal_version_num, __pyx_k_get_gdal_version_num, sizeof(__pyx_k_get_gdal_version_num), 0, 0, 1, 1},
  {&__pyx_n_s_get_internal_encoding, __pyx_k_get_internal_encoding, sizeof(__pyx_k_get_internal_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_get_schema, __pyx_k_get_schema, sizeof(__pyx_k_get_schema), 0, 0, 1, 1},
  {&__pyx_n_s_getpreferredencoding, __pyx_k_getpreferredencoding, sizeof(__pyx_k_getpreferredencoding), 0, 0, 1, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_n_s_hex, __pyx_k_hex, sizeof(__pyx_k_hex), 0, 0, 1, 1},
  {&__pyx_n_s_hour, __pyx_k_hour, sizeof(__pyx_k_hour), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_u_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 1, 0, 1},
  {&__pyx_n_s_ignore_fields, __pyx_k_ignore_fields, sizeof(__pyx_k_ignore_fields), 0, 0, 1, 1},
  {&__pyx_n_s_ignore_geometry, __pyx_k_ignore_geometry, sizeof(__pyx_k_ignore_geometry), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
  {&__pyx_n_u_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 1, 0, 1},
  {&__pyx_n_s_initial_bytes, __pyx_k_initial_bytes, sizeof(__pyx_k_initial_bytes), 0, 0, 1, 1},
  {&__pyx_n_u_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 1, 0, 1},
  {&__pyx_n_u_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 1, 0, 1},
  {&__pyx_n_u_int64, __pyx_k_int64, sizeof(__pyx_k_int64), 0, 1, 0, 1},
  {&__pyx_n_s_integer_types, __pyx_k_integer_types, sizeof(__pyx_k_integer_types), 0, 0, 1, 1},
  {&__pyx_n_s_isactive, __pyx_k_isactive, sizeof(__pyx_k_isactive), 0, 0, 1, 1},
  {&__pyx_kp_u_iso_8859_1, __pyx_k_iso_8859_1, sizeof(__pyx_k_iso_8859_1), 0, 1, 0, 0},
  {&__pyx_n_s_isoformat, __pyx_k_isoformat, sizeof(__pyx_k_isoformat), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_json, __pyx_k_json, sizeof(__pyx_k_json), 0, 0, 1, 1},
  {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
  {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
  {&__pyx_n_s_layer, __pyx_k_layer, sizeof(__pyx_k_layer), 0, 0, 1, 1},
  {&__pyx_n_s_layer_count, __pyx_k_layer_count, sizeof(__pyx_k_layer_count), 0, 0, 1, 1},
  {&__pyx_n_s_layer_index, __pyx_k_layer_index, sizeof(__pyx_k_layer_index), 0, 0, 1, 1},
  {&__pyx_n_s_layer_names, __pyx_k_layer_names, sizeof(__pyx_k_layer_names), 0, 0, 1, 1},
  {&__pyx_n_s_layer_str, __pyx_k_layer_str, sizeof(__pyx_k_layer_str), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_listlayers, __pyx_k_listlayers, sizeof(__pyx_k_listlayers), 0, 0, 1, 1},
  {&__pyx_n_s_loads, __pyx_k_loads, sizeof(__pyx_k_loads), 0, 0, 1, 1},
  {&__pyx_n_s_locale, __pyx_k_locale, sizeof(__pyx_k_locale), 0, 0, 1, 1},
  {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_u_long, __pyx_k_long, sizeof(__pyx_k_long), 0, 1, 0, 1},
  {&__pyx_n_s_lstrip, __pyx_k_lstrip, sizeof(__pyx_k_lstrip), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_major, __pyx_k_major, sizeof(__pyx_k_major), 0, 0, 1, 1},
  {&__pyx_n_s_map, __pyx_k_map, sizeof(__pyx_k_map), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_kp_u_mask_and_bbox_can_not_be_set_tog, __pyx_k_mask_and_bbox_can_not_be_set_tog, sizeof(__pyx_k_mask_and_bbox_can_not_be_set_tog), 0, 1, 0, 0},
  {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_minute, __pyx_k_minute, sizeof(__pyx_k_minute), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_month, __pyx_k_month, sizeof(__pyx_k_month), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_name_b, __pyx_k_name_b, sizeof(__pyx_k_name_b), 0, 0, 1, 1},
  {&__pyx_n_s_name_c, __pyx_k_name_c, sizeof(__pyx_k_name_c), 0, 0, 1, 1},
  {&__pyx_n_s_namedtuple, __pyx_k_namedtuple, sizeof(__pyx_k_namedtuple), 0, 0, 1, 1},
  {&__pyx_kp_u_negative_seek_position, __pyx_k_negative_seek_position, sizeof(__pyx_k_negative_seek_position), 0, 1, 0, 0},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_n_s_next, __pyx_k_next, sizeof(__pyx_k_next), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_normalize_field_type, __pyx_k_normalize_field_type, sizeof(__pyx_k_normalize_field_type), 0, 0, 1, 1},
  {&__pyx_n_s_offset, __pyx_k_offset, sizeof(__pyx_k_offset), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_parse_date, __pyx_k_parse_date, sizeof(__pyx_k_parse_date), 0, 0, 1, 1},
  {&__pyx_n_s_parse_datetime, __pyx_k_parse_datetime, sizeof(__pyx_k_parse_datetime), 0, 0, 1, 1},
  {&__pyx_n_s_parse_time, __pyx_k_parse_time, sizeof(__pyx_k_parse_time), 0, 0, 1, 1},
  {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
  {&__pyx_n_s_path_b, __pyx_k_path_b, sizeof(__pyx_k_path_b), 0, 0, 1, 1},
  {&__pyx_n_s_path_c, __pyx_k_path_c, sizeof(__pyx_k_path_c), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_pop, __pyx_k_pop, sizeof(__pyx_k_pop), 0, 0, 1, 1},
  {&__pyx_n_s_pos, __pyx_k_pos, sizeof(__pyx_k_pos), 0, 0, 1, 1},
  {&__pyx_n_u_properties, __pyx_k_properties, sizeof(__pyx_k_properties), 0, 1, 0, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_FeatureBuilder, __pyx_k_pyx_unpickle_FeatureBuilder, sizeof(__pyx_k_pyx_unpickle_FeatureBuilder), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_MemoryFileBase, __pyx_k_pyx_unpickle_MemoryFileBase, sizeof(__pyx_k_pyx_unpickle_MemoryFileBase), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_OGRFeatureBuilder, __pyx_k_pyx_unpickle_OGRFeatureBuilder, sizeof(__pyx_k_pyx_unpickle_OGRFeatureBuilder), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_u_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 1, 0, 1},
  {&__pyx_n_s_record, __pyx_k_record, sizeof(__pyx_k_record), 0, 0, 1, 1},
  {&__pyx_n_s_records, __pyx_k_records, sizeof(__pyx_k_records), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_remove, __pyx_k_remove, sizeof(__pyx_k_remove), 0, 0, 1, 1},
  {&__pyx_n_s_remove_layer, __pyx_k_remove_layer, sizeof(__pyx_k_remove_layer), 0, 0, 1, 1},
  {&__pyx_n_s_remove_virtual_file, __pyx_k_remove_virtual_file, sizeof(__pyx_k_remove_virtual_file), 0, 0, 1, 1},
  {&__pyx_n_s_result, __pyx_k_result, sizeof(__pyx_k_result), 0, 0, 1, 1},
  {&__pyx_n_s_schema, __pyx_k_schema, sizeof(__pyx_k_schema), 0, 0, 1, 1},
  {&__pyx_n_s_second, __pyx_k_second, sizeof(__pyx_k_second), 0, 0, 1, 1},
  {&__pyx_kp_u_seek_position_past_end_of_file, __pyx_k_seek_position_past_end_of_file, sizeof(__pyx_k_seek_position_past_end_of_file), 0, 1, 0, 0},
  {&__pyx_kp_s_self_cogr_ds_self_cogr_layer_can, __pyx_k_self_cogr_ds_self_cogr_layer_can, sizeof(__pyx_k_self_cogr_ds_self_cogr_layer_can), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_session, __pyx_k_session, sizeof(__pyx_k_session), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_u_slice_step_cannot_be_zero, __pyx_k_slice_step_cannot_be_zero, sizeof(__pyx_k_slice_step_cannot_be_zero), 0, 1, 0, 0},
  {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_startswith, __pyx_k_startswith, sizeof(__pyx_k_startswith), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_n_u_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 1, 0, 1},
  {&__pyx_n_s_string_types, __pyx_k_string_types, sizeof(__pyx_k_string_types), 0, 0, 1, 1},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_strip, __pyx_k_strip, sizeof(__pyx_k_strip), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_text_type, __pyx_k_text_type, sizeof(__pyx_k_text_type), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
  {&__pyx_n_u_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 1, 0, 1},
  {&__pyx_n_u_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 1, 0, 1},
  {&__pyx_n_s_unlink, __pyx_k_unlink, sizeof(__pyx_k_unlink), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_kp_u_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 1, 0, 0},
  {&__pyx_n_s_uuid, __pyx_k_uuid, sizeof(__pyx_k_uuid), 0, 0, 1, 1},
  {&__pyx_n_s_uuid4, __pyx_k_uuid4, sizeof(__pyx_k_uuid4), 0, 0, 1, 1},
  {&__pyx_n_s_valid_geom_types, __pyx_k_valid_geom_types, sizeof(__pyx_k_valid_geom_types), 0, 0, 1, 1},
  {&__pyx_n_s_validate_geometry_type, __pyx_k_validate_geometry_type, sizeof(__pyx_k_validate_geometry_type), 0, 0, 1, 1},
  {&__pyx_n_s_vsi_cfilename, __pyx_k_vsi_cfilename, sizeof(__pyx_k_vsi_cfilename), 0, 0, 1, 1},
  {&__pyx_n_s_vsi_filename, __pyx_k_vsi_filename, sizeof(__pyx_k_vsi_filename), 0, 0, 1, 1},
  {&__pyx_n_s_vsi_handle, __pyx_k_vsi_handle, sizeof(__pyx_k_vsi_handle), 0, 0, 1, 1},
  {&__pyx_n_s_vsi_path, __pyx_k_vsi_path, sizeof(__pyx_k_vsi_path), 0, 0, 1, 1},
  {&__pyx_kp_u_vsimem, __pyx_k_vsimem, sizeof(__pyx_k_vsimem), 0, 1, 0, 0},
  {&__pyx_kp_u_vsimem_0, __pyx_k_vsimem_0, sizeof(__pyx_k_vsimem_0), 0, 1, 0, 0},
  {&__pyx_kp_u_vsimem_0_1, __pyx_k_vsimem_0_1, sizeof(__pyx_k_vsimem_0_1), 0, 1, 0, 0},
  {&__pyx_n_u_w, __pyx_k_w, sizeof(__pyx_k_w), 0, 1, 0, 1},
  {&__pyx_n_s_warn, __pyx_k_warn, sizeof(__pyx_k_warn), 0, 0, 1, 1},
  {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
  {&__pyx_n_s_warnings, __pyx_k_warnings, sizeof(__pyx_k_warnings), 0, 0, 1, 1},
  {&__pyx_n_s_whence, __pyx_k_whence, sizeof(__pyx_k_whence), 0, 0, 1, 1},
  {&__pyx_n_s_writerecs_locals_validate_geomet, __pyx_k_writerecs_locals_validate_geomet, sizeof(__pyx_k_writerecs_locals_validate_geomet), 0, 0, 1, 1},
  {&__pyx_n_s_xyz, __pyx_k_xyz, sizeof(__pyx_k_xyz), 0, 0, 1, 1},
  {&__pyx_n_s_year, __pyx_k_year, sizeof(__pyx_k_year), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 111, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 112, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 112, __pyx_L1_error)
  __pyx_builtin_KeyError = __Pyx_GetBuiltinName(__pyx_n_s_KeyError); if (!__pyx_builtin_KeyError) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 113, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 179, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 182, __pyx_L1_error)
  __pyx_builtin_UnicodeDecodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeDecodeError); if (!__pyx_builtin_UnicodeDecodeError) __PYX_ERR(0, 226, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 506, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(0, 851, __pyx_L1_error)
  __pyx_builtin_OSError = __Pyx_GetBuiltinName(__pyx_n_s_OSError); if (!__pyx_builtin_OSError) __PYX_ERR(0, 897, __pyx_L1_error)
  __pyx_builtin_map = __Pyx_GetBuiltinName(__pyx_n_s_map); if (!__pyx_builtin_map) __PYX_ERR(0, 1111, __pyx_L1_error)
  __pyx_builtin_UnicodeEncodeError = __Pyx_GetBuiltinName(__pyx_n_s_UnicodeEncodeError); if (!__pyx_builtin_UnicodeEncodeError) __PYX_ERR(0, 1135, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(0, 1198, __pyx_L1_error)
  __pyx_builtin_RuntimeWarning = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeWarning); if (!__pyx_builtin_RuntimeWarning) __PYX_ERR(0, 1298, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 1319, __pyx_L1_error)
  __pyx_builtin_IOError = __Pyx_GetBuiltinName(__pyx_n_s_IOError); if (!__pyx_builtin_IOError) __PYX_ERR(0, 1608, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "fiona/ogrext.pyx":182
 *             fdefn = OGR_F_GetFieldDefnRef(feature, i)
 *             if fdefn == NULL:
 *                 raise ValueError("Null feature definition")             # <<<<<<<<<<<<<<
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_Null_feature_definition); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "fiona/ogrext.pyx":185
 *             key_c = OGR_Fld_GetNameRef(fdefn)
 *             if key_c == NULL:
 *                 raise ValueError("Null field name reference")             # <<<<<<<<<<<<<<
 *             key_b = key_c
 *             key = key_b.decode(encoding)
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_u_Null_field_name_reference); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "fiona/ogrext.pyx":319
 *         cdef void *cogr_layer = session.cogr_layer
 *         if cogr_layer == NULL:
 *             raise ValueError("Null layer")             # <<<<<<<<<<<<<<
 *         cdef void *cogr_featuredefn = OGR_L_GetLayerDefn(cogr_layer)
 *         if cogr_featuredefn == NULL:
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_u_Null_layer); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "fiona/ogrext.pyx":325
 *         cdef void *cogr_feature = OGR_F_Create(cogr_featuredefn)
 *         if cogr_feature == NULL:
 *             raise ValueError("Null feature")             # <<<<<<<<<<<<<<
 * 
 *         if feature['geometry'] is not None:
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Null_feature); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "fiona/ogrext.pyx":429
 *     cdef void *cogr_geometry = OGR_F_GetGeometryRef(cogr_feature)
 *     if cogr_geometry == NULL:
 *         raise ValueError("Null geometry")             # <<<<<<<<<<<<<<
 *     result = FeatureBuilder().build(
 *         cogr_feature,
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_u_Null_geometry); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "fiona/ogrext.pyx":579
 *         cdef void *cogr_driver = GDALGetDatasetDriver(self.cogr_ds)
 *         if cogr_driver == NULL:
 *             raise ValueError("Null driver")             # <<<<<<<<<<<<<<
 *         cdef const char *name = OGR_Dr_GetName(cogr_driver)
 *         driver_name = name
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_u_Null_driver); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "fiona/ogrext.pyx":611
 *             cogr_fielddefn = OGR_FD_GetFieldDefn(cogr_featuredefn, i)
 *             if cogr_fielddefn == NULL:
 *                 raise ValueError("Null field definition")             # <<<<<<<<<<<<<<
 * 
 *             key_c = OGR_Fld_GetNameRef(cogr_fielddefn)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_Null_field_definition); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 611, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "fiona/ogrext.pyx":719
 *                     OSRExportToProj4(cogr_crs, &proj_c)
 *                     if proj_c == NULL:
 *                         raise ValueError("Null projection")             # <<<<<<<<<<<<<<
 *                     proj_b = proj_c
 *                     log.debug("Params: %s", proj_b)
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_u_Null_projection); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "fiona/ogrext.pyx":851
 *                 ftcount = OGR_L_GetFeatureCount(self.cogr_layer, 0)
 *                 if ftcount == -1:
 *                     raise IndexError(             # <<<<<<<<<<<<<<
 *                         "collection's dataset does not support negative indexes")
 *                 index += ftcount
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_u_collection_s_dataset_does_not_su); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 851, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_self_cogr_ds_self_cogr_layer_can); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "(tree fragment)":4
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_self_cogr_ds_self_cogr_layer_can); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "fiona/ogrext.pyx":1046
 *                 options = CSLAddNameValue(options, <const char *>k, <const char *>v)
 * 
 *             geometry_type = collection.schema.get("geometry", "Unknown")             # <<<<<<<<<<<<<<
 *             if not isinstance(geometry_type, string_types) and geometry_type is not None:
 *                 geometry_types = set(geometry_type)
 */
  __pyx_tuple__20 = PyTuple_Pack(2, __pyx_n_u_geometry, __pyx_n_u_Unknown); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 1046, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "fiona/ogrext.pyx":1169
 * 
 *         valid_geom_types = collection._valid_geom_types
 *         def validate_geometry_type(record):             # <<<<<<<<<<<<<<
 *             if record["geometry"] is None:
 *                 return True
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_record); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 1169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(1, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_validate_geometry_type, 1169, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 1169, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1223
 *         cdef void *cogr_layer = self.cogr_layer
 *         if cogr_ds == NULL:
 *             raise ValueError("Null data source")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_u_Null_data_source); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 1223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_self_cogr_ds_self_cogr_layer_can); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "(tree fragment)":4
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("self.cogr_ds,self.cogr_layer cannot be converted to a Python object for pickling")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_self_cogr_ds_self_cogr_layer_can); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "fiona/ogrext.pyx":1247
 *                   bbox=None, mask=None):
 *         if collection.session is None:
 *             raise ValueError("I/O operation on closed collection")             # <<<<<<<<<<<<<<
 *         self.collection = collection
 *         cdef Session session
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_u_I_O_operation_on_closed_collecti); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 1247, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "fiona/ogrext.pyx":1258
 * 
 *         if bbox and mask:
 *             raise ValueError("mask and bbox can not be set together")             # <<<<<<<<<<<<<<
 * 
 *         if bbox:
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_u_mask_and_bbox_can_not_be_set_tog); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "fiona/ogrext.pyx":1279
 *         if ftcount == -1 and ((start is not None and start < 0) or
 *                               (stop is not None and stop < 0)):
 *             raise IndexError(             # <<<<<<<<<<<<<<
 *                 "collection's dataset does not support negative slice indexes")
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_kp_u_collection_s_dataset_does_not_su_2); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "fiona/ogrext.pyx":1294
 *             step = 1
 *         if step == 0:
 *             raise ValueError("slice step cannot be zero")             # <<<<<<<<<<<<<<
 *         if step < 0 and not self.fastindex:
 *             warnings.warn("Layer does not support" \
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_slice_step_cannot_be_zero); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(0, 1294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "fiona/ogrext.pyx":1296
 *             raise ValueError("slice step cannot be zero")
 *         if step < 0 and not self.fastindex:
 *             warnings.warn("Layer does not support" \             # <<<<<<<<<<<<<<
 *                     "OLCFastSetNextByIndex, negative step size may" \
 *                     " be slow", RuntimeWarning)
 */
  __pyx_tuple__33 = PyTuple_Pack(2, __pyx_kp_u_Layer_does_not_supportOLCFastSet, __pyx_builtin_RuntimeWarning); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 1296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "fiona/ogrext.pyx":1543
 * 
 *     if vsi_handle == NULL:
 *         raise OSError('failed to map buffer to file')             # <<<<<<<<<<<<<<
 *     if VSIFCloseL(vsi_handle) != 0:
 *         raise OSError('failed to close mapped file handle')
 */
  __pyx_tuple__41 = PyTuple_Pack(1, __pyx_kp_u_failed_to_map_buffer_to_file); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 1543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);

  /* "fiona/ogrext.pyx":1545
 *         raise OSError('failed to map buffer to file')
 *     if VSIFCloseL(vsi_handle) != 0:
 *         raise OSError('failed to close mapped file handle')             # <<<<<<<<<<<<<<
 * 
 *     return vsi_filename
 */
  __pyx_tuple__42 = PyTuple_Pack(1, __pyx_kp_u_failed_to_close_mapped_file_hand); if (unlikely(!__pyx_tuple__42)) __PYX_ERR(0, 1545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__42);
  __Pyx_GIVEREF(__pyx_tuple__42);

  /* "fiona/ogrext.pyx":1579
 *                 initial_bytes = file_or_bytes
 *             if not isinstance(initial_bytes, (bytearray, bytes)):
 *                 raise TypeError(             # <<<<<<<<<<<<<<
 *                     "Constructor argument must be a file opened in binary "
 *                     "mode or bytes/bytearray.")
 */
  __pyx_tuple__43 = PyTuple_Pack(1, __pyx_kp_u_Constructor_argument_must_be_a_f); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 1579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);

  /* "fiona/ogrext.pyx":1608
 * 
 *             if vsi_handle == NULL:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to create in-memory file using initial bytes.")
 * 
 */
  __pyx_tuple__44 = PyTuple_Pack(1, __pyx_kp_u_Failed_to_create_in_memory_file); if (unlikely(!__pyx_tuple__44)) __PYX_ERR(0, 1608, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__44);
  __Pyx_GIVEREF(__pyx_tuple__44);

  /* "fiona/ogrext.pyx":1612
 * 
 *             if VSIFCloseL(vsi_handle) != 0:
 *                 raise IOError(             # <<<<<<<<<<<<<<
 *                     "Failed to properly close in-memory file.")
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(1, __pyx_kp_u_Failed_to_properly_close_in_memo); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 1612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);

  /* "fiona/ogrext.pyx":95
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
  __pyx_tuple__46 = PyTuple_Pack(3, __pyx_n_s_coords, __pyx_n_s_e, __pyx_n_s_f); if (unlikely(!__pyx_tuple__46)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__46);
  __Pyx_GIVEREF(__pyx_tuple__46);
  __pyx_codeobj_ = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__46, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_explode, 95, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj_)) __PYX_ERR(0, 95, __pyx_L1_error)

  /* "fiona/ogrext.pyx":108
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry"""
 *     try:
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_s_geometry, __pyx_n_s_xyz); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_bounds, 108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 108, __pyx_L1_error)

  /* "fiona/ogrext.pyx":424
 * 
 * 
 * def featureRT(feature, collection):             # <<<<<<<<<<<<<<
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 */
  __pyx_tuple__49 = PyTuple_Pack(5, __pyx_n_s_feature, __pyx_n_s_collection, __pyx_n_s_cogr_feature, __pyx_n_s_cogr_geometry, __pyx_n_s_result); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_featureRT, 424, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 424, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1434
 * 
 * 
 * def _remove(path, driver=None):             # <<<<<<<<<<<<<<
 *     """Deletes an OGR data source
 *     """
 */
  __pyx_tuple__51 = PyTuple_Pack(6, __pyx_n_s_path, __pyx_n_s_driver, __pyx_n_s_cogr_driver, __pyx_n_s_cogr_ds, __pyx_n_s_result, __pyx_n_s_driver_c); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_remove, 1434, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 1434, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1464
 * 
 * 
 * def _remove_layer(path, layer, driver=None):             # <<<<<<<<<<<<<<
 *     cdef void *cogr_ds
 *     cdef int layer_index
 */
  __pyx_tuple__53 = PyTuple_Pack(8, __pyx_n_s_path, __pyx_n_s_layer, __pyx_n_s_driver, __pyx_n_s_cogr_ds, __pyx_n_s_layer_index, __pyx_n_s_layer_str, __pyx_n_s_layer_names, __pyx_n_s_result); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(3, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_remove_layer, 1464, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 1464, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1496
 * 
 * 
 * def _listlayers(path, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *     """Provides a list of the layers in an OGR data source.
 */
  __pyx_tuple__55 = PyTuple_Pack(11, __pyx_n_s_path, __pyx_n_s_kwargs, __pyx_n_s_cogr_ds, __pyx_n_s_cogr_layer, __pyx_n_s_path_c, __pyx_n_s_name_c, __pyx_n_s_path_b, __pyx_n_s_layer_count, __pyx_n_s_layer_names, __pyx_n_s_i, __pyx_n_s_name_b); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_listlayers, 1496, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 1496, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1531
 * 
 * 
 * def buffer_to_virtual_file(bytesbuf, ext=''):             # <<<<<<<<<<<<<<
 *     """Maps a bytes buffer to a virtual file.
 * 
 */
  __pyx_tuple__57 = PyTuple_Pack(5, __pyx_n_s_bytesbuf, __pyx_n_s_ext, __pyx_n_s_vsi_filename, __pyx_n_s_vsi_cfilename, __pyx_n_s_vsi_handle); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_buffer_to_virtual_file, 1531, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 1531, __pyx_L1_error)

  /* "fiona/ogrext.pyx":1550
 * 
 * 
 * def remove_virtual_file(vsi_filename):             # <<<<<<<<<<<<<<
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 *     return VSIUnlink(vsi_cfilename)
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_s_vsi_filename, __pyx_n_s_vsi_cfilename); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(1, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_fiona_ogrext_pyx, __pyx_n_s_remove_virtual_file, 1550, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 1550, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__61 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_FeatureBuilder, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__63 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_OGRFeatureBuilder, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_tuple__65 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MemoryFileBase, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  __pyx_umethod_PyDict_Type_get.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_umethod_PyList_Type_index.type = (PyObject*)&PyList_Type;
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_20000 = PyInt_FromLong(20000L); if (unlikely(!__pyx_int_20000)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_222419149 = PyInt_FromLong(222419149L); if (unlikely(!__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_5fiona_6ogrext_FeatureBuilder = &__pyx_vtable_5fiona_6ogrext_FeatureBuilder;
  __pyx_vtable_5fiona_6ogrext_FeatureBuilder.build = (PyObject *(*)(struct __pyx_obj_5fiona_6ogrext_FeatureBuilder *, void *, struct __pyx_opt_args_5fiona_6ogrext_14FeatureBuilder_build *__pyx_optional_args))__pyx_f_5fiona_6ogrext_14FeatureBuilder_build;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_FeatureBuilder) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_FeatureBuilder.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_FeatureBuilder.tp_dictoffset && __pyx_type_5fiona_6ogrext_FeatureBuilder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_FeatureBuilder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5fiona_6ogrext_FeatureBuilder.tp_dict, __pyx_vtabptr_5fiona_6ogrext_FeatureBuilder) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_FeatureBuilder, (PyObject *)&__pyx_type_5fiona_6ogrext_FeatureBuilder) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_FeatureBuilder) < 0) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_FeatureBuilder = &__pyx_type_5fiona_6ogrext_FeatureBuilder;
  __pyx_vtabptr_5fiona_6ogrext_OGRFeatureBuilder = &__pyx_vtable_5fiona_6ogrext_OGRFeatureBuilder;
  __pyx_vtable_5fiona_6ogrext_OGRFeatureBuilder.build = (void *(*)(struct __pyx_obj_5fiona_6ogrext_OGRFeatureBuilder *, PyObject *, PyObject *))__pyx_f_5fiona_6ogrext_17OGRFeatureBuilder_build;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_OGRFeatureBuilder) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_OGRFeatureBuilder.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_OGRFeatureBuilder.tp_dictoffset && __pyx_type_5fiona_6ogrext_OGRFeatureBuilder.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_OGRFeatureBuilder.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5fiona_6ogrext_OGRFeatureBuilder.tp_dict, __pyx_vtabptr_5fiona_6ogrext_OGRFeatureBuilder) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_OGRFeatureBuilder, (PyObject *)&__pyx_type_5fiona_6ogrext_OGRFeatureBuilder) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_OGRFeatureBuilder) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_OGRFeatureBuilder = &__pyx_type_5fiona_6ogrext_OGRFeatureBuilder;
  __pyx_vtabptr_5fiona_6ogrext_Session = &__pyx_vtable_5fiona_6ogrext_Session;
  __pyx_vtable_5fiona_6ogrext_Session.stop = (PyObject *(*)(struct __pyx_obj_5fiona_6ogrext_Session *, int __pyx_skip_dispatch))__pyx_f_5fiona_6ogrext_7Session_stop;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_Session) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_Session.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_Session.tp_dictoffset && __pyx_type_5fiona_6ogrext_Session.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_Session.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5fiona_6ogrext_Session.tp_dict, __pyx_vtabptr_5fiona_6ogrext_Session) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Session, (PyObject *)&__pyx_type_5fiona_6ogrext_Session) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_Session) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_Session = &__pyx_type_5fiona_6ogrext_Session;
  __pyx_vtabptr_5fiona_6ogrext_WritingSession = &__pyx_vtable_5fiona_6ogrext_WritingSession;
  __pyx_vtable_5fiona_6ogrext_WritingSession.__pyx_base = *__pyx_vtabptr_5fiona_6ogrext_Session;
  __pyx_type_5fiona_6ogrext_WritingSession.tp_base = __pyx_ptype_5fiona_6ogrext_Session;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_WritingSession) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_WritingSession.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_WritingSession.tp_dictoffset && __pyx_type_5fiona_6ogrext_WritingSession.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_WritingSession.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_5fiona_6ogrext_WritingSession.tp_dict, __pyx_vtabptr_5fiona_6ogrext_WritingSession) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_WritingSession, (PyObject *)&__pyx_type_5fiona_6ogrext_WritingSession) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_WritingSession) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_WritingSession = &__pyx_type_5fiona_6ogrext_WritingSession;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_Iterator) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_Iterator.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_Iterator.tp_dictoffset && __pyx_type_5fiona_6ogrext_Iterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_Iterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_Iterator, (PyObject *)&__pyx_type_5fiona_6ogrext_Iterator) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_Iterator) < 0) __PYX_ERR(0, 1229, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_Iterator = &__pyx_type_5fiona_6ogrext_Iterator;
  __pyx_type_5fiona_6ogrext_ItemsIterator.tp_base = __pyx_ptype_5fiona_6ogrext_Iterator;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_ItemsIterator) < 0) __PYX_ERR(0, 1381, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_ItemsIterator.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_ItemsIterator.tp_dictoffset && __pyx_type_5fiona_6ogrext_ItemsIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_ItemsIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_ItemsIterator, (PyObject *)&__pyx_type_5fiona_6ogrext_ItemsIterator) < 0) __PYX_ERR(0, 1381, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_ItemsIterator) < 0) __PYX_ERR(0, 1381, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_ItemsIterator = &__pyx_type_5fiona_6ogrext_ItemsIterator;
  __pyx_type_5fiona_6ogrext_KeysIterator.tp_base = __pyx_ptype_5fiona_6ogrext_Iterator;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_KeysIterator) < 0) __PYX_ERR(0, 1412, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_KeysIterator.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_KeysIterator.tp_dictoffset && __pyx_type_5fiona_6ogrext_KeysIterator.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_KeysIterator.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_KeysIterator, (PyObject *)&__pyx_type_5fiona_6ogrext_KeysIterator) < 0) __PYX_ERR(0, 1412, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_KeysIterator) < 0) __PYX_ERR(0, 1412, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_KeysIterator = &__pyx_type_5fiona_6ogrext_KeysIterator;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext_MemoryFileBase) < 0) __PYX_ERR(0, 1555, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext_MemoryFileBase.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext_MemoryFileBase.tp_dictoffset && __pyx_type_5fiona_6ogrext_MemoryFileBase.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext_MemoryFileBase.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_5fiona_6ogrext_MemoryFileBase, "__init__"); if (unlikely(!wrapper)) __PYX_ERR(0, 1555, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase___init__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase___init__.doc = __pyx_doc_5fiona_6ogrext_14MemoryFileBase___init__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase___init__;
    }
  }
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON
  {
    PyObject *wrapper = PyObject_GetAttrString((PyObject *)&__pyx_type_5fiona_6ogrext_MemoryFileBase, "__len__"); if (unlikely(!wrapper)) __PYX_ERR(0, 1555, __pyx_L1_error)
    if (Py_TYPE(wrapper) == &PyWrapperDescr_Type) {
      __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase_4__len__ = *((PyWrapperDescrObject *)wrapper)->d_base;
      __pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase_4__len__.doc = __pyx_doc_5fiona_6ogrext_14MemoryFileBase_4__len__;
      ((PyWrapperDescrObject *)wrapper)->d_base = &__pyx_wrapperbase_5fiona_6ogrext_14MemoryFileBase_4__len__;
    }
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MemoryFileBase, (PyObject *)&__pyx_type_5fiona_6ogrext_MemoryFileBase) < 0) __PYX_ERR(0, 1555, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_5fiona_6ogrext_MemoryFileBase) < 0) __PYX_ERR(0, 1555, __pyx_L1_error)
  __pyx_ptype_5fiona_6ogrext_MemoryFileBase = &__pyx_type_5fiona_6ogrext_MemoryFileBase;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext___pyx_scope_struct___explode) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext___pyx_scope_struct___explode.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext___pyx_scope_struct___explode.tp_dictoffset && __pyx_type_5fiona_6ogrext___pyx_scope_struct___explode.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext___pyx_scope_struct___explode.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5fiona_6ogrext___pyx_scope_struct___explode = &__pyx_type_5fiona_6ogrext___pyx_scope_struct___explode;
  if (PyType_Ready(&__pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs) < 0) __PYX_ERR(0, 1154, __pyx_L1_error)
  __pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs.tp_print = 0;
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs.tp_dictoffset && __pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  __pyx_ptype_5fiona_6ogrext___pyx_scope_struct_1_writerecs = &__pyx_type_5fiona_6ogrext___pyx_scope_struct_1_writerecs;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("fiona._geometry"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5fiona_9_geometry_GeomBuilder = __Pyx_ImportType(__pyx_t_1, "fiona._geometry", "GeomBuilder", sizeof(struct __pyx_obj_5fiona_9_geometry_GeomBuilder), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5fiona_9_geometry_GeomBuilder) __PYX_ERR(2, 110, __pyx_L1_error)
  __pyx_vtabptr_5fiona_9_geometry_GeomBuilder = (struct __pyx_vtabstruct_5fiona_9_geometry_GeomBuilder*)__Pyx_GetVtable(__pyx_ptype_5fiona_9_geometry_GeomBuilder->tp_dict); if (unlikely(!__pyx_vtabptr_5fiona_9_geometry_GeomBuilder)) __PYX_ERR(2, 110, __pyx_L1_error)
  __pyx_ptype_5fiona_9_geometry_OGRGeomBuilder = __Pyx_ImportType(__pyx_t_1, "fiona._geometry", "OGRGeomBuilder", sizeof(struct __pyx_obj_5fiona_9_geometry_OGRGeomBuilder), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_5fiona_9_geometry_OGRGeomBuilder) __PYX_ERR(2, 129, __pyx_L1_error)
  __pyx_vtabptr_5fiona_9_geometry_OGRGeomBuilder = (struct __pyx_vtabstruct_5fiona_9_geometry_OGRGeomBuilder*)__Pyx_GetVtable(__pyx_ptype_5fiona_9_geometry_OGRGeomBuilder->tp_dict); if (unlikely(!__pyx_vtabptr_5fiona_9_geometry_OGRGeomBuilder)) __PYX_ERR(2, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(3, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(4, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(5, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("fiona._shim"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "is_field_null", (void (**)(void))&__pyx_f_5fiona_5_shim_is_field_null, "int (void *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "set_field_null", (void (**)(void))&__pyx_f_5fiona_5_shim_set_field_null, "void (void *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "gdal_flush_cache", (void (**)(void))&__pyx_f_5fiona_5_shim_gdal_flush_cache, "void (void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "gdal_open_vector", (void (**)(void))&__pyx_f_5fiona_5_shim_gdal_open_vector, "void *(char const *, int, PyObject *, PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "gdal_create", (void (**)(void))&__pyx_f_5fiona_5_shim_gdal_create, "void *(void *, char const *, PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "gdal_start_transaction", (void (**)(void))&__pyx_f_5fiona_5_shim_gdal_start_transaction, "OGRErr (void *, int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "gdal_commit_transaction", (void (**)(void))&__pyx_f_5fiona_5_shim_gdal_commit_transaction, "OGRErr (void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_field_subtype", (void (**)(void))&__pyx_f_5fiona_5_shim_get_field_subtype, "OGRFieldSubType (void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "set_field_subtype", (void (**)(void))&__pyx_f_5fiona_5_shim_set_field_subtype, "void (void *, OGRFieldSubType)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "check_capability_create_layer", (void (**)(void))&__pyx_f_5fiona_5_shim_check_capability_create_layer, "int (void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "get_linear_geometry", (void (**)(void))&__pyx_f_5fiona_5_shim_get_linear_geometry, "void *(void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "osr_get_name", (void (**)(void))&__pyx_f_5fiona_5_shim_osr_get_name, "char const *(OGRSpatialReferenceH)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "osr_set_traditional_axis_mapping_strategy", (void (**)(void))&__pyx_f_5fiona_5_shim_osr_set_traditional_axis_mapping_strategy, "void (OGRSpatialReferenceH)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = PyImport_ImportModule("fiona._geometry"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "geometry_type_code", (void (**)(void))&__pyx_f_5fiona_9_geometry_geometry_type_code, "unsigned int (PyObject *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "normalize_geometry_type_code", (void (**)(void))&__pyx_f_5fiona_9_geometry_normalize_geometry_type_code, "PyObject *(unsigned int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "base_geometry_type_code", (void (**)(void))&__pyx_f_5fiona_9_geometry_base_geometry_type_code, "unsigned int (unsigned int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_3 = PyImport_ImportModule("fiona._err"); if (!__pyx_t_3) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "exc_wrap_int", (void (**)(void))&__pyx_f_5fiona_4_err_exc_wrap_int, "int (int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "exc_wrap_pointer", (void (**)(void))&__pyx_f_5fiona_4_err_exc_wrap_pointer, "void *(void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_3, "exc_wrap_vsilfile", (void (**)(void))&__pyx_f_5fiona_4_err_exc_wrap_vsilfile, "VSILFILE *(VSILFILE *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initogrext(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initogrext(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_ogrext(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_ogrext(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_ogrext(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'ogrext' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_ogrext(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("ogrext", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_fiona__ogrext) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "fiona.ogrext")) {
      if (unlikely(PyDict_SetItemString(modules, "fiona.ogrext", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() != 0)) goto __pyx_L1_error;
  if (unlikely(__Pyx_modinit_type_import_code() != 0)) goto __pyx_L1_error;
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely(__Pyx_modinit_function_import_code() != 0)) goto __pyx_L1_error;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "fiona/ogrext.pyx":5
 * from __future__ import absolute_import
 * 
 * import datetime             # <<<<<<<<<<<<<<
 * import json
 * import locale
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_datetime, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":6
 * 
 * import datetime
 * import json             # <<<<<<<<<<<<<<
 * import locale
 * import logging
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_json, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_json, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":7
 * import datetime
 * import json
 * import locale             # <<<<<<<<<<<<<<
 * import logging
 * import os
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_locale, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_locale, __pyx_t_1) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":8
 * import json
 * import locale
 * import logging             # <<<<<<<<<<<<<<
 * import os
 * import warnings
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_logging, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_1) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":9
 * import locale
 * import logging
 * import os             # <<<<<<<<<<<<<<
 * import warnings
 * import math
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_os, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_1) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":10
 * import logging
 * import os
 * import warnings             # <<<<<<<<<<<<<<
 * import math
 * import uuid
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_warnings, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_warnings, __pyx_t_1) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":11
 * import os
 * import warnings
 * import math             # <<<<<<<<<<<<<<
 * import uuid
 * from collections import namedtuple
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_math, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_1) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":12
 * import warnings
 * import math
 * import uuid             # <<<<<<<<<<<<<<
 * from collections import namedtuple
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_uuid, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_uuid, __pyx_t_1) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":13
 * import math
 * import uuid
 * from collections import namedtuple             # <<<<<<<<<<<<<<
 * 
 * from six import integer_types, string_types, text_type
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_namedtuple);
  __Pyx_GIVEREF(__pyx_n_s_namedtuple);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_namedtuple);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_namedtuple); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_namedtuple, __pyx_t_1) < 0) __PYX_ERR(0, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":15
 * from collections import namedtuple
 * 
 * from six import integer_types, string_types, text_type             # <<<<<<<<<<<<<<
 * 
 * from fiona._shim cimport *
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_integer_types);
  __Pyx_GIVEREF(__pyx_n_s_integer_types);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_integer_types);
  __Pyx_INCREF(__pyx_n_s_string_types);
  __Pyx_GIVEREF(__pyx_n_s_string_types);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_string_types);
  __Pyx_INCREF(__pyx_n_s_text_type);
  __Pyx_GIVEREF(__pyx_n_s_text_type);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_text_type);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_six, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_integer_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_integer_types, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_string_types); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_string_types, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_text_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_text_type, __pyx_t_2) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":24
 * from fiona._err cimport exc_wrap_int, exc_wrap_pointer, exc_wrap_vsilfile
 * 
 * import fiona             # <<<<<<<<<<<<<<
 * from fiona._env import GDALVersion, get_gdal_version_num
 * from fiona._err import cpl_errs, FionaNullPointerError, CPLE_BaseError, CPLE_OpenFailedError
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_fiona, __pyx_t_1) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":25
 * 
 * import fiona
 * from fiona._env import GDALVersion, get_gdal_version_num             # <<<<<<<<<<<<<<
 * from fiona._err import cpl_errs, FionaNullPointerError, CPLE_BaseError, CPLE_OpenFailedError
 * from fiona._geometry import GEOMETRY_TYPES
 */
  __pyx_t_1 = PyList_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_GDALVersion);
  __Pyx_GIVEREF(__pyx_n_s_GDALVersion);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_GDALVersion);
  __Pyx_INCREF(__pyx_n_s_get_gdal_version_num);
  __Pyx_GIVEREF(__pyx_n_s_get_gdal_version_num);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_get_gdal_version_num);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fiona__env, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_GDALVersion); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GDALVersion, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_get_gdal_version_num); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_get_gdal_version_num, __pyx_t_1) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":26
 * import fiona
 * from fiona._env import GDALVersion, get_gdal_version_num
 * from fiona._err import cpl_errs, FionaNullPointerError, CPLE_BaseError, CPLE_OpenFailedError             # <<<<<<<<<<<<<<
 * from fiona._geometry import GEOMETRY_TYPES
 * from fiona import compat
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_cpl_errs);
  __Pyx_GIVEREF(__pyx_n_s_cpl_errs);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_cpl_errs);
  __Pyx_INCREF(__pyx_n_s_FionaNullPointerError);
  __Pyx_GIVEREF(__pyx_n_s_FionaNullPointerError);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_FionaNullPointerError);
  __Pyx_INCREF(__pyx_n_s_CPLE_BaseError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_BaseError);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_CPLE_BaseError);
  __Pyx_INCREF(__pyx_n_s_CPLE_OpenFailedError);
  __Pyx_GIVEREF(__pyx_n_s_CPLE_OpenFailedError);
  PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_CPLE_OpenFailedError);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona__err, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_cpl_errs); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_cpl_errs, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_FionaNullPointerError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaNullPointerError, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CPLE_BaseError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_BaseError, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_CPLE_OpenFailedError); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CPLE_OpenFailedError, __pyx_t_2) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":27
 * from fiona._env import GDALVersion, get_gdal_version_num
 * from fiona._err import cpl_errs, FionaNullPointerError, CPLE_BaseError, CPLE_OpenFailedError
 * from fiona._geometry import GEOMETRY_TYPES             # <<<<<<<<<<<<<<
 * from fiona import compat
 * from fiona.errors import (
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_GEOMETRY_TYPES);
  __Pyx_GIVEREF(__pyx_n_s_GEOMETRY_TYPES);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_GEOMETRY_TYPES);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fiona__geometry, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEOMETRY_TYPES, __pyx_t_1) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":28
 * from fiona._err import cpl_errs, FionaNullPointerError, CPLE_BaseError, CPLE_OpenFailedError
 * from fiona._geometry import GEOMETRY_TYPES
 * from fiona import compat             # <<<<<<<<<<<<<<
 * from fiona.errors import (
 *     DriverError, DriverIOError, SchemaError, CRSError, FionaValueError,
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_compat);
  __Pyx_GIVEREF(__pyx_n_s_compat);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_compat);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_compat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_compat, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":30
 * from fiona import compat
 * from fiona.errors import (
 *     DriverError, DriverIOError, SchemaError, CRSError, FionaValueError,             # <<<<<<<<<<<<<<
 *     TransactionError, GeometryTypeValidationError, DatasetDeleteError,
 *     FionaDeprecationWarning)
 */
  __pyx_t_1 = PyList_New(9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_DriverError);
  __Pyx_GIVEREF(__pyx_n_s_DriverError);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_DriverError);
  __Pyx_INCREF(__pyx_n_s_DriverIOError);
  __Pyx_GIVEREF(__pyx_n_s_DriverIOError);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_DriverIOError);
  __Pyx_INCREF(__pyx_n_s_SchemaError);
  __Pyx_GIVEREF(__pyx_n_s_SchemaError);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_SchemaError);
  __Pyx_INCREF(__pyx_n_s_CRSError);
  __Pyx_GIVEREF(__pyx_n_s_CRSError);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_CRSError);
  __Pyx_INCREF(__pyx_n_s_FionaValueError);
  __Pyx_GIVEREF(__pyx_n_s_FionaValueError);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_FionaValueError);
  __Pyx_INCREF(__pyx_n_s_TransactionError);
  __Pyx_GIVEREF(__pyx_n_s_TransactionError);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_TransactionError);
  __Pyx_INCREF(__pyx_n_s_GeometryTypeValidationError);
  __Pyx_GIVEREF(__pyx_n_s_GeometryTypeValidationError);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_s_GeometryTypeValidationError);
  __Pyx_INCREF(__pyx_n_s_DatasetDeleteError);
  __Pyx_GIVEREF(__pyx_n_s_DatasetDeleteError);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_s_DatasetDeleteError);
  __Pyx_INCREF(__pyx_n_s_FionaDeprecationWarning);
  __Pyx_GIVEREF(__pyx_n_s_FionaDeprecationWarning);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_s_FionaDeprecationWarning);

  /* "fiona/ogrext.pyx":29
 * from fiona._geometry import GEOMETRY_TYPES
 * from fiona import compat
 * from fiona.errors import (             # <<<<<<<<<<<<<<
 *     DriverError, DriverIOError, SchemaError, CRSError, FionaValueError,
 *     TransactionError, GeometryTypeValidationError, DatasetDeleteError,
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fiona_errors, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DriverError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DriverError, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DriverIOError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DriverIOError, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_SchemaError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SchemaError, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_CRSError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CRSError, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FionaValueError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaValueError, __pyx_t_1) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TransactionError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TransactionError, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_GeometryTypeValidationError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GeometryTypeValidationError, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DatasetDeleteError); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DatasetDeleteError, __pyx_t_1) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FionaDeprecationWarning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaDeprecationWarning, __pyx_t_1) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":33
 *     TransactionError, GeometryTypeValidationError, DatasetDeleteError,
 *     FionaDeprecationWarning)
 * from fiona.compat import OrderedDict             # <<<<<<<<<<<<<<
 * from fiona.rfc3339 import parse_date, parse_datetime, parse_time
 * from fiona.rfc3339 import FionaDateType, FionaDateTimeType, FionaTimeType
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_OrderedDict);
  __Pyx_GIVEREF(__pyx_n_s_OrderedDict);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderedDict);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona_compat, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_OrderedDict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderedDict, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":34
 *     FionaDeprecationWarning)
 * from fiona.compat import OrderedDict
 * from fiona.rfc3339 import parse_date, parse_datetime, parse_time             # <<<<<<<<<<<<<<
 * from fiona.rfc3339 import FionaDateType, FionaDateTimeType, FionaTimeType
 * from fiona.schema import FIELD_TYPES, FIELD_TYPES_MAP, normalize_field_type
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_parse_date);
  __Pyx_GIVEREF(__pyx_n_s_parse_date);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_parse_date);
  __Pyx_INCREF(__pyx_n_s_parse_datetime);
  __Pyx_GIVEREF(__pyx_n_s_parse_datetime);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_parse_datetime);
  __Pyx_INCREF(__pyx_n_s_parse_time);
  __Pyx_GIVEREF(__pyx_n_s_parse_time);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_parse_time);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fiona_rfc3339, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_parse_date); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_date, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_parse_datetime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_datetime, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_parse_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_parse_time, __pyx_t_1) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":35
 * from fiona.compat import OrderedDict
 * from fiona.rfc3339 import parse_date, parse_datetime, parse_time
 * from fiona.rfc3339 import FionaDateType, FionaDateTimeType, FionaTimeType             # <<<<<<<<<<<<<<
 * from fiona.schema import FIELD_TYPES, FIELD_TYPES_MAP, normalize_field_type
 * from fiona.path import vsi_path
 */
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_FionaDateType);
  __Pyx_GIVEREF(__pyx_n_s_FionaDateType);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_FionaDateType);
  __Pyx_INCREF(__pyx_n_s_FionaDateTimeType);
  __Pyx_GIVEREF(__pyx_n_s_FionaDateTimeType);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_FionaDateTimeType);
  __Pyx_INCREF(__pyx_n_s_FionaTimeType);
  __Pyx_GIVEREF(__pyx_n_s_FionaTimeType);
  PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_FionaTimeType);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona_rfc3339, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_FionaDateType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaDateType, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_FionaDateTimeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaDateTimeType, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_FionaTimeType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FionaTimeType, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":36
 * from fiona.rfc3339 import parse_date, parse_datetime, parse_time
 * from fiona.rfc3339 import FionaDateType, FionaDateTimeType, FionaTimeType
 * from fiona.schema import FIELD_TYPES, FIELD_TYPES_MAP, normalize_field_type             # <<<<<<<<<<<<<<
 * from fiona.path import vsi_path
 * 
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_FIELD_TYPES);
  __Pyx_GIVEREF(__pyx_n_s_FIELD_TYPES);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_FIELD_TYPES);
  __Pyx_INCREF(__pyx_n_s_FIELD_TYPES_MAP);
  __Pyx_GIVEREF(__pyx_n_s_FIELD_TYPES_MAP);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_FIELD_TYPES_MAP);
  __Pyx_INCREF(__pyx_n_s_normalize_field_type);
  __Pyx_GIVEREF(__pyx_n_s_normalize_field_type);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_normalize_field_type);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fiona_schema, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FIELD_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FIELD_TYPES, __pyx_t_1) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FIELD_TYPES_MAP); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FIELD_TYPES_MAP, __pyx_t_1) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_normalize_field_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_normalize_field_type, __pyx_t_1) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fiona/ogrext.pyx":37
 * from fiona.rfc3339 import FionaDateType, FionaDateTimeType, FionaTimeType
 * from fiona.schema import FIELD_TYPES, FIELD_TYPES_MAP, normalize_field_type
 * from fiona.path import vsi_path             # <<<<<<<<<<<<<<
 * 
 * from fiona._shim cimport is_field_null, osr_get_name, osr_set_traditional_axis_mapping_strategy
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_vsi_path);
  __Pyx_GIVEREF(__pyx_n_s_vsi_path);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_vsi_path);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_fiona_path, __pyx_t_2, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_vsi_path); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_vsi_path, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":57
 * 
 * 
 * log = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * 
 * DEFAULT_TRANSACTION_SIZE = 20000
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_logging); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_log, __pyx_t_3) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":59
 * log = logging.getLogger(__name__)
 * 
 * DEFAULT_TRANSACTION_SIZE = 20000             # <<<<<<<<<<<<<<
 * 
 * # OGR Driver capability
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DEFAULT_TRANSACTION_SIZE, __pyx_int_20000) < 0) __PYX_ERR(0, 59, __pyx_L1_error)

  /* "fiona/ogrext.pyx":62
 * 
 * # OGR Driver capability
 * cdef const char * ODrCCreateDataSource = "CreateDataSource"             # <<<<<<<<<<<<<<
 * cdef const char * ODrCDeleteDataSource = "DeleteDataSource"
 * 
 */
  __pyx_v_5fiona_6ogrext_ODrCCreateDataSource = ((char const *)"CreateDataSource");

  /* "fiona/ogrext.pyx":63
 * # OGR Driver capability
 * cdef const char * ODrCCreateDataSource = "CreateDataSource"
 * cdef const char * ODrCDeleteDataSource = "DeleteDataSource"             # <<<<<<<<<<<<<<
 * 
 * # OGR Layer capability
 */
  __pyx_v_5fiona_6ogrext_ODrCDeleteDataSource = ((char const *)"DeleteDataSource");

  /* "fiona/ogrext.pyx":66
 * 
 * # OGR Layer capability
 * cdef const char * OLC_RANDOMREAD = "RandomRead"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_SEQUENTIALWRITE = "SequentialWrite"
 * cdef const char * OLC_RANDOMWRITE = "RandomWrite"
 */
  __pyx_v_5fiona_6ogrext_OLC_RANDOMREAD = ((char const *)"RandomRead");

  /* "fiona/ogrext.pyx":67
 * # OGR Layer capability
 * cdef const char * OLC_RANDOMREAD = "RandomRead"
 * cdef const char * OLC_SEQUENTIALWRITE = "SequentialWrite"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_RANDOMWRITE = "RandomWrite"
 * cdef const char * OLC_FASTSPATIALFILTER = "FastSpatialFilter"
 */
  __pyx_v_5fiona_6ogrext_OLC_SEQUENTIALWRITE = ((char const *)"SequentialWrite");

  /* "fiona/ogrext.pyx":68
 * cdef const char * OLC_RANDOMREAD = "RandomRead"
 * cdef const char * OLC_SEQUENTIALWRITE = "SequentialWrite"
 * cdef const char * OLC_RANDOMWRITE = "RandomWrite"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_FASTSPATIALFILTER = "FastSpatialFilter"
 * cdef const char * OLC_FASTFEATURECOUNT = "FastFeatureCount"
 */
  __pyx_v_5fiona_6ogrext_OLC_RANDOMWRITE = ((char const *)"RandomWrite");

  /* "fiona/ogrext.pyx":69
 * cdef const char * OLC_SEQUENTIALWRITE = "SequentialWrite"
 * cdef const char * OLC_RANDOMWRITE = "RandomWrite"
 * cdef const char * OLC_FASTSPATIALFILTER = "FastSpatialFilter"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_FASTFEATURECOUNT = "FastFeatureCount"
 * cdef const char * OLC_FASTGETEXTENT = "FastGetExtent"
 */
  __pyx_v_5fiona_6ogrext_OLC_FASTSPATIALFILTER = ((char const *)"FastSpatialFilter");

  /* "fiona/ogrext.pyx":70
 * cdef const char * OLC_RANDOMWRITE = "RandomWrite"
 * cdef const char * OLC_FASTSPATIALFILTER = "FastSpatialFilter"
 * cdef const char * OLC_FASTFEATURECOUNT = "FastFeatureCount"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_FASTGETEXTENT = "FastGetExtent"
 * cdef const char * OLC_FASTSETNEXTBYINDEX = "FastSetNextByIndex"
 */
  __pyx_v_5fiona_6ogrext_OLC_FASTFEATURECOUNT = ((char const *)"FastFeatureCount");

  /* "fiona/ogrext.pyx":71
 * cdef const char * OLC_FASTSPATIALFILTER = "FastSpatialFilter"
 * cdef const char * OLC_FASTFEATURECOUNT = "FastFeatureCount"
 * cdef const char * OLC_FASTGETEXTENT = "FastGetExtent"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_FASTSETNEXTBYINDEX = "FastSetNextByIndex"
 * cdef const char * OLC_CREATEFIELD = "CreateField"
 */
  __pyx_v_5fiona_6ogrext_OLC_FASTGETEXTENT = ((char const *)"FastGetExtent");

  /* "fiona/ogrext.pyx":72
 * cdef const char * OLC_FASTFEATURECOUNT = "FastFeatureCount"
 * cdef const char * OLC_FASTGETEXTENT = "FastGetExtent"
 * cdef const char * OLC_FASTSETNEXTBYINDEX = "FastSetNextByIndex"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_CREATEFIELD = "CreateField"
 * cdef const char * OLC_CREATEGEOMFIELD = "CreateGeomField"
 */
  __pyx_v_5fiona_6ogrext_OLC_FASTSETNEXTBYINDEX = ((char const *)"FastSetNextByIndex");

  /* "fiona/ogrext.pyx":73
 * cdef const char * OLC_FASTGETEXTENT = "FastGetExtent"
 * cdef const char * OLC_FASTSETNEXTBYINDEX = "FastSetNextByIndex"
 * cdef const char * OLC_CREATEFIELD = "CreateField"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_CREATEGEOMFIELD = "CreateGeomField"
 * cdef const char * OLC_DELETEFIELD = "DeleteField"
 */
  __pyx_v_5fiona_6ogrext_OLC_CREATEFIELD = ((char const *)"CreateField");

  /* "fiona/ogrext.pyx":74
 * cdef const char * OLC_FASTSETNEXTBYINDEX = "FastSetNextByIndex"
 * cdef const char * OLC_CREATEFIELD = "CreateField"
 * cdef const char * OLC_CREATEGEOMFIELD = "CreateGeomField"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_DELETEFIELD = "DeleteField"
 * cdef const char * OLC_REORDERFIELDS = "ReorderFields"
 */
  __pyx_v_5fiona_6ogrext_OLC_CREATEGEOMFIELD = ((char const *)"CreateGeomField");

  /* "fiona/ogrext.pyx":75
 * cdef const char * OLC_CREATEFIELD = "CreateField"
 * cdef const char * OLC_CREATEGEOMFIELD = "CreateGeomField"
 * cdef const char * OLC_DELETEFIELD = "DeleteField"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_REORDERFIELDS = "ReorderFields"
 * cdef const char * OLC_ALTERFIELDDEFN = "AlterFieldDefn"
 */
  __pyx_v_5fiona_6ogrext_OLC_DELETEFIELD = ((char const *)"DeleteField");

  /* "fiona/ogrext.pyx":76
 * cdef const char * OLC_CREATEGEOMFIELD = "CreateGeomField"
 * cdef const char * OLC_DELETEFIELD = "DeleteField"
 * cdef const char * OLC_REORDERFIELDS = "ReorderFields"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_ALTERFIELDDEFN = "AlterFieldDefn"
 * cdef const char * OLC_DELETEFEATURE = "DeleteFeature"
 */
  __pyx_v_5fiona_6ogrext_OLC_REORDERFIELDS = ((char const *)"ReorderFields");

  /* "fiona/ogrext.pyx":77
 * cdef const char * OLC_DELETEFIELD = "DeleteField"
 * cdef const char * OLC_REORDERFIELDS = "ReorderFields"
 * cdef const char * OLC_ALTERFIELDDEFN = "AlterFieldDefn"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_DELETEFEATURE = "DeleteFeature"
 * cdef const char * OLC_STRINGSASUTF8 = "StringsAsUTF8"
 */
  __pyx_v_5fiona_6ogrext_OLC_ALTERFIELDDEFN = ((char const *)"AlterFieldDefn");

  /* "fiona/ogrext.pyx":78
 * cdef const char * OLC_REORDERFIELDS = "ReorderFields"
 * cdef const char * OLC_ALTERFIELDDEFN = "AlterFieldDefn"
 * cdef const char * OLC_DELETEFEATURE = "DeleteFeature"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_STRINGSASUTF8 = "StringsAsUTF8"
 * cdef const char * OLC_TRANSACTIONS = "Transactions"
 */
  __pyx_v_5fiona_6ogrext_OLC_DELETEFEATURE = ((char const *)"DeleteFeature");

  /* "fiona/ogrext.pyx":79
 * cdef const char * OLC_ALTERFIELDDEFN = "AlterFieldDefn"
 * cdef const char * OLC_DELETEFEATURE = "DeleteFeature"
 * cdef const char * OLC_STRINGSASUTF8 = "StringsAsUTF8"             # <<<<<<<<<<<<<<
 * cdef const char * OLC_TRANSACTIONS = "Transactions"
 * 
 */
  __pyx_v_5fiona_6ogrext_OLC_STRINGSASUTF8 = ((char const *)"StringsAsUTF8");

  /* "fiona/ogrext.pyx":80
 * cdef const char * OLC_DELETEFEATURE = "DeleteFeature"
 * cdef const char * OLC_STRINGSASUTF8 = "StringsAsUTF8"
 * cdef const char * OLC_TRANSACTIONS = "Transactions"             # <<<<<<<<<<<<<<
 * 
 * # OGR integer error types.
 */
  __pyx_v_5fiona_6ogrext_OLC_TRANSACTIONS = ((char const *)"Transactions");

  /* "fiona/ogrext.pyx":84
 * # OGR integer error types.
 * 
 * OGRERR_NONE = 0             # <<<<<<<<<<<<<<
 * OGRERR_NOT_ENOUGH_DATA = 1    # not enough data to deserialize *[inserted by cython to avoid comment closer]/
 * OGRERR_NOT_ENOUGH_MEMORY = 2
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_NONE, __pyx_int_0) < 0) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "fiona/ogrext.pyx":85
 * 
 * OGRERR_NONE = 0
 * OGRERR_NOT_ENOUGH_DATA = 1    # not enough data to deserialize *[inserted by cython to avoid comment closer]/             # <<<<<<<<<<<<<<
 * OGRERR_NOT_ENOUGH_MEMORY = 2
 * OGRERR_UNSUPPORTED_GEOMETRY_TYPE = 3
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_NOT_ENOUGH_DATA, __pyx_int_1) < 0) __PYX_ERR(0, 85, __pyx_L1_error)

  /* "fiona/ogrext.pyx":86
 * OGRERR_NONE = 0
 * OGRERR_NOT_ENOUGH_DATA = 1    # not enough data to deserialize *[inserted by cython to avoid comment closer]/
 * OGRERR_NOT_ENOUGH_MEMORY = 2             # <<<<<<<<<<<<<<
 * OGRERR_UNSUPPORTED_GEOMETRY_TYPE = 3
 * OGRERR_UNSUPPORTED_OPERATION = 4
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_NOT_ENOUGH_MEMORY, __pyx_int_2) < 0) __PYX_ERR(0, 86, __pyx_L1_error)

  /* "fiona/ogrext.pyx":87
 * OGRERR_NOT_ENOUGH_DATA = 1    # not enough data to deserialize *[inserted by cython to avoid comment closer]/
 * OGRERR_NOT_ENOUGH_MEMORY = 2
 * OGRERR_UNSUPPORTED_GEOMETRY_TYPE = 3             # <<<<<<<<<<<<<<
 * OGRERR_UNSUPPORTED_OPERATION = 4
 * OGRERR_CORRUPT_DATA = 5
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_UNSUPPORTED_GEOMETRY_TYPE, __pyx_int_3) < 0) __PYX_ERR(0, 87, __pyx_L1_error)

  /* "fiona/ogrext.pyx":88
 * OGRERR_NOT_ENOUGH_MEMORY = 2
 * OGRERR_UNSUPPORTED_GEOMETRY_TYPE = 3
 * OGRERR_UNSUPPORTED_OPERATION = 4             # <<<<<<<<<<<<<<
 * OGRERR_CORRUPT_DATA = 5
 * OGRERR_FAILURE = 6
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_UNSUPPORTED_OPERATION, __pyx_int_4) < 0) __PYX_ERR(0, 88, __pyx_L1_error)

  /* "fiona/ogrext.pyx":89
 * OGRERR_UNSUPPORTED_GEOMETRY_TYPE = 3
 * OGRERR_UNSUPPORTED_OPERATION = 4
 * OGRERR_CORRUPT_DATA = 5             # <<<<<<<<<<<<<<
 * OGRERR_FAILURE = 6
 * OGRERR_UNSUPPORTED_SRS = 7
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_CORRUPT_DATA, __pyx_int_5) < 0) __PYX_ERR(0, 89, __pyx_L1_error)

  /* "fiona/ogrext.pyx":90
 * OGRERR_UNSUPPORTED_OPERATION = 4
 * OGRERR_CORRUPT_DATA = 5
 * OGRERR_FAILURE = 6             # <<<<<<<<<<<<<<
 * OGRERR_UNSUPPORTED_SRS = 7
 * OGRERR_INVALID_HANDLE = 8
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_FAILURE, __pyx_int_6) < 0) __PYX_ERR(0, 90, __pyx_L1_error)

  /* "fiona/ogrext.pyx":91
 * OGRERR_CORRUPT_DATA = 5
 * OGRERR_FAILURE = 6
 * OGRERR_UNSUPPORTED_SRS = 7             # <<<<<<<<<<<<<<
 * OGRERR_INVALID_HANDLE = 8
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_UNSUPPORTED_SRS, __pyx_int_7) < 0) __PYX_ERR(0, 91, __pyx_L1_error)

  /* "fiona/ogrext.pyx":92
 * OGRERR_FAILURE = 6
 * OGRERR_UNSUPPORTED_SRS = 7
 * OGRERR_INVALID_HANDLE = 8             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OGRERR_INVALID_HANDLE, __pyx_int_8) < 0) __PYX_ERR(0, 92, __pyx_L1_error)

  /* "fiona/ogrext.pyx":95
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_1_explode, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_explode, __pyx_t_3) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":108
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry"""
 *     try:
 */
  __pyx_t_3 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_4_bounds, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bounds, __pyx_t_3) < 0) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "fiona/ogrext.pyx":117
 * 
 * 
 * cdef int GDAL_VERSION_NUM = get_gdal_version_num()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_get_gdal_version_num); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_5fiona_6ogrext_GDAL_VERSION_NUM = __pyx_t_4;

  /* "fiona/ogrext.pyx":424
 * 
 * 
 * def featureRT(feature, collection):             # <<<<<<<<<<<<<<
 *     # For testing purposes only, leaks the JSON data
 *     cdef void *cogr_feature = OGRFeatureBuilder().build(feature, collection)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_6featureRT, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_featureRT, __pyx_t_1) < 0) __PYX_ERR(0, 424, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":835
 *             raise KeyError("There is no feature with fid {!r}".format(fid))
 * 
 *     get = get_feature             # <<<<<<<<<<<<<<
 * 
 *     # TODO: Make this an alias for get_feature in a future version.
 */
  __Pyx_GetNameInClass(__pyx_t_1, (PyObject *)__pyx_ptype_5fiona_6ogrext_Session, __pyx_n_s_get_feature); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem((PyObject *)__pyx_ptype_5fiona_6ogrext_Session->tp_dict, __pyx_n_s_get, __pyx_t_1) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  PyType_Modified(__pyx_ptype_5fiona_6ogrext_Session);

  /* "fiona/ogrext.pyx":1434
 * 
 * 
 * def _remove(path, driver=None):             # <<<<<<<<<<<<<<
 *     """Deletes an OGR data source
 *     """
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_8_remove, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_remove, __pyx_t_1) < 0) __PYX_ERR(0, 1434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1464
 * 
 * 
 * def _remove_layer(path, layer, driver=None):             # <<<<<<<<<<<<<<
 *     cdef void *cogr_ds
 *     cdef int layer_index
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_10_remove_layer, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_remove_layer, __pyx_t_1) < 0) __PYX_ERR(0, 1464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1496
 * 
 * 
 * def _listlayers(path, **kwargs):             # <<<<<<<<<<<<<<
 * 
 *     """Provides a list of the layers in an OGR data source.
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_12_listlayers, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_listlayers, __pyx_t_1) < 0) __PYX_ERR(0, 1496, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1531
 * 
 * 
 * def buffer_to_virtual_file(bytesbuf, ext=''):             # <<<<<<<<<<<<<<
 *     """Maps a bytes buffer to a virtual file.
 * 
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_14buffer_to_virtual_file, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_buffer_to_virtual_file, __pyx_t_1) < 0) __PYX_ERR(0, 1531, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1550
 * 
 * 
 * def remove_virtual_file(vsi_filename):             # <<<<<<<<<<<<<<
 *     vsi_cfilename = vsi_filename if not isinstance(vsi_filename, string_types) else vsi_filename.encode('utf-8')
 *     return VSIUnlink(vsi_cfilename)
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_16remove_virtual_file, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_remove_virtual_file, __pyx_t_1) < 0) __PYX_ERR(0, 1550, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_FeatureBuilder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_18__pyx_unpickle_FeatureBuilder, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_FeatureBuilder, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_FeatureBuilder__set_state(<FeatureBuilder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_FeatureBuilder__set_state(FeatureBuilder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_20__pyx_unpickle_OGRFeatureBuilder, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_OGRFeatureBuilder, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MemoryFileBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_1 = PyCFunction_NewEx(&__pyx_mdef_5fiona_6ogrext_22__pyx_unpickle_MemoryFileBase, NULL, __pyx_n_s_fiona_ogrext); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MemoryFileBase, __pyx_t_1) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fiona/ogrext.pyx":1
 * # These are extension functions and classes using the OGR C API.             # <<<<<<<<<<<<<<
 * 
 * from __future__ import absolute_import
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init fiona.ogrext", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init fiona.ogrext");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return dict ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
        dictptr = (offset > 0) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (!dict || tp_dict_version != __PYX_GET_DICT_VERSION(dict))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyObjectFormatAndDecref */
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatSimpleAndDecref(PyObject* s, PyObject* f) {
    if (unlikely(!s)) return NULL;
    if (likely(PyUnicode_CheckExact(s))) return s;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyString_CheckExact(s))) {
        PyObject *result = PyUnicode_FromEncodedObject(s, NULL, "strict");
        Py_DECREF(s);
        return result;
    }
    #endif
    return __Pyx_PyObject_FormatAndDecref(s, f);
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_FormatAndDecref(PyObject* s, PyObject* f) {
    PyObject *result = PyObject_Format(s, f);
    Py_DECREF(s);
    return result;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      CYTHON_UNUSED Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind;
    Py_ssize_t i, char_pos;
    void *result_udata;
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely(char_pos + ulength < 0))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + char_pos * result_ukind, udata, (size_t) (ulength * result_ukind));
        } else {
            #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    result_ulength++;
    value_count++;
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* UnpackUnboundCMethod */
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    }
#endif
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    if (likely(cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return (*(cfunc->func))(self, arg);
        } else if (PY_VERSION_HEX >= 0x030600B1 && flag == METH_FASTCALL) {
            if (PY_VERSION_HEX >= 0x030700A0) {
                return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, &arg, 1);
            } else {
                return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
            }
        } else if (PY_VERSION_HEX >= 0x030700A0 && flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, &arg, 1, NULL);
        }
    }
    return __Pyx__CallUnboundCMethod1(cfunc, self, arg);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(1);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 1, arg);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(2, self, arg);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030600B1
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    if (likely(cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            #if PY_VERSION_HEX >= 0x030700A0
            return (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)cfunc->func)(self, args, 2);
            #else
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
            #endif
        }
        #if PY_VERSION_HEX >= 0x030700A0
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, 2, NULL);
        #endif
    }
    return __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        args = PyTuple_New(2);
        if (unlikely(!args)) goto bad;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = (*(PyCFunctionWithKeywords)(void*)(PyCFunction)cfunc->func)(self, args, NULL);
        else
            result = (*cfunc->func)(self, args);
    } else {
        args = PyTuple_New(3);
        if (unlikely(!args)) goto bad;
        Py_INCREF(self);
        PyTuple_SET_ITEM(args, 0, self);
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 1, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 2, arg2);
        result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    }
#else
    args = PyTuple_Pack(3, self, arg1, arg2);
    if (unlikely(!args)) goto bad;
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
#endif
bad:
    Py_XDECREF(args);
    return result;
}

/* dict_getitem_default */
static PyObject* __Pyx_PyDict_GetItemDefault(PyObject* d, PyObject* key, PyObject* default_value) {
    PyObject* value;
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (unlikely(PyErr_Occurred()))
            return NULL;
        value = default_value;
    }
    Py_INCREF(value);
    if ((1));
#else
    if (PyString_CheckExact(key) || PyUnicode_CheckExact(key) || PyInt_CheckExact(key)) {
        value = PyDict_GetItem(d, key);
        if (unlikely(!value)) {
            value = default_value;
        }
        Py_INCREF(value);
    }
#endif
    else {
        if (default_value == Py_None)
            value = __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_get, d, key);
        else
            value = __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_get, d, key, default_value);
    }
    return value;
}

/* py_dict_pop */
static CYTHON_INLINE PyObject *__Pyx_PyDict_Pop(PyObject *d, PyObject *key, PyObject *default_value) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B3
    if ((1)) {
        return _PyDict_Pop(d, key, default_value);
    } else
#endif
    if (default_value) {
        return __Pyx_CallUnboundCMethod2(&__pyx_umethod_PyDict_Type_pop, d, key, default_value);
    } else {
        return __Pyx_CallUnboundCMethod1(&__pyx_umethod_PyDict_Type_pop, d, key);
    }
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* PyIntCompare */
static CYTHON_INLINE PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED long inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = Py_SIZE(op1);
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        if (intval == 0) {
            if (size == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
        } else if (intval < 0) {
            if (size >= 0)
                Py_RETURN_FALSE;
            intval = -intval;
            size = -size;
        } else {
            if (size <= 0)
                Py_RETURN_FALSE;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        if (unequal == 0) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
        if ((double)a == (double)b) Py_RETURN_TRUE; else Py_RETURN_FALSE;
    }
    return (
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
#if PY_MAJOR_VERSION < 3 || CYTHON_USE_PYLONG_INTERNALS
#define __Pyx_PyInt_RemainderObjC_ZeroDivisionError(operand)\
    if (unlikely(zerodivision_check && ((operand) == 0))) {\
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");\
        return NULL;\
    }
#endif
static PyObject* __Pyx_PyInt_RemainderObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            __Pyx_PyInt_RemainderObjC_ZeroDivisionError(b)
            x = a % b;
            x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyInt_FromLong(x);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_remainder(op1, op2);
            }
        }
                __Pyx_PyInt_RemainderObjC_ZeroDivisionError(b)
                x = a % b;
                x += ((x != 0) & ((x ^ b) < 0)) * b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla % llb;
                llx += ((llx != 0) & ((llx ^ llb) < 0)) * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceRemainder : PyNumber_Remainder)(op1, op2);
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* FetchCommonType */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
#include <structmember.h>
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, CYTHON_UNUSED void *context) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *context) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_cython); if (unlikely(!reduce_cython)) goto BAD;
            ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto BAD;
            ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto BAD;
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate_cython); if (unlikely(!setstate_cython)) goto BAD;
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto BAD;
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto GOOD;
BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* GetNameInClass */
static PyObject *__Pyx_GetGlobalNameAfterAttributeLookup(PyObject *name) {
    PyObject *result;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    result = __Pyx_PyObject_GetAttrStr(nmspace, name);
    if (!result) {
        result = __Pyx_GetGlobalNameAfterAttributeLookup(name);
    }
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) ((PY_LONG_LONG) 0 - (PY_LONG_LONG) 1), const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_OGRErr(OGRErr value) {
    const OGRErr neg_one = (OGRErr) ((OGRErr) 0 - (OGRErr) 1), const_zero = (OGRErr) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(OGRErr) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(OGRErr) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(OGRErr) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(OGRErr) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(OGRErr) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(OGRErr),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_vsi_l_offset(vsi_l_offset value) {
    const vsi_l_offset neg_one = (vsi_l_offset) ((vsi_l_offset) 0 - (vsi_l_offset) 1), const_zero = (vsi_l_offset) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(vsi_l_offset) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(vsi_l_offset) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(vsi_l_offset) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(vsi_l_offset) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(vsi_l_offset) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(vsi_l_offset),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) ((int) 0 - (int) 1), const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) ((long) 0 - (long) 1), const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) ((PY_LONG_LONG) 0 - (PY_LONG_LONG) 1), const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (PY_LONG_LONG) -1;
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntFromPy */
static CYTHON_INLINE OGRFieldType __Pyx_PyInt_As_OGRFieldType(PyObject *x) {
    const OGRFieldType neg_one = (OGRFieldType) ((OGRFieldType) 0 - (OGRFieldType) 1), const_zero = (OGRFieldType) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(OGRFieldType) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(OGRFieldType, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (OGRFieldType) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (OGRFieldType) 0;
                case  1: __PYX_VERIFY_RETURN_INT(OGRFieldType, digit, digits[0])
                case 2:
                    if (8 * sizeof(OGRFieldType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) >= 2 * PyLong_SHIFT) {
                            return (OGRFieldType) (((((OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(OGRFieldType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) >= 3 * PyLong_SHIFT) {
                            return (OGRFieldType) (((((((OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(OGRFieldType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) >= 4 * PyLong_SHIFT) {
                            return (OGRFieldType) (((((((((OGRFieldType)digits[3]) << PyLong_SHIFT) | (OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (OGRFieldType) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(OGRFieldType) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(OGRFieldType, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(OGRFieldType) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(OGRFieldType, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (OGRFieldType) 0;
                case -1: __PYX_VERIFY_RETURN_INT(OGRFieldType, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(OGRFieldType,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(OGRFieldType) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 2 * PyLong_SHIFT) {
                            return (OGRFieldType) (((OGRFieldType)-1)*(((((OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(OGRFieldType) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 2 * PyLong_SHIFT) {
                            return (OGRFieldType) ((((((OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(OGRFieldType) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 3 * PyLong_SHIFT) {
                            return (OGRFieldType) (((OGRFieldType)-1)*(((((((OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(OGRFieldType) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 3 * PyLong_SHIFT) {
                            return (OGRFieldType) ((((((((OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(OGRFieldType) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 4 * PyLong_SHIFT) {
                            return (OGRFieldType) (((OGRFieldType)-1)*(((((((((OGRFieldType)digits[3]) << PyLong_SHIFT) | (OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(OGRFieldType) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(OGRFieldType, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(OGRFieldType) - 1 > 4 * PyLong_SHIFT) {
                            return (OGRFieldType) ((((((((((OGRFieldType)digits[3]) << PyLong_SHIFT) | (OGRFieldType)digits[2]) << PyLong_SHIFT) | (OGRFieldType)digits[1]) << PyLong_SHIFT) | (OGRFieldType)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(OGRFieldType) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(OGRFieldType, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(OGRFieldType) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(OGRFieldType, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            OGRFieldType val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (OGRFieldType) -1;
        }
    } else {
        OGRFieldType val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (OGRFieldType) -1;
        val = __Pyx_PyInt_As_OGRFieldType(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to OGRFieldType");
    return (OGRFieldType) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to OGRFieldType");
    return (OGRFieldType) -1;
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) ((size_t) 0 - (size_t) 1), const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE vsi_l_offset __Pyx_PyInt_As_vsi_l_offset(PyObject *x) {
    const vsi_l_offset neg_one = (vsi_l_offset) ((vsi_l_offset) 0 - (vsi_l_offset) 1), const_zero = (vsi_l_offset) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(vsi_l_offset) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(vsi_l_offset, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (vsi_l_offset) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (vsi_l_offset) 0;
                case  1: __PYX_VERIFY_RETURN_INT(vsi_l_offset, digit, digits[0])
                case 2:
                    if (8 * sizeof(vsi_l_offset) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) >= 2 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((((vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(vsi_l_offset) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) >= 3 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((((((vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(vsi_l_offset) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) >= 4 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((((((((vsi_l_offset)digits[3]) << PyLong_SHIFT) | (vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (vsi_l_offset) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(vsi_l_offset) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(vsi_l_offset, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(vsi_l_offset) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(vsi_l_offset, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (vsi_l_offset) 0;
                case -1: __PYX_VERIFY_RETURN_INT(vsi_l_offset, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(vsi_l_offset,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(vsi_l_offset) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 2 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((vsi_l_offset)-1)*(((((vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(vsi_l_offset) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 2 * PyLong_SHIFT) {
                            return (vsi_l_offset) ((((((vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(vsi_l_offset) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 3 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((vsi_l_offset)-1)*(((((((vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(vsi_l_offset) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 3 * PyLong_SHIFT) {
                            return (vsi_l_offset) ((((((((vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(vsi_l_offset) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 4 * PyLong_SHIFT) {
                            return (vsi_l_offset) (((vsi_l_offset)-1)*(((((((((vsi_l_offset)digits[3]) << PyLong_SHIFT) | (vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(vsi_l_offset) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(vsi_l_offset, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(vsi_l_offset) - 1 > 4 * PyLong_SHIFT) {
                            return (vsi_l_offset) ((((((((((vsi_l_offset)digits[3]) << PyLong_SHIFT) | (vsi_l_offset)digits[2]) << PyLong_SHIFT) | (vsi_l_offset)digits[1]) << PyLong_SHIFT) | (vsi_l_offset)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(vsi_l_offset) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(vsi_l_offset, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(vsi_l_offset) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(vsi_l_offset, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            vsi_l_offset val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (vsi_l_offset) -1;
        }
    } else {
        vsi_l_offset val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (vsi_l_offset) -1;
        val = __Pyx_PyInt_As_vsi_l_offset(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to vsi_l_offset");
    return (vsi_l_offset) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to vsi_l_offset");
    return (vsi_l_offset) -1;
}

/* PyObjectCallMethod1 */
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
}

/* CoroutineBase */
#include <structmember.h>
#include <frameobject.h>
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(CYTHON_UNUSED __pyx_CoroutineObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(CYTHON_UNUSED PyObject *gen) {
    const char *msg;
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(CYTHON_UNUSED PyObject *gen, PyObject *value, CYTHON_UNUSED int closing) {
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_type) {
        #if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
        #else
        if (exc_state->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_state->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = tstate->frame;
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body((PyObject *) self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
    PyObject *exc_tb = exc_state->exc_traceback;
    if (likely(exc_tb)) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
#endif
    }
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(CYTHON_UNUSED PyObject* gen, PyObject *retval) {
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = _PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, CYTHON_UNUSED PyObject *arg) {
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = PyObject_CallObject(meth, args);
            } else {
                ret = PyObject_CallFunctionObjArgs(meth, typ, val, tb, NULL);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, CYTHON_UNUSED void *context)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, CYTHON_UNUSED void *context)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* FunctionImport */
#ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
